{$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 8192,0,0}

Program Aliens;

{ This program is a version of Space Invaders.  Its primary purpose is as
  a mock-up for porting to Z80 assembler.
  For this reason, there are many non-Pascal aspects:
    1. All conditionals end with gotos,
    2. There are no local variables, no functions, and only byte types
        (except in the computer specific code section).
    3. Minimum code size is acheived at the expense of CPU time.
    4. Numeric expressions are evaluated in several small steps.
    5. All identifiers have game related names to avoid conflict with existing
         identifiers (if any) in the target computer.
    6. Video output is limited to one character at a time.
    7. There are few multiple statements on a line.
  Globally replace the constants and constant expressions with fixed values
  approriate for the target video output device.  }

USES Dos, Crt;

CONST
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;
  SaveVMode : Byte = 0;
  NAlienColumns = 5;  NAlienRows = 3;  { format of alien army }
  AlienXGap = 4;  AlienYGap = 2;       { spacing of alien army ranks }
  GameMaxX = 39;  GameMaxY = 18;       { absolute screen size }
  PlayerMaxX = GameMaxX - AlienXGap;   { farthest player right }
  PlayerY = GameMaxY - 1;              { row player exists in }
  ShootY = PlayerY - 1;                { row bullets start in }
  ShieldY = ShootY - 1;                { row shield are drawn in }
  AlienMaxY = PlayerY;                 { aliens win when they reach here }
  AlienMaxYStart =                  { lowest start position of alien army }
    AlienMaxY - ( NAlienRows * AlienYGap ) - 2;
  MaxGun = 3;  MaxBomb = 2;            { maximum simultaneous weapons }
  MaxSuperBlast = 2;          { maximum number of player super weapon uses }
  ShieldSize = GameMaxX div 6;         { size of each shield }
  AlienUpdateFreq = 2;             { aliens move n times slower than player }
  BombMaxY = GameMaxY + 7;         { point when new bomb is created }
  SuperBombFreq = 5;      { aliens drop one super bomb for every n regular }
  MaxLife = 3;                         { total player lives }
  MaxAlien =                           { total aliens at game start }
    NAlienColumns * NAlienRows;
  AlienChar1 = Ord( 'M' );  AlienChar2 = Ord( 'W' );  { aliens }
  BombChar = Ord( 'v' );  BulletChar = Ord( '^' );    { bomb and bullet }
  SuperBombChar = Ord( 'x' );                         { super bomb }
  BlankChar = Ord( ' ' );  HitChar = Ord( '*' );      { blank and hit }
  PlayerChar = Ord( 'A' );  ShieldChar = Ord( '-' );  { player and shield }
{ THESE MOVEMENT KEYS ARE COMPUTER SPECIFIC }
  SuperBlastKey = 13; { enter }                       { super blast bomb }
  LeftKey = Ord( ',' );  RightKey = Ord( '.' );       { left and right }
  FireKey = Ord( ' ' );  QuitKey = 27; { esc }        { fire and quit }
  NoKey = 0;                                          { no key pressed }
  Silent : Boolean = False;

VAR
  AlienUpdateTimer : Byte;             { countdown to next alien move }
  AlienX, AlienY : Byte;               { current alien army offset }
  { all alien x positioning is offset by GameMaxX to allow for destruction }
  { of outer alien army columns, without using negative numbers }
  AlienMaxX, AlienMinX : Byte;         { limits of alien army movement }
  AlienMinColumn, ALienMaxColumn : Byte;  { outer columns of alien army }
  AlienNewX, AlienNewY : Byte;         { used to calculate next alien move }
  AlienDirection : Byte;      { alien direction = 0/1/2 = left/right/none }
  AliensAlive : Byte;                  { number of aliens currently alive }
  AlienHit : Byte;                     { flag that an alien has been hit }
  PlayerX : Byte;                      { current player column position }
  PlayerDirection : Byte;     { player direction = 0/1/2 = left/right/none }
  AlienAlive :                { flag for each alien = 0/1 = dead/alive }
    Array [ 1..MaxAlien ] of Byte;
  BulletX, BulletY :                       { position of each bullet }
    Array [ 1..MaxGun ] of Byte;
  BombX, BombY, BombStrength :          { position and type of each bomb }
    Array [ 1..MaxBomb ] of Byte;
  BombRow : Byte;                          { lowest aliens in column }
  WhichGun : Byte;                              { next available gun }
  SuperBlastCount : Byte;   { number of player super weapon uses remaining }
  SuperBombTimer : Byte;                   { countdown to next super bomb }
  SuperRound : Byte;         { flag that player has reached expert rounds }
  BombLoop, GameLoop, ScoreLoop, GameTemp : Byte; { various loop counters }
  GameX, GameY, GetChar, PutChar : Byte;   { pos. and char. for video out }
  LivesRemaining : Byte;                   { number of player lives left }
  PlayerHit : Byte;                        { flag that player has been hit }
  Score, HiScore : Word;                   { only word types in program }
  HiName : String[3];               { initials, only string type in program }
  Round : Byte;                            { number of alien waves }
  SaveMode : Word;
  Con : Text;

{***  THESE PROCEDURES ARE COMPUTER SPECIFIC  *************************}

Procedure SetVMode( M : Byte );
  Var
    Regs : Registers;
  Begin
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $00;  Regs.AL := M;
    Intr( $10, Regs );
  End;

Function GetVMode : Byte;
  Var
    Regs : Registers;
  Begin
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := 15;
    Intr( $10, Regs );
    GetVMode := Regs.AL;
  End;

{----------------------------------------------------------------------}

PROCEDURE ShowHiScore;
  VAR
    HiFile:Text;  HiResult:Integer;
    HiScrStr : String[20];
  BEGIN
    Assign( HiFile, 'ALIENS.HIS' );
    {$I-} Reset( HiFile ); {$I+} HiResult := IOResult;
    If HiResult = 0
      THEN BEGIN
        {$I-} Readln( HiFile, HiScrStr ); {$I+} HiResult := IOResult;
        Val( HiScrStr, HiScore, HiResult );
        If HiResult = 0
          THEN BEGIN
            {$I-} Readln( HiFile, HiName ); {$I+} HiResult := IOResult;
          END;
      END;
    {$I-} Close( HiFile ); {$I+} HiResult := IOResult;
    If HiResult <> 0
      THEN BEGIN  HiScore := 0;  HiName := '';  END;
    GotoXY( 20, GameMaxY );
    If HiScore > 0
      THEN Write( 'HI: ' + HiScrStr + '000 ' + HiName );
  END;

{----------------------------------------------------------------------}

PROCEDURE CheckHiScore;
  VAR  HiFile:Text;  HiResult:Integer;
  BEGIN
    If Score > HiScore
      THEN BEGIN
        HiName := '???';
        If ParamCount > 0 Then
          Begin
            If Silent and ( ParamCount > 1 ) Then
              HiName := ParamStr( 2 )
            Else
              If not Silent Then
                HiName := ParamStr( 1 );
          End;
        For HiResult := 1 to Length( HiName ) do
          HiName[ HiResult ] := UpCase( HiName[ HiResult ] );
        Assign( HiFile, 'ALIENS.HIS' );
        {$I-} ReWrite( HiFile ); {$I+} HiResult := IOResult;
        If HiResult = 0
          THEN BEGIN
            {$I-} Writeln( HiFile, Score ); {$I+} HiResult := IOResult;
            If HiResult = 0
              THEN BEGIN
                {$I-} Writeln( HiFile, HiName ); {$I+} HiResult := IOResult;
              END;
          END;
        {$I-} Close( HiFile ); {$I+} HiResult := IOResult;
        If HiResult = 0 THEN HiScore := Score;
      END;
  END;

{----------------------------------------------------------------------}

PROCEDURE GameClearScreen;
  VAR Regs:Registers;
  BEGIN
    If ( ParamCount > 0 ) and ( ParamStr( 1 ) = '?' )
      THEN BEGIN
        For CpyRtLp := 1 to 26 do
          If CpyRtLp = 1 Then
            CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
          Else
            CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                     Ord( CpyRt[ CpyRtLp - 1 ] ) );
        Assign( Con, '' );  ReWrite( Con );
        Writeln( Con, 'ALIENS ver 3.1 rev 91/06/29 - ' + CpyRt + '.' );
        Writeln( Con, 'Basic marching ranks of invaders game.' );
        Writeln( Con, '  Steer with <- and -> keys,' );
        Writeln( Con, '  <SPACE> fires, <ESC> quits,' );
        Writeln( Con, '  <ENTER> for super blast clears all bombs.' );
        Writeln( Con, 'ALIENS [/S] [name]' );
        Writeln( Con, '  /S    Silence all sound effects.' );
        Writeln( Con, '  name  Store name if high score exceeded.' );
        Writeln( Con, 'High score is stored in ALIENS.HIS file.' );
        Writeln( Con, 'CGA compatible graphics is required.' );
        Close( Con );  Halt( 1 );
      END;
    If SaveMode = $07 Then
      Begin
        Assign( Con, '' );  ReWrite( Con );
        Writeln( 'CGA graphics capability required.' );
        Close( Con );
        Halt( 1 );
      End;
    TextMode( BW40 );  { 320x200x4 }
  { get current cursor }
    Regs.AH := $03;
    Regs.BH := $00;
    Intr( $10, Regs );
  { set bits to make it invisible }
    Regs.AH := $01;
    Regs.CX := Regs.CX or $2000;
    Intr( $10, Regs );
  END;

{----------------------------------------------------------------------}

PROCEDURE GameRestoreScreen;
  VAR Regs:Registers;
  BEGIN
    SetVMode( SaveMode );
  { get current cursor }
    Regs.AH := $03;
    Regs.BH := $00;
    Intr( $10, Regs );
  { reset bits to make it visible again }
    Regs.AH := $01;
    Regs.CX := Regs.CX and $DFFF;
    Intr( $10, Regs );
  END;

{----------------------------------------------------------------------}

PROCEDURE GetKey;
  BEGIN
    GetChar := NoKey;
    If not KeyPressed THEN Exit;
    GetChar := Ord( UpCase( ReadKey ) );
    If ( GetChar = 0 ) and KeyPressed Then
      Begin
        GetChar := Ord( ReadKey );
        If GetChar = 75 Then
          GetChar := LeftKey
        Else
          If GetChar = 77 Then
            GetChar := RightKey
          Else
            GetChar := NoKey;
      End;
  END;

{----------------------------------------------------------------------}

PROCEDURE DoGetChar;
  VAR Regs:Registers;
  BEGIN
    GotoXY( GameX, GameY );
    Regs.AH := $08;
    Regs.BH := $00;
    Intr( $10, Regs );
    GetChar := Regs.AL;
  END;

{----------------------------------------------------------------------}

PROCEDURE DoPutChar;
  BEGIN  { assumes top left is 1, 1 }
    GotoXY( GameX, GameY );  Write( Chr( PutChar ) );
  END;

{----------------------------------------------------------------------}

PROCEDURE WriteScore;
  Var
    ScrStr : String[20];
  BEGIN
    GotoXY( GameX, GameY );
    Str( Score, ScrStr );
    Write( ScrStr );
    If Score > 0 THEN Write( '000' );
  END;

{----------------------------------------------------------------------}

PROCEDURE GameCycleDelay;
  BEGIN  Delay( 40 );  END;

{----------------------------------------------------------------------}

PROCEDURE Beep( Freq, Dur:Word );
  BEGIN
    If not Silent Then Sound( Freq );  Delay( Dur );  NoSound;
  END;

PROCEDURE ShootSound;
  BEGIN  Beep( 8000, 50 );  END;

PROCEDURE KillSound;
  BEGIN  Beep( 2000, 70 );  Beep( 4000, 70 );  END;

PROCEDURE DieSound;
  BEGIN
    Beep( 150, 400 );  Beep( 140, 400 );
    Beep( 130, 400 );  Beep( 120, 400 );
  END;

PROCEDURE NewRoundSound;
  BEGIN
    Beep( 392, 200 );  Beep( 523, 200 );  Beep( 659, 200 );
    Beep( 784, 400 );  Beep( 659, 200 );  Beep( 784, 200 );
  END;

PROCEDURE SmashSound;
  BEGIN  Beep( 600, 50 );  Beep( 300, 50 );  Beep( 600, 50 );  END;

PROCEDURE AlienMoveSound;
  BEGIN  Beep( 200, 20 );  END;

PROCEDURE BombSound;
  BEGIN  Beep( 400, 20 );  END;

PROCEDURE SuperBombSound;
  BEGIN  Beep( 300, 20 );  Beep( 290, 20 );  Beep( 280, 20 );  END;

PROCEDURE SuperBlastSound;
  BEGIN
    Beep( 100, 70 );  Beep( 400, 70 );
    Beep( 1600, 70 );  Beep( 6400, 70 );
  END;

{----------------------------------------------------------------------}

{***  BEGINNING OF GENERIC PROGRAM  ***********************************}

{----------------------------------------------------------------------}

PROCEDURE ShowScore;
  LABEL LifeLoop, LifeMark;
  BEGIN
  { show score at screen bottom }
    GameY := GameMaxY;
    ScoreLoop := 0;
  LifeLoop:
    Inc( ScoreLoop );
  { show one hit character for each life lost }
    PutChar := HitChar;
    If ScoreLoop > LivesRemaining THEN Goto LifeMark;
  { show one player character for each life remaining }
    PutChar := PlayerChar;
  LifeMark:
    GameX := ScoreLoop;  DoPutChar;
    If ScoreLoop < MaxLife THEN Goto LifeLoop;
  { write score }
    GameX := GameX + 4;
    WriteScore;
  END;

{----------------------------------------------------------------------}

PROCEDURE ClearWeapons;
  LABEL ClearGuns, ClearBombs, ClearDone;
  BEGIN
    PutChar := BlankChar;
    GameLoop := 0;
    ScoreLoop := 0;
  ClearGuns:
  { blank all live bullets }
    Inc( GameLoop );
    If GameLoop > MaxGun THEN Goto ClearBombs;
    If BulletY[ GameLoop ] = 0 THEN Goto ClearGuns;
    GameY := BulletY[ GameLoop ];
    GameX := BulletX[ GameLoop ];
    DoPutChar;
    Goto ClearGuns;
  ClearBombs:
  { blank all live bombs }
    Inc( ScoreLoop );
    If ScoreLoop > MaxBomb THEN Goto ClearDone;
    If BombY[ ScoreLoop ] > PlayerY THEN Goto ClearBombs;
    GameY := BombY[ ScoreLoop ];
    GameX := BombX[ ScoreLoop ];
    DoPutChar;
    Goto ClearBombs;
  ClearDone:
  { set all positions out of range }
    FillChar( BulletY, MaxGun, 0 );
    FillChar( BombY, MaxBomb, PlayerY + 1 );
  END;

{----------------------------------------------------------------------}

PROCEDURE FindGun;
  LABEL GunCountLoop, GunCountExit;
  BEGIN
  { find first available gun }
    WhichGun := 1;
  GunCountLoop:
    If BulletY[ WhichGun ] = 0 THEN Goto GunCountExit;
    Inc( WhichGun );
    If WhichGun <= MaxGun THEN Goto GunCountLoop;
  GunCountExit:
  END;

{----------------------------------------------------------------------}

PROCEDURE MovePlayer;
  { used by player update }
  LABEL  MoveSkipDec, MoveSkipInc, MovePutPlayer;
  BEGIN
  { blank previous player position }
    GameX := PlayerX;  PutChar := BlankChar;  GameY := PlayerY;
    DoPutChar;
  { if direction = 0, move left }
    If PlayerDirection <> 0 THEN Goto MoveSkipDec;
    Dec( GameX );
  MoveSkipDec:
  { if direction = 1, move right }
    If PlayerDirection <> 1 THEN Goto MoveSkipInc;
    Inc( GameX );
  MoveSkipInc:
  { range check player position }
    If GameX < 1 THEN Goto MovePutPlayer;
    If GameX > PlayerMaxX THEN Goto MovePutPlayer;
  { if range ok, accept new position }
    PlayerX := GameX;
  MovePutPlayer:
  { replace player }
    GameX := PlayerX;  PutChar := PlayerChar;  DoPutChar;
  END;

{----------------------------------------------------------------------}

PROCEDURE UpdatePlayer;
  LABEL PlayerCheckKey, PlayerMove, PlayerExit, DoGun, DoSuperBlast;
  BEGIN
  { check if move or fire key pressed }
  PlayerCheckKey:
    PlayerDirection := 2;  GetKey;
    If GetChar = NoKey THEN Goto PlayerExit;
    If GetChar = FireKey THEN Goto DoGun;
    If GetChar = SuperBlastKey THEN Goto DoSuperBlast;
    PlayerDirection := 0;
    If GetChar = LeftKey THEN Goto PlayerMove;
    PlayerDirection := 1;
    If GetChar = RightKey THEN Goto PlayerMove;
    If GetChar <> QuitKey THEN Goto PlayerExit;
  { if quit, force end of lives }
    LivesRemaining := 1;
    PlayerHit := 1;
    Goto PlayerExit;
  PlayerMove:
    MovePlayer;
    Goto PlayerCheckKey;
  DoGun:
    FindGun;
    If WhichGun > MaxGun THEN Goto PlayerExit;
    ShootSound;
  { find a gun and create a bullet }
    GameX := PlayerX;  GameY := ShootY;  PutChar := BulletChar;
    DoPutChar;
    BulletX[ WhichGun ] := GameX;
    BulletY[ WhichGun ] := GameY;
    Goto PlayerExit;
  DoSuperBlast:
    If SuperBlastCount = 0 THEN Goto PlayerExit;
    ClearWeapons;
    Dec( SuperBlastCount );
  PlayerExit:
  END;

{----------------------------------------------------------------------}

PROCEDURE UpdateBullets;
  LABEL BulletLoop, BulletMove, BulletShieldHit, BulletAlienHit,
        BulletBombHit, BulletHitDone, BulletExit;
  BEGIN
    GameLoop := 0;  AlienHit := 0;
  BulletLoop:
  { loop through all active bullets, update position and check for hits }
    Inc( GameLoop );
    If GameLoop > MaxGun THEN Goto BulletExit;
    GameX := BulletX[ GameLoop ];
    GameY := BulletY[ GameLoop ];
  { if bullet is not live, return to top of loop }
    If GameY < 1 THEN Goto BulletLoop;
  { check if something has hit bullet }
    DoGetChar;
    If GetChar = ShieldChar THEN Goto BulletShieldHit;
    If GetChar = AlienChar1 THEN Goto BulletAlienHit;
    If GetChar = AlienChar2 THEN Goto BulletAlienHit;
    If GetChar = BombChar THEN Goto BulletBombHit;
    If GetChar = HitChar THEN Goto BulletBombHit;
    If GetChar = SuperBombChar THEN Goto BulletMove;
  { blank previous bullet position }
    PutChar := BlankChar;
    DoPutChar;
  { move bullet up }
   BulletMove:
    Dec( BulletY[ GameLoop ] );
    GameY := BulletY[ GameLoop ];
  { check if it has moved off screen }
    If GameY < 1 THEN Goto BulletLoop;
    DoGetChar;
  { check if bullet has hit something }
    If GetChar = ShieldChar THEN Goto BulletShieldHit;
    If GetChar = AlienChar1 THEN Goto BulletAlienHit;
    If GetChar = AlienChar2 THEN Goto BulletAlienHit;
    If GetChar = BombChar THEN Goto BulletBombHit;
    If GetChar = HitChar THEN Goto BulletBombHit;
    If GetChar = SuperBombChar THEN Goto BulletLoop;
  { replace bullet }
    PutChar := BulletChar;
    DoPutChar;
    Goto BulletLoop;
  BulletShieldHit:
  { hit with shield, change bullet to space }
    SmashSound;
    PutChar := BlankChar;
    Goto BulletHitDone;
  BulletBombHit:
  { hit with bomb, change bullet to hit }
    PutChar := HitChar;
    Goto BulletHitDone;
  BulletAlienHit:
  { hit with alien, change bullet to hit }
    KillSound;
    PutChar := HitChar;
    Inc( Score );
    AlienHit := 1;
  BulletHitDone:
  { finish handling hit, clear bullet }
    DoPutChar;
    ShowScore;
    BulletY[ GameLoop ] := 0;
    Goto BulletLoop;
  BulletExit:
  END;

{----------------------------------------------------------------------}

PROCEDURE CheckBombColumn;
  LABEL BombColumnLoop, CheckBombExit;
  BEGIN
  { find lowest living alien in column, starting with highest row }
    BombRow := NAlienRows + 1;
    ScoreLoop := MaxAlien + BombLoop;
  BombColumnLoop:
    Dec( BombRow );
    If BombRow = 0 THEN Goto CheckBombExit;
    ScoreLoop := ScoreLoop - NAlienColumns;
  { exit at first live alien }
    If AlienAlive[ ScoreLoop ] = 0 THEN Goto BombColumnLoop;
  CheckBombExit:
  END;

{----------------------------------------------------------------------}

PROCEDURE UpdateBombs;
  LABEL BombUpdateLoop, BombIsDead, BombFindLoop, BombFindLeft,
        BombFindNearest, BombFindDone, BombCalcY, BombCopyCheck,
        BombCreate, BombCreateNotSuper, BombPut, BombNotSuper,
        BombShieldHit, BombBulletHit, BombPlayerHit, BombHitDone, BombExit;
  BEGIN
    GameLoop := 0;
  BombUpdateLoop:
  { loop through all active bombs, update position and check for hits }
    Inc( GameLoop );
    If GameLoop > MaxBomb THEN Goto BombExit;
    GameX := BombX[ GameLoop ];
    GameY := BombY[ GameLoop ];
  { if bomb is live, update it }
    If GameY <= PlayerY THEN Goto BombPut;
  { if bomb is past player, countdown to create a new one }
    If GameY > BombMaxY THEN Goto BombIsDead;
    Inc( BombY[ GameLoop ] );
    Goto BombUpdateLoop;
  { if bomb is not live, create another }
  BombIsDead:
    AlienNewX := 255;
    GameTemp := 0;
    BombLoop := 0;
  { find an alien column which has some living aliens to do bombing }
  BombFindLoop:
    Inc( BombLoop );
    If BombLoop > NAlienColumns THEN Goto BombFindDone;
    CheckBombColumn;
    If BombRow = 0 THEN Goto BombFindLoop;
  { find position of alien column }
    GameX := BombLoop - 1;
    GameX := GameX * AlienXGap;
    GameX := GameX + AlienX - GameMaxX;
  { determine absolute distance of column from player }
    If GameX < PlayerX THEN Goto BombFindLeft;
    GameX := GameX - PlayerX;  Goto BombFindNearest;
  BombFindLeft:
    GameX := PlayerX - GameX;
  BombFindNearest:
  { check if alien column is closest to player }
    If GameX > AlienNewX THEN Goto BombFindLoop;
    GameTemp := BombLoop;
    AlienNewX := GameX;
    Goto BombFindLoop;
  BombFindDone:
    If GameTemp = 0 THEN Goto BombUpdateLoop;
  { calculate position of closest alien column }
    BombLoop := GameTemp;
    CheckBombColumn;
    GameX := GameTemp - 1;
    GameX := GameX * AlienXGap;
    GameX := GameX + AlienX - GameMaxX;
  { if drop coordinate is 1 left of player, use player coordinate }
    If GameX + 1 <> PlayerX THEN Goto BombCalcY;
    GameX := PlayerX;
  BombCalcY:
    GameY := BombRow - 1;
    GameY := GameY * AlienYGap;
    GameY := GameY + AlienY;
    Inc( GameY );
    BombLoop := 0;
  BombCopyCheck:
  { check if new bomb is in same position as any other bomb }
    Inc( BombLoop );
    If BombLoop > MaxBomb THEN Goto BombCreate;
    If BombY[ BombLoop ] > BombMaxY THEN Goto BombCopyCheck;
    If BombX[ BombLoop ] <> GameX THEN Goto BombCopyCheck;
    If BombY[ BombLoop ] <> GameY THEN Goto BombCopyCheck;
    If GameX = GameMaxX THEN Goto BombCreate;
  { if new bomb is a duplicate, move it left }
    Inc( GameX );
    Goto BombCopyCheck;
  BombCreate:
  { create the new bomb }
    BombSound;
    BombX[ GameLoop ] := GameX;
    BombY[ GameLoop ] := GameY;
    BombStrength[ GameLoop ] := 0;
    PutChar := BombChar;
    If SuperRound = 0 THEN Goto BombCreateNotSuper;
    Dec( SuperBombTimer );
    If SuperBombTimer >= SuperRound THEN Goto BombCreateNotSuper;
    BombStrength[ GameLoop ] := 1;
    PutChar := SuperBombChar;
    SuperBombTimer := SuperBombFreq;
  BombCreateNotSuper:
    DoPutChar;
    Goto BombUpdateLoop;
  BombPut:
  { check if something has hit bomb }
    DoGetChar;
    If GetChar = ShieldChar THEN Goto BombShieldHit;
    If GetChar = PlayerChar THEN Goto BombPlayerHit;
    If GetChar = HitChar THEN Goto BombBulletHit;
    If GetChar = BulletChar THEN Goto BombHitDone;
  { blank previous bomb position }
    PutChar := BlankChar;  DoPutChar;
  { move bomb down }
    Inc( BombY[ GameLoop ] );
    GameY := BombY[ GameLoop ];
    If GameY > PlayerY THEN Goto BombUpdateLoop;
    DoGetChar;
  { check if bomb has hit something }
    If GetChar = ShieldChar THEN Goto BombShieldHit;
    If GetChar = PlayerChar THEN Goto BombPlayerHit;
    If GetChar = HitChar THEN Goto BombBulletHit;
    If GetChar = BulletChar THEN Goto BombHitDone;
  { replace bomb }
    PutChar := BombChar;
    If BombStrength[ GameLoop ] = 0 THEN Goto BombNotSuper;
    PutChar := SuperBombChar;
  BombNotSuper:
    DoPutChar;
    Goto BombUpdateLoop;
  BombBulletHit:
  { hit with bullet }
    SmashSound;
    Goto BombHitDone;
  BombShieldHit:
  { hit with shield }
    SmashSound;
    If BombStrength[ GameLoop ] = 0 THEN Goto BombHitDone;
    PutChar := BlankChar;
    DoPutChar;
    Goto BombUpdateLoop;
  BombPlayerHit:
  { hit with player }
    PutChar := HitChar;
    DoPutChar;
    DieSound;
    PlayerHit := 1;
  { if player hit, leave explosion }
    Goto BombUpdateLoop;
  BombHitDone:
  { do nothing if super bomb }
    If BombStrength[ GameLoop ] = 1 THEN Goto BombUpdateLoop;
  { finish handling hit, clear bomb }
    PutChar := HitChar;
    DoPutChar;
  { if hit bullet, cleanup on next cycle }
    If GetChar = BulletChar THEN Goto BombUpdateLoop;
    PutChar := BlankChar;
    DoPutChar;
    BombY[ GameLoop ] := PlayerY + 1;
    Goto BombUpdateLoop;
  BombExit:
  END;

{----------------------------------------------------------------------}

PROCEDURE PutAlien;
  { used by alien update }
  LABEL CalcNotBlankX, CalcAlienY, CalcNotHit, CalcNotBlankY, PutAlien,
        NotDead1, NotDead2, FinishAlien;
  BEGIN
  { determine x position given alien number }
    GameX := GameLoop - 1;
    GameX := GameX mod NAlienColumns;
    GameX := GameX * AlienXGap;
    If PutChar <> BlankChar THEN Goto CalcNotBlankX;
  { if blank was passed, use previous position }
    GameX := GameX + AlienX - GameMaxX;
    Goto CalcAlienY;
  CalcNotBlankX:
  { if not blank, use new position }
    GameX := GameX + AlienNewX - GameMaxX;
  CalcAlienY:
  { determine y position given alien number }
    GameY := GameLoop - 1;
    GameY := GameY div NAlienColumns;
    GameY := GameY * AlienYGap;
    If PutChar <> BlankChar THEN Goto CalcNotBlankY;
  { if blank was passed, use previous position }
    GameY := GameY + AlienY;
    If GameY < AlienMaxY THEN Goto CalcNotHit;
  { flag player hit if alien has reached shield }
    PlayerHit := 1;
  CalcNotHit:
    Goto PutAlien;
  CalcNotBlankY:
  { if not blank, use new position }
    GameY := GameY + AlienNewY;
  PutAlien:
  { check if left part of an alien was hit }
    DoGetChar;
    If GetChar <> HitChar THEN Goto NotDead1;
  { if hit, flag death }
    AlienAlive[ GameLoop ] := 0;
  NotDead1:
  { check if right part of an alien was hit }
    Inc( GameX );
    DoGetChar;
    Dec( GameX );
    If GetChar <> HitChar THEN Goto NotDead2;
  { if hit, flag death }
    AlienAlive[ GameLoop ] := 0;
  NotDead2:
  { if alien has died, clear it }
    If AlienAlive[ GameLoop ] <> 0 THEN Goto FinishAlien;
    PutChar := BlankChar;
  FinishAlien:
  { draw alien, two characters wide }
    DoPutChar;
    Inc( GameX );
    DoPutChar;
  END;

{----------------------------------------------------------------------}

PROCEDURE FindOuterColumns;
  LABEL FindOuterLoop, FindOuterRight, FindOuterDone;
  BEGIN
  { determine farthest left and right alien army columns }
  { so that living aliens can move edge to edge on the screen }
  { set edges to out of range values to start }
    AlienMinColumn := NAlienColumns + 1;
    AlienMaxColumn := 0;
    BombLoop := 0;
  FindOuterLoop:
  { loop through all alien army columns, check for non-vacant ones }
    Inc( BombLoop );
    If BombLoop > NAlienColumns THEN Goto FindOuterDone;
    CheckBombColumn;
    If BombRow = 0 THEN Goto FindOuterLoop;
  { if alien army column is not vacant, check if limits are effected }
    If BombLoop >= AlienMinColumn THEN Goto FindOuterRight;
    AlienMinColumn := BombLoop;
  FindOuterRight:
    If BombLoop <= AlienMaxColumn THEN Goto FindOuterLoop;
    AlienMaxColumn := BombLoop;
    Goto FindOuterLoop;
  FindOuterDone:
  { after checking all columns, recalculate limits }
  { AlienMaxX is GameMaxX + ( GameMaxX - ( AlienMaxColumn * AlienXGap ) ) }
    AlienMaxX := AlienMaxColumn * AlienXGap;
    AlienMaxX := GameMaxX - AlienMaxX;
    AlienMaxX := AlienMaxX + GameMaxX;
  { AlienMinX is GameMaxX - ( ( AlienMinColumn - 1 ) * AlienXGap ) }
    AlienMinX := AlienMinColumn - 1;
    AlienMinX := AlienMinX * AlienXGap;
    AlienMinX := GameMaxX - AlienMinX + 1;
  END;

{----------------------------------------------------------------------}

PROCEDURE UpdateAliens;
  LABEL AlienSkipDec, AlienSkipInc, AlienNotRight, AlienNotLeft,
        AlienUpdate, AlienLoop, ReplaceAlien,
        AlienDoPut, UpdateDone, AlienExit;
  BEGIN
    AlienNewX := AlienX;
    AlienNewY := AlienY;
  { if hit check, skip update countdown }
    If AlienHit > 0 THEN Goto AlienUpdate;
  { check if ready to update }
    Dec( AlienUpdateTimer );
    If AlienUpdateTimer > 0 THEN Goto AlienExit;
    AlienUpdateTimer := AlienUpdateFreq;
  { calculate alien army next position }
    GameLoop := AlienDirection;
    AlienMoveSound;
    If AlienDirection <> 0 THEN Goto AlienSkipDec;
    Dec( AlienNewX );
  AlienSkipDec:
    If AlienDirection <> 1 THEN Goto AlienSkipInc;
    Inc( AlienNewX );
  AlienSkipInc:
  { change direction if edge reached }
    If AlienNewX >= AlienMinX THEN Goto AlienNotRight;
    AlienDirection := 1;
    AlienNewX := AlienMinX;
  AlienNotRight:
    If AlienNewX <= AlienMaxX THEN Goto AlienNotLeft;
    AlienDirection := 0;
    AlienNewX := AlienMaxX;
  AlienNotLeft:
  { if direction has changed, move down }
    If GameLoop = AlienDirection THEN Goto AlienUpdate;
    Inc( AlienNewY );
  AlienUpdate:
    GameLoop := 0;
  AlienLoop:
  { update aliens, skipping dead ones }
    Inc( GameLoop );
    If GameLoop > MaxAlien THEN Goto UpdateDone;
    If AlienAlive[ GameLoop ] = 0 THEN Goto AlienLoop;
  { blank previous alien position }
    PutChar := BlankChar;
    PutAlien;
    If AlienAlive[ GameLoop ] > 0 THEN Goto ReplaceAlien;
  { if hit, increase alien casualty count }
    Dec( AliensAlive );
  { check if outer alien army columns have changed }
    FindOuterColumns;
    Goto AlienLoop;
  ReplaceAlien:
  { replace alien }
    PutChar := AlienChar1;
    If AlienX mod 2 <> 1 THEN Goto AlienDoPut;
    PutChar := AlienChar2;
  AlienDoPut:
    PutAlien;
    Goto AlienLoop;
  UpdateDone:
    If AlienHit = 1 THEN Goto AlienExit;
  { update alien position }
    AlienX := AlienNewX;
    AlienY := AlienNewY;
  AlienExit:
  END;

{----------------------------------------------------------------------}

PROCEDURE PutShields;
  LABEL ShieldPieceLoop, ShieldXLoop, ShieldDraw;
  BEGIN
  { draw the shields }
    GameY := ShieldY;
    GameX := 0;
    PlayerDirection := 1;
  ShieldPieceLoop:
  { alternate blank and shield segments }
    GameLoop := ShieldSize;
    Inc( PlayerDirection );
    PlayerDirection := PlayerDirection mod 2;
  ShieldXLoop:
    PutChar := ShieldChar;
    If PlayerDirection > 0 THEN Goto ShieldDraw;
    PutChar := BlankChar;
  ShieldDraw:
    Inc( GameX );
    DoPutChar;
    Dec( GameLoop );
    If GameLoop = 0 THEN Goto ShieldPieceLoop;
  { until edge of screen reached}
    If GameX < GameMaxX THEN Goto ShieldXLoop;
  END;

{----------------------------------------------------------------------}

PROCEDURE Reset;
  BEGIN
    ClearWeapons;
  { initialize round variables and display }
    PlayerHit := 0;
    AlienHit := 0;
    AlienUpdateTimer := 1;
    SuperBombTimer := SuperBombFreq;
    FindOuterColumns;
    ShowScore;
    ShowHiScore;
    PlayerDirection := 2;
    SuperBlastCount := MaxSuperBlast;
    MovePlayer;
  END;

{----------------------------------------------------------------------}

PROCEDURE Init;
  LABEL InitSameLevel;
  BEGIN
  { initialize game variables and display }
    AlienX := GameMaxX + 1;
    AlienY := Round + 1;
    If AlienY <= AlienMaxYStart THEN Goto InitSameLevel;
    AlienY := 1;
    Inc( SuperRound );
    Round := 0;
  InitSameLevel:
    AlienDirection := 1;
    FillChar( AlienAlive, MaxAlien, 1 );
    AliensAlive := MaxAlien;
    GameClearScreen;  PutShields;
    PlayerX := GameMaxX div 2;
    Reset;
  END;

{----------------------------------------------------------------------}

LABEL GameMainLoop, GameSubLoop, GameNotNewRound;

BEGIN

  SaveMode := GetVMode;
  If ( ParamCount > 0 ) and
     ( ( ParamStr( 1 ) = '/s' ) or ( ParamStr( 1 ) = '/S' ) ) Then
    Silent := True;
  LivesRemaining := MaxLife;
  SuperRound := 0;
  HiScore := 0;
  Score := 0;
  Round := 0;

  GameMainLoop:
  { setup game }
    Init;
  GameSubLoop:
    UpdateAliens;
    UpdateBombs;
    UpdateBullets;
    UpdatePlayer;
    GameCycleDelay;
    If AliensAlive > 0 THEN Goto GameNotNewRound;
  { all aliens killed, start next round }
    Inc( Round );
    CheckHiScore;
    NewRoundSound;
    Goto GameMainLoop;
  GameNotNewRound:
    If PlayerHit = 0 THEN Goto GameSubLoop;
  { one player killed }
    Dec( LivesRemaining );
    Reset;
    If LivesRemaining > 0 THEN Goto GameSubLoop;
    CheckHiScore;
    GameRestoreScreen;

END.





