{$A-,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 1024,0,0}

Uses Dos;

Const
  NewVec : Byte = $80;
  OldVec = $1C;  { timer tick }
  UpdCnt : Byte = 0;
  Msg_InUse : String[ 18 ] =
    'Already loaded.'#13#10'$';
  Msg_Loaded : String[ 16 ] =
    'CLOCK loaded.'#13#10'$';
  Msg_Help : String[ 84 ] =
    'CLOCK ver 3.1 rev 91/06/29 - (C) Andrew Bedno - Chicago.'#13#10 +
    'Resident clock display.'#13#10'$';

Var
  VAddr : Word;
  H, M : Byte;
  Regs : Registers;

Procedure TsrEntry( Flags, CS, IP,
                    AX, BX, CX, DX, SI, DI, DS, ES, BP : Word );
  Interrupt;
  Begin
    InLine( $FA );  { CLI }
    If ( AX = $FFFF ) and
       ( CX = Ord( 'C' ) ) and ( DX = Ord( 'L' ) ) Then
      DX := $ABCD    { residency check }
    Else
      Begin    { regular call }
        If UpdCnt = 18 Then
          Begin  { needs update }
            UpdCnt := 0;
            { read current CRT mode }
            If Mem[ $40 : $49 ] = $07 Then
              VAddr := $B000       { mono }
            Else
              If Mem[ $40 : $49 ] < 4 Then
                VAddr := $B800     { color }
              Else
                VAddr := 0;        { graphics }
            If VAddr > 0 Then
              Begin  { non graphics }
                H := MemW[ $40 : $6E ];   { hours }
                M := MemW[ $40 : $6C ] div 1092;  { minutes }
                If H > 12 Then H := H - 12;
                If H < 10 Then
                  Mem[ VAddr : 150 ] := 32   { blank leading zero }
                Else
                  Mem[ VAddr : 150 ] := H div 10 + 48;       { Hh }
                Mem[ VAddr : 151 ] := $07;
                Mem[ VAddr : 152 ] := H mod 10 + 48;         { Hl }
                Mem[ VAddr : 153 ] := $07;
                If Mem[ VAddr : 154 ] = 32 Then    { blinking ... }
                  Mem[ VAddr : 154 ] := 58                { colon }
                Else                                     { or ... }
                  Mem[ VAddr : 154 ] := 32;               { space }
                Mem[ VAddr : 155 ] := $07;
                Mem[ VAddr : 156 ] := M div 10 + 48;         { Mh }
                Mem[ VAddr : 157 ] := $07;
                Mem[ VAddr : 158 ] := M mod 10 + 48;         { Ml }
                Mem[ VAddr : 159 ] := $07;
              End;  { non graphics }
          End;  { needs update }
        Inc( UpdCnt );
        { call original code }
        Regs.AX := AX;  Regs.BX := BX;  Regs.CX := CX;  Regs.DX := DX;
        Regs.BP := BP;  Regs.SI := SI;  Regs.DI := DI;
        Regs.DS := DS;  Regs.ES := ES;  Regs.Flags := Flags;
        Intr( NewVec, Regs );
        AX := Regs.AX;  BX := Regs.BX;  CX := Regs.CX;  DX := Regs.DX;
        BP := Regs.BP;  SI := Regs.SI;  DI := Regs.DI;
        DS := Regs.DS;  ES := Regs.ES;  Flags := Regs.Flags;
      End;  { regular call }
    InLine( $FB );  { STI }
  End;

{------------------------------------------------------------------------}

Begin

  { check if help requested }
  If ( Mem[ PrefixSeg : $80 ] > 1 ) and
     ( Mem[ PrefixSeg : $82 ] = Ord( '?' ) ) Then
    Begin
      { show already loaded message }
      Regs.AH := $09;  { display string }
      Regs.DS := Seg( Msg_Help[ 1 ] );
      Regs.DX := Ofs( Msg_Help[ 1 ] );
      Intr( $21, Regs );  { display message }
      Halt( 1 );
    End;

  { check if already resident }
  Regs.AX := $FFFF;
  Regs.CX := Ord( 'C' );  Regs.DX := Ord( 'L' );
  Intr( OldVec, Regs );
  If Regs.DX = $ABCD Then
    Begin
      { show already loaded message }
      Regs.AH := $09;  { display string }
      Regs.DS := Seg( Msg_InUse[ 1 ] );
      Regs.DX := Ofs( Msg_InUse[ 1 ] );
      Intr( $21, Regs );  { display message }
      Halt( 1 );
    End;

  { find a free vector }
  NewVec := $7F;
  Repeat
    Inc( NewVec );
    Regs.AH := $35;  { get vector }
    Regs.AL := NewVec;
    Intr( $21, Regs );
  Until ( ( Regs.ES = 0 ) and ( Regs.BX = 0 ) ) or ( NewVec = $F0 );
  If NewVec = $F0 Then Halt( 1 );

  { get current address of vector to be displaced }
  Regs.AH := $35;  { get vector }
  Regs.AL := OldVec;
  Intr( $21, Regs );

  { copy it to new vector }
  Regs.AH := $25;  { set vector }
  Regs.AL := NewVec;
  Regs.DS := Regs.ES;
  Regs.DX := Regs.BX;
  Intr( $21, Regs );

  { replace original with new address }
  {$F+}
  Regs.DS := Seg( TsrEntry );
  Regs.DX := Ofs( TsrEntry );
  {$F-}
  Regs.AH := $25;  { set vector }
  Regs.AL := OldVec;  { user procedure }
  Intr( $21, Regs );

  { show now loaded message }
  Regs.AH := $09;  { display string }
  Regs.DS := Seg( Msg_Loaded[ 1 ] );
  Regs.DX := Ofs( Msg_Loaded[ 1 ] );
  Intr( $21, Regs );  { display message }

  { exit but keep resident }
  Regs.AH := $31;     { keep }
  Regs.AL := 0;   { exit code }
  { size in pargraphs = ( code + data + stack bytes ) div 16 + 3 extra }
  Regs.DX := 247;
  Intr( $21, Regs );

End.


