{$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 32000,0,0}

Program CRC;


Uses Dos;


Const
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;
  ProgramName = 'CRC';
  Version = '3.1';
  Revision = '91/06/29';
  BufSize = 8192;
  _GetSum = 0;
  _MatchSum = 1;
  Mode : Byte = _GetSum;
  AttrSave : Byte = $80;
  MaxNames = 200;
  NameCount : Integer = 0;
  UseXor : Boolean = False;
  CRCSize : Byte = 2;  { number of bytes in output }
  Noun : String[20] = 'crc'; 

Type
  String80 = String[ 80 ];
  String20 = String[ 20 ];
  String8 = String[ 8 ];
  String2 = String[ 2 ];
  CRCString = String[ 4 ];

Var
  FileName : String80;
  TargetSum : CRCString;
  FileSize : LongInt;
  FileLoop : Integer;
  WorkSum : Word;
  ExitSave : Pointer;
  Buf : Array[ 1..BufSize ] of Byte;  { global for less stack }
  Names : Array [ 1..MaxNames ] of String[12];
  Mask : String[79];
  Path : string[67];
  Name : string[8];
  Ext : string[4];
  Dir : SearchRec;


{------------------------------------------------------------------------}

Function I_S( InNum : LongInt ) : String20;
  Var
    NumStr : String20;
  Begin
    Str( InNum, NumStr );
    I_S := NumStr;
  End;

{------------------------------------------------------------------------}

Procedure Stop( Msg:String );
  Begin
    If Msg <> '' Then
      Writeln( Msg + '.' );
    Halt( 1 );
  End;

{------------------------------------------------------------------------}

Procedure Error( Num : Integer;  Msg : String80 );
  Begin
    Stop( 'Error # ' + I_S( Num ) + ' ' + Msg );
  End;

{------------------------------------------------------------------------}

Procedure ChkErr( Msg:String );
  Var
    ChkResult : Integer;
  Begin
    ChkResult := IOResult;
    If ChkResult <> 0 Then Error( ChkResult, Msg );
  End;

{------------------------------------------------------------------------}

Procedure FindFile( Name : String80 );
  Var
    ChkFile : Text;  Result : Integer;
  Begin
    Assign( ChkFile, Name );
    Reset( ChkFile ); Result := IOResult;
    If Result = 2 Then
      Stop( '"' + Name + '" file not found' );
    If Result <> 0 Then
      Error( Result, ' checking file "' + Name + '"' );
    Close( ChkFile ); Result := IOResult;
  End;

{------------------------------------------------------------------------}

Procedure Help;
  Begin
    For CpyRtLp := 1 to 26 do
      If CpyRtLp = 1 Then
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
      Else
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                 Ord( CpyRt[ CpyRtLp - 1 ] ) );
    Writeln( ProgramName, ' ver ', Version, ' rev ', Revision,
             ' - ', CpyRt, '.' );
    Writeln( 'Report or verify file crc or checksum.' );
    Writeln( '  Display file attributes, size, and crc or checksum.' );
    Writeln( ProgramName + ' filename' );
    Writeln( '  Show attributes, size, and crc of file.' );
    Writeln( ProgramName + ' [/X] filename [target]' );
    Writeln( '  Target   Return errorlevel if not current value.' );
    Writeln( '  /X       Use simple XOR checksum rather than CRC.' );
    Writeln( 'Use wildcard name to process in directory order.' );
    Writeln( 'Files are processed even if hidden or system.' );
    Halt( 1 ); 
  End;

{------------------------------------------------------------------------}

Procedure ReadOptions;
  Var
    UpLoop : Byte;
    NumStr : String[20];
    ParamNum : Byte;
  Begin
    ParamNum := 1;
    If ( ParamStr( ParamNum ) = '/X' ) or
       ( ParamStr( ParamNum ) = '/x' ) Then
      Begin
        UseXor := True;
        CrcSize := 1;
        Noun := 'chksum';
        Inc( ParamNum );
      End;
    If ParamNum > ParamCount Then
      Stop( 'No files specified.' );
    FileName := ParamStr( ParamNum );
    If ParamCount > ParamNum Then
      Begin
        Inc( ParamNum );
        NumStr := ParamStr( ParamNum );
        For UpLoop := 1 to Length( NumStr ) do
          NumStr[ UpLoop ] := UpCase( NumStr[ UpLoop ] );
        If Length( NumStr ) <> CRCSize * 2 Then
          Stop( 'Target ' + Noun + ' must be ' +
                I_S( CRCSize * 2 ) + ' hex digits' );
        TargetSum := NumStr;
        For UpLoop := 1 to Length( TargetSum ) do
          If not ( TargetSum[ UpLoop ] in [ '0'..'9', 'A'..'F' ] ) Then
            Stop( 'Only 0..9, A..F allowed in target ' + Noun + '.' );
        Mode := _MatchSum;
        If ( Pos( '?', FileName ) > 0 ) or
           ( Pos( '*', FileName ) > 0 ) Then
          Begin
            Writeln( 'Testing ' + Noun + ' not allowed ' +
                     'with wildcard file name.' );
            Halt( 1 );
          End;
      End;
  End;

{------------------------------------------------------------------------}

Function ByteHex( B : Byte ) : String2;
  Const
    HexChar : Array [0..15] of Char = '0123456789ABCDEF';
  Begin
    ByteHex := HexChar[ ( B and $F0 ) shr 4 ] + HexChar[ B and $0F ];
  End;

{------------------------------------------------------------------------}

Function CRCHex( CRCIn : Word ) : CRCString;
  Var
    OutStr : CRCString;
    CharLoop : Byte;
    CRCInArr : Array [ 1..2 ] of Byte absolute CRCIn;
  Begin
    OutStr := '';
    For CharLoop := 1 to CRCSize do
      OutStr := OutStr + ByteHex( CRCInArr[ CharLoop ] );
    CRCHex := Copy( OutStr, 1, CRCSize * 2 );
  End;

{------------------------------------------------------------------------}

Procedure ForceAccess( Name : String80 );
  Var
    Regs : Registers;
  Begin
    Name[ Length( Name ) + 1 ] := #0;
    Regs.AH := $43;  { attr function }
    Regs.AL := $00;  { get }
    Regs.DS := Seg( Name[ 1 ] );
    Regs.DX := Ofs( Name[ 1 ] );
    MSDOS( Regs );
    If Regs.Flags and FCarry <> 0 Then
      Error( Regs.AX, 'checking "' + Name + '" file attributes' );
    AttrSave := Regs.CX and $7F;
    If AttrSave and $27 <> 0 Then
      Begin
        Write( '  attr: ' );
        If AttrSave and $01 <> 0 Then Write( 'R' );
        If AttrSave and $02 <> 0 Then Write( 'H' );
        If AttrSave and $04 <> 0 Then Write( 'S' );
        If AttrSave and $20 <> 0 Then Write( 'A' );
      End;
    If AttrSave and $01 <> 0 Then  { make read/write in order to scan }
      Begin
        Regs.AH := $43;  { attr function }
        Regs.AL := $01;  { set }
        Regs.DS := Seg( Name[ 1 ] );
        Regs.DX := Ofs( Name[ 1 ] );
        Regs.CX := AttrSave and $FE;  { allow write }
        MSDOS( Regs );
        If Regs.Flags and FCarry <> 0 Then
          Begin
            AttrSave := $80;
            If Regs.AX = 150 Then
              Error( Regs.AX, 'setting "' + Name + '" file attributes' );
          End;
      End
    Else
      AttrSave := $80;
    End;

{------------------------------------------------------------------------}

Procedure RestoreAttr( Name : String80 );
  Var
    Regs : Registers;
  Begin
    If AttrSave and $80 <> $00 Then Exit;
    Name[ Length( Name ) + 1 ] := #0;
    Regs.AH := $43;  { attr function }
    Regs.AL := $01;  { set }
    Regs.DS := Seg( Name[ 1 ] );
    Regs.DX := Ofs( Name[ 1 ] );
    Regs.CX := AttrSave;  { restore }
    MSDOS( Regs );
    If Regs.Flags and FCarry <> 0 Then
      Error( Regs.AX, 'restoring "' + Name + '" file attributes' );
    AttrSave := $80;
  End;

{------------------------------------------------------------------------}

Procedure UpdCRC( B : Byte;  Var CRC : Word );
  Var
    CRCLoop : Integer;
    Flag : Boolean;
  Begin
    For CRCLoop := 1 to 8 Do
      Begin
        Flag := CRC and $8000 <> 0;
        CRC := CRC shl 1;
        If ( B and $80 ) <> 0 Then CRC := CRC or 1;
        B := ( B and $7F ) shl 1;
        If Flag Then CRC := CRC xor $1021;
      End;
  End;

{------------------------------------------------------------------------}

Procedure CheckFile;
{ Requires prior setting of global variable FileName. }
{ returns global variable Sum. }
  Var
    Result : Integer;
    BufLoop, BufLen : Word;
    CRCTextHandle : Text;
    CRCFile : File;
  Begin
    WorkSum := 0;  BufLen := 0;  FileSize := 0;
    ForceAccess( FileName );
    Assign( CRCFile, FileName );
    Reset( CRCFile, 1 );  ChkErr( 'opening "' + FileName + '" file' );
    Repeat
      BlockRead( CRCFile, Buf, BufSize, BufLen );
      FileSize := FileSize + BufLen;
      ChkErr( 'reading "' + FileName + '" file' );
      For BufLoop := 1 to BufLen do
        Begin
          If UseXor Then
            WorkSum := WorkSum xor Buf[ BufLoop ]
          Else
            UpdCRC( Buf[ BufLoop ], WorkSum );
        End;
    Until EOF( CRCFile );
    If UseXor Then WorkSum := WorkSum and $00FF;
    Close( CRCFile );  ChkErr( 'closing "' + FileName + '" file' );
    RestoreAttr( FileName );
  End;

{------------------------------------------------------------------------}

{$F+}
Procedure TidyExit;
  Begin
    ExitProc := ExitSave;
    RestoreAttr( FileName );
    Halt( ExitCode );
  End;
{$F-}

{------------------------------------------------------------------------}

Procedure GetDirArr;
  Begin
    FSplit( FileName, Path, Name, Ext );
    Mask := Name + Ext;
    FindFirst( Path + Mask, AnyFile, Dir );
    While DosError = 0 do
      Begin
        If NameCount >= MaxNames Then
          Begin
            Writeln( 'Too many files matching "' + FileName + '"' );
            Halt( 1 );
          End;
        If Dir.Attr and Directory = 0 Then
          Begin
            Inc( NameCount );
            Names[ NameCount ] := Dir.Name;
          End;
        FindNext( Dir );
      End;
    If NameCount < 1 Then
      Begin
        Writeln( 'No files matching "' + FileName + '"' );
        Halt( 1 );
      End;
  End;

{------------------------------------------------------------------------}

Begin

  If ( ParamCount < 1 ) or
     ( ( ParamCount > 0 ) and ( ParamStr( 1 ) = '?' ) ) Then Help;
  ExitSave := ExitProc;  ExitProc := @TidyExit;
  ReadOptions;
  GetDirArr;

  For FileLoop := 1 to NameCount do
    Begin
      FileName := Path + Names[ FileLoop ];
      Write( 'File: ' + FileName );
      CheckFile;
      Write( '  size: ' + I_S( FileSize ) );
      Case Mode of
        _GetSum :
           Writeln( '  ' + Noun + ': ' + CRCHex( WorkSum ) );
        _MatchSum :
           If CRCHex( WorkSum ) = TargetSum Then
             Writeln( '  ' + Noun + ' matches ' + TargetSum )
           Else
             Begin
               Writeln( '  ' + Noun + ' DOES NOT MATCH ' + TargetSum );
               Halt( 1 );
             End;
      End;
    End;

End.



