(*********************************************************************
**********************************************************************
****                                                              ****
****                       DATAMAN.PAS                            ****
****                                                              ****
****                Data Manager for files and disk.              ****
****                                                              ****
**********************************************************************
**********************************************************************)


Unit DataMan;

{$I LIBRARY.CFG }

INTERFACE

Uses Dos, Crt, Printer,
     Global, Sounds, Math, Strings, Video, Time, Key, Disk, Init,
     InputMan;

Const
  MaxNDCB   = 30;
  LastDCB   : Byte = 0;
  _UnSorted   : Word = $0001;   { search from begining }
  _Sorted     : Word = $0002;   { binary search }
  _Create     : Word = $0004;   { create if non-existent }
  _Typed      : Word = $0008;   { fixed record format }
  _Text       : Word = $0010;   { text line format }
  _Disk       : Word = $0020;   { data is on disk }
  _Memory     : Word = $0040;   { data is in memory }
  _Fatal      : Word = $0080;   { file is critical }
  _Serious    : Word = $0100;   { file is significant }
  _Silent     : Word = $0200;   { file is unimportant }
  _Sequential : Word = $0400;   { increment position after read/write }
  _Random     : Word = $0800;   { do not alter position after read/write }
  _Terse      : Word = $1000;   { no message during sort/purge }
  _Reverse    : Word = $2000;   { sort high to low }
  _Temporary  : Word = $4000;   { erase when closed }
  _Read     =  'R';
  _Write    =  'W';
  _Append   =  'A';
  _Find     =  'F';
  _Sort     =  'S';
  _Purge    =  'P';
  _Delete   =  'D';
  _Release  =  'L';  { for liberate }
  NilPtr    : Pointer = Nil;
  NullLet = '';   {or something}
  DataMgrDebugEnable : Boolean = True;
  CompareChar :
    Array [ 0..2 { _Equal, _Lower, _Greater } ] of Char = ( '=', '<', '>' );


Type
  DCBType =  { data control block }
    Record
      FileName : GenlString;  { physical file name }
      Desc : SmallString;     { brief description of file purpose }
      FileHandle : File;      { O.S. disk file handle used internally }
      ArrayStart : Pointer;   { start of array area if data in memory }
      Buffer : Pointer;       { pointer to one record data buffer }
      RecSize : Integer;      { byte size of a single record }
      Key : Pointer;          { pointer to key field in buffer }
      KeyPos : Integer;       { byte position in record of key field }
      KeyType : Integer;      { type code of key field }
      KeySize : Integer;      { size of key field type }
      Key2 : Pointer;         { secondary key }
      KeyPos2 : Integer;
      KeyType2 : Integer;
      KeySize2 : Integer;
      Key3 : Pointer;         { terciary key }
      KeyPos3 : Integer;
      KeyType3 : Integer;
      KeySize3 : Integer;
      Pos : LongInt;          { current record number in data (0..n) }
      BytePos : LongInt;      { current byte position in the data }
      Size : LongInt;         { number of data records }
      ByteSize : LongInt;     { size of file in bytes }
      MaxSize : LongInt;      { maximum allowed number of data records }
      Mode : Word;            { access mode }
      Status : Integer;       { current error status of data }
      SortMe : Boolean;       { does the file need to be resorted ? }
      PurgeMe : Boolean;      { does the file contain any deleted records ? }
    End;


Var
  DCBList : Array [ 1..MaxNDCB ] of ^DCBType;
  DataMgrExitSave : Pointer;


Function DataCompare( Var Var1, Var2;  VarType : Integer ) : Integer;

Procedure SetKey( Var DCB : DCBType;
                  Var KeyVar;
                  KeyType, KeyNum : Integer );

Procedure DataSetup( Var DCB : DCBType;
                     FileName : GenlString;
                     Desc : SmallString;
                     Var VarArray;
                     RecordSize : LongInt;
                     Var BufVar;  Var KeyVar;
                     KeyType : Integer;
                     Mode : Word;
                     Size : LongInt;  MaxSize : LongInt );

Procedure FileSetup( Var DCB : DCBType;
                     FileName : GenlString;
                     Desc : SmallString;
                     RecordSize : LongInt;
                     Var BufVar;  Var KeyVar;
                     KeyType : Integer;
                     Mode : Word );

Procedure ArraySetup( Var DCB : DCBType;
                      Desc : SmallString;
                      Var VarArray;
                      RecordSize : LongInt;
                      Var BufVar;  Var KeyVar;
                      KeyType : Integer;
                      Mode : Word;
                      Size, MaxSize : LongInt );

Procedure DataRelease( Var DCB : DCBType );

Procedure DataMgr( Var DCB : DCBType;  Func : Char );

Procedure DataCleanup;

Procedure PrepareText( Var DCB : DCBType );


IMPLEMENTATION


{------------------------------------------------------------------------}
{ Function DataCompare                                                   }
{                                                                        }
{ Returns _Equal if contents of Var1 equals the contents of Var2.        }
{ Returns _Lower  if contents of Var1 is less than the contents of Var2  }
{ Returns _Greater if contents of Var1 is greater than contents of Var2  }
{ Pass Nil^ as Var2 to return _Equal if Var1 is null.  (0 or '')         }
{------------------------------------------------------------------------}

Function DataCompare( Var Var1, Var2;  VarType : Integer ) : Integer;

  Var
    CmpByte1  : Byte absolute Var1;
    CmpByte2  : Byte absolute Var2;
    CmpWord1  : Word absolute Var1;
    CmpWord2  : Word absolute Var2;
    CmpInt1   : Integer absolute Var1;
    CmpInt2   : Integer absolute Var2;
    CmpInt3_1 : Int3 absolute Var1;
    CmpInt3_2 : Int3 absolute Var2;
    CmpLint1  : LongInt absolute Var1;
    CmpLint2  : LongInt absolute Var2;
    CmpReal1  : Real absolute Var1;
    CmpReal2  : Real absolute Var2;
    CmpDate1  : String8 absolute Var1;
    CmpDate2  : String8 absolute Var2;
    CmpStr1   : String absolute Var1;
    CmpStr2   : String absolute Var2;
    CmpPtr    : Pointer absolute Var2;
    R1, R2 : Real;

  Begin
    If NumericVar( VarType ) Then
      Begin
        If VarType = _Byte Then
          Begin  R1 := CmpByte1;  R2 := CmpByte2;  End
        Else
          If VarType = _Integer Then
            Begin  R1 := CmpInt1;  R2 := CmpInt2;  End
          Else
            If VarType = _Word Then
              Begin  R1 := CmpWord1;  R2 := CmpWord2;  End
            Else
              If VarType = _LongInt Then
                Begin  R1 := CmpLint1;  R2 := CmpLint2;  End
              Else
                If VarType = _Int3 Then
                  Begin
                    R1 := Int3Real( CmpInt3_1 );
                    R2 := Int3Real( CmpInt3_2 );
                  End
                Else
                  Begin  R1 := CmpReal1;  R2 := CmpReal2;  End;
        If CmpPtr = Nil Then R2 := 0.0;
        If R1 > R2 Then
          DataCompare := _Greater
        Else
          If R1 = R2 Then
            DataCompare := _Equal
          Else
            DataCompare := _Lower;
      End
    Else
      Begin
        If CmpPtr = Nil Then
          Begin
            If CmpDate1 = '' Then
              DataCompare := _Equal
            Else
              DataCompare := _Greater;
            Exit;
          End;
        If VarType = _Date Then
          Begin
            If GreaterDate( CmpDate1, CmpDate2 ) Then
              DataCompare := _Greater
            Else
              If EqualDate( CmpDate1, CmpDate2 ) Then
                DataCompare := _Equal
              Else
                DataCompare := _Lower
          End
        Else
          If ( VarType >= _String ) and ( VarType <= MaxStringType ) Then
            Begin
              If CmpStr1 > CmpStr2 Then
                DataCompare := _Greater
              Else
                If CmpStr1 = CmpStr2 Then
                  DataCompare := _Equal
                Else
                  DataCompare := _Lower;
            End;
      End;
  End;


{----------------------------------------------------------------------}
{ Procedure DataMgrError                                               }
{                                                                      }
{ Central error logger for data manager functions.                     }
{----------------------------------------------------------------------}

Procedure DataMgrError( Var DCB : DCBType;  Func : Char );
  Begin
    If DCB.Status = OK Then Exit;
    If UpCase( Func ) = 'S' Then Func := 'u';  { convert sorting to using }
    If DCB.Mode and _Fatal = _Fatal Then
      Func := LowCase( Func );
    If ( DCB.Mode and _Serious = _Serious ) or
       ( DCB.Mode and _Fatal = _Fatal ) Then
      ChkErr( DCB.Status, Func, DCB.FileName, DCB.Desc + ' data' );
  End;


{----------------------------------------------------------------------}
{ Procedure SetKey                                                     }
{                                                                      }
{ Set the key field descriptions in a DCB.                             }
{ Key 1 is primary default key.                                        }
{ Set keys 2 or 3 for multi-key sorts.                                 }
{----------------------------------------------------------------------}

Procedure SetKey( Var DCB : DCBType;
                  Var KeyVar;
                  KeyType, KeyNum : Integer );
  Begin
  { KeyPos is position of first byte of key field in buffer: 0..RecSize-1 }
    Case KeyNum of
      2: Begin
           DCB.Key2 := @KeyVar;
           DCB.KeyPos2 := PtrInt( DCB.Key2 ) - PtrInt( DCB.Buffer );
           DCB.KeyType2 := KeyType;
           DCB.KeySize2 := VarSize( KeyType );
         End;
      3: Begin
           DCB.Key3 := @KeyVar;
           DCB.KeyPos3 := PtrInt( DCB.Key3 ) - PtrInt( DCB.Buffer );
           DCB.KeyType3 := KeyType;
           DCB.KeySize3 := VarSize( KeyType );
         End;
      Else
        Begin
           DCB.Key := @KeyVar;
           DCB.KeyPos := PtrInt( DCB.Key ) - PtrInt( DCB.Buffer );
           DCB.KeyType := KeyType;
           DCB.KeySize := VarSize( KeyType );
         End;
    End;
  End;

{----------------------------------------------------------------------}
{ Procedure DataSetup                                                  }
{                                                                      }
{ This is the first step in preparing the data manager to access data. }
{ It initializes a handle and sets a pointer to a buffer.              }
{ If data in memory, ArrayStart, Size, and MaxSize must be passed in.  }
{ This is meant to be called only by FileSetup and ArraySetup.         }
{ For text data in RAM, Size and MaxSize are in bytes, not lines.      }
{----------------------------------------------------------------------}

Procedure DataSetup( Var DCB : DCBType;
                     FileName : GenlString;
                     Desc : SmallString;
                     Var VarArray;
                     RecordSize : LongInt;
                     Var BufVar;
                     Var KeyVar;
                     KeyType : Integer;
                     Mode : Word;
                     Size : LongInt;  MaxSize : LongInt );

  Var
    CRLn : String;  CRLoop : LongInt;

  Begin
    FillChar( DCB, SizeOf( DCB ), 0 );
    DCB.FileName := FileName;
    DCB.Desc := Desc;
    If DCB.Desc = '' Then
      DCB.Desc := DCB.FileName;
    DCB.ArrayStart := @VarArray;
    DCB.MaxSize := MaxSize;
    DCB.Mode := Mode;
    If ( DCB.Mode and _Text <> _Text ) and
       ( DCB.Mode and _Typed <> _Typed ) Then
      DCB.Mode := DCB.Mode or _Typed;
    If ( DCB.Mode and _Sequential <> _Sequential ) and
       ( DCB.Mode and _Random <> _Random ) Then
      DCB.Mode := DCB.Mode or _Sequential;
    If ( DCB.Mode and _Sorted <> _Sorted ) and
       ( DCB.Mode and _UnSorted <> _UnSorted ) Then
      DCB.Mode := DCB.Mode or _UnSorted;
    If ( DCB.Mode and _Fatal <> _Fatal ) and
       ( DCB.Mode and _Serious <> _Serious ) and
       ( DCB.Mode and _Silent <> _Silent ) Then
      DCB.Mode := DCB.Mode or _Serious;
    DCB.RecSize := RecordSize;
    DCB.Buffer := @BufVar;
    SetKey( DCB, KeyVar, KeyType, 1 );
    If LastDCB = MaxNDCB Then ChkErr( -95, 'u', DCB.FileName, DCB.Desc );
    Inc( LastDCB );  DCBList[ LastDCB ] := @DCB;
    If DCB.Mode and _Disk = _Disk Then
      Begin  { disk access }
        { close just in case }
        {$I-} Close( DCB.FileHandle ); {$I+}  CloseResult := IOResult;
        Assign( DCB.FileHandle, DCB.FileName );
        If not FileExists( DCB.FileName ) then
          Begin
            If DCB.Mode and _Create = _Create Then
              Begin  { create if necessarry }
                If DCB.Mode and _Text = _Text Then
                  {$I-} Rewrite( DCB.FileHandle, 1 ) {$I+}
                Else
                  {$I-} Rewrite( DCB.FileHandle, DCB.RecSize );{$I+}
                DCB.Status := IOResult;
              End
            Else
              DCB.Status := _FileNotFound;
          End
        Else  { file exists }
          If ( SizeOfFile(DCB.FileName) mod DCB.RecSize > 0 ) and
             ( DCB.Mode and _Text <> _Text ) then
            DCB.Status := -98 { not a multiple of the record size error # }
          Else  { file exist, size OK }
            Begin
              {$I-}
              If DCB.Mode and _Text = _Text Then
                Reset( DCB.FileHandle, 1 )
              Else
                Reset( DCB.FileHandle, DCB.RecSize );
              {$I+}
              DCB.Status := IOResult;
            End;
        If DCB.Status = OK Then
          Begin
            DCB.ByteSize := SizeOfFile( DCB.FileName );
            DCB.Size := FileSize( DCB.FileHandle );
          End;
        DataMgrError( DCB, 'O' );
      End  { disk access }
    Else
      Begin  { memory access }
        DCB.Status := OK;
        If DCB.FileName = '' Then DCB.FileName := DCB.Desc;
        If DCB.Mode and _Text  = _Text Then
          DCB.ByteSize := Size
        Else
          Begin
            DCB.Size := Size;
            DCB.ByteSize := Size * DCB.RecSize;
          End;
      End;   { memory access }
    PrepareText( DCB );
  End;

{----------------------------------------------------------------------}
{ Procedure FileSetup                                                  }
{                                                                      }
{ Initialization procedure for using DataMgr with disk files.          }
{----------------------------------------------------------------------}

Procedure FileSetup( Var DCB : DCBType;
                     FileName : GenlString;
                     Desc : SmallString;
                     RecordSize : LongInt;
                     Var BufVar;
                     Var KeyVar;
                     KeyType : Integer;
                     Mode : Word );

  Var
    Junk : Pointer;

  Begin
    Mode := Mode or _Disk and ( not _Memory );
    Junk := NIL;
    DataSetup( DCB, FileName, Desc, Junk^,
               RecordSize, BufVar, KeyVar, KeyType, Mode, 0, 0 );
  End;


{----------------------------------------------------------------------}
{ Procedure ArraySetup                                                 }
{                                                                      }
{ Initialization procedure for using DataMgr with memory arrays.       }
{ For text data in RAM, Size and MaxSize are in bytes, not lines.      }
{----------------------------------------------------------------------}

Procedure ArraySetup( Var DCB : DCBType;
                      Desc : SmallString;
                      Var VarArray;
                      RecordSize : LongInt;
                      Var BufVar;  Var KeyVar;
                      KeyType : Integer;
                      Mode : Word;
                      Size, MaxSize : LongInt );

  Begin
    Mode := Mode or _Memory and ( not _Disk );
    DataSetup( DCB, '', Desc, VarArray,
               RecordSize, BufVar, KeyVar, KeyType, Mode, Size, MaxSize );
  End;

{----------------------------------------------------------------------}
{ Procedure DataRelease                                                }
{                                                                      }
{----------------------------------------------------------------------}

Procedure DataRelease(Var DCB : DCBType);

Var
  DCBLoop, DCBNum : Byte;
  Found : Boolean;

Begin
  DCBNum := 0;  Found := False;
  While ( not Found ) and ( DCBNum <= LastDCB ) do
    Begin
      Inc( DCBNum );
      Found := DCBList[ DCBNum ] = @DCB;
    End;
  If not Found Then
    ChkErr( -97, 'o', 'data manager', '' );
  If DCB.Mode and _Disk = _Disk Then
    Begin
      {$I-} Close( DCB.FileHandle ); {$I+} DCB.Status := IOResult;
      If DCB.Mode and _Temporary = _Temporary Then
        JunkResult := EraseFile( DCB.FileName );
    End;
  DCBList[ DCBNum ] := NIL;  Dec( LastDCB );
  For DCBLoop := DCBNum to LastDCB do
    DCBList[ DCBLoop ]:= DCBList[ DCBLoop + 1 ];
End;

{----------------------------------------------------------------------}
{ Procedure TextSeek                                                   }
{                                                                      }
{ Seek to the begining of line specified by DCB.Pos.                   }
{----------------------------------------------------------------------}

Procedure TextSeek( Var DCB : DCBType );

  Var
    BufStr : String;
    BufCarry : String[ 2 ];
    SeekLoop : Byte;
    ReadLen : Integer;
    SeekPos : LongInt;
    MAddr : LongInt;  AddrOfs, AddrSeg : LongInt;

  Begin
    If DCB.Status <> OK Then
      Exit;
    DCB.BytePos := 0;
    {$I-} Reset(DCB.FileHandle, 1); {$I+} DCB.Status := IOResult;
    If (DCB.Status <> OK) or (DCB.Pos = 0) Then
      Exit;
    SeekPos := 0;
    BufCarry := '';
    If DCB.Mode and _Disk = _Disk Then
      Begin
        Repeat
          { ReadOfs is used to check for and read an LF after CR }
          {$I-} BlockRead( DCB.FileHandle,
                           BufStr, SizeOf(BufStr) - 2, ReadLen); {$I+}
          BufStr[0] := Chr(ReadLen);
          DCB.Status := IOResult;
          If DCB.Status <> OK Then
            Exit;
          If BufCarry <> '' Then
            Begin
              BufStr := BufCarry + BufStr;
              BufCarry := '';
            End;
           { If buffer ends with CR, read one more char to check for LF }
          If (BufStr[Length( BufStr)] = CR) and
             (not EOF(DCB.FileHandle)) Then
            Begin
              {$I-}
              BlockRead( DCB.FileHandle, BufCarry[ 1 ], 1, ReadLen );{$I+}
              BufCarry[ 0 ] := Chr( ReadLen );
              DCB.Status := IOResult;
              If DCB.Status <> OK Then
                Exit;
              { if LF, append to buffer, else carry to next buffer }
              If BufCarry = LF Then
                Begin
                  BufStr := BufStr + BufCarry;
                  BufCarry := '';
                End;
            End;
          SeekLoop := 1;
          While ( SeekLoop < Length( BufStr ) ) and
                ( BufStr[SeekLoop] <> EF ) and
                ( SeekPos < DCB.Pos ) do
            Begin
              If BufStr[ SeekLoop ] = CR Then
                Begin
                  Inc( SeekPos );
                  { consume LF if any }
                  If SeekLoop < Length( BufStr ) Then
                    If BufStr[ SeekLoop + 1 ] = LF Then
                      Begin
                        Inc( DCB.BytePos );
                        Inc( SeekLoop );
                      End;
                End;
              Inc( DCB.BytePos );
              Inc( SeekLoop );
            End;
          If (SeekLoop = Length(BufStr)) and
             (BufStr[SeekLoop] <> EF) and (SeekPos < DCB.Pos) then
            BufCarry := BufStr[SeekLoop];
          If SeekPos = DCB.Pos then
            Inc(DCB.BytePos);
        Until ( SeekPos = DCB.Pos ) or EOF( DCB.FileHandle );
        If (SeekPos < DCB.Pos) or (BufStr[SeekLoop] = EF) Then
          DCB.Status := _SeekError
        Else
          Begin  { found }
            { seek back to line start, from last block read end }
            {$I-} Seek( DCB.FileHandle, DCB.BytePos ); {$I+}
            DCB.Status := IOResult;
          End;
        DCB.Pos := SeekPos;
      End  { disk text seek }
    Else
      Begin  { RAM text seek }
        MAddr := PtrInt( DCB.ArrayStart );
        AddrSeg := MAddr div $10;  AddrOfs := MAddr mod $10;
        ChkPtr( Mem[ AddrSeg : AddrOfs ], DCB.ByteSize,
                'DataMgr text seek' );
        While ( DCB.BytePos < DCB.ByteSize ) and
              ( SeekPos < DCB.Pos ) do
          Begin
            MAddr := PtrInt( DCB.ArrayStart ) + DCB.BytePos;
            AddrSeg := MAddr div $10;  AddrOfs := MAddr mod $10;
            If Chr( Mem[ AddrSeg : AddrOfs ] ) = CR Then
              Begin
                Inc( SeekPos );
                { consume LF if any }
                If DCB.BytePos < DCB.ByteSize Then
                  If Chr( Mem[ AddrSeg : AddrOfs + 1 ] ) = LF Then
                    Inc( DCB.BytePos );
              End;
            Inc( DCB.BytePos );
          End;
        If SeekPos < DCB.Pos Then
          DCB.Status := _SeekError;
      End;  { RAM text seek }
  End;


{----------------------------------------------------------------------}
{ Procedure PrepareText                                                }
{                                                                      }
{ Checks text data, adds a CR at the end of data if no CR or CR/LF     }
{ already there, and counts number of lines.                           }
{----------------------------------------------------------------------}

Procedure PrepareText( Var DCB : DCBType );

  Var
    LastChars : String[ 2 ];
    CharLen : Byte;
    MAddr, AddrOfs, AddrSeg : LongInt;

  Begin
    If ( DCB.Status <> OK ) or
       ( DCB.Mode and _Text <> _Text ) or
       ( DCB.ByteSize = 0 ) Then
      Exit;
    If DCB.Size < 2 Then
      CharLen := 1
    Else
      CharLen := 2;
    If DCB.Mode and _Disk = _Disk Then
      Begin  { disk }
        { go to end of file }
        {$I-} Seek( DCB.FileHandle, DCB.ByteSize - CharLen ); {$I+}
        DCB.Status := IOResult;
        If DCB.Status <> OK Then
          Exit;
        { read last two characters, or only character if size = 1 }
        {$I-} BlockRead( DCB.FileHandle, LastChars[ 1 ], CharLen ); {$I+}
        DCB.Status := IOResult;
        If DCB.Status <> OK Then
          Exit;
        LastChars[ 0 ] := Chr( CharLen );
        { exit if last character is an EOLN }
        If not ((LastChars = CR) or (LastChars = NL)) Then
          Begin  {append a CR to the end}
            {$I-} Seek( DCB.FileHandle, DCB.ByteSize );  {$I+}
            DCB.Status := IOResult;
            If DCB.Status <> OK Then
              Exit;
            LastChars := CR;
            {$I-} BlockWrite( DCB.FileHandle, LastChars, 1 );  {$I+}
            DCB.Status := IOResult;
            If DCB.Status <> OK Then Exit;
            Inc( DCB.ByteSize );
          End;
      End  { disk }
    Else
      Begin  { RAM }
        MAddr := PtrInt( DCB.ArrayStart ) + DCB.ByteSize - CharLen;
        AddrSeg := MAddr div $10;  AddrOfs := MAddr mod $10;
        { read last two characters, or only character if size = 1 }
        Move( Mem[ AddrSeg : AddrOfs ], LastChars[ 1 ], CharLen );
        LastChars[ 0 ] := Chr( CharLen );
        If LastChars <> NL Then  { append a CR/LF to the end }
          Begin
            If LastChars = CR Then
              LastChars := LF
            Else
              LastChars := NL;
            CharLen := Length( LastChars );
            MAddr := PtrInt( DCB.ArrayStart ) + DCB.ByteSize;
            AddrSeg := MAddr div $10;  AddrOfs := MAddr mod $10;
            Move( LastChars[ 1 ], Mem[ AddrSeg : AddrOfs ], CharLen );
           DCB.ByteSize := DCB.ByteSize + CharLen;
          End;
      End;  { RAM }
    { find number of text lines }
    DCB.Pos := MaxLongint;
    TextSeek( DCB );
    DCB.Size := DCB.Pos - 1;
{    DCB.ByteSize := DCB.BytePos;  don't know yet if that's correct}
    If DCB.Status = _SeekError Then
      DCB.Status := OK;
  End;


{----------------------------------------------------------------------}
{ Procedure TextDataRecord                                             }
{                                                                      }
{ Reads or writes a text line from line DCB.Pos in memory or file      }
{ and, if read, puts it in the record data buffer of the Handle.  Func }
{ indicates whether this will be a read or write.  The IOResult of     }
{ the operation will be put into Handle.Status.                        }
{----------------------------------------------------------------------}

Procedure TextDataRecord( Var DCB : DCBType;  Func : Char );

  Var
    DataLn : String;
    DataCh : Char;
    WriteLoop : Byte;
    Dummy, EndFound : Boolean;

{-------------}
  Function TextChar( Var Ch : Char ) : Boolean;

  {Reads 1 character from a text file and returns it in Ch. If the character}
  {read is an EOF, TextChar returns True. Otherwise TextChar returns False. }

    Var
      MAddr, AddrOfs, AddrSeg : LongInt;

    Begin
      TextChar := False;
      If DCB.Mode and _Disk =_Disk Then
        Begin
          If Func = _Read Then
            Begin
              {$I-} BlockRead( DCB.FileHandle, Ch, 1 ); {$I+}
              If Ch = EF then
                Begin
                  TextChar := True;
                  Dec(DCB.BytePos); {to compensate for later inc}
                End;
            End
          Else
            {$I-} BlockWrite( DCB.FileHandle, Ch, 1 ); {$I+}
          DCB.Status := IOResult;
        End
      Else
        Begin
          MAddr := PtrInt( DCB.ArrayStart ) + DCB.ByteSize;
          AddrSeg := MAddr div $10;
          AddrOfs := MAddr mod $10;
          If Func = _Read Then
            Begin
              Ch := Chr( Mem[ AddrSeg : AddrOfs ] );
              If Ch = EF then
                Begin
                  TextChar := True;
                  Dec(DCB.BytePos);  {to compensate for later inc}
                End;
            End
          Else
            Mem[ AddrSeg : AddrOfs ] := Ord( Ch );
        End;
      Inc( DCB.BytePos );
      If ( Func = _Write ) and ( DCB.BytePos = DCB.ByteSize ) Then
        Inc( DCB.ByteSize );
    End;
{------------}

  Begin
    {have to do a TextSeek here - otherwise you can't know DCB.BytePos
     which is used to detect EOF}
    TextSeek(DCB);
    If DCB.Status = OK then
      Begin
        {DCB.BytePos is correct}
        If Func = _Read Then
          Begin  { read }
            FillChar( DCB.Buffer^ , DCB.RecSize, 0 );
            FillChar( DataLn, SizeOf( DataLn ), 0 );
            EndFound := False;
            { This loop appends characters to DataLn until EOF, or }
            { error, or buffer full, or CR read. }
            While (DCB.BytePos < DCB.ByteSize) and (not EndFound) and
                  (DCB.Status = OK) and (Length(DataLn) < DCB.RecSize) and
                  ((DataLn = '') or (DataLn[Length(DataLn)] <> CR)) do
              Begin
                DataLn[0] := Succ(DataLn[0]);
                EndFound := TextChar(DataLn[Ord(DataLn[0])]);
                If EndFound Then
                  Dec(DataLn[0]);
              End;
            { check for and read LF after CR }
            If ( DCB.BytePos < DCB.ByteSize ) and (not EndFound) and
               ( DCB.Status = OK ) and ( Length( DataLn ) < DCB.RecSize ) and
               ( DataLn[ Length( DataLn ) ] = CR ) Then
              Begin
                DataLn := DataLn + LF;
                EndFound := TextChar( DataCh );
                If DataCh <> LF Then
                  Begin  { backstep from non LF }
                    Dec( DCB.BytePos );
                    If DCB.Mode and _Disk = _Disk Then
                      Begin
                        {$I-} Seek( DCB.FileHandle, DCB.BytePos ); {$I+}
                        DCB.Status := IOResult;
                      End;
                  End;
              End;  { check for LF }
            Move( DataLn, DCB.Buffer^, DCB.RecSize );
          End  { read }
        Else
          Begin  { write }
            Move( DCB.Buffer^, DataLn, DCB.RecSize );
            { strip LF's }
            For WriteLoop := Length( DataLn ) DownTo 1 do
              If DataLn[ WriteLoop ] = LF Then
                Delete( DataLn, WriteLoop, 1 );
            { force line to end with CR }
            If DataLn[ Length( DataLn ) ] <> CR Then
              DataLn := DataLn + CR;
            { write line loop }
            WriteLoop := 1;
            While ( DCB.Status = OK ) and
                  ( WriteLoop <= Length( DataLn ) ) and
                  ( ( DCB.Mode and _Disk = _Disk ) or
                    ( DCB.ByteSize < DCB.MaxSize ) ) do
              Begin
                DataCh := DataLn[ WriteLoop ];
                Dummy := TextChar(DataCh);
                If DataCh = CR Then
                  Begin
                    Inc( DCB.Size );  { line appended }
                    {add LF after CR}
                    DataCh := LF;
                    If DCB.Status = OK Then
                      Dummy := TextChar( DataCh );
                  End;
                Inc( WriteLoop );
              End;
          End;  { write }
      End;
  End;


{----------------------------------------------------------------------}
{ Procedure DataRecord                                                 }
{                                                                      }
{ Reads or writes a record from position DCB.Pos in memory or file     }
{ and, if read, puts it in the record data buffer of the Handle.  Func }
{ indicates whether this will be a read or write.  The IOResult of     }
{ the operation will be put into Handle.Status.                        }
{----------------------------------------------------------------------}

Procedure DataRecord( Var DCB : DCBType;  Func : Char );

  Var
    MAddr : LongInt;
    AddrOfs, AddrSeg : LongInt;
    BlockResult, LineNum : Integer;

  Begin
    CheckStack;
    If DCB.Status <> OK Then Exit;
    If DCB.Mode and _Text = _Text Then
      TextDataRecord( DCB, Func )
    Else
      Begin  { not text data }
        If ( ( Func = _Write ) and ( DCB.Pos > DCB.Size ) ) or
           ( ( Func = _Read ) and ( DCB.Pos > DCB.Size - 1 ) ) Then
          Begin
            DCB.Status := _SeekError; { seek past end error # }
            Exit;
          End;
        If DCB.Mode and _Disk = _Disk Then
          Begin  { disk record access }
            If FilePos( DCB.FileHandle ) <> DCB.Pos Then
              Begin  { new position required }
                {$I-} Seek( DCB.FileHandle, DCB.Pos ); {$I+}
                DCB.Status := IOResult;
                If DCB.Status <> OK Then Exit;
              End;  { new position required }
            If Func = _Read Then  { read }
              Begin
                {$I-} BlockRead( DCB.FileHandle, DCB.Buffer^, 1 ); {$I+}
                DCB.Status := IOResult;
                If Sys.DebugEnabled and DataMgrDebugEnable Then
                  Begin
                    WaitKeyChar := NoKey;
                    Debug( 'Read ' + DCB.Filename +
                           ', record ' + I_S( DCB.Pos ) +
                           ' of ' + I_S( DCB.Size ) +
                           OKMsg( DCB.Status ) + NL + NL +
                           'Press <ESC> to silence DataMgr debug.' );
                    DataMgrDebugEnable := WaitKeyChar <> EscKey;
                  End;
              End
            Else  { write }
              Begin
                {$I-} BlockWrite( DCB.FileHandle, DCB.Buffer^, 1 ); {$I+}
                DCB.Status := IOResult;
                If ( DCB.Status = OK ) and ( DCB.Pos >= DCB.Size ) Then
                  Inc( DCB.Size );
                If Sys.DebugEnabled and DataMgrDebugEnable Then
                  Begin
                    WaitKeyChar := NoKey;
                    Debug( 'Write ' + DCB.Filename +
                           ', record ' + I_S( DCB.Pos ) +
                           ' of ' + I_S( DCB.Size ) +
                           OKMsg( DCB.Status ) + NL + NL +
                           'Press <ESC> to silence DataMgr debug.' );
                    DataMgrDebugEnable := WaitKeyChar <> EscKey;
                  End;
              End;
            If DCB.Status <> OK Then Exit;
          End  { disk record access }
        Else
          Begin  { memory access }
            MAddr := PtrInt( DCB.ArrayStart ) + DCB.Pos * DCB.RecSize;
            AddrSeg := MAddr div $10;  AddrOfs := MAddr mod $10;
            ChkPtr( Mem[ AddrSeg : AddrOfs ], DCB.RecSize, 'DataMgr rec' );
            If Func = 'R' Then  { read }
              Move( Mem[ AddrSeg : AddrOfs ], DCB.Buffer^, DCB.RecSize )
            Else
              Begin  { write }
                Move( DCB.Buffer^, Mem[ AddrSeg : AddrOfs ], DCB.RecSize );
                If DCB.Pos >= DCB.Size Then Inc( DCB.Size );
              End;
          End;  { memory access }
        DCB.ByteSize := DCB.Size * DCB.RecSize;
        DCB.BytePos := DCB.Pos * DCB.RecSize;
      End;  { not text data }
  End;

{----------------------------------------------------------------------}
{ Procedure DataSort                                                   }
{                                                                      }
{ Called by DataMgr to sorts a data file or array.                     }
{ Modified Shell-Metzner sort uses just two records of heap.           }
{ Leaves DCB.Pos on new position of original DCB.Pos record.           }
{----------------------------------------------------------------------}

Procedure DataSort( Var DCB : DCBType );

  Label
    SortOut;

  Var
    SavePos, Temp, Index, Jump, Gap : LongInt;
    TotalRecs, RecCount, KeyAddr : LongInt;
    Rec1, Rec2, KeyA1, KeyA2, KeyB1, KeyB2, KeyC1, KeyC2 : Pointer;
    KeyACompare, KeyBCompare, KeyCCompare : Integer;
    CommaStrSave : String2;
    SortDebugEnable : Boolean;

  Begin
    If MaxAvail < DCB.RecSize * 2 Then
      Begin  DCB.Status := -94;  Exit;  End;
    If DCB.Size < 2 Then Goto SortOut;
    { get memory for swap records }
    GetMem( Rec1, DCB.RecSize );  GetMem( Rec2, DCB.RecSize );
    CheckStack;
    SavePos := DCB.Pos;
    { calculate pointer to key fields }
    KeyAddr := PtrInt( @Rec1^ ) + DCB.KeyPos;
    KeyA1 := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    KeyAddr := PtrInt( @Rec2^ ) + DCB.KeyPos;
    KeyA2 := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    KeyAddr := PtrInt( @Rec1^ ) + DCB.KeyPos2;
    KeyB1 := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    KeyAddr := PtrInt( @Rec2^ ) + DCB.KeyPos2;
    KeyB2 := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    KeyAddr := PtrInt( @Rec1^ ) + DCB.KeyPos3;
    KeyC1 := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    KeyAddr := PtrInt( @Rec2^ ) + DCB.KeyPos3;
    KeyC2 := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    { prepare to sort }
    Gap := DCB.Size Div 2;
    Temp := Gap;  RecCount := 0;  TotalRecs := 0;
    While Temp > 0 do  { figure required iterations }
      Begin  Inc( TotalRecs );  Temp := Temp div 2;  End;
    If DCB.Mode and _Terse <> _Terse Then
      Begin
        Message( 'Sorting ' + DCB.Desc + ' data . . .' + NL, Msg_PlainPut );
        ShowPercent( 0.0 );
      End;
    SortDebugEnable := True;
    While Gap > 0 Do
      Begin  { gap }
        For Index := ( Gap + 1 ) to DCB.Size do
          Begin  { index }
            Jump := Index - Gap;
            While Jump > 0 Do
              Begin  { jump }
                { read record 1 }
                DCB.Pos := Jump - 1;  { record one position }
                DataRecord( DCB, _Read );
                If DCB.Status <> OK Then Goto SortOut;
                Move( DCB.Buffer^, Rec1^, DCB.RecSize );
                { read record 2 }
                DCB.Pos := Jump + Gap - 1;  { record two position }
                DataRecord( DCB, _Read );
                If DCB.Status <> OK Then Goto SortOut;
                Move( DCB.Buffer^, Rec2^, DCB.RecSize );
                If DCB.KeyType = NoType Then  { prepare key 1 comparison }
                  KeyACompare := _Equal
                Else
                  Begin
                    KeyACompare := DataCompare( KeyA1^, KeyA2^, DCB.KeyType );
                    If DCB.Mode and _Reverse = _Reverse Then
                      If KeyACompare =_Greater Then
                        KeyACompare := _Lower
                      Else
                        KeyACompare := _Greater;
                  End;
                If DCB.KeyType2 = NoType Then  { prepare key 2 comparison }
                  KeyBCompare := _Equal
                Else
                  Begin
                    KeyBCompare := DataCompare( KeyB1^, KeyB2^, DCB.KeyType );
                    If DCB.Mode and _Reverse = _Reverse Then
                      If KeyBCompare =_Greater Then
                        KeyBCompare := _Lower
                      Else
                        KeyBCompare := _Greater;
                  End;
                If DCB.KeyType3 = NoType Then  { prepare key 3 comparison }
                  KeyCCompare := _Equal
                Else
                  Begin
                    KeyCCompare := DataCompare( KeyC1^, KeyC2^, DCB.KeyType );
                    If DCB.Mode and _Reverse = _Reverse Then
                      If KeyCCompare =_Greater Then
                        KeyCCompare := _Lower
                      Else
                        KeyCCompare := _Greater;
                  End;
                { prepare debug window }
                If Sys.DebugEnabled and SortDebugEnable Then
                  Begin  { debug }
                    WaitKeyChar := NoKey;
                    { this line supresses commas in Show(number) }
                    CommaStrSave := CommaStr;  CommaStr := '';
                    PushScreen;
                    BorderStyle := MessageBorderStyle;
                    BorderColor := MessageBorderColor;
                    FrameType := Frame_InsetWindow;
                    Frame( 10, 7, 70, 16 );
                    TextColor( MessageTextColor );
                    DispLn( 'Comparing records ' + I_S( Jump - 1 ) +
                            ' and ' + I_S( Jump + Gap - 1 ) +
                            ' in ' + DCB.FileName );
                    If DCB.KeyType <> NoType Then
                      DispLn( 'Key A: DataCompare(' +
                              Show( KeyA1^, DCB.KeyType ) + ',' +
                              Show( KeyA2^, DCB.KeyType ) + ',' +
                              VarName( DCB.KeyType ) + ')="' +
                              CompareChar[ KeyACompare ] + '"');
                    If DCB.KeyType2 <> NoType Then
                      DispLn( 'Key B: DataCompare(' +
                              Show( KeyB1^, DCB.KeyType2 ) + ',' +
                              Show( KeyB2^, DCB.KeyType2 ) + ',' +
                              VarName( DCB.KeyType2 ) + ')="' +
                              CompareChar[ KeyBCompare ] + '"' );
                    If DCB.KeyType3 <> NoType Then
                      DispLn( 'Key C: DataCompare(' +
                              Show( KeyC1^, DCB.KeyType3 ) + ',' +
                              Show( KeyC2^, DCB.KeyType3 ) + ',' +
                              VarName( DCB.KeyType3 ) + ')="' +
                              CompareChar[ KeyCCompare ] + '"' );
                    DispLn( NL + ' Press <ESC> to silence sort debug.' );
                    WindowTitle( 0, FrameY2, Bright + Blink,
                                 ' PRESS ANY KEY ' );
                    WaitKeyPress;
                    PopScreen;
                    CommaStr := CommaStrSave;
                    SortDebugEnable := WaitKeyChar <> EscKey;
                  End;  { debug }
                { compare records }
                If ( KeyACompare = _Greater ) or
                   ( ( KeyACompare = _Equal ) and
                     ( KeyBCompare = _Greater ) ) or
                   ( ( KeyACompare = _Equal ) and
                     ( KeyBCompare = _Equal ) and
                     ( KeyCCompare = _Greater ) ) Then
                  Begin
                    { write record 1 at previous record 2 position }
                    DCB.Pos := Jump + Gap - 1;
                    Move( Rec1^, DCB.Buffer^, DCB.RecSize );
                    DataRecord( DCB, _Write );
                    If DCB.Status <> OK Then Goto SortOut;
                    { write record 2 at previous record 1 position }
                    DCB.Pos := Jump - 1;
                    Move( Rec2^, DCB.Buffer^, DCB.RecSize );
                    DataRecord( DCB, _Write );
                    If DCB.Status <> OK Then Goto SortOut;
                    { check to adjust saved starting position }
                    If SavePos = Jump - 1 Then
                      SavePos := Jump + Gap - 1
                    Else
                      If SavePos = Jump + Gap - 1 Then
                        SavePos := Jump - 1;
                    Jump := Jump - Gap;
                  End
                Else
                  Jump := 0;
              End;  { junmp }
          End;  { index }
        Gap := Gap div 2;
        If DCB.Status <> 0 Then Goto SortOut;
        Inc( RecCount );
        If DCB.Mode and _Terse <> _Terse Then
          ShowPercent( RecCount / TotalRecs );
      End;  { gap }
  SortOut:
    FreeMem( Rec1, DCB.RecSize );  FreeMem( Rec2, DCB.RecSize );
    If DCB.Status = OK Then
      Begin
        DCB.SortMe := False;
        DCB.Pos := SavePos;
        DataRecord( DCB, _Read );
      End;
    AuditLog( 'File ' + DCB.FileName + ' sorted' + OKMsg( DCB.Status ) );
    PopScreen;
  End;


{--------------------------------------------------------------------------}
{ Function DataPurge                                                       }
{                                                                          }
{ This function purges all records having a null key field.                }
{ A temporary file is created and erased if the purge was successful.      }
{ Leaves DCB.Pos on new position of original DCB.Pos record.               }
{--------------------------------------------------------------------------}

Procedure DataPurge( Var DCB : DCBType );

Const
  PurgeFileSpec : String40 = '\PURGE.$$$';

Var
  PurgedOK : Boolean;
  KeyCompare : Integer;
  PosLoop, PurgedRecs, SavePos, SavePosOut : LongInt;
  PurgeDCB : DCBType;
  PurgeDebugEnable : Boolean;
  CommaStrSave : String2;
  DCBSave : DCBType;
  PurgeDrive : String2;

Begin
  If ( DCB.Status <> OK ) or ( DCB.Size = 0 ) then Exit;
  { purge from memory returns bad function result }
  If DCB.Mode and _Disk <> _Disk Then
    Begin  DCB.Status := -96;  Exit;  End;
  If ( Length( DCB.FileName ) < 3 ) or
     ( DCB.FileName[ 2 ] <> ':' ) Then
    PurgeDrive := ''
  Else
    PurgeDrive := Copy( DCB.FileName, 1, 2 );
  PurgeFileSpec := PurgeDrive + PurgeFileSpec;
  JunkResult := EraseFile( PurgeFileSpec );
  DCBSave := DCB;
  CheckStack;
  SavePos := DCB.Pos;  SavePosOut := -1;
  PosLoop := 0;  PurgedRecs := 0;
  FileSetup( PurgeDCB, PurgeFileSpec, 'purge temporary',
             DCB.RecSize, DCB.Buffer^, DCB.Key^, DCB.KeyType,
             _Disk + _Create + _Sequential + _Temporary );
  PurgedOK := False;
  If PurgeDCB.Status = OK Then
    Begin
      If DCB.Mode and _Terse <> _Terse Then
        Begin
          Message( 'Purging empty records in ' +
                   DCB.Desc + ' data . . .' + NL, Msg_PlainPut );
          ShowPercent( 0.0 );
        End;
      PurgeDebugEnable := True;
      While ( DCB.Status = OK ) and ( PurgeDCB.Status = OK ) and
            ( PosLoop < DCB.Size ) do
        Begin
          DCB.Pos := PosLoop;
          DataMgr( DCB, _Read );
          KeyCompare := DataCompare( DCB.Key^, NilPtr, DCB.KeyType );
          { prepare debug window }
          If Sys.DebugEnabled and PurgeDebugEnable Then
            Begin  { debug }
              WaitKeyChar := NoKey;
              { this line supresses commas in Show(number) }
              CommaStrSave := CommaStr;  CommaStr := '';
              PushScreen;
              BorderStyle := MessageBorderStyle;
              BorderColor := MessageBorderColor;
              FrameType := Frame_InsetWindow;
              Frame( 10, 7, 70, 16 );
              TextColor( MessageTextColor );
              DispLn( 'Checking record ' + I_S( PosLoop ) +
                      ' in ' + DCB.FileName );
              DispLn( 'DataCompare(' +
                      Show( DCB.Key^, DCB.KeyType ) +
                      ',NilPtr,' + VarName( DCB.KeyType ) + ')="' +
                      CompareChar[ KeyCompare ] + '"' );
              DispLn( NL + ' Press <ESC> to silence purge debug.' );
              WindowTitle( 0, FrameY2, Bright + Blink, ' PRESS ANY KEY ' );
              WaitKeyPress;
              PopScreen;
              CommaStr := CommaStrSave;
              PurgeDebugEnable := WaitKeyChar <> EscKey;
            End;  { debug }
          If ( KeyCompare <> _Equal ) and
             ( DCB.Status = OK ) Then
            Begin
              PurgeDCB.Pos := PurgeDCB.Size;
              If PosLoop = SavePos Then SavePosOut := PurgeDCB.Pos;
              DataRecord( PurgeDCB, _Write )
            End
          Else
            Inc( PurgedRecs );
          If DCB.Mode and _Terse <> _Terse Then
            ShowPercent( DCB.Pos / DCB.Size );
          Inc( PosLoop );
        End;
    End;
  PurgedOK :=( DCB.Status = OK ) and ( PurgeDCB.Status = OK );
  DataRelease( PurgeDCB );
  If PurgedOK and ( PurgedRecs > 0 ) Then
    Begin
      { Released, so erase/rename don't bother file handles. }
      DataRelease( DCB );
      JunkResult := EraseFile( DCBSave.FileName );
      JunkResult := RenameFile( PurgeFileSpec, DCBSave.FileName );
      FileSetup( DCB, DCBSave.FileName, DCBSave.Desc,
                 DCBSave.RecSize, DCBSave.Buffer^,
                 DCBSave.Key^, DCBSave.KeyType, DCBSave.Mode );
      PurgedOK := DCB.Status = OK;
      If SavePosOut < 0 Then SavePosOut := DCB.Size;
      If PurgedOK Then DCB.Pos := SavePosOut;
    End
  Else
    JunkResult := EraseFile( PurgeFileSpec );
  DCB.PurgeMe := not PurgedOK;
  AuditLog( 'File ' + DCB.FileName + ' purged' + OKMsg( DCB.Status ) );
  PopScreen;
End;


{----------------------------------------------------------------------}
{ Procedure DataSearch                                                 }
{                                                                      }
{ Return the record if found, the position in the "file" where it was  }
{ found. If the record was not found then Handle.Pos is placed one     }
{ record past end of file, ready to append.                            }
{----------------------------------------------------------------------}

Procedure DataSearch( Var DCB : DCBType );

  Var
    Found : Boolean;  Target, SaveRec : Pointer;
    UpperPos, LowerPos : Integer;  KeyAddr : LongInt;

  Begin
    If DCB.Status <> OK Then Exit;
    DCB.Pos := 0;  { set to beginning of file }
    Found := False;
    GetMem( SaveRec, DCB.RecSize );  { preserve original record }
    Move( DCB.Key^, SaveRec^, DCB.RecSize );
    KeyAddr := PtrInt( @SaveRec^ ) + DCB.KeyPos;
    Target := Ptr( KeyAddr div $10, KeyAddr mod $10 );
    UpperPos := DCB.Size - 1;  LowerPos := 0;
    If ( DCB.Mode and _Sorted <> _Sorted ) or DCB.SortMe Then
      While ( not Found ) and ( DCB.Pos < DCB.Size ) do
        Begin  { sequential search }
          DataRecord( DCB, 'R' );
          If DataCompare( Target^, DCB.Key^, DCB.KeyType ) = _Equal Then
            Found := True
          Else
            Inc( DCB.Pos );
        End  { sequential search }
    Else
      While ( not Found ) and ( UpperPos <> LowerPos ) do
        Begin  { binary search }
          DCB.Pos := Trunc( ( UpperPos - LowerPos ) / 2 );
          DataRecord( DCB, 'R' );
          If DataCompare( Target^, DCB.Key^, DCB.KeyType ) = _Equal Then
            Found := True
          Else
            If DataCompare( DCB.Key^, Target^, DCB.KeyType ) = _Greater Then
              UpperPos := DCB.Pos
            Else
              LowerPos := DCB.Pos;
        End;  { end binary search }
    If not Found Then DCB.Pos := DCB.Size;  { key not found in file }
    Move( SaveRec^, DCB.Key^, DCB.RecSize );  { restore original }
    FreeMem( SaveRec, DCB.RecSize );
  End;


{--------------------------------------------------------------------------}
{ Procedure DataMgr                                                        }
{                                                                          }
{ This procedure and the DataSetup procedure should be all that's needed   }
{ for the data manager.  This allows you to read and write to disk or to   }
{ memory using the same "file" handle.  Func determines whether this is a  }
{ read or a write, sort by key field, or delete a record.                  }
{ To append a record to the file, the Handle.Pos has to equal the FileSize }
{ of the file.  If it is larger, the procedure will put up an error window }
{ and will not append.  To insert a record into a sorted file, you have to }
{ first append the record and then sort the entire file.                   }
{--------------------------------------------------------------------------}

Procedure DataMgr( Var DCB : DCBType;  Func : Char );

  Var
    DCBNum : Byte;  Found : Boolean;

  Begin
    If DCB.Status <> OK Then
      Exit;
    Func := UpCase( Func );
    DCBNum := 0;
    Found := False;
    While ( not Found ) and ( DCBNum <= LastDCB ) do
      Begin
        Inc( DCBNum );
        Found := DCBList[ DCBNum ] = @DCB;
      End;
    If not Found Then
      ChkErr( -97, 'o', 'data manager', '' );
    Operation( 'DataMgr' + UpCase( Func ) );
    Case Func of
      _Read :
         Begin
           DataRecord( DCB, _Read );
           If ( DCB.Status = OK ) and
              ( DCB.Mode and _Sequential = _Sequential ) Then
             Inc( DCB.Pos );
         End;
      _Write :
         Begin
           { Note: if overwriting an existing record, calling program is }
           { responsible for forcing re-sort if key field is changed. }
           DataRecord( DCB, _Write );
           If ( DCB.Status = OK ) and
              ( DCB.Mode and _Sequential = _Sequential ) Then
             Inc( DCB.Pos );
         End;
      _Append :
         Begin
           DCB.Pos := DCB.Size;
           DataRecord( DCB, _Write );
           If ( DCB.Status = OK ) and
              ( DCB.Mode and _Sequential = _Sequential ) Then
             Inc( DCB.Pos );
         End;
      _Find :
         DataSearch( DCB );
      _Sort :
         DataSort( DCB );
      _Purge :
         DataPurge( DCB );
      _Delete :
         Begin
           FillChar( DCB.Buffer^, DCB.RecSize, 0 );
           DataRecord( DCB, _Write );
           DCB.PurgeMe := True;
         End;
      _Release :
         DataRelease(DCB);
      Else
         ChkErr( -96, 'u', DCB.FileName, DCB.Desc );
    End;  {case}
    DataMgrError( DCB, Func );
    Operation( '' );
  End;


{--------------------------------------------------------------------------}
{ Procedure DataCleanup                                                    }
{                                                                          }
{ Sort if necessary, close the file and clear out the handle.              }
{--------------------------------------------------------------------------}

Procedure DataCleanup;

  Var
    DCBLoop : Byte;  DCB : ^DCBType;

   Begin
    ExitProc := DataMgrExitSave;
    If TidyCode = 0 Then TidyCode := ExitCode;
    If TidyCode <> 0 Then Exit;
    Operation( 'DataCln' );
    For DCBLoop := 1 to LastDCB do
      Begin
        DCB := @DCBList[ DCBLoop ]^;
        If ( DCB^.Mode and _Temporary = _Temporary ) and
           ( DCB^.Mode and _Disk = _Disk ) Then
          Begin
            {$I-} Close( DCB^.FileHandle ); {$I+} JunkResult := IOResult;
            JunkResult := EraseFile( DCB^.FileName );
          End
        Else
          Begin  { not temporary }
            If ( DCB^.Status = OK ) and DCB^.PurgeMe then
              Begin
                DataPurge( DCB^ );
                DataMgrError( DCB^, 'C' );
              End;
            If ( DCB^.Status = OK ) and DCB^.SortMe Then
              Begin
                DataSort( DCB^ );
                DataMgrError( DCB^, 'C' );
              End;
            If DCB^.Mode and _Disk = _Disk Then
              Begin
                {$I-} Close( DCB^.FileHandle ); {$I+}
                CloseResult := IOResult;
              End;
          End;  { not temporary }
      End;
    Operation( '' );
  End;


{--------------------------------------------------------------------------}

Procedure RealDataMgrView;

  Var
    DCBLoop : Byte;  DCB : ^DCBType;

  Begin
    For DCBLoop := 1 to LastDCB do
      Begin
        DCB := @DCBList[ DCBLoop ]^;
        DispLn( 'DCB #' + I_S( DCBLoop ) + ', Desc="' + DCB^.Desc + '"' +
                ', Name="' + DCB^.FileName + '"' );
        DispLn( '  Size=' + I_S( DCB^.Size ) + ', Pos=' + I_S( DCB^.Pos ) +
                ', Status=' + OKMsg( DCB^.Status ) +
                ', Mode=' + ByteBinary( Hi( DCB^.Mode ) ) + ':' +
                ByteBinary( Lo( DCB^.Mode ) ) );
        Disp( '  KeyType=' + I_S( DCB^.KeyType ) +
              ', KeyType2=' + I_S( DCB^.KeyType2 ) +
              ', KeyType3=' + I_S( DCB^.KeyType3 ) );
        If DCB^.SortMe Then Disp( ', SortMe' );
        If DCB^.PurgeMe Then Disp( ', PurgeMe' );
        DispLn( '' );
      End;
  End;

{--------------------------------------------------------------------------}

Begin

  FillChar( DCBList, SizeOf( DCBList ), 0 );
  DataMgrView := RealDataMgrView;
  DataMgrExitSave := ExitProc;
  ExitProc := @DataCleanup;

End.

