{-------------------------------------------------------------------------}
{ DTSec returns seconds since 00:00:00 80-01-01                           }
{ as a real number, given date and time strings YY-MM-DD and HH:MM:SS.    }
{-------------------------------------------------------------------------}

Function DTSec( DateStr, TimeStr : String8 ) : Real;
  Var
    Yr, Mo, Da, Hr, Mn, Sc, DaTotal, ScTotal : LongInt;
    DTLp, DTResult : Integer;
  Const
    DaysPerMonth : array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    { initialize function to zero, flags conversion error }
    DTSec := -1.0;
    { remove invalid characters }
    For DTLp := Length( DateStr ) DownTo 1 do
      If not ( DateStr[ DTLp ] in [ '0'..'9' ] ) Then
        Delete( DateStr, DTLp, 1 );
    For DTLp := Length( TimeStr ) DownTo 1 do
      If not ( TimeStr[ DTLp ] in [ '0'..'9' ] ) Then
        Delete( TimeStr, DTLp, 1 );
    { exit if remaining length too small }
    If ( Length( DateStr ) <> 6 ) or ( Length( TimeStr ) <> 6 ) Then
      Exit;
    { convert digit pairs to numbers }
    Val( Copy( DateStr, 1, 2 ), Yr, DTResult );
    Val( Copy( DateStr, 3, 2 ), Mo, DTResult );
    Val( Copy( DateStr, 5, 2 ), Da, DTResult );
    Val( Copy( TimeStr, 1, 2 ), Hr, DTResult );
    Val( Copy( TimeStr, 3, 2 ), Mn, DTResult );
    Val( Copy( TimeStr, 5, 2 ), Sc, DTResult );
    { range check all values }
    If ( Yr < 0 ) or ( Yr > 99 ) or
       ( Mo < 1 ) or ( Mo > 12 ) or
       ( Da < 1 ) or ( Da > 31 ) or
       ( ( not ( Da in [ 1 .. DaysPerMonth[ Mo ] ] ) ) and
         ( not ( ( Da = 29 ) and ( Mo = 2 ) and ( Yr mod 4 = 0 ) ) ) ) or
       ( Hr < 0 ) or ( Hr > 23 ) or
       ( Mn < 0 ) or ( Mn > 59 ) or
       ( Sc < 0 ) or ( Sc > 59 ) Then
      Exit;
    { convert date to days since 01-01-80 }
    { assume years before 80 are in 21st century }
    If Yr < 80 Then Yr := Yr + 100;
    Yr := Yr - 80;
    { start with leap year days }
    DaTotal := Yr div 4;
    { compensate for fact that the base year, 1980, is a leap year }
    Inc( DaTotal );
    { compensate if current year is leap year, but leap day is not past }
    If ( Yr mod 4 = 0 ) and
       ( ( Mo < 2 ) or ( ( Mo = 2 ) and ( Da < 29 ) ) ) Then
      Dec( DaTotal );
    { add days for past years }
    DaTotal := DaTotal + ( Yr * 365 );
    DTLp := 1;
    { add days for elapsed months in year }
    While DTLp < Mo do
      Begin
        DaTotal := DaTotal + DaysPerMonth[ DTLp ];
        Inc( DTLp );
      End;
    { add elapsed days in month to finish }
    DaTotal := DaTotal + Da - 1;
    { convert time to seconds since 00:00:00 }
    ScTotal := Hr * 3600 + Mn * 60 + Sc;
    DTSec := DaTotal * 86400.0 + ScTotal * 1.0;
  End;

{-------------------------------------------------------------------------}
{ SecDT returns a string containing date and time "YY-MM-DD HH:MM:SS"     }
{ given a real number of seconds since 00:00:00 80-01-01.                 }
{-------------------------------------------------------------------------}

Function SecDT( InSec : Real ) : String17;
  Var
    Yr, Mo, Da, Hr, Mn, Sc, DaTotal, ScTotal : LongInt;
    DTLoop, DaysInCurrMonth : Integer;
  Const
    DaysPerMonth : Array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    SecDT := '';
    ScTotal := Trunc( InSec - ( Int( InSec / 86400.0 ) * 86400.0 ) );
    DaTotal := Trunc( InSec / 86400.0 ) + 1;
    Hr := ScTotal div 3600;
    Mn := ( ScTotal - ( Hr * 3600 ) ) div 60;
    Sc := ScTotal - ( Hr * 3600 ) - ( Mn * 60 );
    Yr := 1980;  Mo := 1;  Da := 1;
    If ( Mo <> 2 ) or ( Yr mod 4 <> 0 ) Then
      DaysInCurrMonth := DaysPerMonth[ Mo ]
    Else
      DaysInCurrMonth := 29;
    While DaTotal > DaysInCurrMonth do
      Begin
        DaTotal := DaTotal - DaysInCurrMonth;
        Inc( Mo );
        If Mo > 12 Then
          Begin
            Mo := 1;
            Inc( Yr );
          End;
        If ( Mo <> 2 ) or ( Yr mod 4 <> 0 ) Then
          DaysInCurrMonth := DaysPerMonth[ Mo ]
        Else
          DaysInCurrMonth := 29;
      End;
    Da := DaTotal;
    SecDT :=
      Chr( Yr mod 100 div 10 + 48 ) + Chr( Yr mod 10 + 48 ) + '/' +
      Chr( Mo div 10 + 48 ) + Chr( Mo mod 10 + 48 ) + '/' +
      Chr( Da div 10 + 48 ) + Chr( Da mod 10 + 48 ) + ' ' +
      Chr( Hr div 10 + 48 ) + Chr( Hr mod 10 + 48 ) + ':' +
      Chr( Mn div 10 + 48 ) + Chr( Mn mod 10 + 48 ) + ':' +
      Chr( Sc div 10 + 48 ) + Chr( Sc mod 10 + 48 );
  End;

{-------------------------------------------------------------------------}

Function ShowDate( InDate : String8;  Style : Integer ) : String40;
  Var
    Yr, Mo, Da, DaTotal, FullYr : LongInt;
    DTLp, DTResult, DayOfWk : Integer;
    DaStr, YrStr : String8;
  Const
    DayName : Array [ 1..7 ] of String[ 12 ] =
      ( 'Sunday', 'Monday', 'Tuesday', 'Wednesday',
        'Thursday', 'Friday', 'Saturday' );
    MonthName : Array [ 1..12 ] of String[ 12 ] =
      ( 'January', 'February', 'March', 'April', 'May', 'June',
        'July', 'August', 'September', 'October', 'November', 'December' );
    DaysPerMonth : Array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    { initialize function to null string, flags conversion error }
    ShowDate := '';
    { remove invalid characters }
    For DTLp := Length( InDate ) DownTo 1 do
      If not ( InDate[ DTLp ] in [ '0'..'9' ] ) Then
        Delete( InDate, DTLp, 1 );
    { exit if remaining length too small }
    If Length( InDate ) <> 6 Then Exit;
    { convert digit pairs to numbers }
    Val( Copy( InDate, 1, 2 ), Yr, DTResult );
    Val( Copy( InDate, 3, 2 ), Mo, DTResult );
    Val( Copy( InDate, 5, 2 ), Da, DTResult );
    { range check all values }
    If ( Yr < 0 ) or ( Yr > 99 ) or
       ( Mo < 1 ) or ( Mo > 12 ) or
       ( Da < 1 ) or ( Da > 31 ) or
       ( ( not ( Da in [ 1 .. DaysPerMonth[ Mo ] ] ) ) and
         ( not ( ( Da = 29 ) and ( Mo = 2 ) and ( Yr mod 4 = 0 ) ) ) ) Then
      Exit;
    { convert date to days since 01-01-80 }
    { assume years before 80 are in 21st century }
    If Yr < 80 Then
      FullYr := Yr + 2000
    Else
      FullYr := Yr + 1900;
    If Yr < 80 Then Yr := Yr + 100;
    Yr := Yr - 80;
    { start with leap year days }
    DaTotal := Yr div 4;
    { compensate for fact that the base year, 1980, is a leap year }
    Inc( DaTotal );
    { compensate if current year is leap year, but leap day is not past }
    If ( Yr mod 4 = 0 ) and
       ( ( Mo < 2 ) or ( ( Mo = 2 ) and ( Da < 29 ) ) ) Then
      Dec( DaTotal );
    { add days for past years }
    DaTotal := DaTotal + ( Yr * 365 );
    DTLp := 1;
    { add days for elapsed months in year }
    While DTLp < Mo do
      Begin
        DaTotal := DaTotal + DaysPerMonth[ DTLp ];
        Inc( DTLp );
      End;
    { add elapsed days in month to finish }
    DaTotal := DaTotal + Da - 1;
    { format into string }
    If Style > 0 Then
      Begin
        { styles:
            0 : no presentation date generated
            1 : Tue., Jan. 1, 1990
            else : Tuesday, January 1st., 1990  }
        DayOfWk := ( DaTotal   { 01-01-80 = Tuesday = 0 }
                     mod 7     { modulo into weeks }
                     + 2 )     { offset for Tuesday in 0..6 range }
                     mod 7     { correct range }
                     + 1;      { offset to 1..7 range }
        Str( Da, DaStr );
        Str( FullYr, YrStr );
        If Style = 1 Then
          ShowDate := Copy( DayName[ DayOfWk ], 1, 3 ) + '., ' +
                      Copy( MonthName[ Mo ], 1, 3 ) + '. ' +
                      DaStr + ', ' + YrStr
        Else
          Begin
            Da := Da mod 10;
            If Da = 1 Then
              DaStr := DaStr + 'st'
            Else
              If Da = 2 Then
                DaStr := DaStr + 'nd'
              Else
                If Da = 3 Then
                  DaStr := DaStr + 'rd'
                Else
                  DaStr := DaStr + 'th';
            ShowDate := DayName[ DayOfWk ] + ', ' +
                        MonthName[ Mo ] + ' ' + DaStr + '., ' + YrStr;
          End;
      End;
  End;

{-------------------------------------------------------------------------}

Function KeyPressed : Boolean;
  Var
    Regs : Registers;
  Begin
    Regs.AH := $0B;  MSDos( Regs );
    KeyPressed := Regs.AL = $FF;
  End;

{-------------------------------------------------------------------------}

Procedure Dly( Dur : Integer );
  Var
    Regs : Registers;
    StartSec100, EndSec100, CurrSec100 : LongInt;
  Begin
    Regs.AH := $2C;  MSDos( Regs );  { get time }
    StartSec100 :=
      Regs.CH * 360000 + Regs.CL * 6000 + Regs.DH * 100 + Regs.DL;
    Dur := Dur div 10;
    EndSec100 := StartSec100 + Dur;
    Repeat
      Regs.AH := $2C;  MSDos( Regs );  { get time }
      CurrSec100 :=
        Regs.CH * 360000 + Regs.CL * 6000 + Regs.DH * 100 + Regs.DL;
    Until ( CurrSec100 >= EndSec100 ) or ( CurrSec100 < StartSec100 );
  End;

{-------------------------------------------------------------------------}

Function Str2( InNum : Word ) : String2;
  Begin
    Str2 := Chr( InNum div 10 + 48 ) + Chr( InNum mod 10 + 48 );
  End;

{-------------------------------------------------------------------------}

Function DateString : String8;
  Var
    Regs : Registers;
  Begin
    Regs.AH := $2A;  MSDos( Regs );
    DateString := Str2(Regs.CX mod 100)+'/'+Str2(Regs.DH)+'/'+Str2(Regs.DL);
  End;

{-------------------------------------------------------------------------}

Function TimeString : String8;
  Var
    Regs : Registers;
  Begin
    Regs.AH := $2C;  MSDos( Regs );
    TimeString := Str2(Regs.CH)+':'+Str2(Regs.CL)+':'+Str2(Regs.DH);
  End;

{-------------------------------------------------------------------------}

Function GetKey : Char;
  Var
    Regs : Registers;
  Begin
    Regs.AH := $08;  { ConInputNoEcho }
    MsDos( Regs );
    GetKey := UpCase( Chr( Regs.AL ) );
  End;

{-------------------------------------------------------------------------}

Function SelectCh( Msg : String;  Mask : String30 ) : Char;
  Var
    SlctCh : Char;
  Begin
    Write( Msg, ' > ' );
    Repeat
      SlctCh := GetKey;  SelectCh := SlctCh;
      If ( Pos( SlctCh, Mask ) = 0 ) and ( Mask <> #0 ) Then
        Write( ^G );  { bell }
    Until ( Pos( SlctCh, Mask ) > 0 ) or ( Mask = #0 );
  End;

{-------------------------------------------------------------------------}

Function Upper( S : String80 ) : String80;
  Var
    Loop : Byte;
  Begin
    For Loop := 1 to Length( S ) do
      S[ Loop ] := UpCase( S[ Loop ] );
    Upper := S;
  End;

{-------------------------------------------------------------------------}

Function Strip( S : String80 ) : String80;
  Begin
    While ( S[ 1 ] = ' ' ) and ( S <> '' ) do
      Delete( S, 1, 1 );
    While ( S[ Length( S ) ] = ' ' ) and ( S <> '' ) do
      Delete( S, Length( S ), 1 );
    Strip := S;
  End;

{-------------------------------------------------------------------------}

Function ErrorName( Err : Integer ):String80;
  Var
    ErrNumMsg : String[10];
  Begin
    Str( Err, ErrNumMsg );
    ErrorName := 'Error # ' + ErrNumMsg;
    Case Err of
         0 : ErrorName := 'No error';
         2 : ErrorName := 'Error, file not found';
         3 : ErrorName := 'Error, path not found';
         4 : ErrorName := 'Error, too many files open';
         5 : ErrorName := 'Error, access denied';
         6 : ErrorName := 'Invalid file handle';
        12 : ErrorName := 'Invalid file access code';
        15 : ErrorName := 'Invalid drive number';
        16 : ErrorName := 'Cannot remove current path';
        17 : ErrorName := 'Error, inter-device rename';
       100 : ErrorName := 'Error, access past file end';
       101 : ErrorName := 'Error, disk full';
       102 : ErrorName := 'Error, file not assigned';
       103 : ErrorName := 'Error, file not open';
       104 : ErrorName := 'Error, not open for input';
       105 : ErrorName := 'Error, not open for output';
       106 : ErrorName := 'Bad numeric format';
       150 : ErrorName := 'Error, write protected';
       151 : ErrorName := 'Unknown unit error';
       152 : ErrorName := 'Error, drive not ready';
       154 : ErrorName := 'CRC error in data';
       155 : ErrorName := 'Drive request error';
       156 : ErrorName := 'Seek error';
       157 : ErrorName := 'Error, unknown disk type';
       158 : ErrorName := 'Error, sector not found';
       159 : ErrorName := 'Printer out of paper';
       160 : ErrorName := 'Printer off line';
       161 : ErrorName := 'Device read error';
       162 : ErrorName := 'Hardware failure';
       200 : ErrorName := 'Error, division by zero';
       201 : ErrorName := 'Range check error';
       202 : ErrorName := 'Stack overflow error';
       203 : ErrorName := 'Heap overflow error';
       204 : ErrorName := 'Bad pointer operation';
       205 : ErrorName := 'Numeric overflow error';
       206 : ErrorName := 'Numeric underflow error';
       207 : ErrorName := 'Real number error';
       255 : ErrorName := 'User break';
     End;
  End;

{-------------------------------------------------------------------------}

Function HexByte( InByt : Byte ) : String2;
  Const
    HexSet : String[ 16 ] = '0123456789ABCDEF';
  Begin
    HexByte := HexSet[ ( InByt shr 4 ) and $0F + 1 ] +
               HexSet[ InByt and $0F + 1 ];
  End;

{-------------------------------------------------------------------------}

Procedure ChkErr( Result : Integer;  Msg : String80 );
  Begin
    If Result <> 0 Then
      Begin
        Writeln( NL );
        If Result > 0 Then
          WriteLn( ErrorName( Result ), ' ', Msg, ' file.' )
        Else
          WriteLn( Msg, '.' );
        If FromMenu Then
          Writeln( 'Process terminated.', NL )
        Else
          Writeln( 'Program terminated.', NL );
        Write( ^G^G^G );
        (****  Close all files just in case.  ****)
        Close( SortFile );  Result := IOResult;
        Erase( SortFile );  Result := IOResult;
        Close( ODv );  Result := IOResult;
        Close( FindProjNameFile );  Result := IOResult;
        Close( ProjNameFile );  Result := IOResult;
        Close( DoLog );  Result := IOResult;
        Close( DoneLog );  Result := IOResult;
        Close( SrcLog );  Result := IOResult;
        Close( CutLog );  Result := IOResult;
        ExitProc := ExitSave;
        If FromMenu Then
          IOError := True
        Else
          Halt( 1 );
      End;
  End;

{-------------------------------------------------------------------------}

{$F+}
Procedure TidyExit( Result : Integer;  Msg : String80 );
  Var
    TermCode : Integer;
    TermAddrPtr : Pointer;
    TermAddr : Array [ 1..4 ] of Byte absolute TermAddrPtr;
  Begin
    TermCode := ExitCode;
    TermAddrPtr := ErrorAddr;
    ExitProc := ExitSave;
    If TermCode = 255 Then Writeln( '** User Break **' );
    If ( TermCode <= 1 ) or ( TermCode = 255 ) Then Halt;
    FromMenu := False;
    ChkErr( -1, ErrorName( TermCode ) + ' at address $' +
               HexByte( TermAddr[ 4 ] ) + HexByte( TermAddr[ 3 ] ) + ':' +
               HexByte( TermAddr[ 2 ] ) + HexByte( TermAddr[ 1 ] ) );
  End;
{$F-}

{-------------------------------------------------------------------------}

Function CheckKey : Boolean;
  Var
    ChkCh : Char;
  Begin
    CheckKey := False;
    If not KeyPressed Then Exit;
    Repeat
      ChkCh := GetKey;
    Until not KeyPressed;
    If ChkCh = #27 Then
      Begin
        Writeln( '**** TERMINATED ****' );
        CheckKey := True;
      End
    Else
      Begin
        Write( #13, '**** PAUSED ****' );
        Repeat
          ChkCh := GetKey;
        Until not KeyPressed;
        Write( #13, '                ', #13 );
        If ChkCh = #27 Then
          Begin
            Writeln( '**** TERMINATED ****' );
            CheckKey := True;
          End
      End
  End;

{-------------------------------------------------------------------------}

Function FindProjName( Project : String8 ) : String30;
  Var
    ProjName : String30;
    Result, ProjNameLoop : Integer;
  Begin
    { finds a full project name given a three letter code }
    FindProjName := '';
    Assign( FindProjNameFile, ProjNameFileSpec );
    { loads file if null name passed }
    If Project = '' Then
      Begin
        ProjNames := 0;
        FillChar( ProjNameArr, SizeOf( ProjNameArr ), 0 );
        Writeln( 'Loading project names . . .' );
        Reset( FindProjNameFile );  Result := IOResult;
        While ( Result = 0 ) and ( not EOF( FindProjNameFile ) ) do
          Begin
            ReadLn( FindProjNameFile, ProjName );
            ChkErr( IOResult, 'reading project name' );
            If IOError Then Exit;
            ProjName := Strip( ProjName );
            If Length( ProjName ) > 4 Then
              Begin
                If ProjNames = MaxProjName Then
                  Begin
                    ChkErr( -1, 'Too many project names' );
                    If IOError Then Exit;
                  End;
                Inc( ProjNames );
                ProjNameArr[ ProjNames ] := ProjName;
              End;
          End;
        Close( FindProjNameFile );  Result := IOResult;
      End;
    { search through name array }
    ProjNameLoop := 1;
    While ProjNameLoop <= ProjNames do
      Begin
        If Project = Copy( ProjNameArr[ ProjNameLoop ], 1, 3 ) Then
          Begin
            FindProjName := Copy( ProjNameArr[ ProjNameLoop ], 5, 255 );
            ProjNameLoop := ProjNames;
          End;
        Inc( ProjNameLoop );
      End;
  End;

{-------------------------------------------------------------------------}

{ 0        1         2         3          }
{ 123456789012345678901234567890123456789 }
{ _yy-mm-dd_hh:mm:ss_ccc_notes........... }  { before posting }
{ uuu_yy-mm-dd_hh:mm:ss_nn.d_ccc_notes... }  { after posting }

Procedure PostLog;
  Const
    InvalidFileNameCharSet : Set of Char =
      [ #0..#20, #128..#255, '.', '"', '/', '\', '[', ']',
        ':', '|', '<', '>', '+', '=', ';', ',' ];
  Var
    HasCR : Boolean;
    LeftOverCh : Char;
    ProjName, _DoLogSpec : String80;
    PrevLn, CurrLn, OutLn : String;
    PrevDT, CurrDT, Hours : Real;
    User, NumStr, Project, PrevProject : String8;
    LineNum : LongInt;
    Result, JunkResult : Integer;
    PrevStat, CurrStat : String30;
    SelectChar : Char; 

  Procedure FixLine( Var InLn : String );
    Begin
      { allow reposting by removing user identifier }
      If ( Length( InLn ) > 3 ) and
         ( InLn[ 1 ] in [ 'A'..'Z' ] ) and
         ( InLn[ 2 ] in [ 'A'..'Z' ] ) and
         ( InLn[ 3 ] in [ 'A'..'Z' ] ) and
         ( InLn[ 4 ] = ' ' ) Then
        Delete( InLn, 1, 3 );
      { correct for formats not having leading space }
      If ( InLn <> '' ) and ( InLn[ 1 ] <> ' ' ) Then
        InLn := ' ' + InLn;
      { allow reposting by removing elapsed time }
      If ( Copy( InLn, 22, 1 ) = '.' ) and
         ( Copy( InLn, 19, 1 ) = ' ' ) and
         ( Copy( InLn, 24, 1 ) = ' ' ) Then
        Delete( InLn, 20, 5 );
      If Copy( InLn, 20, 1 ) = ';' Then
        Begin
          { if comment, remove characters upto space after CCC }
          Delete( InLn, 20, 1 );
          If Copy( InLn, 20, 1 ) <> ' ' Then Delete( InLn, 20, 1 );
          If Copy( InLn, 20, 1 ) <> ' ' Then Delete( InLn, 20, 1 );
          { delete extra spaces if any }
          While Copy( InLn, 20, 1 ) = ' ' do Delete( InLn, 20, 1 );
          { insert current project }
          Insert( Project + ' ', InLn, 20 );
        End;
    End;

  { allows reading of unix LF delimited lines }
  Procedure X_ReadLn( Var InLn : String;  Msg : String );
    Var
      Ch : Char;
    Begin  
      If HasCR Then
        Begin
          ReadLn( DoLog, InLn );
          ChkErr( IOResult, Msg );
          If IOError Then Exit;
          If LeftOverCh <> #0 Then
            InLn := LeftOverCh + '';
          LeftOverCh := #0;
        End
      Else
        Begin
          InLn := '';
          If LeftOverCh <> #0 Then
            InLn := LeftOverCh + '';
          Repeat
            Read( DoLog, Ch );
            ChkErr( IOResult, Msg );
            If IOError Then Exit;
            If not ( Ch in [ #13, #10 ] ) Then
              InLn := InLn + Ch;
          Until ( Ch in [ #13, #10 ] ) or EOF( DoLog );
          Repeat
            HasCR := Ch = #13;
            Read( DoLog, Ch );
            ChkErr( IOResult, Msg );
            If IOError Then Exit;
          Until ( not ( Ch in [ #13, #10 ] ) ) or EOF( DoLog );
          LeftOverCh := #0;
          If not EOF( DoLog ) Then LeftOverCh := Ch;
        End;
    End;

Begin
    HasCR := False;
    LeftOverCh := #0;
    User := Upper( ParamArr[ 2 ] );
    If ( Length( User ) <> 3 ) or
       ( not ( User[ 1 ] in [ 'A'..'Z' ] ) ) or
       ( not ( User[ 2 ] in [ 'A'..'Z' ] ) ) or
       ( not ( User[ 3 ] in [ 'A'..'Z' ] ) ) Then
      Begin
        ChkErr( -1, 'Invalid user name "' + User + '".' );
        If IOError Then Exit;
      End;
    { open DO log from beginning }
    If ParamCnt > 2 Then
      _DoLogSpec := ParamArr[ 3 ]
    Else
      _DoLogSpec := DoLogSpec;
    Assign( DoLog, _DoLogSpec );
    SetTextBuf( DoLog, DoBuf );
    Reset( DoLog );  Result := IOResult;
    If Result = 2 Then
      ChkErr( -1, 'No DO log found.' )
    Else
      ChkErr( Result, 'opening DO' );
    If IOError Then Exit;
    Writeln( FindProjName( '' ) );  { load project name array }
    If IOError Then Exit;
    Writeln( 'Checking DO log for errors . . .' );
    Writeln( '(press any key to pause, <ESC> to terminate)' );
    CurrDT := 0.0;  CurrLn := 'Start of file.';  Project := '';
    LineNum := 0;
    PrevStat := '';
    While not EOF( DoLog ) do
      Begin  { scan for format errors }
        PrevDT := CurrDT;  PrevLn := CurrLn;
        X_ReadLn( CurrLn, 'scanning DO' );
        FixLine( CurrLn );
        Inc( LineNum );
        { show scanning progress }
        CurrStat := Copy( CurrLn, 2, 11 );
        If CurrStat <> PrevStat Then
          Begin
            Write( #13, ' ', CurrStat, ':00' );
            PrevStat := CurrStat;
            If CheckKey Then Exit;
          End;
        { get date/time of current line }
        CurrDt := DTSec( Copy( CurrLn, 2, 8 ), Copy( CurrLn, 11, 8 ) );
        If CurrDt < 1.0 Then
          Begin
            Write( #13, Copy( CurrLn, 1, 75 ) );
            Str( LineNum, NumStr );
            ChkErr( -1, 'Bad date or time in line ' + NumStr );
            If IOError Then Exit;
          End;
        { check for valid project identifier }
        PrevProject := Project;
        Project := Copy( CurrLn, 19, 5 );
        If ( Length( Project ) <> 5 ) or
           ( Project[ 1 ] <> ' ' ) or ( Project[ 5 ] <> ' ' ) or
           ( Project[ 2 ] in InvalidFileNameCharSet ) or
           ( Project[ 3 ] in InvalidFileNameCharSet ) or
           ( Project[ 4 ] in InvalidFileNameCharSet ) Then
          Begin
            Write( #13, Copy( CurrLn, 1, 75 ) );
            Str( LineNum, NumStr );
            ChkErr( -1, 'Bad project name format in line ' + NumStr );
            If IOError Then Exit;
          End;
        { check for if project file previously started }
        Project := Upper( Copy( Project, 2, 3 ) );
        If FindProjName( Project ) = '' Then
          Begin  { unknown project }
            If IOError Then Exit;
            Write( #13, Copy( CurrLn, 1, 75 ) );
            Str( LineNum, NumStr );
            Writeln( NL, 'Unknown project "', Project,
                     '" in line ' + NumStr );
            SelectChar := SelectCh( 'Press  Y  to accept', #0 );
            If SelectChar <> 'Y' Then
              Begin
                Writeln( 'No', NL, 'Posting cancelled.' );
                Exit;
              End;
            Writeln( 'Yes' );
            { write project name to first line of log }
            Write( 'Enter full name for project "', Project, '" > ' );
            ReadLn( ProjName );  ProjName := Strip( ProjName );
            If Length( ProjName ) >= 30 Then
              ProjName := Copy( ProjName, 1, 29 );
            { add new entry to project name file }
            If ProjNames = MaxProjName Then
              Begin
                ChkErr( -1, 'Too many project names' );
                If IOError Then Exit;
              End;
            Inc( ProjNames );
            ProjNameArr[ ProjNames ] := Project + ' ' + ProjName;
            Assign( ProjNameFile, ProjNameFileSpec );
            Append( ProjNameFile );  Result := IOResult;
            If Result <> 0 Then
              Begin
                ReWrite( ProjNameFile );
                ChkErr( IOResult, 'creating project name' );
                If IOError Then Exit;
              End;
            WriteLn( ProjNameFile, Project + ' ' + ProjName );
            ChkErr( IOResult, 'writing project name' );
            If IOError Then Exit;
            Close( ProjNameFile );
            ChkErr( IOResult, 'closing project name' );
            If IOError Then Exit;
          End;  { unknown project }
        If PrevDT > 1.0 Then
          Begin  { valid elapsed time }
            Hours := Round( ( CurrDt - PrevDT ) / 36.0 ) / 100.0;
            If Hours < 0.0 Then
              Begin
                Write( #13, Copy( CurrLn, 1, 75 ) );
                Str( LineNum, NumStr );
                ChkErr( -1, 'Time or date out of sequence in line ' +
                            NumStr );
                If IOError Then Exit;
              End;
            If ( Hours > MaxSeqHours ) and
               ( Copy( PrevProject, 1, 1 ) <> NonWork ) Then
              Begin
                Write( #13, Copy( PrevLn, 1, 75 ) );
                Writeln( NL, Hours:1:1, ' consecutive hours on ',
                             'project "', PrevProject, '".' );
                SelectChar := SelectCh( 'Press  Y  to accept', #0 );
                If SelectChar <> 'Y' Then
                  Begin
                    Writeln( 'No', NL, 'Posting cancelled.' );
                    Exit;
                  End;
                Writeln( 'Yes' );
              End;
          End;  { valid elapsed time }
        If CheckKey Then Exit;
      End;  { scan for format errors }
    If LineNum < 2 Then
      Begin
        ChkErr( -1, 'Minimum of two lines required for posting' );
        If IOError Then Exit;
      End;
    { no errors, start posting }
    Writeln( NL, 'Posting DO log . . .' );
    { reset DO log to beginning }
    Reset( DoLog );
    ChkErr( IOResult, 'opening DO' );
    If IOError Then Exit;
    { open DONE log appending to end }
    Assign( DoneLog, DoneLogSpec );
    SetTextBuf( DoneLog, DoneBuf );
    Append( DoneLog );
    If IOResult <> 0 Then ReWrite( DoneLog );
    ChkErr( IOResult, 'opening DONE' );
    If IOError Then Exit;
    { preload the current line }
    X_ReadLn( PrevLn, 'reading DO' );
    FixLine( PrevLn );
    { preset project to nothing }
    PrevProject := '';
    PrevStat := '';
    While not EOF( DoLog ) do
      Begin  { while not EOF }
        { show posting progress }
        CurrStat := Copy( PrevLn, 2, 11 );
        If CurrStat <> PrevStat Then
          Begin
            Write( #13, ' ', CurrStat, ':00' );
            PrevStat := CurrStat;
          End;
        { get date/time of previous line }
        PrevDt := DTSec( Copy( PrevLn, 2, 8 ), Copy( PrevLn, 11, 8 ) );
        If PrevDT = 0 Then
          Begin
            ChkErr( -1, 'Bad date or time in DO file' );
            If IOError Then Exit;
          End;
        { load current line }
        X_ReadLn( CurrLn, 'reading DO' );
        FixLine( CurrLn );
        { get date/time of current line }
        CurrDt := DTSec( Copy( CurrLn, 2, 8 ), Copy( CurrLn, 11, 8 ) );
        If CurrDT = 0 Then
          Begin
            ChkErr( -1, 'Bad date or time in DO file' );
            If IOError Then Exit;
          End;
        { calculate elapsed time in hours with one decimal place }
        Hours := Round( ( CurrDt - PrevDT ) / 36.0 ) / 100.0;
        { correct for allowed ranges }
        If Hours > 99.9 Then
          Hours := 99.9
        Else
          If Hours <= 0.01 Then
            Hours := 0.0
          Else
            If Hours < 0.1 Then
              Hours := 0.1;
        Str( Hours:4:1, NumStr );
        { create output from previous line with inserted elapsed time }
        OutLn := User +
          Copy( PrevLn, 1, 19 ) + NumStr + ' ' + Copy( PrevLn, 20, 255 );
        { write line to done file }
        Writeln( DoneLog, OutLn );
        Project := Upper( Copy( PrevLn, 20, 3 ) );
        { make current line to previous for next reference }
        PrevLn := CurrLn;
      End;  { while not EOF }
    Close( DoneLog );
    ChkErr( IOResult, 'closing DONE');
    If IOError Then Exit;
    Close( DoLog );
    ChkErr( IOResult, 'closing DO' );
    If IOError Then Exit;
    { erase original DO log except unposted last line }
    Assign( DoLog, _DoLogSpec );
    SetTextBuf( DoLog, DoBuf );
    ReWrite( DoLog );
    ChkErr( IOResult, 'opening new DO' );
    If IOError Then Exit;
    If HasCR Then
      CurrLn := CurrLn + #13 + #10
    Else
      CurrLn := CurrLn + #10;
    Repeat  { writes in same format as read }
      Write( DoLog, CurrLn[ 1 ] );
      ChkErr( IOResult, 'writing new DO' );
      If IOError Then Exit;
      Delete( CurrLn, 1, 1 );
    Until CurrLn = '';
    Close( DoLog );
    ChkErr( IOResult, 'closing new DO' );
    If IOError Then Exit;
    Writeln( NL, NL, 'Posting completed.' );
  End;

{-------------------------------------------------------------------------}


