{$B-,D-,F-,I-,L-,N-,R-,S-,V-}
{$M 64000,0,0}

Program Done;

Uses
  Dos;

Type
  SortString = String;
  String80 = String[80];
  String40 = String[40];
  String30 = String[30];
  String17 = String[17];
  String8 = String[8];
  String2 = String[2];
  { text file buffer type }
  TBuf = Array [ 1..512 ] of Char;

Const
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      'fwvs)1sz~';
  CpyRtLp : Byte = 1;
  NL = #13#10;
  Esc = #27;
  CtrlC = #3;
  { file names }
  DoLogSpec : String[30] = 'C:\DO.DAT';
  DoneLogSpec : String[30] = 'DONE.DAT';
  CutLogSpec : String[30] = 'CUT.DO';
  ProjNameFileSpec : String[30] = 'DONEPROJ.DTA';
  { elapsed time requiring confirmation when posting }
  MaxSeqHours : Integer = 5;
  { symbol for non-work group }
  NonWork = '^';
  { used for project name array }
  MaxProjName = 500;
  ProjNames : Integer = 0;
  { replaces command line to alternatively control by menu }
  MaxParam = 12;
  ParamCnt : Byte = 0;
  FromMenu : Boolean = False;
  IOError : Boolean = False;

Var
  VerMsg : String[80];
  { used by tidy exit }
  ExitSave : Pointer;
  { buffers for faster text file access }
  DoBuf, DoneBuf, CutBuf, PrnBuf, SrcBuf : TBuf;
  { project name array }
  ProjNameArr : Array [ 1..MaxProjName ] of String30;
  ParamArr : Array [ 1..MaxParam ] of String30;
  (****  All files are global so that they can be closed  ****)
  (****  error handling procedure.                        ****)
  { temporary file for sorting }
  SortFile : File of SortString;
  { report and cut output file }
  ODv : Text;
  { used by FindProjName procedure }
  FindProjNameFile : Text;
  { used by post procedure }
  ProjNameFile : Text;
  { used by post and log procedures }
  DoLog : Text;
  { used by post and report procedures }
  DoneLog : Text;
  { used by cut procedure }
  SrcLog, CutLog : Text;

{-------------------------------------------------------------------------}

  {$I DONE.INC }

{-------------------------------------------------------------------------}

Procedure ReportLog( StartDate, EndDate : String8 );
  Var
    ProjName : String80;
    SrcNm : String8;
    CurrDT, StartDT, EndDT : Real;
    CurrHour, ProjHours, GroupHours, TotalHours, WorkHours : Real;
    PrevProject, Project, HourStr, FirstDate, LastDate, SwapDate : String8;
    Result : Integer;
    Index, Jump, Gap, Size, Temp, RecCount, TotalRecs : LongInt;
    OutLn, Ln1, Ln2 : SortString;
    ODvName : String80;
    PrevGroup, Group, Ch : Char;
    ODvType : Byte;  { 0/1/2 = screen/device/file }
    EventCnt : LongInt;
    ParamLoop : Byte;
    Terse, UseCut, ShowStat : Boolean;
    PrevStat, CurrStat : String30;
    SelectChar, PauseKey : Char;

  Procedure ShowProjectHours;
    Begin
      If not Terse Then
        Writeln( ODv, '                     ------' );
      ChkErr( IOResult, 'writing report' );
      If IOError Then Exit;
      Write( ODv, 'Project hours ---->', ProjHours:7:1, '   ' );
      ChkErr( IOResult, 'writing report' );
      If IOError Then Exit;
      If Copy( PrevProject, 1, 1 ) = NonWork Then
        Write( ODv, ( ( ProjHours * 100 ) / TotalHours ):5:1,
                    ' % of total hours' )
      Else
        Write( ODv, ( ( ProjHours * 100 ) / WorkHours ):5:1,
                    ' % of worked hours' );
      ChkErr( IOResult, 'writing report' );
      If IOError Then Exit;
      { show number of events, average duration, and average frequency }
      If Copy( PrevProject, 1, 1 ) <> NonWork Then
        Writeln( ODv, ' (/', EventCnt, '=', ( ProjHours / EventCnt ):1:1,
                 '@', ( WorkHours / EventCnt ):1:1, ')', NL )
      Else
        Writeln( ODv, NL );
      ChkErr( IOResult, 'writing report' );
      If IOError Then Exit;
    End;

  Procedure ShowGroupHours;
    Begin
      If Copy( PrevProject, 1, 1 ) = NonWork Then
        Writeln( ODv, 'Group hours ------>', GroupHours:7:1,
                      '   ', ( ( GroupHours * 100 ) / TotalHours ):5:1,
                      ' % of total hours.', NL )
      Else
        Writeln( ODv, 'Group hours ------>', GroupHours:7:1,
                      '   ', ( ( GroupHours * 100 ) / WorkHours ):5:1,
                      ' % of work hours.', NL );
      Writeln( ODv, '-----------------------------------',
                    '-----------------------------------', NL );
      ChkErr( IOResult, 'writing report' );
      If IOError Then Exit;
    End;

  Begin
    { check validity of options }
    Terse := False;  UseCut := False;
    If Upper( ParamArr[ 1 ] ) = 'REPORT' Then
      For ParamLoop := 4 to ParamCnt do
        If Upper( ParamArr[ ParamLoop ] ) = '/T' Then
          Terse := True
        Else
          If Upper( ParamArr[ ParamLoop ] ) = '/C' Then
            UseCut := True
          Else
            If ( ParamLoop > 4 ) or
               ( Copy( ParamArr[ ParamLoop ], 1, 1 ) = '/' ) Then
              Begin
                Writeln( NL, 'Unknown option: "',
                             Upper( ParamArr[ ParamLoop ] ), '"', NL,
                             'Enter  DONE ?  for help.', NL );
                Halt( 1 );
              End;
    { check validity of dates }
    If ( DTSec( StartDate, '00:00:00' ) < 1.0 ) or
       ( DTSec( EndDate, '00:00:00' ) < 1.0 ) Then
      Begin
        ChkErr( -1, 'Bad start or end date' );
        If IOError Then Exit;
      End;
    { swap dates if necessary }
    If DTSec( EndDate, '00:00:00' ) < DTSec( StartDate, '00:00:00' ) Then
      Begin
        SwapDate := EndDate;  EndDate := StartDate;  StartDate := SwapDate;
      End;
    { get output destination name and type }
    If ( ParamCnt > 3 ) and ( Copy( ParamArr[ 4 ], 1, 1 ) <> '/' ) Then
      ODvName := Upper( ParamArr[ 4 ] )
    Else
      ODvName := '';
    If ODvName = '' Then
      ODvType := 0
    Else
      If ( ODvName = 'PRN' ) or
         ( ODvName = 'LPT1' ) or ( ODvName = 'LPT2' ) or
         ( ODvName = 'COM1' ) or ( ODvName = 'COM2' ) Then
        ODvType := 1
      Else
        ODvType := 2;
    { open done log }
    If UseCut Then
      Begin
        SrcNm := 'CUT';
        Assign( DoneLog, CutLogSpec );
        SetTextBuf( DoneLog, DoneBuf );
      End
    Else
      Begin
        SrcNm := 'DONE';
        Assign( DoneLog, DoneLogSpec );
        SetTextBuf( DoneLog, DoneBuf );
      End;
    Reset( DoneLog );
    ChkErr( IOResult, 'opening ' + SrcNm );
    If IOError Then Exit;
    { open temporary sort file }
    Assign( SortFile, 'DOSORT.$$$' );
    ReWrite( SortFile );
    ChkErr( IOResult, 'opening sort' );
    If IOError Then Exit;
    { convert start and end dates to reals }
    StartDT := DTSec( StartDate, '00:00:00' );
    EndDT := DTSec( EndDate, '00:00:00' );
    FirstDate := '';  LastDate := '';
    TotalHours := 0.0;  WorkHours := 0.0;
    { copy target range to sort file }
    Writeln( FindProjName( '' ) );  { load project name array }
    If IOError Then Exit;
    Writeln( 'Scanning ' + SrcNm + ' log . . .' );
    Writeln( '(press any key to pause, <ESC> to terminate)' );
    PrevStat := '';  ShowStat := False;
    While not EOF( DoneLog ) do
      Begin
        ReadLn( DoneLog, OutLn );
        ChkErr( IOResult, 'reading ' + SrcNm );
        If IOError Then Exit;
        { show scan progress }
        CurrStat := Copy( OutLn, 5, 11 );
        If PrevStat <> CurrStat Then
          Begin
            Write( #13, ' ', CurrStat, ':00' );
            PrevStat := CurrStat;
            ShowStat := True;
            If CheckKey Then Exit;
          End;
        If Length( OutLn ) < 30 Then
          Begin
            ChkErr( -1, 'Data missing in ' + SrcNm + ' file' );
            If IOError Then Exit;
          End;
        CurrDT := DTSec( Copy( OutLn, 5, 8 ), '00:00:00' );
        If ( CurrDT >= StartDT ) and ( CurrDT <= EndDT ) Then
          Begin
            If ShowStat Then
              Begin  Write( ' *' );  ShowStat := False;  End;
            LastDate := Copy( OutLn, 5, 8 );
            If FirstDate = '' Then FirstDate := LastDate;
            Write( SortFile, OutLn );
            ChkErr( IOResult, 'writing sort' );
            If IOError Then Exit;
            { collect total time for percentile figures }
            Project := Copy( OutLn, 28, 3 );
            Val( Copy( OutLn, 23, 4 ), CurrHour, Result );
            If Result <> 0 Then
              Begin
                ChkErr( -1, 'Bad elapsed time in scan file' );
                If IOError Then Exit;
              End;
            If Copy( Project, 1, 1 ) <> NonWork Then
              WorkHours := WorkHours + CurrHour;
            TotalHours := TotalHours + CurrHour;
          End
        Else
          Begin
            If ShowStat Then
              Begin  Write( '  ' );  ShowStat := False;  End;
          End;
      End;
    Close( DoneLog );
    ChkErr( IOResult, 'closing ' + SrcNm );
    If IOError Then Exit;
    If FileSize( SortFile ) < 1 Then
      Begin
        ChkErr( -1, 'No data in range for report' );
        If IOError Then Exit;
      End;
    If ( WorkHours < 1.0 ) or ( TotalHours < 1.0 ) Then
      Begin
        ChkErr( -1, 'No elapsed times found.' );
        If IOError Then Exit;
      End;
    { sort file of entries in range by project identifier field }
    Size := FileSize( SortFile );
    If Size < 2 Then Exit;
    Gap := Size Div 2;
    Temp := Gap;  RecCount := 0;  TotalRecs := 0;
    While Temp > 0 do  { figure required iterations }
      Begin  Inc( TotalRecs );  Temp := Temp div 2;  End;
    Write( NL, 'Sorting report data . . .', NL, 0.0:5:0, ' % done.' );
    While Gap > 0 Do
      Begin
        For Index := ( Gap + 1 ) to Size do
          Begin
            Jump := Index - Gap;
            While Jump > 0 Do
              Begin
                Seek( SortFile, Jump - 1 );
                ChkErr( IOResult, 'moving sort' );
                If IOError Then Exit;
                Read( SortFile, Ln1 );
                ChkErr( IOResult, 'reading sort' );
                If IOError Then Exit;
                Seek( SortFile, Jump + Gap - 1 );
                ChkErr( IOResult, 'moving sort' );
                If IOError Then Exit;
                Read( SortFile, Ln2 );
                ChkErr( IOResult, 'reading sort' );
                If IOError Then Exit;
                { swap }
                If Copy( Ln1, 28, 3 ) + Copy( Ln1, 5, 17 ) >
                   Copy( Ln2, 28, 3 ) + Copy( Ln2, 5, 17 ) Then
                  Begin
                    Seek( SortFile, Jump - 1 );
                    ChkErr( IOResult, 'moving sort' );
                    If IOError Then Exit;
                    Write( SortFile, Ln2 );
                    ChkErr( IOResult, 'writing sort' );
                    If IOError Then Exit;
                    Seek( SortFile, Jump + Gap - 1 );
                    ChkErr( IOResult, 'moving sort' );
                    If IOError Then Exit;
                    Write( SortFile, Ln1 );
                    ChkErr( IOResult, 'writing sort' );
                    If IOError Then Exit;
                    Jump := Jump - Gap;
                  End
                Else
                  Jump := 0;
              End;
            If CheckKey Then Exit;
          End;
        Gap := Gap div 2;  Inc( RecCount );
        Write( #13, Trunc( ( RecCount * 100.0 ) / TotalRecs ):5 );
      End;  { sort }
    Writeln( NL );
    { reset for output }
    Seek( SortFile, 0 );
    ChkErr( IOResult, 'preparing output' );
    If IOError Then Exit;
    { open output device }
    Assign( ODv, ODvName );
    SetTextBuf( ODv, PrnBuf );
    If ODvType = 2 Then
      Begin
        Append( ODv );  Result := IOResult;
        If Result = 0 Then
          Begin  { file exists }
            SelectChar :=
              SelectCh( 'Report file exists - ' +
                        'select: [O]verwrite/[A]ppend/[Q]uit',
                        'OAQ' + Esc + CtrlC );
            Case SelectChar of
              'Q', Esc, CtrlC :
                Begin
                  Writeln( 'Quit' + NL + 'Report cancelled.' );
                  Close( ODv );  Result := IOResult;
                  Exit;
                End;
              'O' :
                Begin
                  Writeln( 'Overwriting' );
                  Close( ODv );  Result := IOResult;
                  ReWrite( ODv );
                  ChkErr( IOResult, 'opening report' );
                  If IOError Then Exit;
                End;
              'A' : Writeln( 'Appending' );
            End;  { case }
          End  { file exists }
        Else
          Begin
            Close( ODv );  Result := IOResult;
            Writeln( 'Creating report file . . .' );
            ReWrite( ODv );
            ChkErr( IOResult, 'opening report' );
            If IOError Then Exit;
          End;
        Writeln( 'Writing report to file . . .' );
      End
    Else
      Begin
        ReWrite( ODv );
        ChkErr( IOResult, 'opening report' );
        If IOError Then Exit;
        If ODvType = 1 Then
          Writeln( 'Printing report . . .' )
        Else
          Begin
            Writeln( 'Ready to display report . . .' );
            Write( '****  PRESS ANY KEY  ****', ^G );
            PauseKey := GetKey;
            Write( #13 );
            Writeln( 'Displaying report . . .' );
            Writeln( '(press any key to pause, <ESC> to terminate)' );
            Dly( 700 );  Writeln;
          End;
      End;
    PrevProject := '';  PrevGroup := #0;
    ProjHours := 0.0;  EventCnt := 0;
    If Terse Then
      Writeln( ODv, '***  SUMMARY TIME LOG REPORT  ***' )
    Else
      Writeln( ODv, '***  DETAILED TIME LOG REPORT  ***' );
    Writeln( ODv, NL, ShowDate( FirstDate, 2 ), ' - ',
                  ShowDate( LastDate, 2 ), NL );
    Writeln( ODv, '-----------------------------------',
                  '-----------------------------------', NL );
    While not EOF( SortFile ) do
      Begin  { output loop }
        Read( SortFile, OutLn );
        ChkErr( IOResult, 'reading output' );
        If IOError Then Exit;
        Val( Copy( OutLn, 23, 4 ), CurrHour, Result );
        If Result <> 0 Then
          Begin
            ChkErr( -1, 'Bad elapsed time in sort file' );
            If IOError Then Exit;
          End;
        Project := Copy( OutLn, 28, 3 );
        Group := Project[ 1 ];
        If PrevGroup = #0 Then PrevGroup := Group;
        { report hours each time project changes }
        If Project = PrevProject Then
          Begin
            ProjHours := ProjHours + CurrHour;
            GroupHours := GroupHours + CurrHour;
          End
        Else
          Begin
            If PrevProject <> '' Then
              Begin  { report in progress }
                ShowProjectHours;
                EventCnt := 0;
                { detect if group changed }
                { change of non-alpha, or change between alpha and non }
                If ( ( not ( PrevGroup in [ 'A'..'Z' ] ) ) and
                     ( not ( Group in [ 'A'..'Z' ] ) ) and
                     ( Group <> PrevGroup ) ) or
                   ( ( ( not ( PrevGroup in [ 'A'..'Z' ] ) ) and
                       ( Group in [ 'A'..'Z' ] ) ) or
                     ( ( PrevGroup in [ 'A'..'Z' ] ) and
                       ( not ( Group in [ 'A'..'Z' ] ) ) ) ) Then
                  Begin
                    ShowGroupHours;
                    GroupHours := CurrHour;
                  End
                Else
                  GroupHours := GroupHours + CurrHour;
              End;  { report in progress }
            If PrevProject = '' Then
              Begin
                PrevProject := Project;
                GroupHours := CurrHour;
              End;
            ProjName := FindProjName( Project );
            If IOError Then Exit;
            If ProjName = '' Then
              ProjName := 'UNKNOWN PROJECT (' + Project + ')';
            ProjName := Upper( Strip( ProjName ) );
            Writeln( ODv, ProjName, ':' );
            ChkErr( IOResult, 'writing report' );
            If IOError Then Exit;
            PrevProject := Project;  ProjHours := CurrHour;
            PrevGroup := Group;
          End;
        If not Terse Then
          Begin
            Writeln( ODv, OutLn );
            ChkErr( IOResult, 'writing report' );
            If IOError Then Exit;
          End;
        Inc( EventCnt );
        If CheckKey Then Exit;
      End;
    { finish report }
    ShowProjectHours;
    ShowGroupHours;
    Writeln( ODv, 'Work Hours  ------>', WorkHours:7:1, '   ',
                  ( ( WorkHours * 100 ) / TotalHours ):5:1,
                  ' % of total hours.', NL );
    ChkErr( IOResult, 'writing report' );
    If IOError Then Exit;
    Writeln( ODv, 'Total hours ------>', TotalHours:7:1, NL );
    ChkErr( IOResult, 'writing report' );
    If IOError Then Exit;
    { close output file }
    If ODvTYpe = 1 Then
      Begin  { formfeed }
        Write( ODv, ^L );
        ChkErr( IOResult, 'writing report' );
        If IOError Then Exit;
      End;
    Close( ODv );
    ChkErr( IOResult, 'closing report' );
    If IOError Then Exit;
    { erase temporary file }
    Close( SortFile );
    ChkErr( IOResult, 'closing sort' );
    If IOError Then Exit;
    Erase( SortFile );
    ChkErr( IOResult, 'erasing sort' );
    If IOError Then Exit;
  End;

{---------------------------------------------------------------------}

Procedure DoCutLog;
  Const
    InvalidFileNameCharSet : Set of Char =
      [ #0..#20, #128..#255, '.', '"', '/', '\', '[', ']',
        ':', '|', '<', '>', '+', '=', ';', ',' ];
  Var
    Found, UseCut, ToVideo, AnyFound : Boolean;
    TgtUser, TgtProject, TgtStartDate, TgtEndDate, SwapDate : String8;
    TgtNote : String80;
    SrcNm : String80;
    ScanLn : String;
    Result, ParamLoop : Integer;
    Ch : Char;
    CurrStat, PrevStat : String30;
    ShowStat : Boolean;
    SelectChar : Char;
  Begin
    ParamLoop := 2;  UseCut := False;  ToVideo := False;
    TgtUser := '';  TgtProject := '';  TgtNote := '';
    TgtStartDate := '';  TgtEndDate := '';
    { collect command line parameters }
    Repeat
      Found := False;
      If Upper( ParamArr[ ParamLoop ] ) = '/U' Then
        Begin
          Found := ( ParamLoop < ParamCnt ) and
                   ( Copy( ParamArr[ ParamLoop + 1 ], 1, 1 ) <> '/' );
          If Found Then
            Begin
              Inc( ParamLoop );
              TgtUser := Upper( ParamArr[ ParamLoop ] );
              If ( Length( TgtUser ) > 3 ) or
                 ( ( Length( TgtProject ) >= 1 ) and
                   ( not ( TgtUser[ 1 ] in [ 'A'..'Z' ] ) ) ) or
                 ( ( Length( TgtProject ) >= 2 ) and
                   ( not ( TgtUser[ 2 ] in [ 'A'..'Z' ] ) ) ) or
                 ( ( Length( TgtProject ) = 3 ) and
                   ( not ( TgtUser[ 3 ] in [ 'A'..'Z' ] ) ) ) Then
                Begin
                  ChkErr( -1, 'Invalid target user name "' +
                              TgtUser + '".' );
                  If IOError Then Exit;
                End;
            End;
        End
      Else
        If Upper( ParamArr[ ParamLoop ] ) = '/P' Then
          Begin
            Found := ( ParamLoop < ParamCnt ) and
                     ( Copy( ParamArr[ ParamLoop + 1 ], 1, 1 ) <> '/' );
            If Found Then
              Begin
                Inc( ParamLoop );
                TgtProject := Upper( ParamArr[ ParamLoop ] );
                If ( Length( TgtProject ) > 3 ) or
                   ( ( Length( TgtProject ) >= 1 ) and
                     ( TgtProject[ 1 ] in InvalidFileNameCharSet ) ) or
                   ( ( Length( TgtProject ) >= 2 ) and
                     ( TgtProject[ 2 ] in InvalidFileNameCharSet ) ) or
                   ( ( Length( TgtProject ) = 3 ) and
                     ( TgtProject[ 3 ] in InvalidFileNameCharSet ) ) Then
                  Begin
                    ChkErr( -1, 'Invalid target project name "' +
                                TgtProject + '".' );
                    If IOError Then Exit;
                  End;
              End;
          End
        Else
          If Upper( ParamArr[ ParamLoop ] ) = '/N' Then
            Begin
              Found := ( ParamLoop < ParamCnt ) and
                       ( Copy( ParamArr[ ParamLoop + 1 ], 1, 1 ) <> '/' );
              If Found Then
                Begin
                  Repeat
                    Inc( ParamLoop );
                    TgtNote := TgtNote + ' ' +
                               Upper( ParamArr[ ParamLoop ] );
                  Until ( ParamLoop >= ParamCnt ) or
                        ( Copy( ParamArr[ ParamLoop + 1 ], 1, 1 ) = '/' );
                  Delete( TgtNote, 1, 1 );  { remove first space }
                  If Copy( TgtNote, 1, 1 ) = '"' Then
                    Delete( TgtNote, 1, 1 );
                  If Copy( TgtNote, Length( TgtNote ), 1 ) = '"' Then
                    Delete( TgtNote, Length( TgtNote ), 1 );
                End;
            End
          Else
            If Upper( ParamArr[ ParamLoop ] ) = '/D' Then
              Begin
                Inc( ParamLoop );
                If ParamLoop <= ParamCnt Then
                  TgtStartDate := ParamArr[ ParamLoop ];
                If ( ParamLoop < ParamCnt ) and
                   ( Copy( ParamArr[ ParamLoop + 1 ], 1, 1 ) <> '/' ) Then
                  Begin
                    Inc( ParamLoop );
                    TgtStartDate := ParamArr[ ParamLoop ];
                  End;
                If ( TgtStartDate <> '' ) and
                   ( DTSec( TgtStartDate, '00:00:00' ) < 1.0 ) Then
                  Begin
                    ChkErr( -1, 'Bad start date' );
                    If IOError Then Exit;
                  End;
                If ( TgtEndDate <> '' ) and
                   ( DTSec( TgtEndDate, '00:00:00' ) < 1.0 ) Then
                  Begin
                    ChkErr( -1, 'Bad end date' );
                    If IOError Then Exit;
                  End;
                { swap dates if necessary }
                If DTSec( TgtEndDate, '00:00:00' ) <
                     DTSec( TgtStartDate, '00:00:00' ) Then
                  Begin
                    SwapDate := TgtEndDate;  TgtEndDate := TgtStartDate;
                    TgtStartDate := SwapDate;
                  End;
                Found := True;
              End
            Else
              If Upper( ParamArr[ ParamLoop ] ) = '/C' Then
                Begin
                  Found := True;  UseCut := True;
                End
              Else
                If Upper( ParamArr[ ParamLoop ] ) = '/V' Then
                  Begin
                    Found := True;  ToVideo := True;
                  End
                Else
                  Begin
                    Writeln( NL, 'Unknown option: "',
                                 Upper( ParamArr[ ParamLoop ] ), '"', NL,
                                 'Enter  DONE ?  for help.', NL );
                    Halt( 1 );
                  End;
      If not Found Then
        Begin
          Writeln( NL, 'Parameter missing following ',
                       Upper( ParamArr[ ParamLoop ] ), ' option.', NL,
                       'Enter  DONE ?  for help.', NL );
          Halt( 1 );
        End;
      Inc( ParamLoop );
    Until ParamLoop > ParamCnt;
    { open source file }
    If UseCut Then
      Begin
        SrcNm := 'CUT';
        Assign( SrcLog, CutLogSpec );
        SetTextBuf( SrcLog, SrcBuf );
      End
    Else
      Begin
        SrcNm := 'DONE';
        Assign( SrcLog, DoneLogSpec );
        SetTextBuf( SrcLog, SrcBuf );
      End;
    Reset( SrcLog );
    ChkErr( IOResult, 'opening ' + SrcNm );
    If IOError Then Exit;
    If ToVideo Then
      Begin  { to video }
        Assign( CutLog, '' );
        SetTextBuf( CutLog, CutBuf );
        ReWrite( CutLog );
        ChkErr( IOResult, 'opening output' );
        If IOError Then Exit;
      End  { to video }
    Else
      If UseCut Then
        Begin  { cut source }
          Assign( CutLog, 'DOCUT.$$$' );
          SetTextBuf( CutLog, CutBuf );
          ReWrite( CutLog );
          ChkErr( IOResult, 'opening CUT' );
          If IOError Then Exit;
        End  { cut source }
      Else
        Begin  { done source }
          Assign( CutLog, CutLogSpec );
          SetTextBuf( CutLog, CutBuf );
          Reset( CutLog );  Result := IOResult;
          If Result = 0 Then
            Begin  { file exists }
              SelectChar :=
                SelectCh( 'Cut file exists - ' +
                          'select: [O]verwrite/[A]ppend/[Q]uit',
                          'OAQ' + Esc + CtrlC );
              Case SelectChar of
                'Q', Esc, CtrlC :
                  Begin
                    Writeln( 'Quit' + NL + 'Cut cancelled.' );
                    Close( CutLog );  Result := IOResult;
                    Exit;
                  End;
                'O' :
                  Begin
                    Writeln( 'Overwriting' );
                    Close( CutLog );  Result := IOResult;
                    ReWrite( CutLog );
                    ChkErr( IOResult, 'opening cut' );
                    If IOError Then Exit;
                  End;
                'A' :
                  Begin
                    Writeln( 'Appending' );
                    Close( CutLog );  Result := IOResult;
                    Append( CutLog );
                    ChkErr( IOResult, 'appending cut' );
                    If IOError Then Exit;
                  End;
              End; { case }
            End  { file exists }
          Else
            Begin  { file is new }
              Close( CutLog );  Result := IOResult;
              Writeln( 'Creating cut file . . .' );
              ReWrite( CutLog );
              ChkErr( IOResult, 'opening cut' );
              If IOError Then Exit;
            End;  { file is new }
        End;  { done source source }
    { ready to scan }
    Write( 'Target' );
    If TgtUser <> '' Then Write( ' user: "', TgtUser, '"' );
    If TgtProject <> '' Then Write( ' project: "', TgtProject, '"' );
    If TgtNote <> '' Then Write( ' notes: "', TgtNote, '"' );
    If TgtStartDate <> '' Then
      Write( ' date: "', TgtStartDate, '"' );
    If TgtEndDate <> '' Then
      Write( '..', TgtEndDate, '"' );
    Writeln;
    Writeln( 'Scanning ' + SrcNm + ' log . . .' );
    Writeln( '(press any key to pause, <ESC> to terminate)' );
    PrevStat := '';  ShowStat := False;
    While not EOF( SrcLog ) do
      Begin
        Found := False;
        ReadLn( SrcLog, ScanLn );
        ChkErr( IOResult, 'reading ' + SrcNm );
        If IOError Then Exit;
        { show scan progress }
        CurrStat := Copy( ScanLn, 5, 11 );
        If CurrStat <> PrevStat Then
          Begin
            Write( #13, ' ', CurrStat, ':00' );
            PrevStat := CurrStat;
            ShowStat := True;
            If CheckKey Then Exit;
          End;
        If TgtUser <> '' Then
          Found := Pos( TgtUser, Upper( Copy( ScanLn, 1, 3 ) ) ) > 0;
        If ( not Found ) and ( TgtProject <> '' ) Then
          Found := Pos( TgtProject, Upper( Copy( ScanLn, 28, 3 ) ) ) > 0;
        If ( not Found ) and ( TgtNote <> '' ) Then
          Found := Pos( TgtNote, Upper( Copy( ScanLn, 31, 255 ) ) ) > 0;
        If ( not Found ) and
           ( TgtStartDate <> '' ) and ( TgtEndDate = '' ) Then
          Found := DTSec( TgtStartDate, '00:00:00' ) =
                   DTSec( Copy( ScanLn, 5, 8 ), '00:00:00' );
        If ( not Found ) and
           ( TgtStartDate <> '' ) and ( TgtEndDate <> '' ) Then
          Found := ( DTSec( TgtStartDate, '00:00:00' ) <=
                     DTSec( Copy( ScanLn, 5, 8 ), '00:00:00' ) ) and
                   ( DTSec( TgtEndDate, '00:00:00' ) >=
                     DTSec( Copy( ScanLn, 5, 8 ), '00:00:00' ) );
        If Found Then
          Begin
            If ToVideo Then Write( #13 );
            Writeln( CutLog, ScanLn );
            ChkErr( IOResult, 'writing output' );
            If IOError Then Exit;
          End;
        If ShowStat and ( not ToVideo ) Then
          If Found Then
            Write( ' *' )
          Else
            Write( '  ' );
        AnyFound := AnyFound or Found;
      End;
    Writeln;
    If not AnyFound Then
      Writeln( 'No matching entries found.' );
    Close( CutLog );
    ChkErr( IOResult, 'closing output' );
    If IOError Then Exit;
    Close( SrcLog );
    ChkErr( IOResult, 'closing ' + SrcNm );
    If IOError Then Exit;
    If UseCut and ( not ToVideo ) and AnyFound Then
      Begin
        Assign( CutLog, CutLogSpec );
        Erase( CutLog );  Result := IOResult;
        Assign( CutLog, 'DOCUT.$$$' );
        Rename( CutLog, CutLogSpec );
        ChkErr( IOResult, 'renaming DO' );
        If IOError Then Exit;
      End;
  End;

{---------------------------------------------------------------------}

Procedure WriteLog;
  Var
    Result, PLoop : Integer;
  Begin
    Assign( DoLog, DoLogSpec );
    SetTextBuf( DoLog, DoBuf );
    Append( DoLog );  Result := IOResult;
    If Result = 2 Then  { file not found }
      Begin  ReWrite( DoLog );  Result := IOResult;  End;
    ChkErr( Result, 'appending DO' );
    If IOError Then Exit;
    Write( DoLog, ' ', DateString, ' ', TimeString );
    ChkErr( Result, 'appending DO' );
    If IOError Then Exit;
    For PLoop := 2 to ParamCnt do
      Begin
        Write( DoLog, ' ', ParamArr[ PLoop ] );
        ChkErr( Result, 'appending DO' );
        If IOError Then Exit;
      End;
    Writeln( DoLog );
    ChkErr( Result, 'appending DO' );
    If IOError Then Exit;
    Close( DoLog );
    ChkErr( IOResult, 'closing DO' );
    If IOError Then Exit;
  End;

{---------------------------------------------------------------------}

Procedure Help;
  Begin
    Writeln( VerMsg );
    Writeln( 'Report and maintenance functions for time tracking system.' );
    Writeln( 'DONE [??]   No arguments for menu.  ?? for extended help.' );
    Writeln( 'DONE POST uuu [fn]' );
    Writeln( '  Post DO log to DONE log.' );
    Writeln( '    uuu   User identifier specifying originator of DO log.' );
    Writeln( '    fn    Alternate source log, default ' + DoLogSpec + '.' );
    Writeln( 'DONE REPORT yymmdd yymmdd [fn] [/T] [/C]' );
    Writeln( '  Generate report summarizing DONE log for period.' );
    Writeln( '    yymmdd   Start and end dates.' );
    Writeln( '    fn   Output destination.' );
    Writeln( '    /T   Terse report showing totals only.' );
    Writeln( '    /C   Report from CUT file instead of DONE log.' );
    Writeln( 'DONE LOG msg...' );
    Writeln( '  Append entry to DO log, include project identifier.' );
    Writeln( 'DONE CUT [/C] [/D dt [dt] ] [/N ...] [/P ppp] [/U uuu] [/V]' );
    Writeln( '  Excerpt from DONE log to CUT file.' );
    Writeln( '    /C   Cut from CUT file instead of DONE log.' );
    Writeln( '    /D   Cut entries for date yymmdd [through second date].' );
    Writeln( '    /N   Cut entries containing ... in notes section.' );
    Writeln( '    /P   Cut entries for project ppp.' );
    Writeln( '    /U   Cut entries for user uuu.' );
    Writeln( '    /V   Send cut entries to screen instead of file.' );
  End;

{---------------------------------------------------------------------}

Procedure Help2;
  Procedure DoPause;
    Var
      PauseKey : Char;
    Begin
      If ( ParamCount = 1 ) and ( ParamStr( 1 ) = '???' ) Then Exit;
      Write( 'PRESS ANY KEY . . .' );  PauseKey := GetKey;  Write( #13 );
    End;
  Begin
    Writeln( VerMsg );
    Writeln( 'This program is intended for use with the program DO' );
    Writeln( 'which is a resident time logging utility.' );
    Writeln( 'Together these perform time tracking and reporting.' );
    Writeln( 'If one uses DO faithfully, DONE can then be used' );
    Writeln( 'weekly to generate reports of total time by project' );
    Writeln( 'as well as keep diaries by project.  In service' );
    Writeln( 'industries such as programming, law, or medicine, these' );
    Writeln( 'programs can be very useful, as they combine ease of' );
    Writeln( 'data entry, report generation for billing, and diary' );
    Writeln( 'maintenance.  If one finds the DO program too limited,' );
    Writeln( 'the DO log can be maintained using any simple text editor' );
    Writeln( 'and still be manipulated with this DONE program.' );
    Writeln;  DoPause;
    Writeln( 'DO - generates a log named "', DoLogSpec, '" in the form' );
    Writeln( '  YY/MM/DD HH:MM:SS PPP Notes . . .' );
    Writeln( 'where PPP is a project or customer identifier.' );
    Writeln( 'DO shows a simple message when run.  It becomes resident' );
    Writeln( 'and linked to the keyboard by displacing interrupt 16h.' );
    Writeln( 'It cannot be unloaded without rebooting.  The hot key' );
    Writeln( 'and file name are user specifiable on the command line:' );
    Writeln( 'DO [c [fn]]  Where c is 0..9, A..Z for alt key.' );
    Writeln( 'DO can be hot key activated at any time (Alt-0 default).' );
    Writeln( 'Once activated, type PPP, space, notes, and press <enter>' );
    Writeln( 'to make a log entry.   PPP must be ; or three characters.'  );
    Writeln( 'Using ; in place of PPP assumes same PPP as prior entry.' );
    Writeln( 'Press <esc> to cancel an entry, <enter> to finish,' );
    Writeln( '<backspace> to make changes, or <ctrl>-<enter> to continue' );
    Writeln( 'an entry to the next line.  DO uses about 7K of memory.' );
    Writeln( 'DO only displays text entry line if video is in a text mode.' );
    Writeln;  DoPause;
    Writeln( 'DONE POST - processes the DO log, inserting elapsed times' );
    Writeln( 'and user identifier.' );
    Writeln( 'The lines are removed from the DO log, and appended' );
    Writeln( 'to a file named "', DoneLogSpec, '", short for done log.' );
    Writeln( 'Each line must contain a project identifier PPP.' );
    Writeln( '; in place of PPP is converted to same PPP as prior entry.' );
    Writeln( 'All output files are generated to current directory.' );
    Writeln( 'The last entry in the DO log is not posted, as the elapsed' );
    Writeln( 'time cannot be determined until another entry is made.' );
    Writeln( 'You will be prompted to enter the full name for projects' );
    Writeln( 'not previously used.  Project names are stored in' );
    Writeln( ProjNameFileSpec, ' file.  You will also be asked to confirm' );
    Writeln( 'entries having elapsed times greater than ',
             MaxSeqHours, ' hours.' );
    Writeln( 'Note that if corrections must be made, a done log can be' );
    Writeln( 'renamed and re-posted as if it was a do log, but do not' );
    Writeln( 'attempt this with a done log containing multiple users' );
    Writeln( 'as the elapsed time calculations will be incorrect.' );
    Writeln;  DoPause;
    Writeln( 'DONE REPORT - summarizes the DONE or CUT log by project' );
    Writeln( 'showing total times, percentages, and averages.' );
    Writeln( 'First character of PPP, if non A..Z, indicates a group for' );
    Writeln( 'combined subtotaling.  Group ^ is excluded from work hours,' );
    Writeln( 'use for day in, day out, break, sick, and lunch entries.' );
    Writeln( '(/X=Y@Z) on totals lines indicates X occurences, averaging' );
    Writeln( 'Y hours each, at an average of once every Z hours.' );
    Writeln;  DoPause;
    Writeln( 'LOG - writes entries to DO log.  Useful if you cannot load' );
    Writeln( 'the DO program, or for logging entries from batch files.' );
    Writeln;
    Writeln( 'DONE CUT - excerpts entries for more specific reports,' );
    Writeln( 'or find specific entries.  Scans the DONE log for entries' );
    Writeln( 'by a certain user, for a given project, containing' );
    Writeln( 'specified text, or any combination of these.' );
    Writeln( 'Scan is case insensitive.  After cutting, use REPORT to' );
    Writeln( 'view data for selected date range.  Use /C option to cut' );
    Writeln( 'from CUT file, creating a compound cut.  When searching' );
    Writeln( 'contents of notes section, enclose target in quotes "..."' );
    Writeln( 'to include leading or trailing spaces in search.' );
    Writeln( 'Entries are cut to a file named "', CutLogSpec, '".' );
    Writeln;  DoPause;
    Writeln( 'This program, DONE, and its partner program, DO, may be' );
    Writeln( 'freely distributed in accordance with the following terms:' );
    Writeln;
    Writeln( '1. They may be distributed only in unmodified form.' );
    Writeln( '2. Only the executable files, documentation, and sample' );
    Writeln( '   data may be distributed, not the source code.' );
    Writeln( '3. No charge may be made for the programs, beyond' );
    Writeln( '   reasonable charge for media and shipping if applicable.' );
    Writeln( '4. The programs may not be included with other software' );
    Writeln( '   not distributed on a similar free basis.' );
    Writeln;
    Writeln( 'For more information, contact author:' );
    Writeln( '   Andrew Bedno, 7319 N. Ashland, Chicago IL 60626.' );
    Writeln;
  End;

{-------------------------------------------------------------------------}

Procedure DoMenu;
  Var
    InpLn : String80;
    Menu, SubMenu, PauseKey : Char;
  Procedure DoCase;
    Var
      Parse : Byte;
    Begin
      FillChar( ParamArr, SizeOf( ParamArr ), 0 );
      ParamCnt := 0;
      Case Menu of
        '?' : 
          Begin
            Writeln( '*****  BASIC HELP  *****' );
            Help;
          End;
        '!' :
          Begin
            Writeln( '*****  EXTENDED HELP  *****' );
            Help2;
          End;
        'P' :
          Begin
            Writeln( '*****  POST  *****' );
            Inc( ParamCnt );
            ParamArr[ ParamCnt ] := 'POST';
            Write( 'Enter user identifier > ' );
            Inc( ParamCnt );
            ReadLn( ParamArr[ ParamCnt ] );
            If ParamArr[ ParamCnt ] = '' Then Exit;
            Write( 'Enter source log (<enter>=', DoLogSpec, ') > ' );
            ReadLn( ParamArr[ 3 ] );
            ParamArr[ 3 ] := Strip( ParamArr[ 3 ] );
            If ParamArr[ 3 ] <> '' Then Inc( ParamCnt );
            PostLog;
          End;
        'C' :
          Begin  { cut }
            Writeln( '*****  CUT  *****' );
            Inc( ParamCnt );
            ParamArr[ ParamCnt ] := 'CUT';
            SubMenu :=
              SelectCh( 'Press  Y  to cut from existing cut file', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/C';
              End
            Else
              Writeln( 'No' );
            SubMenu :=
              SelectCh( 'Press  Y  to search in dates section', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/D';
                Write( 'Enter starting date to search for (YYMMDD) > ' );
                Inc( ParamCnt );
                ReadLn( ParamArr[ ParamCnt ] );
                If ParamArr[ ParamCnt ] = '' Then Exit;
                Write( 'Enter ending date to search for (YYMMDD) > ' );
                Inc( ParamCnt );
                ReadLn( ParamArr[ ParamCnt ] );
                If ParamArr[ ParamCnt ] = '' Then Exit;
              End
            Else
              Writeln( 'No' );
            SubMenu :=
              SelectCh( 'Press  Y  to search in notes section', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/N';
                Write( 'Enter notes to search for > ' );
                Inc( ParamCnt );
                ReadLn( ParamArr[ ParamCnt ] );
                If ParamArr[ ParamCnt ] = '' Then Exit;
              End
            Else
              Writeln( 'No' );
            SubMenu :=
              SelectCh( 'Press  Y  to search by project', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/P';
                Write( 'Enter project to search for > ' );
                Inc( ParamCnt );
                ReadLn( ParamArr[ ParamCnt ] );
                If ParamArr[ ParamCnt ] = '' Then Exit;
              End
            Else
              Writeln( 'No' );
            SubMenu :=
              SelectCh( 'Press  Y  to search by user', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/U';
                Write( 'Enter user to search for > ' );
                Inc( ParamCnt );
                ReadLn( ParamArr[ ParamCnt ] );
                If ParamArr[ ParamCnt ] = '' Then Exit;
              End
            Else
              Writeln( 'No' );
            SubMenu :=
              SelectCh( 'Press  Y  to output to screen', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/V';
              End
            Else
              Writeln( 'No' );
            DoCutLog;
          End;  { cut }
        'R' :
          Begin  { report }
            Writeln( '*****  REPORT  *****' );
            Inc( ParamCnt );
            ParamArr[ ParamCnt ] := 'REPORT';
            Write( 'Enter report start date (YYMMDD) > ' );
            Inc( ParamCnt );
            ReadLn( ParamArr[ ParamCnt ] );
            If ParamArr[ ParamCnt ] = '' Then Exit;
            Write( 'Enter report end date (YYMMDD) > ' );
            Inc( ParamCnt );
            ReadLn( ParamArr[ ParamCnt ] );
            If ParamArr[ ParamCnt ] = '' Then Exit;
            Write( 'Enter output file name ' +
                   '(PRN=Printer <ENTER>=screen) > ' );
            InpLn := '';  ReadLn( InpLn );
            If InpLn <> '' Then
              Begin
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := InpLn;
              End
            Else
              Writeln( 'Screen' );
            SubMenu :=
              SelectCh( 'Press letter to select:' +
                        '  <S>ummary or <D>etailed report',
                        'SDQ' + Esc + CtrlC );
            If SubMenu = 'S' Then
              Begin
                Writeln( 'Summary' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/T';
              End
            Else
              Writeln( 'Detailed' );
            SubMenu :=
              SelectCh( 'Press  Y  to report from CUT file', #0 );
            If SubMenu = 'Y' Then
              Begin
                Writeln( 'Yes' );
                Inc( ParamCnt );
                ParamArr[ ParamCnt ] := '/C';
              End
            Else
              Writeln( 'No' );
            ReportLog( ParamArr[ 2 ], ParamArr[ 3 ] )
          End;  { report }
        'L' :
          Begin  { log }
            Writeln( '*****  LOG  *****' );
            Inc( ParamCnt );
            ParamArr[ ParamCnt ] := 'LOG';
            Write( 'Enter message to log > ' );
            ReadLn( InpLn );
            If InpLn = '' Then Exit;
            InpLn := Strip( InpLn );
            Inc( ParamCnt );
            For Parse := 1 to Length( InpLn ) do
              Begin
                If ( InpLn[ Parse ] = ' ' ) and
                   ( ( Parse = 1 ) or
                     ( ( Parse > 1 ) and ( InpLn[Parse-1] <> ' ' ) ) ) Then
                  Inc( ParamCnt )
                Else
                  ParamArr[ ParamCnt ] :=
                    ParamArr[ ParamCnt ] + InpLn[ Parse ];
              End;
            WriteLog;
          End;  { log }
        Else
          Begin
            Writeln( 'Quit' );
            Halt( 0 );
          End;
      End;  { case }
      IOError := False;
      Write( '****  PRESS ANY KEY  ****' );
      PauseKey := GetKey;
      Write( #13, '                         ', #13 );
    End;
  Begin
   FromMenu := True;
   Repeat
      Writeln( NL, VerMsg, NL, 'TIME LOG PROCESSOR', NL );
      Writeln( '*****  MAIN MENU  *****' );
      Menu := SelectCh( 'Select operation:' + NL +
                        '  <?>  -  basic help' + NL +
                        '  <!>  -  extended help' + NL +
                        '  <P>  -  post' + NL +
                        '  <C>  -  cut' + NL +
                        '  <R>  -  report' + NL +
                        '  <L>  -  log' + NL +
                        '  <ESC>  -  quit' + NL,
                        '?!PCLR' + Esc + CtrlC );
      Write( #13 );
      DoCase;
    Until Menu in [ Esc, CtrlC ];
    Writeln( NL, 'DONE!' );
  End;

{-------------------------------------------------------------------------}

Begin

  ExitSave := ExitProc;
  ExitProc := @TidyExit;

  For CpyRtLp := 1 to 26 do
    If CpyRtLp = 1 Then
      CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
    Else
      CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                               Ord( CpyRt[ CpyRtLp - 1 ] ) );
  VerMsg := 'DONE ver 3.1 rev 91/06/29 - ' + CpyRt + '.';

  If ParamCount = 0 Then DoMenu;

  While ( ParamCnt < ParamCount ) and ( ParamCnt < MaxParam ) do
    Begin
      Inc( ParamCnt );
      ParamArr[ ParamCnt ] := ParamStr( ParamCnt );
    End;

  If ( ParamArr[ 1 ] = '??' ) or ( ParamArr[ 1 ] = '???' ) Then
    Begin
      Help2;
      Halt( 1 );
    End;

  If ( ParamArr[ 1 ] = '?' ) or
     ( ( Upper( ParamArr[ 1 ] ) <> 'POST' ) and
       ( Upper( ParamArr[ 1 ] ) <> 'REPORT' ) and
       ( Upper( ParamArr[ 1 ] ) <> 'LOG' ) and
       ( Upper( ParamArr[ 1 ] ) <> 'CUT' ) ) or
     ( ( Upper( ParamArr[ 1 ] ) = 'POST' ) and
       ( ( ParamCnt < 1 ) or ( ParamCnt > 3 ) ) ) or
     ( ( Upper( ParamArr[ 1 ] ) = 'LOG' ) and ( ParamCnt < 2 ) ) or
     ( ( Upper( ParamArr[ 1 ] ) = 'CUT' ) and ( ParamCnt < 2 ) ) or
     ( ( Upper( ParamArr[ 1 ] ) = 'REPORT' ) and
       ( ( ParamCnt < 3 ) or ( ParamCnt > 6 ) ) ) Then
    Begin
      Help;
      Halt( 1 );
    End;

  Writeln( NL, VerMsg, NL, 'TIME LOG PROCESSOR', NL );

  If Upper( ParamArr[ 1 ] ) = 'POST' Then
    PostLog
  Else
    If Upper( ParamArr[ 1 ] ) = 'REPORT' Then
      ReportLog( ParamArr[ 2 ], ParamArr[ 3 ] )
    Else
      If Upper( ParamArr[ 1 ] ) = 'CUT' Then
        DoCutLog
      Else
        WriteLog;

  Writeln( 'DONE!' );

End.


