{$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 1024,0,110000}

USES Dos;

CONST
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;
  MaxFiles = 5000;
  MaxPaths = 500;
  PathNameSize = 80;
  NL = #13#10;

TYPE
  String2 = String[ 2 ];
  PathString = String[ PathNameSize ];
  FileDataRec =
    Record  PathNum:Byte;  Size:LongInt;  Time:LongInt;  End;
  FileNameArray = Array [ 1..MaxFiles ] of String[12];
  FileDataArray = Array [ 1..MaxFiles ] of FileDataRec;

VAR
  FileName : ^FileNameArray;
  FileData : ^FileDataArray;
  PathName : Array [ 1..MaxPaths ] of PathString;
  Result:Integer;  PrevLineLength:Byte;
  FileLoop, Loop, Files, DelCount, Paths, Drive : Word;
  Mask, PureMask, PureParam, Path, OldMsg, GloMsg : PathString;
  Silent, Global, VeryTerse, Terse, ListOnly, Individual,
  ExactAge, CheckAge, ReStart : Boolean;
  OldAge, DelSize, TotalSize, RamRequired : LongInt;

{-------------------------------------------------------------------------}

FUNCTION Pad( S:String; N:Integer ):String;  { N<0 = LftJust, N>0 = RtJust }
  BEGIN
    IF Length( S ) > Abs( N )
      THEN S[ 0 ] := Chr( Abs( N ) )
      ELSE WHILE Length( S ) < Abs( N ) DO
             IF N < 0 THEN S := S + ' ' ELSE S := ' ' + S;
    Pad := S;
  END;

{-------------------------------------------------------------------------}

PROCEDURE Show( Msg:String; Removable:Boolean );
  VAR  BackCount:Byte;
  BEGIN
    If PrevLineLength > 0
      THEN Write( #13, Pad( ' ', PrevLineLength ), #13 );
    If Removable
      THEN PrevLineLength := Length( Msg ) ELSE PrevLineLength := 0;
    Write( Msg );
  END;

{-------------------------------------------------------------------------}

PROCEDURE Stop( Msg:String );
  BEGIN  Show( Msg + '.' + NL, False );  Halt(1);  END;

{-------------------------------------------------------------------------}

PROCEDURE CheckForBreak;
  VAR  Regs:Registers;
  BEGIN
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $0B;  { ConInputStatus }
    MsDos( Regs );
  END;

{-------------------------------------------------------------------------}

FUNCTION GetKey:Char;
  VAR  Regs:Registers;
  BEGIN
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $08;  { ConInputNoEcho }
    MsDos( Regs );
    GetKey := UpCase( Chr( Regs.AL ) );
  END;

{-------------------------------------------------------------------------}

PROCEDURE ParseMask( Spec:String );
 { Return file name without path or drive. }
  VAR  Temp:PathString;  ParseLoop:Word;
  CONST  BadChars : Set of Char =
           [ '"', '/', '\', '[', ']', ':', '|', '<', '>', '+', '=',
             ';', ',', #0..#20 ];  { invalid filename characters }
  BEGIN
    ParseLoop := Length( Spec );
    While ( ParseLoop > 0 ) and
          ( not ( Spec[ ParseLoop ] in BadChars ) ) do Dec( ParseLoop );
    Mask := Copy( Spec, ParseLoop + 1, 255 );
    If ParseLoop > 0
      THEN Path := Copy( Spec, 1, ParseLoop )
      ELSE BEGIN
        GetDir( 0, Temp );  Result := IOResult;
        If Result = 0 THEN Path := Temp + '\';
      END;
    If ( Length( Path ) = 2 ) and ( Path[ 2 ] = ':' )
      THEN BEGIN
        GetDir( Ord( Path[ 1 ] ) - 64, Temp );
        Result := IOResult;  If Result = 0 THEN Path := Temp + '\';
      END;
    If ( Length( Path ) > 2 ) and
       ( Copy( Path, Length( Path ) - 1, 2 ) = '\\' )
      THEN Delete( Path, Length( Path ), 1 );
    If ( Mask <> '' ) and ( Mask[ Length( Mask ) ] = '\' )
      THEN Delete( Mask, Length( Mask ), 1 );
    PureMask := Mask;
    If Mask = '.' THEN Mask := '*.*';
    If ( Length( Mask ) > 1 ) and
       ( Mask[ 1 ] = '.' ) and ( Mask[ 2 ] <> '.' )
      THEN Mask := '*' + Mask;
  END;

{-------------------------------------------------------------------------}

FUNCTION FileCount( Mask:String ):Word;
  VAR  Dir:SearchRec;  Count:Word;
  BEGIN
    Count := 0;
    FindFirst( Mask, Archive, Dir );
    While DosError = 0 do
      BEGIN
        Inc( Count );
        FindNext( Dir );
      END;
    FileCount := Count;
  END;

{-------------------------------------------------------------------------}

FUNCTION DateInDays( Year, Month, Day : Word ):LongInt; { days since 1-1-80 }
  CONST  DaysPerMonth : Array[ 1..12 ] of Byte =
           ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  VAR  YearLoop, MonthLoop : Integer;  Total : LongInt;
  BEGIN
    If Year < 80 then
      BEGIN Year := 80; Month := 1; Day := 1; END;
    YearLoop := 80;  MonthLoop := 1;  Total := Day;
    While YearLoop < Year do
      BEGIN
        If YearLoop mod 4 = 0 then Total := Total + 366
          ELSE Total := Total + 365;
        YearLoop := YearLoop + 1;
      END;
    While MonthLoop < Month do
      BEGIN
        Total := Total + DaysPerMonth[ MonthLoop ];
        If ( Year mod 4 = 0 ) and ( MonthLoop = 2 )
          then Total := Total + 1;
        MonthLoop := MonthLoop + 1;
      END;
    DateInDays := Total;
  END;

{-------------------------------------------------------------------------}

FUNCTION AgeOK( InPack:LongInt ):Boolean;
  VAR  DT:DateTime;  AgeTemp:LongInt;
  BEGIN
    UnPackTime( InPack, DT );
    AgeTemp := DateInDays( DT.Year, DT.Month, DT.Day );
    AgeOK := ( not CheckAge ) or
             ( ExactAge and ( AgeTemp = OldAge ) ) or
             ( ( not ExactAge ) and
               ( ( ( OldAge >= 0 ) and ( AgeTemp < OldAge ) ) or
                 ( ( OldAge < 0 ) and ( AgeTemp > Abs( OldAge ) ) ) ) );
  END;

{-------------------------------------------------------------------------}

PROCEDURE GetPaths;
  VAR  Dir:SearchRec;  ScanPath:Integer;
  BEGIN
    ScanPath := 1;
    REPEAT
      CheckForBreak;
      FindFirst( PathName[ ScanPath ] + '*.*', Directory, Dir );
      While DosError = 0 do
        BEGIN
          If ( ( Dir.Attr and Directory ) <> 0 ) and
             ( Dir.Name <> '.' ) and ( Dir.Name <> '..' )
            THEN BEGIN
              If Paths >= MaxPaths
                THEN Stop( 'Too many paths.' );
              Inc( Paths );
              If PathNameSize <
                 ( Length( PathName[ ScanPath ] ) + Length( Dir.Name ) + 1 )
                THEN Stop( 'Path name too long.' );
              PathName[ Paths ] := PathName[ ScanPath ] + Dir.Name + '\';
            END;
          FindNext( Dir );
        END;
      Inc( ScanPath );
    UNTIL ScanPath > Paths;
  END;

{-------------------------------------------------------------------------}

PROCEDURE GetFileData( SearchPathNum:Byte;  SearchMask:String );
  VAR  Dir:SearchRec;
  BEGIN
    FindFirst( PathName[ SearchPathNum ] + SearchMask, Archive, Dir );
    While DosError = 0 do
      BEGIN
        CheckForBreak;
        If AgeOK( Dir.Time )
          THEN BEGIN
            If Files >= MaxFiles
              THEN Stop( 'Too many files matching: "'
                         + Path + Mask + '"' + OldMsg + GloMsg );
            Inc( Files );  TotalSize := TotalSize + Dir.Size;
            FileName^[ Files ] := Dir.Name;
            FileData^[ Files ].PathNum := SearchPathNum;
            FileData^[ Files ].Size := Dir.Size;
            FileData^[ Files ].Time := Dir.Time;
          END;
        FindNext( Dir );
      END;
  END;

{-------------------------------------------------------------------------}

PROCEDURE SortFileData;
  VAR  NameSwap:String[12];  DataSwap:FileDataRec;
       Done:Boolean;  SortLoop, DelLoop, Base:Word;
       Gap, Index, Jump:Integer;  First, Next:String;
  BEGIN
    If not Terse THEN Show( 'Sorting . . .', True );
    Base := 1;
    REPEAT  { Remove duplicates. }
      CheckForBreak;
      Done := True;  SortLoop := Base + 1;
      While ( SortLoop < Files ) and Done do
        BEGIN
          First := PathName[ FileData^[ Base ].PathNum ] + FileName^[ Base ];
          Next := PathName[ FileData^[ SortLoop ].PathNum ] +
                    FileName^[ SortLoop ];
          Done := ( First <> Next ) and ( FileName^[ SortLoop ] <> '' );
          Inc( SortLoop );
        END;
      If Base + 1 >= Files
        THEN Done := True
        ELSE If Done
               THEN BEGIN  Inc( Base );  Done := Base >= Files;  END
               ELSE BEGIN
                 Dec( Files );
                 For DelLoop := SortLoop to Files do
                   FileName^[ DelLoop ] := FileName^[ DelLoop + 1 ];
                 For DelLoop := SortLoop to Files do
                   FileData^[ DelLoop ] := FileData^[ DelLoop + 1 ];
               END;
    UNTIL Done;  { Remove duplicates. }
    If Files > 1
      THEN BEGIN  { sort }
        Gap := Files div 2;
        While Gap > 0 do
          BEGIN  { gap loop }
            CheckForBreak;
            For Index := Gap + 1 to Files do
              BEGIN  { index loop }
                Jump := Index - Gap;
                While Jump > 0 do
                  BEGIN  { jump loop }
                    First := PathName[ FileData^[ Jump ].PathNum ] +
                               FileName^[ Jump ];
                    Next := PathName[ FileData^[ Jump + Gap ].PathNum ] +
                               FileName^[ Jump + Gap ];
                    If First > Next
                      THEN BEGIN
                        NameSwap := FileName^[ Jump ];
                        DataSwap := FileData^[ Jump ];
                        FileName^[ Jump ] := FileName^[ Jump + Gap ];
                        FileData^[ Jump ] := FileData^[ Jump + Gap ];
                        FileName^[ Jump + Gap ] := NameSwap;
                        FileData^[ Jump + Gap ] := DataSwap;
                        Jump := Jump - Gap;
                      END ELSE Jump := 0;
                  END;  { jump loop }
              END;  { index loop }
            Gap := Gap div 2;
          END;  { gap loop }
      END;  { sort }
  END;

{-------------------------------------------------------------------------}

FUNCTION IntString( InNum:LongInt ):String;
  VAR  S:String[40];
  BEGIN  Str( InNum, S );  IntString := S;  END;

{-------------------------------------------------------------------------}

PROCEDURE EatKeys;
  VAR  Regs:Registers;
  BEGIN
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $0C;  { ConInputFlush }
    MsDos( Regs );
  END;

{-------------------------------------------------------------------------}

FUNCTION Str2( InNum:Integer; PadCh:Char ):String2;
  VAR  Temp : String[2];
  BEGIN
    Str( Abs( InNum ), Temp );
    If Length( Temp ) < 2
      THEN Temp := PadCh + Temp;
    Str2 := Temp;
  END;

{-------------------------------------------------------------------------}

FUNCTION UnPackedTime( InPack:LongInt ):String;
  VAR  DT:DateTime;
  BEGIN
    UnPackTime( InPack, DT );
    UnPackedTime := Str2( DT.Month, ' ' ) + '-' + Str2( DT.Day, '0' ) +
                    '-' + Str2( DT.Year mod 100, '0' ) + ' ' +
                    Str2( DT.Hour, ' ' ) + ':' + Str2( DT.Min, '0' ) +
                    ':' + Str2( DT.Sec, '0' );
  END;

{-------------------------------------------------------------------------}

PROCEDURE ShowDirectory;
  VAR  Ch:Char;  ShowCount, DirLoop:Integer;  DirRow:Byte;  OutName:String;
  BEGIN
    DirRow := 1;  DirLoop := 1;  ShowCount := 1;
    If not Terse
      THEN Show( IntString( Files - DelCount ) + ' FILE(S) MATCHING "' +
                 Path + Mask + '"' + OldMsg + GloMsg + ':' + NL, False );
    REPEAT
      CheckForBreak;
      If FileName^[ DirLoop ] <> ''
        THEN BEGIN
          If Global
            THEN OutName := '   ' +
                   UnPackedTime( FileData^[ DirLoop ].Time ) +
                   Pad( IntString( FileData^[ DirLoop ].Size ), 7 ) +
                   ' ' + PathName[ FileData^[ DirLoop ].PathNum ] +
                   FileName^[ DirLoop ]
            ELSE OutName := '   ' +
                   Pad( FileName^[ DirLoop ], -13 ) +
                   Pad( IntString( FileData^[ DirLoop ].Size ), 7 ) + ' ' +
                   Copy( UnPackedTime( FileData^[ DirLoop ].Time ), 1, 14 );
          If VeryTerse
            THEN Show( PathName[ FileData^[ DirLoop ].PathNum ] +
                       FileName^[ DirLoop ] + NL, False )
            ELSE Show( OutName, False );
          Inc( ShowCount );
          If ( Terse or Odd( ShowCount ) or Global ) and ( not VeryTerse )
            THEN BEGIN  Inc( DirRow );  Show( NL, False );  END;
        END;
      Inc( DirLoop );
      If ( not Terse ) and
         ( ( DirRow >= 24 ) or
           ( ListOnly and ( ShowCount > Files ) and ( DirRow >= 20 ) ) )
        THEN BEGIN
          If DirLoop < Files
            THEN Show( 'PRESS <ESC> TO FINISH DIRECTORY ' +
                       'OR ANY OTHER KEY TO CONTINUE . . .', True )
            ELSE Show( 'PRESS ANY KEY TO FINISH DIRECTORY . . .', True );
          Ch := UpCase( GetKey );
          If Ch = #27 THEN DirLoop := Files + 1;
          DirRow := 1;  Ch := ' ';
        END;
    UNTIL DirLoop > Files;
    If ( not Terse ) and ( not Odd( ShowCount ) ) and ( not Global )
      THEN Show( NL, False );
    If not Terse THEN Show( NL, False );
  END;

{-------------------------------------------------------------------------}

PROCEDURE Finished( Done:Boolean );
  BEGIN
    If Done THEN Write( 'Done . . .  ' );
    Write( DelCount );
    If DelCount <> Files THEN Write( ' of ', Files );
    Write( ' file(s) deleted.  ', DelSize );
    If DelSize <> TotalSize THEN Write( ' of ', TotalSize );
    Writeln( ' bytes freed.' );
  END;

{-------------------------------------------------------------------------}

PROCEDURE EraseFile( EraseFileName:String );
  VAR  ScrapFile:File;  Ch:Char;  FileSize:LongInt;  Msg:String;
  BEGIN
    FileSize := FileData^[ FileLoop ].Size;
    Assign( ScrapFile, EraseFileName );  Ch := 'Y';
    If Individual
      THEN BEGIN
        REPEAT
          Msg := 'Delete "' + EraseFileName + '" <Y/N/';
          If Files - DelCount > 1 THEN Msg := Msg + 'A/D/';
          If ( FileLoop > 1 ) and ( Files - DelCount > 1 )
            THEN Msg := Msg + 'P/';
          Show( Msg + 'Q> ? ', True );
          REPEAT
            Ch := UpCase( GetKey );
            If ( Ch = 'P' ) and
               ( ( FileLoop <= 1 ) or ( Files - DelCount <= 1 ) )
              THEN Ch := ' ';
            If ( Ch in [ 'A', 'D' ] ) and ( Files - DelCount <= 1 )
              THEN Ch := ' ';
            If not ( Ch in [ 'Y', 'N', 'D', 'A', 'P', 'Q' ] )
              THEN Write( ^G );
          UNTIL Ch in [ 'Y', 'N', 'D', 'A', 'P', 'Q' ];
          If Ch <> 'D' THEN PrevLineLength := 0;
          Case Ch of
            'Y' : Show( 'confirmed.' + NL, False );
            'N' : BEGIN  Show( 'cancelled.' + NL, False );  Exit;  END;
            'Q' : BEGIN
                    Show( 'quit.' + NL, False );  Finished( False );  Halt;
                  END;
            'D' : ShowDirectory;
            'A' : BEGIN
                    Show( 'all mode . . .' + NL, False );
                    Individual := False;  Restart := True;
                    FileLoop := 0;
                    Exit;
                  END;
            'P' : BEGIN
                    Show( 'previous . . .' + NL, False );
                    If ( FileLoop > 1 ) and ( Files - DelCount > 1 )
                      THEN BEGIN
                        If FileName^[ FileLoop - 1 ] <> ''
                          THEN FileLoop := FileLoop - 2
                          ELSE BEGIN
                          { skip over already deleted }
                            Dec( FileLoop );
                            While ( FileName^[ FileLoop ] = '' ) and
                                  ( FileLoop > 1 ) do Dec( FileLoop );
                            Dec( FileLoop );
                          END;
                      END;
                    Exit;
                  END;
           End;
        UNTIL Ch = 'Y';
      END;
    Erase( ScrapFile );
    Result := IOResult;
    If Result = 0
      THEN BEGIN
        Inc( DelCount );  DelSize := DelSize + FileSize;
        If not Individual
          THEN Show( 'Deleted "' + EraseFileName + '"' + NL, False );
        FileName^[ FileLoop ] := '';
      END
      ELSE Show( 'Cannot delete "' + EraseFileName + '"' + NL, False );
  END;

{-------------------------------------------------------------------------}

PROCEDURE Confirm( Msg:String );
  VAR  Ch:Char;  FilesMsg:String[4];
  BEGIN
    REPEAT
      If Files > 1 THEN FilesMsg := '/I/D' ELSE FilesMsg := '';
      Show( Msg + ' <Y/N' + FilesMsg + '> ??? ', True );
      REPEAT
        Ch := UpCase( GetKey );
        If ( not ( Ch in [ 'Y', 'N', 'I', 'D' ] ) ) or
           ( ( Files < 2 ) and ( Ch in [ 'I', 'D' ] ) ) THEN Write( ^G );
      UNTIL ( Ch in [ 'Y', 'N' ] ) or
            ( ( Files > 1 ) and ( Ch in [ 'I', 'D' ] ) );
      If Ch = 'N' THEN Stop( 'No files deleted' );
      If Ch = 'D' THEN ShowDirectory;
      If Ch = 'I' THEN Individual := True;
    UNTIL Ch in [ 'Y', 'I' ];
  END;

{-------------------------------------------------------------------------}

PROCEDURE FigureAgeLimit( VAR AgeIn:LongInt );
  VAR  Year, Month, Day, DayOfWeek : Word;  AgeTemp : LongInt;
  BEGIN
    If ExactAge THEN AgeIn := Abs( AgeIn );
    GetDate( Year, Month, Day, DayOfWeek );
    AgeTemp := DateInDays( Year, Month, Day );
    AgeTemp := AgeTemp - Abs( AgeIn );
    If AgeIn < 0 THEN AgeTemp := 0 - AgeTemp;
    If ExactAge
      THEN OldMsg := ', =' + IntString( AgeIn ) + ' days'
      ELSE If AgeIn < 0
             THEN OldMsg := ', <' + IntString( Abs( AgeIn ) ) + ' days'
             ELSE OldMsg := ', >' + IntString( AgeIn ) + ' days';
    AgeIn := AgeTemp;
  END;

{-------------------------------------------------------------------------}

PROCEDURE ShowHelp;
  BEGIN
    For CpyRtLp := 1 to 26 do
      If CpyRtLp = 1 Then
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
      Else
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                 Ord( CpyRt[ CpyRtLp - 1 ] ) );
    Writeln( 'DOSRM ver 3.1 rev 91/06/29 - ', CpyRt, '.' );
    Writeln( 'Enhanced file delete (DEL) utility.' );
    Writeln( '  Provides group or individual confirmations,' );
    Writeln( '    recursive subdirectory scanning, and age options.' );
    Writeln( '  Prompts for confirmation before any deletions.' );
    Writeln( '  Prompts twice to confirm for whole directory delete' );
    Writeln( '    (if search is not global).' );
    Writeln( 'RM filename [options]' );
    Writeln( 'Options:' );
    Writeln( '  /G  Search sub-directories of path (global).' );
    Writeln( '  /L  List files only, no deletions.' );
    Writeln( '  /T  List only, with very terse output.' );
    Writeln( '  /S  Bypass all confirmations (silent).' );
    Writeln( '  /n  /-n  /=n  Only files more, less,',
             ' or exactly n days old.' );
    Writeln( 'Interactive keys:  (when not silent)' );
    Writeln( '  D   Display directory.' );
    Writeln( '  I   Switch from ALL to INDIVIDUAL deletion mode.' );
    Writeln( '  A   Switch back from INDIVIDUAL to ALL mode.' );
    Writeln( '  P   Backstep to previous skipped individual file.' );
    Writeln( 'Ctrl-C terminates at any time.' );
    Halt(1);
  END;

{-------------------------------------------------------------------------}

PROCEDURE ReadCommandLine;
  VAR  Found:Boolean;  ParamLoop, UpLoop, Code:Integer;  Temp, Param:String;
  BEGIN
    Silent := False;  CheckAge := False;
    Terse := False;  VeryTerse := False;
    ListOnly := False;  Individual := False;
    Global := False;  GloMsg := '';  OldMsg := '';  OldAge := 0;
    DelSize := 0;  TotalSize := 0;
    For ParamLoop := 2 to ParamCount do
      BEGIN
        Found := False;
        Param := ParamStr( ParamLoop );
        For UpLoop := 1 to Length( Param ) do
          Param[ UpLoop ] := UpCase( Param[ UpLoop ] );
        If Param = '/G'
          THEN BEGIN
            Found := True;  Global := True;  GloMsg := ', global';
          END;
        If Param = '/H' THEN BEGIN  Writeln;  ShowHelp;  END;
        If Param = '/L'
          THEN BEGIN
            Found := True;  ListOnly := True;
            Terse := Silent and ListOnly;
          END;
        If Param = '/S'
          THEN BEGIN
            Found := True;  Silent := True;
            Terse := Silent and ListOnly;
          END;
        If Param = '/T'
          THEN BEGIN
            Found := True;  ListOnly := True;
            Terse := True;  VeryTerse := True;
          END;
        If ( not Found ) and ( ParamLoop > 1 ) and ( OldAge = 0 ) and
           ( Length( Param ) > 1 ) and ( Param[ 1 ] = '/' )
          THEN BEGIN
            CheckAge := True;
            Temp := Param;  Delete( Temp, 1, 1 );
            ExactAge := Temp[ 1 ] = '=';
            If ExactAge THEN Delete( Temp, 1, 1 );
            Val( Temp, OldAge, Code );
            If Code = 0
              THEN BEGIN  Found := True;  FigureAgeLimit( OldAge );  END;
          END;
        If not Found
          THEN Stop( 'Bad parameter: "' + Param + '"' + NL +
                     'Enter  "RM /H"  for help.' );
      END;
  END;

{-------------------------------------------------------------------------}

BEGIN

  RamRequired := SizeOf( FileNameArray ) + SizeOf( FileDataArray );
  PrevLineLength := 0;
  If ParamCount < 1 THEN Stop( NL + 'No file(s) specified.  ' +
                               'Enter  "RM /H"  for help' );
  ReadCommandLine;  If not Terse THEN Writeln;
  PureParam := ParamStr( 1 );
  For Loop := 1 to Length( PureParam ) do
    PureParam[ Loop ] := UpCase( PureParam[ Loop ] );
  If ( PureParam = '/H' ) or ( PureParam = '?' ) THEN ShowHelp;
  If PureParam[ 1 ] = '/'
    THEN Stop( 'Bad file name.  Enter  "RM /H"  for help' );
  ParseMask( PureParam );
  Files := 0;  DelCount := 0;
  If MaxAvail < RamRequired
    THEN Stop( 'Insufficient memory, ' +
               IntString( RamRequired div 1000 ) + 'K heap required' );
  New( FileName );
  FillChar( FileName^, SizeOf( FileName^ ), 0 );
  New( FileData );
  FillChar( FileData^, SizeOf( FileData^ ), 0 );
  FillChar( PathName, SizeOf( PathName ), 0 );
  Paths := 1;  PathName[ 1 ] := Path;
  If ( Copy( Path, 2, Length( Path ) ) = ':\' ) and
     ( ( Mask = '*.*' ) or ( Mask = '????????.???' ) )
     and Global and ( not ListOnly ) Then
    Stop( 'Whole disk deletion not allowed' );
  If not Terse THEN Show( 'Scanning . . .', True );
  If Global
    THEN BEGIN
      GetPaths;
      For Loop := 1 to Paths do
        GetFileData( Loop, Mask );
    END
    ELSE BEGIN
      GetFileData( 1, Mask );
      If Files = 0
        THEN BEGIN
          If PureMask <> '' THEN Path := Path + PureMask + '\';
          PathName[ 1 ] := Path;  Mask := '*.*';
          GetFileData( 1, Mask );
        END;
    END;
  If Files = 0
    THEN Stop( 'No file(s) matching: "' + PureParam + '"' +
               OldMsg + GloMsg );
  SortFileData;
  If ListOnly
    THEN BEGIN  ShowDirectory;  Halt;  END;
  REPEAT
    ReStart := False;
    If not Silent
      THEN BEGIN
        If Files < 2
          THEN Confirm( 'Delete "' + PathName[ FileData^[ 1 ].PathNum ] +
                        FileName^[ 1 ] + '"' )
          ELSE Confirm( IntString( Files - DelCount ) +
                        ' files will be deleted . . .  Are you sure' );
        If ( FileCount( Path + '*.*' ) = Files ) and
           ( not Individual ) and ( Paths = 1 )
          THEN Confirm( 'Entire contents of "' + Path +
                        '" will be deleted . . .' + NL +
                        'ARE YOU REALLY SURE' );
      END;
    FileLoop := 0;
    While ( FileLoop < Files ) and ( not ReStart ) do
      BEGIN
        Inc( FileLoop );
        While ( FileName^[ FileLoop ] = '' ) and ( FileLoop <= Files ) do
          Inc( FileLoop );  { skip over already deleted }
        If FileLoop <= Files
          THEN If Global
                 THEN EraseFile( PathName[ FileData^[ FileLoop ].PathNum ] +
                                 FileName^[ FileLoop ] )
                 ELSE EraseFile( Path + FileName^[ FileLoop ] );
      END;
    UNTIL not ReStart;
  Finished( True );

END.


