{$A+,B-,D-,E-,F+,I-,L-,N-,O-,R-,S-,V-}
{$M 16000,0,655360}

Program DskChk;

Uses
  Dos;

Var
  Regs : Registers;
  Dir : SearchRec;
  Result, ChkLp : Integer;
  PrevPath : String[ 80 ];
  DTASpace : Array [ 0..4096 ] of Byte;

Const
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;
  MemSpec = #255;
  TgtPath : String[ 80 ] = '';
  ParamTemp : String[ 40 ] = '';
  TempInt : LongInt = 0;
  Drive : Char = #0;
  SpaceCheck : LongInt = -1;
  NewDiskSpace : LongInt = -1;
  NewMemSpace : LongInt = -1;
  PrevDiskSpace : LongInt = -1;
  PrevMemSpace : LongInt = -1;
  MemOfs : LongInt = 30096;

{----------------------------------------------------------------}

Function Upper( S : String ) : String;
  Var
    UCLoop : Byte;
  Begin
    For UCLoop := 1 to Length( S ) do
      S[ UCLoop ] := UpCase( S[ UCLoop ] );
    Upper := S;
  End;

{----------------------------------------------------------------}

Function FreeMem : LongInt;
  Begin
    FreeMem := MemAvail + MemOfs;
  End;

{----------------------------------------------------------------}

Procedure ReadSpaceFile;
  Var
    SpcFile : Text;
    Result : Integer;
    Drv : String[ 1 ];
  Begin
    If Drive = MemSpec Then
      Drv := ''
    Else
      Drv := Drive + ':';
    Assign( SpcFile, Drv + '\DSKCHK.$$$' );
    Reset( SpcFile );  Result := IOResult;
    If Result = 0 Then
      Begin
        ReadLn( SpcFile, PrevDiskSpace );  Result := IOResult;
        If Result = 0 Then
          Begin
            ReadLn( SpcFile, PrevMemSpace );  Result := IOResult;
          End;
      End;
    If Result <> 0 Then PrevDiskSpace := -1;
    Close( SpcFile );  Result := IOResult;
  End;

{----------------------------------------------------------------}

Procedure WriteSpaceFile;
  Var
    SpcFile : Text;
    Result : Integer;
    Drv : String[ 1 ];
  Begin
    If Drive = MemSpec Then
      Drv := ''
    Else
      Drv := Drive + ':';
    Assign( SpcFile, Drv + '\DSKCHK.$$$' );
    ReWrite( SpcFile );  Result := IOResult;
    If Result = 0 Then
      Begin
        Writeln( SpcFile, NewDiskSpace );
        If Result = 0 Then
          Begin
            Writeln( SpcFile, NewMemSpace );
          End;
      End;
    Close( SpcFile );  Result := Result + IOResult;
  End;

{----------------------------------------------------------------}

Function Mounted( Drive : Char ) : Boolean;
  Var
    TstFile : Text;
  Begin
    Assign( TstFile, Drive + ':\DSKCHK.$$$' );
    Reset( TstFile );  Result := IOResult;
    Mounted := ( Result = 0 ) or ( Result = 2 );
  End;

{----------------------------------------------------------------}

Begin

  If ( ParamCount > 2 ) or
     ( ( ParamCount > 0 ) and ( ParamStr( 1 ) = '?' ) ) Then
    Begin
      For CpyRtLp := 1 to 26 do
        If CpyRtLp = 1 Then
          CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
        Else
          CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                   Ord( CpyRt[ CpyRtLp - 1 ] ) );
      Writeln( 'DSKCHK ver 3.2 rev 92/09/27 - ', CpyRt, '.' );
      Writeln( 'Check for disk space, path, or readiness, and free RAM.' );
      Writeln( '  Return errorlevel if any test fails.' );
      Writeln( 'DSKCHK [drive]' );
      Writeln( '  Check if disk is installed in drive (default a:).' );
      Writeln( 'DSKCHK /M' );
      Writeln( '  Display current amount of free standard memory.' );
      Writeln( 'DSKCHK drive path' );
      Writeln( '  Check for existence of path on disk drive.' );
      Writeln( 'DSKCHK drive space' );
      Writeln( '  Check drive for space disk bytes free.' );
      Writeln( 'DSKCHK drive 0' );
      Writeln( '  Display disk bytes free and change since last check.' );
      Writeln( 'DSKCHK /M space' );
      Writeln( '  Check memory for space bytes free.' );
      Writeln( 'DSKCHK /M 0' );
      Writeln( '  Display memory bytes free and change since last check.' );
      Writeln( '  Bytes free value is stored in drive\DSKCHK.$$$ file.' );
      Halt( 1 );
    End;

  { get the drive }
  If ParamCount = 0 Then
    Drive := 'A'
  Else
    If ( ParamCount > 0 ) and ( Upper( ParamStr( 1 ) ) = '/M' ) Then
      Drive := MemSpec
    Else
      Begin
        ParamTemp := Upper( ParamStr( 1 ) );
        If ( Length( ParamTemp ) = 2 ) and
           ( ParamTemp[ 2 ] = ':' ) Then
          Delete( ParamTemp, 2, 1 );
        If ( ParamTemp = '' ) or ( Length( ParamTemp ) > 1 ) or
           ( not ( ParamTemp[ 1 ] in [ 'A'..'Z' ] ) ) Then
          Begin
            Writeln( 'Invalid drive.' );
            Halt( 1 );
          End
        Else
          Drive := ParamTemp[ 1 ];
      End;

  { check for space parameter }
  If ParamCount > 1 Then
    Begin
      ParamTemp := Upper( ParamStr( 2 ) );
      { check for size parameter }
      Val( ParamTemp, TempInt, Result );
      If Result = 0 Then
        SpaceCheck := TempInt
      Else
        Begin
          If Drive = MemSpec Then
            Begin
              Writeln( 'Invalid option.' );
              Halt( 1 );
            End;
          { check for path parameter }
          TgtPath := ParamTemp;
          For ChkLp := 1 to Length( TgtPath ) do
            If TgtPath[ ChkLp ] in
                 [ #0..#20, #128..#255, '.', '"', '/', '[', ']',
                   ':', '|', '<', '>', '+', '=', ';', ',' ] Then
              Begin
                Writeln( 'Invalid character in path name.' );
                Halt( 1 );
              End;
          If Copy( TgtPath, Length( TgtPath ), 1 ) = '\' Then
            Delete( TgtPath, 1, 1 );
        End;
    End;

  If Drive <> MemSpec Then
    Begin
      FillChar( Regs, SizeOf( Regs ), 0 );
      Regs.AH := $0D;  MsDos( Regs );  { sync }
      If not Mounted( Drive ) Then
        Begin
          Writeln( 'Disk drive ', Drive, ' is not ready.' );
          Halt( 1 );
        End
      Else
        Writeln( 'Disk drive ', Drive, ' is ready.' );
    End;

  If TgtPath <> '' Then
    Begin
      GetDir( 0, PrevPath );  Result := IOResult;
      ChDir( Drive + ':' + TgtPath );  Result := Result + IOResult;
      ChDir( PrevPath );  Result := Result + IOResult;
      If Result <> 0 Then
        Begin
          Writeln( 'Path ' + TgtPath + ' does not exist.' );
          Halt( 1 );
        End
      Else
        Begin
          Writeln( 'Path ' + TgtPath + ' exists.' );
          Halt;
        End;
    End;

  If ( SpaceCheck < 0 ) and ( Drive = MemSpec ) Then
    SpaceCheck := 0;

  If SpaceCheck < 0 Then Halt;

  If Drive <> MemSpec Then
    Begin
      NewDiskSpace := DiskFree( Ord( Drive ) - 64 );
      If SpaceCheck > 0 Then
        Begin
          If NewDiskSpace < SpaceCheck Then
            Begin
              Writeln( 'Less than ', SpaceCheck, ' disk bytes available.' );
              Halt( 1 );
            End
          Else
            Begin
              Writeln( 'At least ', SpaceCheck, ' disk bytes available.' );
              Halt;
            End;
        End
      Else
        Begin
          Writeln( NewDiskSpace, ' disk bytes available.' );
          ReadSpaceFile;
          NewMemSpace := PrevMemSpace;
          If PrevDiskSpace >= 0 Then
            Begin
              Writeln( PrevDiskSpace, ' disk bytes were available.' );
              If PrevDiskSpace > NewDiskSpace Then
                Writeln( PrevDiskSpace - NewDiskSpace,
                         ' disk bytes newly used.' )
              Else
                Writeln( NewDiskSpace - PrevDiskSpace,
                         ' disk bytes newly freed.' );
            End;
          WriteSpaceFile;
        End;
    End  { disk }
  Else
    Begin  { ram }
      NewMemSpace := FreeMem;
      If SpaceCheck > 0 Then
        Begin
          If NewMemSpace < SpaceCheck Then
            Begin
              Writeln( 'Less than ', SpaceCheck,
                       ' memory bytes available.' );
              Halt( 1 );
            End
          Else
            Begin
              Writeln( 'At least ', SpaceCheck,
                       ' memory bytes available.' );
              Halt;
            End;
        End
      Else
        Begin
          Writeln( NewMemSpace, ' memory bytes available.' );
          If ParamCount > 1 Then
            Begin
              ReadSpaceFile;
              NewDiskSpace := PrevDiskSpace;
              If PrevMemSpace >= 0 Then
                Begin
                  Writeln( PrevMemSpace, ' memory bytes were available.' );
                  If PrevMemSpace > NewMemSpace Then
                    Writeln( PrevMemSpace - NewMemSpace,
                             ' memory bytes newly used.' )
                  Else
                    Writeln( NewMemSpace - PrevMemSpace,
                             ' memory bytes newly freed.' );
                End;
            End;
          WriteSpaceFile;
        End;
    End;  { ram }



End.



