{$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 1024,0,0}

Program DtUtl;

Uses
  Dos;

Type
  String2 = String[2];
  String8 = String[8];
  String10 = String[10];
  String17 = String[17];
  String30 = String[30];

Const
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;

Var
  HR24, HR, HR2, MI, SC, AP,
    DW, DW2, MN, MN2, MO, MO2, DA, DA2, TH, CN, YR : String[10];
  DWCalc : String30;
  Date1, Date2, Trgt : String8;
  R_Date1, R_Date2 : Real;
  FmtStr, OutStr : String[80];
  Dif : Real;
  Result, Prs, ParamLoop : Integer;
  O : String;  
  Found : Boolean;

{-------------------------------------------------------------------------}
{ DTSec returns seconds since 80-01-01 00:00:00                           }
{ as a real number, given date and time strings YY-MM-DD and HH:MM:SS.    }
{-------------------------------------------------------------------------}

Function DTSec( DateStr, TimeStr : String8 ) : Real;
  Var
    Yr, Mo, Da, Hr, Mn, Sc, DaTotal, ScTotal : LongInt;
    DTLp, DTResult : Integer;
  Const
    DaysPerMonth : array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    { initialize function to zero, flags conversion error }
    DTSec := -1.0;
    { remove invalid characters }
    For DTLp := Length( DateStr ) DownTo 1 do
      If not ( DateStr[ DTLp ] in [ '0'..'9' ] ) Then
        Delete( DateStr, DTLp, 1 );
    For DTLp := Length( TimeStr ) DownTo 1 do
      If not ( TimeStr[ DTLp ] in [ '0'..'9' ] ) Then
        Delete( TimeStr, DTLp, 1 );
    { exit if remaining length too small }
    If ( Length( DateStr ) <> 6 ) or ( Length( TimeStr ) <> 6 ) Then
      Exit;
    { convert digit pairs to numbers }
    Val( Copy( DateStr, 1, 2 ), Yr, DTResult );
    Val( Copy( DateStr, 3, 2 ), Mo, DTResult );
    Val( Copy( DateStr, 5, 2 ), Da, DTResult );
    Val( Copy( TimeStr, 1, 2 ), Hr, DTResult );
    Val( Copy( TimeStr, 3, 2 ), Mn, DTResult );
    Val( Copy( TimeStr, 5, 2 ), Sc, DTResult );
    { range check all values }
    If ( Yr < 0 ) or ( Yr > 99 ) or
       ( Mo < 1 ) or ( Mo > 12 ) or
       ( Da < 1 ) or ( Da > 31 ) or
       ( ( not ( Da in [ 1 .. DaysPerMonth[ Mo ] ] ) ) and
         ( not ( ( Da = 29 ) and ( Mo = 2 ) and ( Yr mod 4 = 0 ) ) ) ) or
       ( Hr < 0 ) or ( Hr > 23 ) or
       ( Mn < 0 ) or ( Mn > 59 ) or
       ( Sc < 0 ) or ( Sc > 59 ) Then
      Exit;
    { convert date to days since 01-01-80 }
    { assume years before 80 are in 21st century }
    If Yr < 80 Then Yr := Yr + 100;
    Yr := Yr - 80;
    { start with leap year days }
    DaTotal := Yr div 4;
    { compensate for fact that the base year, 1980, is a leap year }
    Inc( DaTotal );
    { compensate if current year is leap year, but leap day is not past }
    If ( Yr mod 4 = 0 ) and
       ( ( Mo < 2 ) or ( ( Mo = 2 ) and ( Da < 29 ) ) ) Then
      Dec( DaTotal );
    { add days for past years }
    DaTotal := DaTotal + ( Yr * 365 );
    DTLp := 1;
    { add days for elapsed months in year }
    While DTLp < Mo do
      Begin
        DaTotal := DaTotal + DaysPerMonth[ DTLp ];
        Inc( DTLp );
      End;
    { add elapsed days in month to finish }
    DaTotal := DaTotal + Da - 1;
    { convert time to seconds since 00:00:00 }
    ScTotal := Hr * 3600 + Mn * 60 + Sc;
    DTSec := DaTotal * 86400.0 + ScTotal * 1.0;
  End;

{-------------------------------------------------------------------------}
{ SecDT returns a string containing date and time "YY-MM-DD HH:MM:SS"     }
{ given a real number of seconds since 80-01-01 00:00:00.                 }
{-------------------------------------------------------------------------}

Function SecDT( InSec : Real ) : String17;
  Var
    Yr, Mo, Da, Hr, Mn, Sc, DaTotal, ScTotal : LongInt;
    DTLoop, DaysInCurrMonth : Integer;
  Const
    DaysPerMonth : Array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    SecDT := '';
    ScTotal := Trunc( InSec - ( Int( InSec / 86400.0 ) * 86400.0 ) );
    DaTotal := Trunc( InSec / 86400.0 ) + 1;
    Hr := ScTotal div 3600;
    Mn := ( ScTotal - ( Hr * 3600 ) ) div 60;
    Sc := ScTotal - ( Hr * 3600 ) - ( Mn * 60 );
    Yr := 1980;  Mo := 1;  Da := 1;
    If ( Mo <> 2 ) or ( Yr mod 4 <> 0 ) Then
      DaysInCurrMonth := DaysPerMonth[ Mo ]
    Else
      DaysInCurrMonth := 29;
    While DaTotal > DaysInCurrMonth do
      Begin
        DaTotal := DaTotal - DaysInCurrMonth;
        Inc( Mo );
        If Mo > 12 Then
          Begin
            Mo := 1;
            Inc( Yr );
          End;
        If ( Mo <> 2 ) or ( Yr mod 4 <> 0 ) Then
          DaysInCurrMonth := DaysPerMonth[ Mo ]
        Else
          DaysInCurrMonth := 29;
      End;
    Da := DaTotal;
    SecDT :=
      Chr( Yr mod 100 div 10 + 48 ) + Chr( Yr mod 10 + 48 ) + '/' +
      Chr( Mo div 10 + 48 ) + Chr( Mo mod 10 + 48 ) + '/' +
      Chr( Da div 10 + 48 ) + Chr( Da mod 10 + 48 ) + ' ' +
      Chr( Hr div 10 + 48 ) + Chr( Hr mod 10 + 48 ) + ':' +
      Chr( Mn div 10 + 48 ) + Chr( Mn mod 10 + 48 ) + ':' +
      Chr( Sc div 10 + 48 ) + Chr( Sc mod 10 + 48 );
  End;

{-------------------------------------------------------------------------}

Function FmtDate( InDate : String8;  Style : Integer ) : String30;
  Var
    Yr, Mo, Da, DaTotal, FullYr : LongInt;
    DTLp, DTResult, DayOfWk : Integer;
    DaStr, YrStr : String8;
  Const
    DayName : Array [ 1..7 ] of String[ 12 ] =
      ( 'Sunday', 'Monday', 'Tuesday', 'Wednesday',
        'Thursday', 'Friday', 'Saturday' );
    MonthName : Array [ 1..12 ] of String[ 12 ] =
      ( 'January', 'February', 'March', 'April', 'May', 'June',
        'July', 'August', 'September', 'October', 'November', 'December' );
    DaysPerMonth : Array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    { initialize function to null string, flags conversion error }
    FmtDate := '';
    { remove invalid characters }
    For DTLp := Length( InDate ) DownTo 1 do
      If not ( InDate[ DTLp ] in [ '0'..'9' ] ) Then
        Delete( InDate, DTLp, 1 );
    { exit if remaining length too small }
    If Length( InDate ) <> 6 Then Exit;
    { convert digit pairs to numbers }
    Val( Copy( InDate, 1, 2 ), Yr, DTResult );
    Val( Copy( InDate, 3, 2 ), Mo, DTResult );
    Val( Copy( InDate, 5, 2 ), Da, DTResult );
    { range check all values }
    If ( Yr < 0 ) or ( Yr > 99 ) or
       ( Mo < 1 ) or ( Mo > 12 ) or
       ( Da < 1 ) or ( Da > 31 ) or
       ( ( not ( Da in [ 1 .. DaysPerMonth[ Mo ] ] ) ) and
         ( not ( ( Da = 29 ) and ( Mo = 2 ) and ( Yr mod 4 = 0 ) ) ) ) Then
      Exit;
    { convert date to days since 01-01-80 }
    { assume years before 80 are in 21st century }
    If Yr < 80 Then
      FullYr := Yr + 2000
    Else
      FullYr := Yr + 1900;
    If Yr < 80 Then Yr := Yr + 100;
    Yr := Yr - 80;
    { start with leap year days }
    DaTotal := Yr div 4;
    { compensate for fact that the base year, 1980, is a leap year }
    Inc( DaTotal );
    { compensate if current year is leap year, but leap day is not past }
    If ( Yr mod 4 = 0 ) and
       ( ( Mo < 2 ) or ( ( Mo = 2 ) and ( Da < 29 ) ) ) Then
      Dec( DaTotal );
    { add days for past years }
    DaTotal := DaTotal + ( Yr * 365 );
    DTLp := 1;
    { add days for elapsed months in year }
    While DTLp < Mo do
      Begin
        DaTotal := DaTotal + DaysPerMonth[ DTLp ];
        Inc( DTLp );
      End;
    { add elapsed days in month to finish }
    DaTotal := DaTotal + Da - 1;
    { format into string }
    If Style > 0 Then
      Begin
        { styles:
            0 : no presentation date generated
            1 : Tue., Jan. 1, 1980
            else : Tuesday, January 1st., 1980  }
        DayOfWk := ( DaTotal   { 01-01-80 = Tuesday = 0 }
                     mod 7     { modulo into weeks }
                     + 2 )     { offset for Tuesday in 0..6 range }
                     mod 7     { correct range }
                     + 1;      { offset to 1..7 range }
        Str( Da, DaStr );
        Str( FullYr, YrStr );
        If Style = 1 Then
          FmtDate := Copy( DayName[ DayOfWk ], 1, 3 ) + '., ' +
                     Copy( MonthName[ Mo ], 1, 3 ) + '. ' +
                     DaStr + ', ' + YrStr
        Else
          Begin
            Da := Da mod 10;
            If Da = 1 Then
              DaStr := DaStr + 'st'
            Else
              If Da = 2 Then
                DaStr := DaStr + 'nd'
              Else
                If Da = 3 Then
                  DaStr := DaStr + 'rd'
                Else
                  DaStr := DaStr + 'th';
            FmtDate := DayName[ DayOfWk ] + ', ' +
                       MonthName[ Mo ] + ' ' + DaStr + ', ' + YrStr;
          End;
      End;
  End;

{-------------------------------------------------------------------------}

Function Str2( InNum : Word ) : String2;
  Begin
    Str2 := Chr( InNum div 10 + 48 ) + Chr( InNum mod 10 + 48 );
  End;

{-------------------------------------------------------------------------}

Function DateString : String8;
  Var
    Regs : Registers;
  Begin
    Regs.AH := $2A;  MSDos( Regs );
    DateString := Str2(Regs.CX mod 100)+'/'+Str2(Regs.DH)+'/'+Str2(Regs.DL);
  End;

{-------------------------------------------------------------------------}

Function TimeString : String10;
  Var
    Regs : Registers;
    TZ, HrStr : String2;
    HR : Integer;
  Begin
    Regs.AH := $2C;  MSDos( Regs );
    If Regs.CH > 11 Then
      TZ := 'pm'
    Else
      TZ := 'am';
    HR := Regs.CH;
    If HR = 0 Then HR := 24;
    If HR > 12 Then HR := HR - 12;
    HrStr := Str2( HR );
    TimeString := HrStr + ':' + Str2( Regs.CL ) + TZ;
  End;

{-------------------------------------------------------------------------}

Function FullTimeString : String10;
  Var
    Regs : Registers;
  Begin
    Regs.AH := $2C;  MSDos( Regs );
    FullTimeString :=
      Str2( Regs.CH ) + ':' + Str2( Regs.CL ) + ':' + Str2( Regs.DH );
  End;

{-------------------------------------------------------------------------}

Procedure DoWait;
  Var
    Seconds : LongInt;
    Code : Integer;
    Sec10, PrevSec10 : Byte;
    Regs : Registers;
  Begin
    Code := 0;
    If ParamCount < 2 Then
      Seconds := 10
    Else
      Val( ParamStr( 2 ), Seconds, Code );
    If Code <> 0 Then
      Begin
        Writeln( 'Invalid number for wait time value.' );
        Halt( 1 );
      End;
    If ( Seconds < 1 ) or ( Seconds > 3600 ) Then
      Begin
        Writeln( 'Time value out of range.' );
        Halt( 1 );
      End;
    Regs.AH := $2C;  MSDos( Regs );
    Sec10 := Regs.DL div 10;
    Repeat
      PrevSec10 := Sec10;
      Write( #13, 'Pausing for ', Seconds, ' seconds. ' );
      { wait for next 1/10 second }
      Repeat
        Regs.AH := $2C;  MSDos( Regs );
        Sec10 := Regs.DL div 10;
      Until Sec10 <> PrevSec10;
      { wait for same 1/10 of next second }
      Repeat
        Regs.AH := $2C;  MSDos( Regs );
        Sec10 := Regs.DL div 10;
      Until Sec10 = PrevSec10;
      Dec( Seconds );
    Until Seconds = 0;
    Writeln;
    Halt;
  End;

{-------------------------------------------------------------------------}

Procedure DoSet;
  Var
    DT_File: Text;
    InDate, InTime : String8;
    R, Result : Integer;
    DT_Real : Real;
    Yr, Mo, Da, Lp : Word;
  Begin
    If ParamCount > 1 Then
      Begin
        InDate := ParamStr( 2 );
        If DTSec( InDate, '00:00:00' ) > 0.0 Then
          Begin
            For Lp := Length( InDate ) DownTo 1 do
              If not ( InDate[ Lp ] in [ '0'..'9' ] ) Then
                Delete( InDate, Lp, 1 );
            Val( Copy( InDate, 1, 2 ), Yr, R );  Result := R;
            Val( Copy( InDate, 3, 2 ), Mo, R );  Result := Result + R;
            Val( Copy( InDate, 5, 2 ), Da, R );  Result := Result + R;
            If Yr < 80 Then
              Yr := Yr + 2000
            Else
              Yr := Yr + 1900;
            SetDate( Yr, Mo, Da );
          End;
      End;
    If Copy( DateString, 1, 2 ) = '80' Then
      Begin  { date is at default }
        Assign( DT_File, '\DTUTL.$$$' );
        Reset( DT_File );  Result := IOResult;
        InDate := '01-01-80';  InTime := '00:00:00';
        If Result = 0 Then
          Begin
            ReadLn( DT_File, InDate );  Result := IOResult;
            If Result = 0 Then
              Begin
                ReadLn( DT_File, InTime );  Result := IOResult;
              End;
          End;
        Close( DT_File );  Result := IOResult;
        If ( Copy( FullTimeString, 1, 2 ) <> '00' ) and
           ( FullTimeString < InTime ) Then
          Begin  { increment date }
            DT_Real := DTSec( InDate, InTime );
            DT_Real := DT_Real + 86400.0;
            InDate := Copy( SecDT( DT_Real ), 1, 8 );
          End;
        Val( Copy( InDate, 1, 2 ), Yr, R );  Result := R;
        Val( Copy( InDate, 4, 2 ), Mo, R );  Result := Result + R;
        Val( Copy( InDate, 7, 2 ), Da, R );  Result := Result + R;
        If Yr < 80 Then
          Yr := Yr + 2000
        Else
          Yr := Yr + 1900;
        SetDate( Yr, Mo, Da );
      End;
    Assign( DT_File, '\DTUTL.$$$' );
    ReWrite( DT_File );  Result := IOResult;
    Writeln( DT_File, DateString, #13, #10, FullTimeString );
    Result := IOResult;
    Close( DT_File );  Result := IOResult;
    Halt;
  End;

{-------------------------------------------------------------------------}

Procedure DoCalendar;
  Var
    MoNm, YrStr, UpMo, StartDt, YrNm, CalHd : String[21];
    Mo, Yr, CalLp, DoWLp, StartDoW, DaTotal, CalResult : Integer;
    WorkDt : Real;
  Const
    MonthName : Array [ 1..12 ] of String[ 12 ] =
      ( 'January', 'February', 'March', 'April', 'May', 'June',
        'July', 'August', 'September', 'October', 'November', 'December' );
    DaysPerMonth : Array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    MoNm := ParamStr( 1 );
    YrStr := ParamStr( 2 );
    For CalLp := 1 to Length( MoNm ) do
      MoNm[ CalLp ] := UpCase( MoNm[ CalLp ] );
    Mo := 0;
    Repeat
      Inc( Mo );
      UpMo := MonthName[ Mo ];
      For CalLp := 1 to Length( UpMo ) do
        UpMo[ CalLp ] := UpCase( UpMo[ CalLp ] );
    Until ( Pos( MoNm, UpMo ) > 0 ) or ( Mo = 12 );
    If Pos( MoNm, UpMo ) = 0 Then
      Begin
        Writeln( 'Bad month name.' );
        Halt( 1 );
      End;
    YrStr := ParamStr( 2 );
    Val( YrStr, Yr, CalResult );
    If CalResult > 0 Then
      Begin
        Writeln( 'Bad year value.' );
        Halt( 1 );
      End;
    If Yr < 80 Then Yr := Yr + 100;
    Yr := Yr + 1900;
    If ( Yr < 1980 ) or ( Yr > 2079 ) Then
      Begin
        Writeln( 'Year out of range.' );
        Halt( 1 );
      End;
    If ( Mo = 2 ) and ( Yr mod 4 = 0 ) Then
      DaysPerMonth[ Mo ] := 29;
    Str( Mo, MoNm );
    If Length( MoNm ) < 2 Then MoNm := '0' + MoNm;
    Str( Yr mod 100, YrStr );
    If Length( YrStr ) < 2 Then YrStr := '0' + YrStr;
    StartDt := YrStr + MoNm + '01';
    WorkDt := DTSec( StartDt, '00:00:00' ); 
    DaTotal := Round( WorkDt / 86400.0 );
    StartDoW := ( DaTotal   { 01-01-80 = Tuesday = 0 }
                  mod 7     { modulo into weeks }
                    + 2 )   { offset for Tuesday in 0..6 range }
                  mod 7     { correct range }
                  + 1;      { offset to 1..7 range }
    Str( Yr, YrNm );
    CalHd := MonthName[ Mo ] + ' ' + YrNm;
    CalLp := Length( CalHd );
    While CalLp < 20 do
      Begin
        If not Odd( CalLp ) Then CalHd := ' ' + CalHd;
        Inc( CalLp );
      End;
    Writeln( CalHd );
    Writeln( 'Su Mo Tu We Th Fr Sa' );
    DoWLp := 1;  { continuously increments, mod 7 for true DoW }
    CalLp := 1;  { starts incrementing only after DoWLp >= StartDoW }
    Repeat
      If DoWLp < StartDow Then
        Begin
          If ( DoWLp - 1 ) mod 7 = 0 Then
            Write( '  ' )
          Else
            Write( '   ' );
        End
      Else
        Begin
          If ( CalLp > 1 ) and ( ( DoWLp - 1 ) mod 7 = 0 ) Then
            Writeln;  { start new line }
          If ( DoWLp - 1 ) mod 7 = 0 Then
            Write( CalLp:2 )
          Else
            Write( CalLp:3 );
          Inc( CalLp );
        End;
      Inc( DoWLp );
    Until ( CalLp > DaysPerMonth[ Mo ] );
    Writeln;
    Halt;
  End;

{-------------------------------------------------------------------------}

Begin

  If ( ParamCount > 0 ) and ( ParamStr( 1 ) = '?' ) Then
    Begin
      For CpyRtLp := 1 to 26 do
        If CpyRtLp = 1 Then
          CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
        Else
          CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                   Ord( CpyRt[ CpyRtLp - 1 ] ) );
      Writeln( 'DTUTL ver 3.1 rev 91/06/29 - ', CpyRt, '.' );
      Writeln( 'Date/time display, calculate, calendar and utilities.' );
      Writeln( 'DTUTL' );
      Writeln( '  Display full, formatted time/day/date.' );
      Writeln( 'DTUTL yymmdd1 yymmdd2' );
      Writeln( '  Display difference between two dates in days.' );
      Writeln( 'DTUTL [yymmdd] [+/-n] ["text"]' );
      Writeln( '  Display current or selected date +/- n days.' );
      Writeln( '  These $xx sequences are substituted in the text:' );
      Writeln( '  Example:  1:23:07pm Saturday April 5th 1980' );
      Writeln( '    $24=13 $HR=01 $Hr=1 $MI=23 $SC=07 $AP=pm' );
      Writeln( '    $DW=Saturday $Dw=Sat $MN=April $Mn=Apr' );
      Writeln( '    $MO=04 $Mo=4 $DA=05 $Da=5 $TH=th $CN=19 $YR=80' );
      Writeln( 'DTUTL mmm yy' );
      Writeln( '  Show calendar for month: ie. JAN 91.' );
      Writeln( 'DTUTL /P [n]' );
      Writeln( '  Pause for n seconds (default 10).' );
      Writeln( 'DTUTL /S [yymmdd]' );
      Writeln( '  Set date for old PCs w/o r.t. clock.' );
      Writeln( '  If date not set, sets to contents of \DTUTL.$$$' );
      Writeln( '    else write current time/date to file.' );
      Writeln( '  Increment date if time in file > current time > 00hr.' );
      Halt( 1 );
    End;
  If ParamCount = 0 Then
    Begin
      OutStr := TimeString;
      If Copy( OutStr, 1, 1 ) = '0' Then Delete( OutStr, 1, 1 );
      Writeln( 'The time is ', OutStr, ' on ',
               FmtDate( DateString, 2 ) );
      Halt;
    End;
  If ParamCount = 2 Then
    Begin
      Date1 := ParamStr( 1 );
      If ( Length( Date1 ) <= 3 ) and
         ( UpCase( Date1[ 1 ] ) in [ 'A'..'Z' ] ) Then
        DoCalendar;
    End;
  If ParamCount > 0 Then
    Begin
      FmtStr := ParamStr( 1 );
      For ParamLoop := 1 to Length( FmtStr ) do
        FmtStr[ ParamLoop ] := UpCase( FmtStr[ ParamLoop ] );
      If FmtStr = '/P' Then DoWait;
      If FmtStr = '/S' Then DoSet;
    End;
  ParamLoop := 1;
  Date1 := '';  Date2 := '';  Dif := 0.0;  FmtStr := '';
  While ParamLoop <= ParamCount do
    Begin
      Found := False;
      If ( DTSec( ParamStr( ParamLoop ), '00:00:00' ) > 0.0 ) and
         ( ( ParamLoop = 1 ) or ( ParamLoop = 2 ) ) Then
        Begin
          If Date1 = '' Then
            Date1 := ParamStr( ParamLoop )
          Else
            Date2 := ParamStr( ParamLoop );
          Inc( ParamLoop );
          Found := True;
        End;
      If ( not Found ) and
         ( ( ( Copy( ParamStr( ParamLoop ), 1, 1 ) = '+' ) or
             ( Copy( ParamStr( ParamLoop ), 1, 1 ) = '-' ) ) and
           ( ( ParamLoop = 1 ) or ( ParamLoop = 2 ) ) ) Then
        Begin
          Val( ParamStr( ParamLoop ), Dif, Result );
          If Result <> 0 Then
            Begin
              Writeln( 'Invalid date difference value.' );
              Halt( 1 );
            End;
          Found := True;
          Inc( ParamLoop );
        End;
      If ( not Found ) and
         ( Copy( ParamStr( ParamLoop ), 1, 1 ) = '"' ) Then
        Begin
          While ParamLoop <= ParamCount do
            Begin
              If FmtStr = '' Then
                FmtStr := ParamStr( ParamLoop )
              Else
                FmtStr := FmtStr + ' ' + ParamStr( ParamLoop );
              Inc( ParamLoop );
            End;
          Delete( FmtStr, 1, 1 );
          If FmtStr <> '' Then
            If Copy( FmtStr, Length( FmtStr ), 1 ) <> '"' Then
              Begin
                Writeln( 'Missing close quote.' );
                Halt( 1 );
              End;
          Delete( FmtStr, Length( FmtStr ), 1 );
          If ( Pos( '$24', FmtStr ) = 0 ) and
             ( Pos( '$Hr', FmtStr ) = 0 ) and
             ( Pos( '$HR', FmtStr ) = 0 ) and
             ( Pos( '$AP', FmtStr ) = 0 ) and
             ( Pos( '$MI', FmtStr ) = 0 ) and
             ( Pos( '$SC', FmtStr ) = 0 ) and
             ( Pos( '$DW', FmtStr ) = 0 ) and
             ( Pos( '$Dw', FmtStr ) = 0 ) and
             ( Pos( '$MN', FmtStr ) = 0 ) and
             ( Pos( '$Mn', FmtStr ) = 0 ) and
             ( Pos( '$MO', FmtStr ) = 0 ) and
             ( Pos( '$Mo', FmtStr ) = 0 ) and
             ( Pos( '$DA', FmtStr ) = 0 ) and
             ( Pos( '$Da', FmtStr ) = 0 ) and
             ( Pos( '$TH', FmtStr ) = 0 ) and
             ( Pos( '$YR', FmtStr ) = 0 ) and
             ( Pos( '$CN', FmtStr ) = 0 ) Then
            Begin
              Writeln( 'No replacements specified in format.' );
              Halt( 1 );
            End;
          Found := True;
        End;
      If not Found Then
        Begin
          Writeln( 'Unknown parameter: ' + ParamStr( ParamLoop ) );
          Halt( 1 );
        End;
    End;
  If Date1 = '' Then Date1 := DateString;

  If ( Date1 <> '' ) and ( Date2 <> '' ) Then
    Begin
      Writeln( 'Starting date: ', FmtDate( Date1, 2 ) );
      Writeln( 'Ending date:   ', FmtDate( Date2, 2 ) );
      R_Date1 := DTSec( Date1, '00:00:00' );
      R_Date2 := DTSec( Date2, '00:00:00' );
      Dif := Round( Abs( R_Date1 - R_Date2 ) / 86400.0 );
      Writeln( 'Days between:  ', Dif:1:0 );
      Halt;
    End;

  Date2 := Date1;
  If Abs( Dif ) > 0.0 Then
    Begin
      R_Date1 := DTSec( Date1, '00:00:00' );
      R_Date1 := R_Date1 + ( Dif * 86400.0 );
      Date2 := Copy( SecDT( R_Date1 ), 1, 8 );
    End;

  If FmtStr = '' Then
    Begin
      If Abs( Dif ) > 0.0 Then
        Begin
          Writeln( 'Starting date:    ', FmtDate( Date1, 2 ) );
          Writeln( 'Difference days:  ', Round( Dif ) );
          Writeln( 'Resulting date:   ', FmtDate( Date2, 2 ) );
        End
      Else
        Writeln( 'Requested date is ', FmtDate( Date2, 2 ) );
      Halt;
    End;

{ Example  1:23:07pm Saturday April 5th 1980 }
{   $24=13 $HR=01 $Hr=1 $MI=23 $SC=07 $AP=pm }
{   $DW=Saturday $Dw=Sat $MN=April $Mn=Apr }
{   $MO=04 $Mo=4 $DA=05 $Da=5 $TH=th $CN=19 $YR=80 }
  HR24 := Copy( FullTimeString, 1, 2 );
  HR := Copy( TimeString, 1, 2 );
  HR2 := HR;
  If Copy( HR2, 1, 1 ) = '0' Then Delete( HR2, 1, 1 );
  MI := Copy( TimeString, 4, 2 );
  SC := Copy( FullTimeString, 7, 2 );
  AP := Copy( TimeString, 6, 2 );
  DWCalc := FmtDate( Date2, 2 );
  DW := Copy( DWCalc, 1, Pos( ',', DWCalc ) - 1 );
  DW2 := Copy( DW, 1, 3 );
  Delete( DWCalc, 1, Length( DW ) + 2 );
  TH := Copy( DWCalc, Pos( ',', DWCalc ) - 2, 2 );
  Prs := Pos( ',', FmtDate( Date2, 2 ) );
  MN := Copy( FmtDate( Date2, 2 ), Prs + 2, 10 );
  Prs := Pos( ' ', MN );
  MN := Copy( MN, 1, Prs - 1 );
  MN2 := Copy( MN, 1, 3 );
  MO := Copy( Date2, 4, 2 );
  MO2 := MO;
  If Copy( MO2, 1, 1 ) = '0' Then Delete( MO2, 1, 1 );
  If YR < '80' Then
    CN := '20'
  Else
    CN := '19';
  YR := Copy( Date2, 1, 2 );
  DA := Copy( Date2, 7, 2 );
  DA2 := DA;
  If Copy( DA2, 1, 1 ) = '0' Then Delete( DA2, 1, 1 );

  OutStr := '';
  While FmtStr <> '' do
    Begin
      Prs := Pos( '$', FmtStr );
      If Prs = 0 Then Prs := Length( FmtStr ) + 1;
      OutStr := OutStr + Copy( FmtStr, 1, Prs - 1 );
      Delete( FmtStr, 1, Prs );
      If FmtStr <> '' Then
        Begin
          Trgt := Copy( FmtStr, 1, 2 );
          Delete( FmtStr, 1, 2 );
          If Trgt = '24' Then
            OutStr := OutStr + HR24
          Else
          If Trgt = 'HR' Then
            OutStr := OutStr + HR
          Else
          If Trgt = 'Hr' Then
            OutStr := OutStr + HR2
          Else
          If Trgt = 'MI' Then
            OutStr := OutStr + MI
          Else
          If Trgt = 'SC' Then
            OutStr := OutStr + SC
          Else
          If Trgt = 'AP' Then
            OutStr := OutStr + AP
          Else
          If Trgt = 'DW' Then
            OutStr := OutStr + DW
          Else
          If Trgt = 'Dw' Then
            OutStr := OutStr + DW2
          Else
          If Trgt = 'MN' Then
            OutStr := OutStr + MN
          Else
          If Trgt = 'Mn' Then
            OutStr := OutStr + MN2
          Else
          If Trgt = 'MO' Then
            OutStr := OutStr + MO
          Else
          If Trgt = 'Mo' Then
            OutStr := OutStr + MO2
          Else
          If Trgt = 'DA' Then
            OutStr := OutStr + DA
          Else
          If Trgt = 'Da' Then
            OutStr := OutStr + DA2
          Else
          If Trgt = 'TH' Then
            OutStr := OutStr + TH
          Else
          If Trgt = 'CN' Then
            OutStr := OutStr + CN
          Else
          If Trgt = 'YR' Then
            OutStr := OutStr + YR
          Else
            OutStr := OutStr + '$' + Trgt;
        End;
    End;
  Writeln( OutStr );

End.



