(*********************************************************************
**********************************************************************
****                                                              ****
****                      GLOBAL.PAS                              ****
****                                                              ****
****        Fundamental and widely shared globals.                ****
****                                                              ****
**********************************************************************
**********************************************************************)


Unit Global;

{$I LIBRARY.CFG }

INTERFACE

Uses Dos, Crt;

CONST
  MaxOperation = 15;
  PrintableCharSet : Set of Char = [ ' '..'~' ];
{ Escape is reset at start of high level input procedures:  Menu, Enter, }
{ and Select, then set in those procedures if <ESC> key is pressed. }
  Escape : Boolean = False;
  Abandon : Boolean = False;
  QuitPrinting : Boolean = False;   {used in preview/printing routines}
  Hell_Freezes_Over : Boolean = False;  { Repeat ... until ... (forever) }
  OK : Integer = 0;
  NL = #13#10;
  CR = #13;
  LF = #10;
  EF = #26;   {end of file}
  RemoteForceNL : Boolean = False;

{ named constant for each possible type }
  NoType      : Integer =   0;
  _Boolean    : Integer =   1;
  _Byte       : Integer =   2;
  _Integer    : Integer =   3;
  _Real       : Integer =   4;
  _Real2      : Integer =   5;
  _Real4      : Integer =   6;
  _Int3       : Integer =   7;
  _SignedInt3 : Integer =   8;
  _Binary     : Integer =   9;
  _PosInt     : Integer =  10;
  _State      : Integer =  11;
  _Word       : Integer =  12;
  _LongInt    : Integer =  13;
  _BCDReal    : Integer =  14;
  _Date       : Integer =  15;
  SimpeTypes    = 15;
  _String     : Integer = 100;  { + dimensioned length }
  MaxStringType  : Integer = 355;
  _Array      : Integer = 500;  { + size in bytes }
  MaxType     : Integer = MaxInt;
  _Time       : Integer = 108;
{ descriptions of simple types }
  VarSizes : Array [ 0..SimpeTypes ] of Byte =   { memory size }
               ( 0, 1, 1, 2, 6, 6, 6, 3, 3, 1, 2, 2, 2, 4, 10, 9 );
  VarWidths : Array [ 0..SimpeTypes ] of Byte =  { displayed width }
               ( 0, 5, 3, 6, 11, 11, 11, 8, 9, 9, 5, 2, 6, 13, 11, 8 );
  VarNames : Array [ 0..SimpeTypes ] of String[20] =
     ( 'unused', 'boolean', 'byte', 'integer', 'real',
       'dollars', 'cents', '3 byte int.',
       'signed 3 byte int.', 'binary byte', 'positive integer',
       'state abbr.', 'word', 'long integer', 'BCD real', 'date' );
  VarRange : Array [ 0..SimpeTypes ] of
      Record  Min, Max : Real;  End =
        ( ( Min : 0.0;  Max : 0.0 ),
          ( Min : 0.0;  Max : 0.0 ),
          ( Min : 0.0;  Max : 255.0 ),
          ( Min : -32767.0;  Max : 32767.0 ),
          ( Min : -1.7E38;  Max : 1.7E38 ),
          ( Min : -1.7E38;  Max : 1.7E38 ),
          ( Min : -1.7E38;  Max : 1.7E38 ),
          ( Min : 0.0;  Max : 16777215.0 ),
          ( Min : -8388607.0;  Max : 8388607.0 ),
          ( Min : 0.0;  Max : 255.0 ),
          ( Min : 0.0;  Max : 32768.0 ),
          ( Min : 0.0;  Max : 255.0 ),
          ( Min : 0.0;  Max : 65535 ),
          ( Min : -2147483648.0;  Max : 2147483647.0 ),
          ( Min : -1.7E38;  Max : 1.7E38 ),
          ( Min : 0.0;  Max : 0.0 ) );

TYPE
  GenlString = String[ 82 ];
  SmallString = String[ 20 ];
  BigString = String[ 150 ];


VAR
  Sys : Record  { system configuration record }
    ProgramName : String[20];
    Version : String[10];
    Revision : String[20];
    SystemType : String[30];    { such as "CUSTOM SOFTWARE" }
    ComputerName : String[20];  { such as "SOIL PC" }
    ProcessName : String[30];   { such as "ACCOUNTS PAYABLE" }
    UseSystemDate : Boolean;
    UseFutureDate : Boolean;
    EuropeanDate : Boolean;
    SoundOn : Boolean;
    PrinterAbandoned : Boolean;
    DebugEnabled : Boolean;
    TankEnable : Boolean;
    PlantName : String[40];
    SoilPath : String[40];
    MiscPath : String[40];
    CleanPath : String[40];
    LogPath : String[40];
    InstallCode : String[5];
    HelpPath : String[40];
    CommPath : String[40];
    DTAPath : String[40];
    DATPath : String[40];
    MinYear, MaxYear : Word;
    Silent : Boolean;  { inhibit logs }
    Demo : Boolean;  { reduced features }
    KeyLogEnabled : Boolean;
    ScreenSaverTime : LongInt;
    IgnoreLowMemory : Boolean;
    Remote : Boolean;
    RemoteOnly : Boolean;
    ForceRemote : Boolean;
    { ShowTimeEnabled if set true by ClearScreen, false by FullWindow }
    ShowTimeEnabled : Boolean;
   End;

  OperationStack : Array[ 1..MaxOperation ] of String[10];
  OperationStackTop : Byte;
  R : Text;  { This is STDOUT access. }
  GlobalExitSave : Pointer;
{ the following are spares, for use in application programs }
  Regs : Registers;
  Dir : SearchRec;
  DT : DateTime;
  B : Boolean;

{ This is used as an inter-unit forward.  It is redefined in DISK.PAS, }
{ after screen and keyboard support have been included. }
  ChkErr : Procedure( Result:integer; Errtype:Char;
                      FSpec:GenlString; Desc:GenlString );



Procedure Remote( InLn : String );
Procedure RemoteLn( InLn : String );
Procedure Disp( InLn : String );
Procedure DispLn( InLn : String );
Procedure Window( X1, Y1, X2, Y2 : Byte );
Procedure TextColor( Color : Byte );
Procedure TextBackground( Color : Byte );
Procedure GotoXY( X, Y : Byte );
Procedure ClrScr;
Procedure Operation( InMsg : SmallString );
Function  FullOperation : GenlString;


IMPLEMENTATION

{-------------------------------------------------------------------------}
{ Procedure Remote                                                        }
{                                                                         }
{ If remote communications are enabled, this procedure writes a string to }
{ the remote device using write() statement.                              }
{-------------------------------------------------------------------------}
Procedure Remote( InLn : String );

  BEGIN
    If not Sys.Remote THEN Exit;
    If RemoteForceNL THEN InLn := NL + InLn;
    RemoteForceNL := False;
    Write( R, InLn );
  END;


{-------------------------------------------------------------------------}
{ Procedure RemoteLn                                                      }
{                                                                         }
{ If remote communications are enabled, this procedure writes a string to }
{ the remote device using writeln() statement.                            }
{-------------------------------------------------------------------------}
Procedure RemoteLn( InLn : String );

  BEGIN
    If not Sys.Remote THEN Exit;
    If RemoteForceNL THEN InLn := NL + InLn;
    RemoteForceNL := False;
    Writeln( R, InLn );
  END;


{-------------------------------------------------------------------------}
{ Procedure CloseRemote                                                   }
{                                                                         }
{ Closes the remote device.                                               }
{-------------------------------------------------------------------------}
Procedure CloseRemote;

  BEGIN
    Close( R );
    ExitProc := GlobalExitSave;
  END;


{-------------------------------------------------------------------------}
{ Procedure Disp                                                          }
{                                                                         }
{ Writes to the screen and the remote device using a write() statement.   }
{ If the Sys.RemoteOnly flag is set, the string is written only to the    }
{ remote device.                                                          }
{-------------------------------------------------------------------------}
Procedure Disp( InLn : String );

  BEGIN
    If not Sys.RemoteOnly
      THEN Write( InLn );
    Remote( InLn );
  END;


{-------------------------------------------------------------------------}
{ Procedure DispLn                                                        }
{                                                                         }
{ Writes to the screen and the remote device using a writeln() statement. }
{ If the Sys.RemoteOnly flag is set, the string is written only to the    }
{ remote device.                                                          }
{-------------------------------------------------------------------------}
Procedure DispLn( InLn : String );

  BEGIN
    If not Sys.RemoteOnly
      THEN WriteLn( InLn );
    RemoteLn( InLn );
  END;


{-------------------------------------------------------------------------}
{ Procedure Window                                                        }
{                                                                         }
{ Sets up a window on the screen with the coordinates (X1,Y1) - (X2,Y2).  }
{-------------------------------------------------------------------------}
Procedure Window( X1, Y1, X2, Y2 : Byte );

 Begin
    If not Sys.RemoteOnly
      THEN Crt.Window( X1, Y1, X2, Y2 );
  End;


{-------------------------------------------------------------------------}
{ Procedure TextColor                                                     }
{                                                                         }
{ Sets the textcolor screen attribute unless the system is running on     }
{ remote only.                                                            }
{-------------------------------------------------------------------------}
Procedure TextColor( Color : Byte );

  Begin
    If not Sys.RemoteOnly
      THEN Crt.TextColor( Color );
  End;


{-------------------------------------------------------------------------}
{ Procedure TextBackground                                                }
{                                                                         }
{ Sets the textbackground screen attribute unless the system is running on}
{ remote only.                                                            }
{-------------------------------------------------------------------------}
Procedure TextBackground( Color : Byte );

  Begin
    If not Sys.RemoteOnly
      THEN Crt.TextBackground( Color );
  End;


{-------------------------------------------------------------------------}
{ Procedure GotoXY                                                        }
{                                                                         }
{ Moves the cursor to the (X,Y) position on the screen unless the system  }
{ is sending output to the remote device only.                            }
{-------------------------------------------------------------------------}
Procedure GotoXY( X, Y : Byte );

  Begin
    If not Sys.RemoteOnly
      THEN Crt.GotoXY( X, Y );
  End;


{-------------------------------------------------------------------------}
{ Procedure ClrScr                                                        }
{                                                                         }
{ Clears the screen.                                                      }
{-------------------------------------------------------------------------}
Procedure ClrScr;

  Begin
    If not Sys.RemoteOnly
      THEN Crt.ClrScr;
  End;


{-------------------------------------------------------------------------}
{ Procedure BasicChkErr                                                   }
{                                                                         }
{ Puts an error message on the screen and prompts for <Enter>. The error  }
{ description must be passed to the procedure in Desc.                    }
{-------------------------------------------------------------------------}
PROCEDURE BasicChkErr( Result:Integer; Errtype:Char;
                       FSpec:GenlString; Desc:GenlString );

  BEGIN
    Str( Result, Desc );
    DispLn( NL + 'Error # ' + Desc + ' with ' + FSpec + NL );
    If ErrType <> UpCase( ErrType ) THEN Halt;
    DispLn( 'Press <ENTER> to continue...' );
    ReadLn( ErrType );
    Halt( 253 );
  END;


{----------------------------------------------------------------------}
{ Procedure Operation                                                  }
{                                                                      }
{ This procedure helps track locations in executed code by pushing     }
{ and poping the names of the operation currently being executed.      }
{ It is used by the CheckStack and TidyExit procedures.                }
{ Operation should be called once at the top of each significant       }
{ operation (it is called in Menu, Select, GetKey, ...), and then      }
{ again at the end of the operation with no name to pop.               }
{----------------------------------------------------------------------}
Procedure Operation( InMsg : SmallString );

  Begin
    If InMsg = '' Then
      Begin
        If OperationStackTop > 0 Then Dec( OperationStackTop )
      End
    Else
      Begin
        Inc( OperationStackTop );
        If OperationStackTop <= MaxOperation Then
          OperationStack[ OperationStackTop ] := InMsg;
      End;
  End;


{----------------------------------------------------------------------}
{ Function FullOperation                                               }
{                                                                      }
{ This returns the operation stack in a displayable format.            }
{----------------------------------------------------------------------}
Function FullOperation : GenlString;

  Var LoopTop, FullLoop : Byte;  FullMsg : GenlString;

  Begin
    If OperationStackTop = 0 Then FullMsg := 'Startup' Else FullMsg := '';
    LoopTop := OperationStackTop;
    If LoopTop > MaxOperation Then LoopTop := MaxOperation;
    For FullLoop := 1 to LoopTop do
      Begin
        If FullMsg <> '' Then FullMsg := FullMsg + ', ';
        FullMsg := FullMsg + OperationStack[ FullLoop ];
      End;
    If OperationStackTop > MaxOperation Then
      FullMsg := FullMsg + ', ...';
    FullOperation := FullMsg;
  End;

{-------------------------------------------------------------------------}

BEGIN

  OperationStackTop := 0;
  ChkErr := BasicChkErr;
  FillChar( Sys, SizeOf( Sys ), 0 );
  {$IFDEF DEMO } Sys.Demo := True; {$ENDIF }
  Sys.InstallCode := '000';
  Randomize;
  Sys.ForceRemote := GetEnv( 'REMOTE' ) <> '';
  If Sys.ForceRemote
    THEN BEGIN
      Sys.Remote := True;
      Sys.RemoteOnly := True;
    END
    ELSE BEGIN
      Sys.Remote := True;
      Sys.RemoteOnly := GetEnv( 'SHARE' ) = '';
    END;
  Assign( R, '' );  ReWrite( R );
  GlobalExitSave := ExitProc;
  ExitProc := @CloseRemote;

END.

