{$A+,B-,D+,E-,F-,I-,L+,N-,O-,R+,S+,V-}
{$M 16384,0,0}

(*********************************************************************
**********************************************************************
****                                                              ****
****                          GRAPH.PAS                           ****
****                                                              ****
****                  Bit mapped grahics support.                 ****
****                                                              ****
**********************************************************************
**********************************************************************)


Uses Dos;

Const
  MaxCrtMode = 7;
  CrtData : Array [ 1..MaxCrtMode ] of
    Record
      Name : String[6];  Mode, MaxC : Byte;  MaxX, MaxY : Integer;
    End =
      ( ( Name:'LoCGA'; Mode:$04; MaxC:3; MaxX:319; MaxY:199 ),
        ( Name:'HiCGA'; Mode:$06; MaxC:1; MaxX:639; MaxY:199 ),
        ( Name:'Tandy1'; Mode:$09; MaxC:15; MaxX:319; MaxY:199 ),
        ( Name:'Tandy2'; Mode:$0A; MaxC:3; MaxX:639; MaxY:199 ),
        ( Name:'EGA'; Mode:$0E; MaxC:15; MaxX:639; MaxY:199 ),
        ( Name:'HiEGA'; Mode:$10; MaxC:15; MaxX:639; MaxY:349 ),
        ( Name:'VGA'; Mode:$12; MaxC:15; MaxX:639; MaxY:479 ) );
  DevMaxX : Integer = 0;  DevMaxY : Integer = 0;
  DevMaxColor : Byte = 0;  GraphDevName : String[6] = 'None';
  CrtMode : Byte = 0;  PrevMode : Byte = 0;
  _LoCGA = $04;  _HiCGA = $06;
  _Tandy1 = $09;  _Tandy2 = $0A;  _EGA = $0E;
  WXFactor : Real = 1.0;  WYFactor : Real = 1.0;
  WX1 : Real = 0.0;   WX2 : Real = 0.0;
  WY1 : Real = 0.0;   WY2 : Real = 0.0;

Var
  GraphExitSave : Pointer;


(* USED FOR DEBUGGING TESTS *)
Var
  Log : Text;
  ExistsResult,  Loop : Integer;

Function I_S( I : Integer ) : String;
  Var S : String;
  Begin  Str( I:1, S );  I_S := S;  End;

Procedure WaitKey;
  Var S : String;
  Begin  ReadLn( S );  End;


{---------------------------------------------------------------------------}

Procedure SetWorld( X1, Y1, X2, Y2 : Real );
  Var VRegs : Registers;
  Begin
    WX1 := X1;  WY1 := Y1;  WX2 := X2;  WY2 := Y2;
    WXFactor := DevMaxX / ( WX2 - WX1 );
    WYFactor := DevMaxX / ( WY2 - WY1 );
  End;

{---------------------------------------------------------------------------}

Procedure Shade( CIndex, CShade:Byte );
  Var VRegs : Registers;
  Begin
    VRegs.AH := 16;        { function }
    VRegs.AL := 0;
    VRegs.BH := CShade;    { color shade }
    VRegs.BL := CIndex;    { color number }
    Intr( $10, VRegs );
    If CIndex <> 0 Then Exit;
    { also change screen border if background shade changed }
    VRegs.AH := 16;        { function }
    VRegs.AL := 1;
    VRegs.BH := CShade;    { color shade }
    VRegs.BL := 0;         { color number }
    Intr( $10, VRegs );
  End;

{---------------------------------------------------------------------------}

Function WDevX( X : Real ) : Integer;
  Begin
    If WXFactor = 1.0 Then
      WDevX := Trunc( X )
    Else
      WDevX := Trunc( X * WXFactor );
  End;

{---------------------------------------------------------------------------}

Function WDevY( Y : Real ) : Integer;
  Begin
    If WYFactor = 1.0 Then
      WDevY := Trunc( Y )
    Else
      WDevY := Trunc( Y * WYFactor );
  End;

{---------------------------------------------------------------------------}

Procedure DevSetDot( X, Y : Integer;  C : Integer );
  Var VRegs : Registers;
  Begin
    VRegs.AH := 12;        { function }
    VRegs.AL := C;         { color }
    VRegs.CX := X;         { x coordinate }
    VRegs.DX := Y;         { y coordinate }
    Intr( $10, VRegs );
  End;

{---------------------------------------------------------------------------}

Function DevGetDot( X, Y : Integer ) : Byte;
  Var VRegs : Registers;
  Begin
    VRegs.AH := 13;        { function }
    VRegs.CX := X;         { x coordinate }
    VRegs.DX := Y;         { y coordinate }
    Intr( $10, VRegs );
    DevGetDot := VRegs.AL;    { returned color }
  End;

{---------------------------------------------------------------------------}

Function  ModColor( C : Byte ) : Byte;
  Begin
    If C > DevMaxColor Then
      ModColor := C mod ( DevMaxColor + 1 )
    Else
      ModColor := C;
  End;

{---------------------------------------------------------------------------}

Procedure SetDot( X, Y : Real;  C : Integer );
  Var VRegs : Registers;
  Begin
    If ( X < WX1 ) or ( X > WX2 ) or
       ( Y < WY1 ) or ( Y > WY2 ) Then Exit;  { clip }
    DevSetDot( WDevX( X ), WDevY( Y ), ModColor( C ) );
  End;

{---------------------------------------------------------------------------}

Function GetDot( X, Y : Real ) : Byte;
  Var VRegs : Registers;
  Begin
    If ( X < WX1 ) or ( X > WX2 ) or
       ( Y < WY1 ) or ( Y > WY2 ) Then
      Begin  GetDot := 0;  Exit;  End;  { clip }
    GetDot := DevGetDot( WDevX( X ), WDevY( Y ) );
  End;

{---------------------------------------------------------------------------}

Procedure GWrite( X, Y : Real;  Attr : Integer;  Msg : String );
  Var  VRegs : Registers;  GLoop, DX, DY : Integer;
  Begin
    If ( X < WX1 / 8 ) or ( X > WX2 / 8 ) or
       ( Y < WY1 / 8 ) or ( Y > WY2 / 8 ) Then Exit;  { clip }
    DY := WDevY( Y );  DX := WDevX( X );  GLoop := 1;
    VRegs.AH := $02;  VRegs.BH := 0;
    VRegs.DH := DY;  VRegs.DL := DX;
    Intr( $10, VRegs );  { position cursor }
    While ( GLoop <= Length( Msg ) ) and
          ( DY <= DevMaxY - 7 ) and ( DX <= DevMaxX - 7 ) do
      Begin
        If Msg[ GLoop ] in [ ' '..'~' ] Then
          Begin
            VRegs.AH := $09;                  { function }
            VRegs.AL := Ord( Msg[ GLoop ] );  { character }
            VRegs.BH := $00;                  { bank }
            VRegs.BL := Lo( Attr );           { attribute }
            VRegs.CX := 1;                    { repetitions }
            Intr( $10, VRegs );
            Inc( DX );
            If DX > DevMaxX - 7 Then   { line wrap }
              Begin  DX := 0;  DY := DY + 8;  End;
            VRegs.AH := $02;  VRegs.BH := 0;
            VRegs.DH := DY;  VRegs.DL := DX;
            Intr( $10, VRegs );  { position cursor }
          End;
        Inc( GLoop );
      End;
  End;

{---------------------------------------------------------------------------}

Function GetMode : Byte;
  Var VRegs : Registers;
  Begin
    VRegs.AH := 15;
    Intr( $10, VRegs );
    GetMode := VRegs.AL;
  End;

{---------------------------------------------------------------------------}

Procedure SetMode( ModeNum : Byte );
  Var VRegs : Registers;
  Begin
    VRegs.AH := 0;         { function }
    VRegs.AL := ModeNum;   { video mode }
    Intr( $10, VRegs );
    { Available modes:
        0=CO40, 1=BW40, 2=CO80, 3=BW80, 4=320x200x4,
        5=320x200x1, 6=640x200x1, 7=mono, D=320x200x16,
        E=640x200x16, F=640x350x4, $10=640x350x16, $12=640x480x16  }
  End;

{---------------------------------------------------------------------------}

Procedure TextMode;
  Begin  SetMode( PrevMode );  End;

{---------------------------------------------------------------------------}

Procedure DevLine( X1, Y1, X2, Y2 : Integer;  C : Byte );
  Var
    DeltaX, DeltaY : Real;
    XLoop, YLoop, LLoop, Dif : Integer;
  BEGIN
    C := ModColor( C );
    { optimize depending on slope and length of line }
    If ( X1 = X2 ) and ( Y1 = Y2 ) Then
      DevSetDot( X1, Y1, C )  { single dot }
    Else
      If X1 = X2 Then
        Begin  { vertical line }
          If Y2 > Y1 Then
            For YLoop := Y1 to Y2 do
              DevSetDot( X1, YLoop, C )
          Else
            For YLoop := Y1 downto Y2 do
              DevSetDot( X1, YLoop, C );
        End  { vertical line }
      Else
        If Y1 = Y2 Then
          Begin  { horizontal line }
            If X2 > X1 Then
              For XLoop := X1 to X2 do
                DevSetDot( XLoop, Y1, C )
            Else
              For XLoop := X1 downto X2 do
                DevSetDot( XLoop, Y1, C );
          End  { horizontal line }
        Else
          If ( X1 <> X2 ) and ( Y1 <> Y2 ) Then
            Begin  { angled line }
              DevSetDot( X1, Y1, C );  DevSetDot( X2, Y2, C );
              If Abs( X1 - X2 ) > Abs( Y1 - Y2 ) Then
                Dif := Abs( X1 - X2 )  { wider than high }
              Else
                Dif := Abs( Y1 - Y2 );  { higher than wide }
              DeltaX := ( X1 - X2 ) / Dif;
              DeltaY := ( Y1 - Y2 ) / Dif;
              For LLoop := 0 TO Dif - 1 do
                Begin
                  XLoop := Trunc( X2 + DeltaX * LLoop );
                  YLoop := Trunc( Y2 + DeltaY * LLoop );
                  DevSetDot( XLoop, YLoop, C );
                End;
            End;  { angled line }
  End;

{---------------------------------------------------------------------------}

Procedure Line( X1, Y1, X2, Y2 : Real;  Wid, C : Byte );
  Var
    M, B, PX1, PY1, PX2, PY2, DX1, DY1, DX2, DY2 : Real;
    DeltaX, DeltaY : Real;
    LX1, LY1, LX2, LY2, XLoop, YLoop, LLoop, Dif : Integer;
  BEGIN
    { check if line is entirely outside of window }
    Writeln( Log, 'World = ', WX1:1:4, ', ', WY1:1:4,
                    ' : ', WX2:1:4, ', ', WY2:1:4 );
    Writeln( Log, 'In = ', X1:1:4, ', ', Y1:1:4,
                    ' : ', X2:1:4, ', ', Y2:1:4 );
    If ( ( X1 < WX1 ) and ( X2 < WX1 ) ) or
       ( ( X1 > WX2 ) and ( X2 > WX2 ) ) or
       ( ( Y1 < WY1 ) and ( Y2 < WY1 ) ) or
       ( ( Y1 > WX2 ) and ( Y2 > WX2 ) ) Then Exit;
    { line might pass through window, calculate entry and exit points }
    { y = mx + b    y - mx = b   y - b = mx   ( y - b ) / m = x   }
    If X1 = X2 Then
      M := 1E20  { vertical line, virtual infinity }
    Else
      M := ( Y1 - Y2 ) / ( X1 - X2 );  { regular slope }
    B := Y1 - M * X1;
    Writeln( Log, 'M = ', M:1:4, ',    B = ', B:1:4 );
    PX1 := ( WY1 - B ) / M;  PX2 := ( WY2 - B ) / M;
    PY1 := M * WX1 + B;  PY2 := M * WX2 + B;
    Writeln( Log, 'Out = ', PX1:1:4, ', ', PY1:1:4,
                     ' : ', PX2:1:4, ', ', PY2:1:4 );
    { check if line doesn't cross through window }
    If ( ( PX1 < WX1 ) and ( PX2 < WX1 ) ) or
       ( ( PX1 > WX2 ) and ( PX2 > WX2 ) ) or
       ( ( PY1 < WY1 ) and ( PY2 < WY1 ) ) or
       ( ( PY1 > WX2 ) and ( PY2 > WX2 ) ) Then Exit;
    { find the valid pair of points from the four calculated intersections }
    If ( PX1 >= WX1 ) and ( PX1 <= WX2 ) Then
      Begin  DX1 := PX1;  DY1 := WY1;  End;
    If ( PX2 >= WX1 ) and ( PX2 <= WX2 ) Then
      Begin  DX1 := PX2;  DY1 := WY2;  End;
    If ( PY1 >= WY1 ) and ( PY1 <= WY2 ) Then
      Begin  DX2 := WX1;  DY2 := PY1;  End;
    If ( PY2 >= WY1 ) and ( PY2 <= WY2 ) Then
      Begin  DX2 := WX2;  DY2 := PY2;  End;
    { coordinates determined, now draw the line }
    Writeln( Log, 'Draw = ', DX1:1:4, ', ', DY1:1:4,
                      ' : ', DX2:1:4, ', ', DY2:1:4 );
    DevLine( WDevX( DX1 ), WDevY( DY1 ),
             WDevX( DX2 ), WDevY( DY2 ), C );
  End;

{---------------------------------------------------------------------------}

Procedure FindGraphMode;
  Var
    ChkMode, ModeLoop, StartMode : Byte;  Found : Boolean;
    VRegs : Registers;
  Begin
    VRegs.AH := $03;  VRegs.BH := 0;  Intr( $10, VRegs );  { get cursor }
    StartMode := MaxCrtMode;
    For ModeLoop := MaxCrtMode downto 1 do
      If CrtMode = CrtData[ ModeLoop ].Mode Then StartMode := ModeLoop;
    ModeLoop := StartMode;  Found := False;
    Found := True;  ModeLoop := MaxCrtMode;
    While ( not Found ) and ( ModeLoop >= 1 ) do
      Begin
        SetMode( $80 or CrtData[ ModeLoop ].Mode );  { don't clear }
        ChkMode := GetMode and $7F;
        Found := ChkMode = CrtData[ ModeLoop ].Mode;
        If not Found Then Dec( ModeLoop );
      End;
    If Found Then
      Begin
        CrtMode := CrtData[ ModeLoop ].Mode;
        DevMaxColor := CrtData[ ModeLoop ].MaxC;
        DevMaxX := CrtData[ ModeLoop ].MaxX;
        DevMaxY := CrtData[ ModeLoop ].MaxY;
        GraphDevName := CrtData[ ModeLoop ].Name;
        WX1 := 0;  WX2 := DevMaxX;  WY1 := 0;  WY2 := DevMaxY;
      End;
    SetMode( $80 or PrevMode );
    VRegs.AH := $02;  Intr( $10, VRegs );  { restore cursor }
  End;

{---------------------------------------------------------------------------}

{$F+}
Procedure GraphExit;
  Begin
    ExitProc := GraphExitSave;
    If ( CrtMode > 0 ) and ( GetMode <> PrevMode ) Then TextMode;
  End;
{$F-}

{---------------------------------------------------------------------------}

Procedure GraphMode;
  Begin
    If CrtMode = 0 Then FindGraphMode;
    If CrtMode <> 0 Then SetMode( CrtMode );
  End;


{---------------------------------------------------------------------------}

Begin

  GraphExitSave := ExitProc;  ExitProc := @GraphExit;
  PrevMode := GetMode;

  Assign( Log, 'DEBUG.LOG' );
  {$I-} Append( Log ); {$I+} ExistsResult := IOResult;
  If ExistsResult <> 0 Then ReWrite( Log );
  Writeln( Log, 'Begin' );

  SetMode( $12 );
  Writeln( GetMode );

(*

  FindGraphMode;
  GraphMode;

  For Loop := 1 to 4 do
    GWrite( Loop - 1, Loop - 1, Loop, '1234567890' );

  GWrite( 1, 5, 1, I_S( DevMaxX ) + ' ' + I_S( DevMaxY ) + ' ' +
                   I_S( DevMaxColor ) + ' ' + GraphDevName );

  For Loop := 1 to 999 do
    DevSetDot( Random( DevMaxX + 1 ), Random( DevMaxY + 1 ),
               Random( DevMaxColor + 1 ) );

  For Loop := 1 to 100 do Shade( 1, Loop );

  SetWorld( 0, 0, 1000, 1000 );

  For Loop := 1 to DevMaxColor do
    DevLine( 0, 0, DevMaxX, 36, Loop );

*)

  Write( ^G );
  WaitKey;
  Writeln( Log, 'End' );
  Close( Log );

End.


