(*********************************************************************
**********************************************************************
****                                                              ****
****                         KEY.PAS                              ****
****                                                              ****
****        High level user input and dialog features.            ****
****                                                              ****
**********************************************************************
**********************************************************************)


Unit Key;


{$I LIBRARY.CFG }


INTERFACE


Uses Dos, Crt, Global, Sounds, Math, Strings, Video, Time;


CONST

{ vacant keyboard translation ranges:
    (28..31 reserved), 128, 130..142, 154..157, 167..171,
    (179..182 mapped),
    183..186, 197..198, 202,
    (206 mapped), (242 mapped), (243..246 mapped), (247 mapped)
    Note that these unused ranges return UnKnownKey from GetKey.  }

{ available keys currently not implemented:
    CtrlCenter5Key, CtrlInsKey, CtrlDelKey, AltBackSpaceKey,
    CtrlBackSpaceKey, AltEnterKey, Ctrl0Key..Ctrl9Key  }

  NoKey = #0;
  UnKnownKey = #255;

{ These are used for tutorial keyboard playback scripts. }
{ Keys between a pair of CommentCode's are consumed and ignored, }
{ MessageCode starts reading a message for display, }
{ end message with another MessageCode for "press any key" prompt, }
{ or end message with one or more PauseCodes then a MessageCode to }
{ wait one second for each PauseCode, then continue. }
  CommentCode = #28;  MessageCode = #29;  PauseCode = #30;
{ Character #31 is reserved for future use. }

{ control keys }            { 0..127, 143 }
  BackSpaceKey = ^H;  TabKey = ^I;  BackTabKey = #143;
  EnterKey = ^M;   ESCKey = ^[;  SpaceKey = ' ';
   AltEscKey = #129;  { used for macro record start/stop }

{ alt A..Z }                { 144..153, 158..166, 172..178 }
  AltQKey = #144;    AltWKey = #145;    AltEKey = #146;    AltRKey = #147;
  AltTKey = #148;    AltYKey = #149;    AltUKey = #150;    AltIKey = #151;
  AltOKey = #152;    AltPKey = #153;    AltAKey = #158;    AltSKey = #159;
  AltDKey = #160;    AltFKey = #161;    AltGKey = #162;    AltHKey = #163;
  AltJKey = #164;    AltKKey = #165;    AltLKey = #166;    AltZKey = #172;
  AltXKey = #173;    AltCKey = #174;    AltVKey = #175;    AltBKey = #176;
  AltNKey = #177;    AltMKey = #178;

{ function keys }           { 187..196 }
  F1Key = #187;    F2Key = #188;    F3Key = #189;    F4Key = #190;
  F5Key = #191;    F6Key = #192;    F7Key = #193;    F8Key = #194;
  F9Key = #195;    F10Key = #196;

{ cursor keys }             { 199..211, 243..246 }
  HomeKey = #199;
  EndKey = #207;          CtrlEndKey = #245;
  UpArKey = #200;         DnArKey = #208;
  LfArKey = #203;         CtrlLfArKey = #243;
  RtArKey = #205;         CtrlRtArKey = #244;
  PgUpKey = #201;
  PgDnKey = #209;         CtrlPgDnKey = #246;
  InsKey = #210;          DelKey = #211;
  Cursor5Key = #204;   { center key on numeric keypad }

{ extended cursor keys }    { mapped to 179..182 from 247, 260, 269, 273 }
  CtrlHomeKey = #179 {was 247};     CtrlPgUpKey = #180 {was 260};
  CtrlUpArKey = #181 {was 269};     CtrlDnArKey = #182 {was 273};

{ shift function keys }     { 212..221 }
  ShftF1Key = #212;    ShftF2Key = #213;    ShftF3Key = #214;
  ShftF4Key = #215;    ShftF5Key = #216;    ShftF6Key = #217;
  ShftF7Key = #218;    ShftF8Key = #219;    ShftF9Key = #220;
  ShftF10Key = #221;

{ control function keys }   { 222..231 }
  CtrlF1Key = #222;    CtrlF2Key = #223;    CtrlF3Key = #224;
  CtrlF4Key = #225;    CtrlF5Key = #226;    CtrlF6Key = #227;
  CtrlF7Key = #228;    CtrlF8Key = #229;    CtrlF9Key = #230;
  CtrlF10Key = #231;

{ alt function keys }       { 232..241 }
  AltF1Key = #232;    AltF2Key = #233;    AltF3Key = #234;
  AltF4Key = #235;    AltF5Key = #236;    AltF6Key = #237;
  AltF7Key = #238;    AltF8Key = #239;    AltF9Key = #240;
  AltF10Key = #241;

{ 243..246 mapped to extended cursor keys, above }

{ alt 0..9 }             { mapped to 247, 248..254, 206, 242, from 248..257 }
  Alt0Key = #247 {mapped from 257};
  Alt1Key = #248;    Alt2Key = #249;    Alt3Key = #250;   Alt4Key = #251;
  Alt5Key = #252;    Alt6Key = #253;    Alt7Key = #254;
  Alt8Key = #206 {mapped from 255};
  Alt9Key = #242 {mapped from 256};

CONST
{ HelpName and window are available to the program to allow use of the }
{ help system as a cross index browser of any type of data. }
  HelpName : String20 = 'help';
  HelpWX1 = 3;  HelpWY1 = 3;  HelpWX2 = 78;  HelpWY2 = 23;
  HelpDisabled : Boolean = True;
  DataViewDisabled : Boolean = False;
  MaxHelpChapterLen = 40;
  MacroRecord : Char = NoKey;
  MacroString : GenlString = '';
  KeyLogOpen : Boolean = False;
  KeyLog_TimeStampInterval = 10 * 60 * 1000; { ten minutes in m.s.'s }
  YesNoSet : Set of Char = [ 'Y', 'y', 'N', 'n' ];
  YesNoEscSet : Set of Char = [ 'Y', 'y', 'N', 'n', ESCKey ];
  AltKeySet : Set of Char = [ Alt0Key..Alt9Key, AltQKey..AltMKey ];
  MacroControlKey = AltEscKey;
{ These constants are used with Message to specify dialog box type. }
  Msg_OopKey = 0;      Msg_OopKeyHalt = 1;
  Msg_ClaxCall = 2;    Msg_ClaxCallHalt = 3;
  Msg_AttnKey = 4;     Msg_AttnYNEsc = 5;
  Msg_PlainPush = 6;   Msg_SirenCallHalt = 7;
  Msg_PlainPut = 8;
{ assists in DOS keypressed detection }
  DosPendingKeyCode : Word = Ord( UnKnownKey );

VAR
  HelpFileSpec : String80;
  HelpChapter : String[ MaxHelpChapterLen ];
  WaitKeyChar : Char;
  KeyLogName : GenlString;
  KeyLogFile : Text;
  DataView : Procedure;
  DataMgrView : Procedure;


Procedure Help;
Procedure WriteKeyLog( KeyLn : GenlString );
Procedure KeyLogComment( KeyLn : GenlString );
Procedure OpenKeyLog;
Procedure EatKeys;
Function  GetKey : Char;
Function  ReadKey : Char;
Function  KeyPressed : Boolean;
Function  NumLock : Boolean;
Function  CapsLock : Boolean;
Function  Select( Prompt : GenlString;  OkChars : CharSet ) : Char;
Procedure SetScreenSaver;
Procedure CheckScreenSaver;
Procedure Message( Msg : String;  Style : Integer );
Procedure Debug( Msg : String );
Procedure WaitKeypress;
Procedure WaitKey;
Procedure BadWaitKey;
Procedure GoodWaitKey;


{-------------------------------------------------------------------------}

IMPLEMENTATION

{-----------------------------------------------------------------------}
{ PROCEDURE RawHelp                                                     }
{                                                                       }
{ Brings up the help screen(s) and implements the cross-referencing     }
{ ability of the help system.                                           }
{-----------------------------------------------------------------------}

Procedure RawHelp;
  (*
      Help file imbeded control codes:
        ..IDX =  Optional quick find index data.
        ..IX  =  Index name for finding topic.
        ..HE  =  Chapter heading.
        ..PA  =  Hard page break.
        ^S topic ^S  =  Cross reference entries.
        ^B text ^B = Bold.
      All other dot commands are ignored.
      Required help file format is as follows:
        ..IDX MM-DD-YY HH:MM:SS
        ..IDX topic_name character_index:7
        ..IX topic_name
        ..HE topic_title
        help text
        ..PA
        more help text
        ..IX
  *)

  VAR
    HelpMsg, HelpTitle : GenlString;
    HelpLn, CheckLn, PendingLn : BigString;
    HelpFile : File Of Char;  HelpResult : Integer;
    HelpQuit, ChapterFound, ChapterDone : Boolean;
    HelpKey : Char;  HelpPageNum, HelpPageNumSave, HelpY : Word;
    HelpPutLoop,  HelpPutAttr : Byte;
    CrossRefX, CrossRefY, CrossRefWY1, CrossRefWY2 : Word;
    CrossRefs : Byte;  Junk : Boolean;
    CrossRef, NewChapter : String[ MaxHelpChapterLen ];
    HelpFileDateTime, HelpIndexDateTime : String20;
    HelpIndex : LongInt;
    HelpPagePosStack : Array[ 1..5 ] of LongInt;

  CONST
    HelpNormColor = Yellow;  HelpHiColor = White;
    CrossRefColor = LightBlue;
    CrossRefChar = ^S;  BoldChar = ^B;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  Procedure HelpWindow;
    { create the window in which help is displayed }
    VAR  CJ : Byte;  OutTitle : GenlString;
    BEGIN
      { reset cross reference variables for new page }
      CrossRefs := 0;  NewChapter := '';
      CrossRefX := HelpWX1;  CrossRefY := HelpWY1;
      { reversed ranges, optimized in histrip }
      CrossRefWY1 := HelpWY2;  CrossRefWY2 := HelpWY1;
      { draw frame and place titles }
      Frame( HelpWX1, HelpWY1, HelpWX2, HelpWY2 );  HelpY := 0;
      Window( 1, 1, 80, 25 );
      If HelpTitle = ''
        THEN OutTitle := Upper( HelpName )
        ELSE OutTitle := HelpTitle + ' ' + Upper( HelpName );
      If HelpPageNum > 1 THEN OutTitle := OutTitle + ' - Cont''d.';
      OutTitle := ' ' + OutTitle + ' ';
      WindowTitle( 0, FrameY1, White, OutTitle );
      OutTitle := ' Page ' + I_S( HelpPageNum ) + ' ';
      HelpPagePosStack[ HelpPageNum ] := FilePos( HelpFile );
      WindowTitle( FrameX2 - Length( OutTitle ) - 1, FrameY1, White,
                   OutTitle );
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ used by search for next cross-reference entry }

  Procedure IncCrossScan;
    BEGIN
      Inc( CrossRefX );
      If CrossRefX > HelpWX2
        THEN BEGIN
          CrossRefX := HelpWX1;
          Inc( CrossRefY );
          If CrossRefY > CrossRefWY2
            THEN CrossRefY := CrossRefWY1;
        END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ Finds cross reference entries on a help screen. }
{ Starting from the last cross reference position, }
{ reverse attributes are changed to cross ref attributes, }
{ while searching for a cross ref attribute.  Upon finding a cross }
{ ref attribute, attributes are set to reverse until an attribute }
{ that is neither reverse nor cross ref is found. }

  Procedure ShowCrossRef;
    Var
      CrossRefXSave, CrossRefYSave : Word;  ScrAttr : Byte;
      Found : Boolean;  FoundX, FoundY : Word;
    Begin
      If CrossRefY < CrossRefWY1 THEN CrossRefY := CrossRefWY1;
      CrossRefXSave := CrossRefX;  CrossRefYSave := CrossRefY;
      CrossRef := '';  Found := False;
      REPEAT
        ScrAttr := Screen^[ CrossRefY, CrossRefX ].Attr;
        If ScrAttr = CrossRefColor
          THEN BEGIN
            If not Found
              THEN BEGIN  FoundX := CrossRefX;  FoundY := CrossRefY;  END;
            Found := True;
            Screen^[ CrossRefY, CrossRefX ].Attr := Reverse;
            CrossRef := CrossRef + Screen^[ CrossRefY, CrossRefX ].Ch;
          END
          ELSE BEGIN
            If ScrAttr = Reverse
              THEN Screen^[ CrossRefY, CrossRefX ].Attr := CrossRefColor;
          END;
        If ( not Found ) or ( ScrAttr in [ CrossRefColor, Reverse ] )
          THEN IncCrossScan;
      UNTIL ( ( CrossRefX = CrossRefXSave ) and
              ( CrossRefY = CrossRefYSave ) ) or
            ( Found and ( not ( ScrAttr in [ CrossRefColor, Reverse ] ) ) );
      If Sys.Remote and ( CrossRefs > 0 ) THEN RemoteLn( CrossRef );
      If Found
        THEN BEGIN  CrossRefX := FoundX;  CrossRefY := FoundY;  END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  Function PageContinue( EndOfTopic : Boolean ) : Boolean;
    { prompt to continue help or quit }
    VAR  PromptKey : Char;  KeyOK, ContDone : Boolean;
    BEGIN
      AttentionSound;
        If EndOfTopic
          THEN BEGIN
            If CrossRefs > 0
              THEN WindowTitle( 0, FrameY2, White,
                                ' End of topic. ' +
                                ' Press <ESC> to exit ' + HelpName + ',' +
                                ' or <SPACE> for new topic. ' )
              ELSE WindowTitle( 0, FrameY2, White,
                                ' End of ' + HelpName + ' on this topic. ' +
                                ' Press <ESC> to exit ' + HelpName + '. ' );
          END
          ELSE BEGIN
            If CrossRefs > 0
              THEN WindowTitle( 0, FrameY2, White,
                                ' Press <PgDn> for next page,' +
                                ' <ESC> to exit ' + HelpName + ',' +
                                ' or <SPACE> for new topic. ' )
              ELSE WindowTitle( 0, FrameY2, White,
                                ' Press  <PgDn> for next page,' +
                                ' or  <ESC>  to exit ' + HelpName + '. ' );
          END;
      REPEAT
        ContDone := True;
        If ( not Sys.RemoteOnly ) and ( CrossRefs > 0 ) THEN ShowCrossRef;
        REPEAT
          EatKeys;
          PromptKey := GetKey;
          KeyOK :=
            ( PromptKey = EscKey ) or
            ( ( not EndOfTopic ) and ( PromptKey in [ ^C, PgDnKey ] ) ) or
            ( ( CrossRefs > 0 ) and ( PromptKey = EnterKey ) ) or
            ( ( CrossRefs > 1 ) and ( PromptKey = SpaceKey ) ) or
            ( ( HelpPageNum > 1 ) and ( PromptKey in [ ^R, PgUpKey ] ) );
          If not KeyOK THEN OopSound;
        UNTIL KeyOK;
        If PromptKey = ESCKey
          THEN BEGIN
            ZoopSound;  PageContinue := False;  HelpQuit := True;
          END;
        If PromptKey in [ PgDnKey, ^C ]
          THEN BEGIN
            ContinueSound;  Inc( HelpPageNum );
            PageContinue := True;  HelpWindow;
          END;
        If PromptKey in [ PgUpKey, ^R ]
          THEN BEGIN
            ContinueSound;  Dec( HelpPageNum );
            {$I-} Seek( HelpFile, HelpPagePosStack[ HelpPageNum ] ); {$I+}
            HelpResult := IOResult;
            PageContinue := True;  HelpWindow;
          END;
        If CrossRefs > 0
          THEN BEGIN
            If PromptKey = EnterKey
              THEN BEGIN
                ContinueSound;  NewChapter := CrossRef;
                PageContinue := False;  HelpQuit := True;
              END;
            If PromptKey = SpaceKey
              THEN BEGIN
                If Sys.Remote and Sys.RemoteOnly
                  THEN BEGIN
                    RemoteLn( 'Cross referenced ' + HelpName +
                              ' is not available to remote users.' );
                    ContDone := False;  OopSound;
                  END
                  ELSE BEGIN
                    DiSound;  ContDone := False;
                  END;
              END;
          END;
      UNTIL ContDone;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ strip bit 7 from string to read WordStar files }

  Procedure HiStrip( VAR HiLn : BigString );
    VAR  HiLoop, TempY : Byte;  CrossStarted : Boolean;
    BEGIN
      CrossStarted := False;
      For HiLoop := Length( HiLn ) DownTo 1 do
        BEGIN
          HiLn[ HILoop ] := Chr( Ord( HiLn[ HiLoop ] ) and $7F );
          If HiLn[ HiLoop ] = #31 THEN HiLn[ HiLoop ] := '-';
          { also count cross references and record their line range }
          If HiLn[ HiLoop ] = CrossRefChar
            THEN BEGIN
              TempY := HelpWY1 + HelpY - 1 + 2 { frame and space };
              If TempY < CrossRefWY1 THEN CrossRefWY1 := TempY;
              If TempY > CrossRefWY2 THEN CrossRefWY2 := TempY;
              If not CrossStarted Then Inc( CrossRefs );
              CrossStarted := not CrossStarted;
            END;
          If ( not ( HiLn[ HiLoop ] in
                       [ BoldChar, CrossRefChar ] ) )
             and ( not ( HiLn[ HiLoop ] in PrintableCharSet ) )
            THEN Delete( HiLn, HiLoop, 1 );
        END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ Read a line from disk upto a #10 character.  This is required }
{ because WordStar doesn't end all lines with a #13 which is }
{ required by Turbo's ReadLn procedure. }

  Procedure GetLn( Var GetLnOut : BigString );
    VAR  GetCh : Char;
    BEGIN
      GetCh := NoKey;  GetLnOut := '';
      While ( HelpResult = OK ) and
            ( not EOF( HelpFile ) ) and ( GetCh <> #10 ) do
        BEGIN
          {$I-} Read( HelpFile, GetCh ); {$I+} HelpResult := IOResult;
          If HelpResult = OK THEN GetLnOut := GetLnOut + GetCh;
        END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  BEGIN
    { search for the requested help topic }
    Operation( 'help' );
    Assign( HelpFile, HelpFileSpec );
    {$I-} Reset( HelpFile ); {$I+} HelpResult := IOResult;
    HelpPageNum := 1;
    If HelpResult = OK
      THEN BEGIN
        HelpMsg := ' Scanning ' + HelpName + ' . . .';
        Frame( 54 - ( Length( HelpMsg ) + 2 ), 9, 54, 11 );
        TextColor( White );  Disp( HelpMsg );  HelpMsg := '';
        RemoteForceNL := True;
      END
      ELSE HelpMsg := 'No ' + HelpName + ' is available for this program.';
    HelpTitle := '';  ChapterFound := False;
    HelpChapter := Upper( HelpChapter );
    { prepare for index detection }
    { index is used if date/time stamp in index section is equal to or }
    { later than file date/time }
    HelpIndexDateTime := '';  HelpFileDateTime := '';
    If HelpResult = OK Then
      Begin
        FindFirst( HelpFileSpec, Archive, Dir );
        If DosError = OK Then
          Begin
            UnPackTime( Dir.Time, DT );
            HelpFileDateTime :=
              YYMMDD( MakeDate( DT.Year, DT.Month, DT.Day ) ) + ' ' +
              MakeTime( DT.Hour, DT.Min, DT.Sec );
          End;
      End;
    { scan for topic }
    While ( HelpResult = OK ) and ( not EOF( HelpFile ) ) and
          ( not ChapterFound ) do
      BEGIN
        {$I-} GetLn( HelpLn ); {$I+} HelpResult := IOResult;
        If HelpResult = 0
          THEN BEGIN
            HiStrip( HelpLn );
            If Copy( HelpLn, 1, 2 ) = '..'
              THEN BEGIN
                CheckLn := HelpLn;  CheckLn := Upper( CheckLn );
                If ( HelpIndexDateTime >= HelpFileDateTime ) and
                   ( Copy( CheckLn, 1, 6 ) = '..IDX ' )
                  THEN BEGIN
                    HelpLn := Copy( HelpLn, 7, 255 );  Strip( HelpLn );
                    HelpLn := Upper( HelpLn );
                    HelpIndex :=
                      S_I( Copy( HelpLn, Length( HelpLn ) - 6, 7 ) );
                    HelpLn := Copy( HelpLn, 1, Length( HelpLn ) - 7 );
                    Strip( HelpLn );
                    If ( not ConversionError ) and
                       ( Copy( HelpLn, 1, MaxHelpChapterLen ) = HelpChapter )
                      THEN BEGIN
                        ChapterFound := True;
                        {$I-} Seek( HelpFile, HelpIndex ); {$I+}
                        HelpResult := IOResult;
                      END;
                  END;
                If ( HelpIndexDateTime = '' ) and
                   ( Copy( CheckLn, 1, 6 ) = '..IDX ' )
                  THEN BEGIN
                    HelpIndexDateTime :=  { MM-DD-YY HH:MM:SS }
                         Copy( CheckLn, 13, 2 ) + '-' + { YY- }
                         Copy( CheckLn, 7, 5 ) + { MM-DD }
                         Copy( CheckLn, 15, 9 );  {  HH:MM:SS }
                    If ( HelpIndexDateTime >= HelpFileDateTime )
                      THEN Debug( 'Help index is current.' )
                      ELSE Debug( 'Help index is out of date.' );
                  END;
                { check if index is requested topic }
                If Copy( CheckLn, 1, 5 ) = '..IX '
                  THEN BEGIN
                    HelpLn := Copy( HelpLn, 6, 255 );  Strip( HelpLn );
                    HelpLn := Upper( HelpLn );
                    ChapterFound :=
                      Copy( HelpLn, 1, MaxHelpChapterLen ) = HelpChapter;
                  END;
              END;
          END;
      END;
    If ( HelpResult <> OK ) and ( HelpMsg = '' )
      THEN HelpMsg := 'Trouble scanning ' + HelpName + ' file.';
    If ( HelpMsg = '' ) and ( not ChapterFound )
      THEN BEGIN
        HelpResult := 99;
        HelpMsg := 'No ' + HelpName + ' is available on this topic.';
      END;
    If HelpResult = OK THEN HelpWindow;
    ChapterDone := False;  HelpQuit := False;  PendingLn := '';
    { display help pages within topic }
    TextColor( HelpNormColor );
    While ( HelpResult = OK ) and
          ( not ( EOF( HelpFile ) and ( PendingLn = '' ) ) ) and
          ( not ChapterDone ) and ( not HelpQuit ) do
      BEGIN
        { PendingLn implements wrap of too long ines }
        Strip( PendingLn );
        If PendingLn = ''
          THEN GetLn( HelpLn )
          ELSE HelpLn := PendingLn;
        If HelpResult = 0
          THEN BEGIN
            HiStrip( HelpLn );  DeTrail( HelpLn );
            CheckLn := HelpLn;  CheckLn := Upper( CheckLn );
            If ( PendingLn = '' ) and ( Copy( CheckLn, 1, 1 ) = '.' )
              THEN BEGIN  { dot line, not displayed }
                If Copy( CheckLn, 1, 4 ) = '..PA'
                  THEN HelpQuit := not PageContinue( False );
                If Copy( CheckLn, 1, 4 ) = '..HE'
                  THEN BEGIN
                    HelpTitle := Copy( CheckLn, 6, 255 );
                    Strip( HelpTitle );
                  END;
                Strip( CheckLn );
                { blank index flags end of topic }
                If CheckLn = '..IX'
                  THEN ChapterDone := True
              END
              ELSE BEGIN  { not dot line, display }
                PendingLn := '';
                If PrintLen( HelpLn ) > 74
                  THEN BEGIN
                    PendingLn := Copy( HelpLn, 75, 255 );
                    HelpLn := Copy( HelpLn, 1, 74 );
                  END;
                Inc( HelpY );
                GotoXY( HelpWX1 + 1, HelpWY1 + 1 + HelpY - 1 );
                { put help line including bolds }
                For HelpPutLoop := 1 to Length( HelpLn ) do
                  Begin
                    If HelpLn[ HelpPutloop ] = BoldChar Then
                      Begin  { bold }
                        If TextAttr and $0F = HelpNormColor Then
                          TextColor( HelpHiColor )
                        Else
                          TextColor( HelpNormColor );
                      End
                    Else
                      If HelpLn[ HelpPutloop ] = CrossRefChar Then
                        Begin  { crossref }
                          If TextAttr and $0F = HelpNormColor Then
                            TextColor( CrossRefColor )
                          Else
                            TextColor( HelpNormColor );
                        End
                    Else
                      Disp( HelpLn[ HelpPutloop ] );
                  End;
                RemoteLn( '' );
                If EOF( HelpFile ) and ( PendingLn = '' )
                  THEN ChapterDone := True;
                If ( not ChapterDone ) and ( HelpY >= 19 )
                  THEN HelpQuit := not PageContinue( False );
              END;  { not dot line }
          END;
        If ChapterDone
          THEN BEGIN
            HelpPageNumSave := HelpPageNum;
            Junk := PageContinue( True );
            ChapterDone := HelpPageNum >= HelpPageNumSave;
          END;
      END;
    If ( HelpResult <> OK ) and ( HelpMsg = '' )
      THEN HelpMsg := 'Trouble reading ' + HelpName + ' file.';
    { report errors if any }
    If HelpResult <> 0
      THEN BEGIN
        PutScreen( ScreenStackPointer );
        Frame( 78 - ( Length( HelpMsg ) + 2 ), 9, 78, 11 );
        WindowTitle( 0, FrameY1, White, ' ' + Upper( HelpName ) + ' ' );
        Disp( HelpMsg );  RemoteForceNL := True;
        WindowTitle( 0, FrameY2, White + Blink, ' PRESS ANY KEY ' );
        NewChapter := '';  HelpChapter := '';
        OopSound;  While not KeyPressed do;  EatKeys;  ContinueSound;
      END;
    {$I-} Close( HelpFile ); {$I+} HelpResult := IOResult;
    HelpChapter := NewChapter;
    Operation( '' );
  END;

{-------------------------------------------------------------------------}

{ help is passed through RawHelp to implement cross-reference feature }
Procedure Help;
  Var HelpChapterSave : String[ MaxHelpChapterLen ];
  Begin
    { prepare screen }
    PushScreen;  Cursor( Off );  NormVideo;  BottomLn( ' ' );
    HelpDisabled := True;  EatKeys;  SpecialSound;
    { execute help }
    HelpChapterSave := HelpChapter;
    { Help returns cross reference index name, if any, in HelpChapter }
    Repeat RawHelp Until HelpChapter = '';
    HelpChapter := HelpChapterSave;
    { restore screen }
    PopScreen;  HelpDisabled := False;
  End;

{-------------------------------------------------------------------------}

Procedure WriteKeyLog( KeyLn : GenlString );
  { Writes to the optional keyboard recording file. }
  Var  KeyLogResult : Integer;
  Begin
    If ( not KeyLogOpen ) or ( not Sys.KeyLogEnabled ) THEN Exit;
    If TimerElapsed( KeyLogTimer )
      THEN BEGIN  { time stamp }
        KeyLn := KeyLn + CommentCode + TimeString + CommentCode;
        SetTimer( KeyLogTimer, KeyLog_TimeStampInterval );
      END;
    {$I-} Write( KeyLogFile, KeyLn ); {$I+} KeyLogResult := IOResult;
    ChkErr( KeyLogResult, 'O', KeyLogName, 'keyboard log.  File closed.' );
    Sys.KeyLogEnabled := KeyLogResult = OK;
  End;

{-------------------------------------------------------------------------}

Procedure KeyLogComment( KeyLn : GenlString );
  { Writes a comment to the keyboard recording file. }
  Begin
    KeyLn := CommentCode + KeyLn + CommentCode;
    WriteKeyLog( KeyLn );
  End;

{-------------------------------------------------------------------------}

Procedure OpenKeyLog;
  { Prepares the optional keyboard recording file. }
  Var  KeyLogResult : Integer;
  Begin
    If not Sys.KeyLogEnabled THEN Exit;
    KeyLogName := Sys.LogPath + 'KY' +
      Copy( DateString, 1, 2 ) + '01' + Copy( DateString, 7, 2 ) + '.LOG';
    Assign( KeyLogFile, KeyLogName );
    {$I-}
    Append( KeyLogFile );  KeyLogResult := IOResult;
    If KeyLogResult <> OK
      THEN BEGIN  ReWrite( KeyLogFile );  KeyLogResult := IOResult;  END;
    KeyLogResult := IOResult;
    ChkErr( KeyLogResult, 'O', KeyLogName, 'keyboard log.  File closed.' );
    Sys.KeyLogEnabled := KeyLogResult = OK;
    SetTimer( KeyLogTimer, KeyLog_TimeStampInterval );
    KeyLogOpen := Sys.KeyLogEnabled;
    WriteKeyLog( CommentCode + Sys.ProgramName + CommentCode );
  End;

{-------------------------------------------------------------------------}

Function BiosKeyPressed : Boolean;
  Begin
    If DosPendingKeyCode <> Ord( UnKnownKey )
      THEN BiosKeyPressed := True
      ELSE BEGIN
        Regs.AH := $01;  Intr( $16, Regs );
        BiosKeyPressed := Regs.Flags and FZero = 0;
      END;
  End;

{-------------------------------------------------------------------------}

Function DosKeyPressed : Boolean;
  Begin
    If DosPendingKeyCode <> Ord( UnKnownKey )
      THEN DosKeyPressed := True
      ELSE BEGIN
        Regs.AH := $06;  Regs.DL := $FF;  MsDos( Regs );
        If Regs.Flags and FZero = 0  { key available }
          THEN BEGIN
            DosKeyPressed := True;
            DosPendingKeyCode := Regs.AL;
            If DosPendingKeyCode = 0
              THEN BEGIN
                Regs.AH := $06;  Regs.DL := $FF;  MsDos( Regs );
                DosPendingKeyCode := Regs.AL * $0100;
              END;
          END
          ELSE DosKeyPressed := False;
      END;
  End;

{-------------------------------------------------------------------------}

Function KeyPressed : Boolean;
  BEGIN
    CheckStack;
    KeyPressed := DosKeyPressed or
      ( ( MacroString <> '' ) and ( MacroRecord = NoKey ) ) or
      ( BiosKeyPressed and ( not Sys.RemoteOnly ) )
  END;

{-------------------------------------------------------------------------}

Function DosGetKeyCode : Word;
  Begin
    If DosPendingKeyCode <> Ord( UnKnownKey )
      THEN BEGIN
        DosGetKeyCode := DosPendingKeyCode;
        DosPendingKeyCode := Ord( UnKnownKey );
      END
      ELSE BEGIN
        Regs.AH := $07;  MsDos( Regs );
        DosGetKeyCode := Regs.AL;
      END;
  End;

{-------------------------------------------------------------------------}

Function BiosGetKeyCode : Word;
  Begin
    If DosPendingKeyCode <> Ord( UnKnownKey )
      THEN BEGIN
        BiosGetKeyCode := DosPendingKeyCode;
        DosPendingKeyCode := Ord( UnKnownKey );
      END
      ELSE BEGIN
        Regs.AH := $00;  Intr( $16, Regs );
        BiosGetKeyCode := Regs.AX;
      END;
  End;

{-------------------------------------------------------------------------}

Function KeyXlate( KeyCode : Word ) : Word;
  { Translate extended key codes into lower ranges. }
  BEGIN
  { Map out of range keys. }
    If KeyCode = 247 THEN KeyCode := Ord( CtrlHomeKey );
    If KeyCode = 260 THEN KeyCode := Ord( CtrlPgUpKey );
    If KeyCode = 269 THEN KeyCode := Ord( CtrlUpArKey );
    If KeyCode = 273 THEN KeyCode := Ord( CtrlDnArKey );
    If KeyCode = 255 THEN KeyCode := Ord( Alt8Key );
    If KeyCode = 256 THEN KeyCode := Ord( Alt9Key );
    If KeyCode = 257 THEN KeyCode := Ord( Alt0Key );
  { Map unknown keys. }
    If KeyCode > 255
      THEN KeyCode := Ord( UnKnownKey );
    If KeyCode = 128
      THEN KeyCode := Ord( UnKnownKey );
    If ( KeyCode >= 130 ) and ( KeyCode <= 142 )
      THEN KeyCode := Ord( UnKnownKey );
    If ( KeyCode >= 154 ) and ( KeyCode <= 157 )
      THEN KeyCode := Ord( UnKnownKey );
    If ( KeyCode >= 167 ) and ( KeyCode <= 171 )
      THEN KeyCode := Ord( UnKnownKey );
    If ( KeyCode >= 183 ) and ( KeyCode <= 186 )
      THEN KeyCode := Ord( UnKnownKey );
    If ( KeyCode >= 197 ) and ( KeyCode <= 198 )
      THEN KeyCode := Ord( UnKnownKey );
    KeyXlate := Lo( KeyCode );
  END;

{-------------------------------------------------------------------------}

Function GetKeyCode : Word;  { assumes keypress true upon entry }
  Var  KeyHold : Word;
  Begin
    If DosKeyPressed and ( not BiosKeyPressed )  { get remote key first }
      THEN BEGIN
        KeyHold := DosGetKeyCode;
        If KeyHold = 0
          THEN KeyHold := KeyXlate( DosGetKeyCode + $80 );
      END
      ELSE BEGIN  { otherwise, get local key }
        KeyHold := BiosGetKeyCode;
        If Lo( KeyHold ) = 0
          THEN KeyHold := KeyXlate( Hi( KeyHold ) + $80 )
          ELSE KeyHold := Lo( KeyHold );
      END;
    GetKeyCode := KeyHold;
  End;

{-------------------------------------------------------------------------}

Procedure EatKeys;  { clear local keyboard buffer }

  Var  Eat : Word;

  Begin
    While BiosKeyPressed do
      Eat := BiosGetKeyCode;
  End;

{-------------------------------------------------------------------------}

Procedure HandlePlaybackKeys( InKey : Char );
  Var  KeyMsg : String;  Junk : Char;  Paused : Boolean;
  Begin
    Case InKey of
     PauseCode : Delay( 1000 );
     CommentCode :
       Repeat
         If DosKeyPressed
           THEN Junk := Chr( Lo( DosGetKeyCode ) )
       Until ( Junk = CommentCode ) or ( not DosKeyPressed );
     MessageCode :
       BEGIN
         KeyMsg := '';  Paused := False;
         Repeat
           If DosKeyPressed
             THEN BEGIN
               Junk := Chr( Lo( DosGetKeyCode ) );
               If ( Junk <> PauseCode ) and ( Junk <> MessageCode )
                 THEN KeyMsg := KeyMsg + Junk;
             END;
         Until ( Junk = PauseCode ) or ( Junk = MessageCode ) or
               ( not DosKeyPressed );
         If Junk = MessageCode
           THEN Message( KeyMsg, Msg_AttnKey )
           ELSE BEGIN
             AttentionSound;
             Message( KeyMsg, Msg_PlainPush );
             Delay( 1000 );
             Repeat
               If DosKeyPressed
                 THEN BEGIN
                   Junk := Chr( Lo( DosGetKeyCode ) );
                   If Junk = PauseCode THEN Delay( 1000 );
                 END;
             Until ( Junk = MessageCode ) or ( not DosKeyPressed );
             PopScreen;
           End;
       End;
     End;
  End;

{-------------------------------------------------------------------------}

Procedure HandleMacroKeys( InKey : Char );
  Var  MacroFile : File;  MacroResult, JunkResult : Integer;
       MacroFileName : GenlString;  MacroLen : Integer;
  Const AltKeySet : Set of Char =
          [ Alt0Key..Alt9Key, AltQKey..AltPKey, AltAKey..AltLKey,
            AltZKey..AltMKey, EscKey ];
  Begin
    If ( InKey = MacroControlKey ) and ( MacroRecord = NoKey )
      THEN BEGIN  { start recording }
        GrikSound;  MacroString := '';
        Message( 'MACRO KEY RECORDING:' + NL +
                 'Press a key, Alt-0..Alt-9, Alt-A..Alt-Z' + NL +
                 'to begin recording (or <ESC> to not record)   ' + ^H^H,
                 Msg_PlainPush );
        TextColor( Blink + White );  Write( '?' );
        TextColor( White );  Write( #250 );
        SetScreenSaver;
        Repeat
          MacroRecord := Chr( Lo( GetKeyCode ) );
          If not ( MacroRecord in AltKeySet ) THEN OopSound;
        Until MacroRecord in AltKeySet;
        If MacroRecord = ESCKey
          THEN BEGIN  ZoopSound;  MacroRecord := NoKey;  END
          ELSE GrikSound;
        PopScreen;  Exit;
      END;  { start recording }
    {$I-}
    If ( ( InKey = MacroControlKey ) and ( MacroRecord <> NoKey ) ) or
       ( MacroRecord = InKey )
      THEN BEGIN  { stop recording }
        MacroFileName := Sys.MiscPath + 'KEY' +
          PadString( I_S( Ord( MacroRecord ) ), '0', 3 ) + '.MAC';
        Assign( MacroFile, MacroFileName );
        GrikSound;
        If MacroString = ''
          THEN BEGIN  { erase if empty }
            Erase( MacroFile );  JunkResult := IOResult;
          END  { erase if empty }
          ELSE BEGIN  { otherwise, write to file }
            ReWrite( MacroFile, 1 );  MacroResult := IOResult;
            ChkErr( MacroResult, 'w', MacroFileName, 'macro file' );
            If MacroResult = OK
              THEN BEGIN  { open ok }
                BlockWrite( MacroFile, MacroString[ 1 ],
                            Length( MacroString ) );  MacroResult := IOResult;
                ChkErr( MacroResult, 'w', MacroFileName, 'macro file' );
              END;  { open ok }
          END;  { write to file }
        Close( MacroFile );  JunkResult := IOResult;
        MacroRecord := NoKey;  MacroString := '';
      END  { stop recording }
      ELSE BEGIN  { playback if exists }
        MacroRecord := NoKey;
        MacroFileName := Sys.MiscPath + 'KEY' +
          PadString( I_S( Ord( InKey ) ), '0', 3 ) + '.MAC';
        Assign( MacroFile, MacroFileName );
        Reset( MacroFile, 1 );  MacroResult := IOResult;
        If MacroResult = OK
          THEN BEGIN  { playback }
            BlockRead( MacroFile, MacroString[ 1 ],
                       SizeOf( MacroString ) - 1, MacroLen );
            MacroResult := IOResult;
            MacroString[ 0 ] := Chr( Lo( MacroLen ) );
            If ( MacroResult <> OK ) or ( MacroString = '' )
              THEN OopSound ELSE GrikSound;
          END  { playback }
          ELSE OopSound;  { unknown macro }
        If MacroResult <> OK THEN MacroString := '';
        Close( MacroFile );  JunkResult := IOResult;
      END;  { playback if exists }
    {$I+}
  End;

{-------------------------------------------------------------------------}

Procedure RawDataView;
  Begin
  End;

{---------------------------------------------------------------------------}

Procedure DoDataView;
  Begin
    If Sys.InstallCode <> '000' Then Exit;
    Operation( 'DataView' );
    PushScreen;
    BottomLn( ' Diagnostic data view window, press any key . . .' );
    Frame( 3, 3, 78, 16 );
    WindowTitle( 0, FrameY1, White, ' DIAGNOSTIC DATA VIEW WINDOW ' );
    DispLn( 'Name=' + FullName );
    DispLn( 'Operation=' + FullOperation );
    DispLn( 'MemAvail=' + I_S( MemAvail ) +
            ', MaxAvail=' + I_S( MaxAvail )  +
            ', ScreenStackPointer=' + I_S( ScreenStackPointer ) );
    DispLn( 'MaxStackUsed=' + I_S( MaxStackUsed ) +
            ' in ' + LowStackMsg );
    DispLn( 'LowRam=' + I_S( LowRam ) + ' in ' + LowRamMsg );
    { DataMan shows DCB's here by assigning procedure variable DataMgrView }
    DispLn( '' );
    DispLn( 'Press: F9 for DCB''s, F10 for program data, any other quits.' );
    DataViewDisabled := True;
    GoodWaitKey;
    If WaitKeyChar = F9Key Then
      Begin  ClrScr;  DataMgrView;  GoodWaitKey;  End;
    If WaitKeyChar = F10Key Then
      Begin  ClrScr;  DataView;  GoodWaitKey;  End;
    DataViewDisabled := False;
    PopScreen;
    Operation( '' );
  End;

{-------------------------------------------------------------------------}

Function GetKey : Char;
  VAR  Done : Boolean;  InKey : Char;
  BEGIN
    Done := False;
    Repeat
      SetScreenSaver;
      While not KeyPressed do CheckScreenSaver;
      If ( MacroString = '' ) or ( MacroRecord <> NoKey )
        THEN InKey := Chr( Lo( GetKeyCode ) )  { get raw key code }
        ELSE BEGIN   { or macro playback }
          InKey := MacroString[ 1 ];
          Delete( MacroString, 1, 1 );
        END;
      GetKey := InKey;
    { log the keypress }
      WriteKeyLog( InKey );
    { record macro character if applicable }
      If ( MacroRecord <> NoKey ) and
         ( not ( InKey in [ MacroRecord, MacroControlKey ] ) )
        THEN MacroString := MacroString + InKey;
    { If Key is help, macro, or script, go back and get another key. }
    { If Alt key, record or playback macro. }
      If InKey in [ MacroControlKey, Alt0Key..Alt9Key, AltQKey..AltPKey,
                    AltAKey..AltLKey, AltZKey..AltMKey ]
        THEN HandleMacroKeys( InKey )
        ELSE BEGIN  { script or help }
        { if script control, handle special codes }
          If InKey in [ CommentCode, MessageCode, PauseCode ]
            THEN HandlePlaybackKeys( InKey )
            ELSE BEGIN  { not script }
            { if help or data view, show it, otherwise return keypress }
              If not ( InKey in [ F1Key, ^J, AltF10Key ] )
                THEN Done := True
                ELSE BEGIN  { possible help or data view }
                  If ( InKey in [ F1Key, ^J ] ) and
                     ( not HelpDisabled )
                    THEN Help
                    ELSE If ( InKey in [ AltF10Key ] ) and
                            ( not DataViewDisabled )
                           THEN DoDataView
                           ELSE Done := True;
                END;  { possible help or data view }
            END;  { not script }
        END;  { script or help }
    Until Done;
  END;

(* ------------------------------------------------------------------------ *)

Function ReadKey : Char;            { synonym to replace Crt.ReadKey }
  Begin  ReadKey := GetKey;  End;

(**  END OF HELP, START OF EXTENDED KEYBOARD SUPPORT  ***********************)

Function NumLock : BOOLEAN;
  BEGIN
    Regs.AX := $0200;  { key board status}
    Intr( $16, Regs );   { do kbd interrupt }
    NumLock := Regs.AX AND $0020 <> 0;
  END;

(* ------------------------------------------------------------------------ *)

Function CapsLock : BOOLEAN;
  BEGIN
    Regs.AX := $0200;  { key board status }
    Intr( $16, Regs );  { do kbd interrupt }
    CapsLock := Regs.AX and $0040 <> 0;
  END;

(* ------------------------------------------------------------------------ *)

{ prompt, and wait for a keypress within the valid set }

Function Select( Prompt : GenlString;  OkChars : CharSet ) : Char;
  VAR  YNOnly : Boolean;  ChLoop : Byte;  Ch : Char;
  BEGIN
    ClrEol;  Disp( Prompt );
    KeyLogComment( 'select' );
    TextColor( White + Blink );  Disp( ' ? ' );
    Escape := False;  NormVideo;
    Disp( #250 + ^H );
    REPEAT
      Ch := GetKey;  Ch := Upcase(Ch);
      If not ( Ch in OkChars ) THEN OopSound;
    UNTIL Ch in OkChars;
    YNOnly := True;
    For ChLoop := 0 to 255 do
      If ( Chr( ChLoop ) in OkChars ) and
         ( not ( Chr( ChLoop ) in YesNoEscSet ) )
        THEN YNOnly := False;
    If not YNOnly
      THEN BEGIN
        Case Ch of
          ESCKey : ZoopSound;
          ELSE DeeDeeSound
         End;
      END
      ELSE BEGIN
        Case Ch of
          ESCKey : ZoopSound;
          'Y' : OkaySound;
          'N' : NOkaySound;
         End;
      END;
    IF ( Ch in PrintableCharSet ) THEN Disp( Ch );
    RemoteForceNL := True;
    Escape := Ch = ESCKey;
    Select := Ch;
  END;

(* ------------------------------------------------------------------------ *)

Procedure SetScreenSaver;
  BEGIN    { SetTimer uses milli-seconds }
    SetTimer( ScreenSaverTimer, Abs( Sys.ScreenSaverTime ) * 1000 );
  END;

{---------------------------------------------------------------------------}

Procedure CheckScreenSaver;
  VAR
    SaveLn : Array [ 1..ScreenWidth ] of Char;
    Row, XLoop, YLoop, DimColor : Word;
    HelpDisableSave : Boolean;
  CONST
    Prompt = '***  SCREEN SAVER MODE  --  PRESS ANY KEY  ***';
  BEGIN
    If Sys.Remote or ( not ScreenCleared ) or
       ( not TimerElapsed( ScreenSaverTimer ) ) THEN Exit;
    If VMode = _Mono Then DimColor := Black Else DimColor := Blue;
    Operation( 'screen_saver' );
    HelpDisableSave := HelpDisabled;  HelpDisabled := True;
    AttentionSound;
    PushScreen;  { save the screen }
    For YLoop := 1 to ScreenHeight do  { dim the attributes }
      For XLoop := 1 to ScreenWidth do
        Screen^[ YLoop, XLoop ].Attr := DimColor;
    Repeat
      Row := Random( ScreenHeight ) + 1;
      For XLoop := 1 to ScreenWidth do  { preserve the line }
        SaveLn[ XLoop ] := Screen^[ Row, XLoop ].Ch;
      VWrite( 40 - Length( Prompt ) div 2, Row, White + Blink, Prompt );
      SetTimer( 1, 15 * 1000 );  { milli-seconds }
      Repeat Until KeyPressed or TimerElapsed( 1 );
      For XLoop := 1 to ScreenWidth do  { restore the line }
        BEGIN
          Screen^[ Row, XLoop ].Attr := DimColor;
          Screen^[ Row, XLoop ].Ch := SaveLn[ XLoop ];
        END;
    Until KeyPressed;
    ContinueSound;
    EatKeys;
    If Sys.ScreenSaverTime < 0 THEN Halt( 254 );
    PopScreen;  HelpDisabled := HelpDisableSave;
    SetTimer( 1, Abs( Sys.ScreenSaverTime ) * 1000 );  { milli-seconds }
    Operation( '' );
  END;

{---------------------------------------------------------------------------}

Procedure Message( Msg : String;  Style : Integer );
  { Puts up a prompt in an automatically sized window. }
  { Style:                             }
  {   Msg_OopKey         = 0 = oop + key                   }
  {   Msg_OopKeyHalt     = 1 = oop + key + terminate       }
  {   Msg_ClaxCall       = 2 = claxon + call               }
  {   Msg_ClaxCallHalt   = 3 = claxon + call + terminate   }
  {   Msg_AttnKey        = 4 = attention + key             }
  {   Msg_AttnYNEsc      = 5 = attention + yes/no/esc      }
  {   Msg_PlainPush      = 6 = no sound, no key (caller must pop screen)  }
  {   Msg_SirenCallHalt  = 7 = siren + call + terminate    }
  {   Msg_PlainPut       = 8 = no sound, no key, no screen push  }
  VAR  WX1, WY1, X, X1, Y, Y1, Longest, CRs, MsgLoop : Integer;
       Prompt : SmallString;  LastX, LastY : Integer;
       Snd : Procedure;
  BEGIN
    Operation( 'message' );
    KeyLogComment( 'message ' + I_S( Style ) );
    If Style = Msg_AttnYNEsc THEN Msg := Msg + ' (Y/N)';
    If not ( Style in [ Msg_PlainPut, Msg_SirenCallHalt, Msg_ClaxCallHalt ] )
      THEN BEGIN  PushScreen;  Cursor( Off );  END;
    NormVideo;  LastX := 0;  LastY := 0;
    If ( Style = Msg_OopKeyHalt ) or
       ( Style = Msg_ClaxCallHalt ) or
       ( Style = Msg_SirenCallHalt )
      THEN Msg := Msg + NL + 'Program terminated.';
    MeasureMessage( Msg, Longest, CRs, LastX );
    Prompt := '';  Snd := NoSound;
    If Style = Msg_AttnYNEsc
      THEN If LastX + 4 > Longest
             THEN Longest := LastX + 4;
    If ( Style = Msg_ClaxCall ) or
       ( Style = Msg_ClaxCallHalt ) or
       ( Style = Msg_SirenCallHalt )
      THEN BEGIN  Prompt := ' ! CALL RSPE ! ';  Snd := ClaxonSound;  END;
    If ( Style = Msg_OopKey ) or ( Style = Msg_OopKeyHalt )
      THEN BEGIN  Prompt := ' PRESS ANY KEY ';  Snd := OopSound;  END;
    If Style = Msg_AttnKey
      THEN BEGIN  Prompt := ' PRESS ANY KEY ';  Snd := AttentionSound;  END;
    If Style = Msg_AttnYNEsc
      THEN BEGIN  Prompt := '';  Snd := AttentionSound;  END;
    If Length( Prompt ) > Longest THEN Longest := Length( Prompt );
    { center vertically }
    Y1 := ( ScreenHeight - TopMargin ) div 2 - ( CRs div 2 );
    X1 := 40 - ( Longest div 2 );  { center horizontally }
    If Style in [ Msg_OopKeyHalt, Msg_ClaxCallHalt, Msg_SirenCallHalt ] Then
      BorderColor := Blink + Reverse
    Else
      BorderColor := MessageBorderColor;
    BorderStyle := MessageBorderStyle;
    Frame( X1 - 1, TopMargin + Y1 - 1,
           X1 + Longest + 1, TopMargin + Y1 + CRs + 1 );
    Disp( Msg );  RemoteForceNL := True;
    If ( Style <> Msg_PlainPush ) and ( Style <> Msg_PlainPut )
      THEN BEGIN
        If Prompt <> ''
          THEN BEGIN
            WindowTitle( 0, FrameY2, White + Blink, Prompt );
            Snd;  WaitKeypress;
          END;
        If Style = Msg_AttnYNEsc
          THEN WaitKeyChar := Select( '', [ 'Y', 'N', ESCKey ] );
        PopScreen;
      END;
    Operation( '' );
  END;

(* ----------------------------------------------------------------------- *)

{ Display a message window of diagnostic data if debugging enabled. }

Procedure Debug( Msg : String );
  Begin
    If not Sys.DebugEnabled Then Exit;
    Message( Msg, Msg_AttnKey );
  End;

(* ----------------------------------------------------------------------- *)

Procedure WaitKeypress;
  BEGIN
    EatKeys;  WaitKeyChar := Getkey;  EatKeys;
    Escape := WaitKeyChar = ESCKey;
    If Escape THEN ZoopSound ELSE ContinueSound;
  END;

(* ----------------------------------------------------------------------- *)

Procedure WaitKey;
  Var  X : Integer;  Str : String100;
  Begin
    PushScreen;
    VWrite( 1, ScreenHeight, Blink + White,
      Center( '*******  PRESS ANY KEY TO CONTINUE  *******', ScreenWidth ) );
    WaitKeypress;  PopScreen;
  End;

(* ----------------------------------------------------------------------- *)

Procedure BadWaitKey;  BEGIN  OopSound;  WaitKey;  END;

(* ----------------------------------------------------------------------- *)

Procedure GoodWaitKey;  BEGIN  AttentionSound;  WaitKey;  END;

{---------------------------------------------------------------------------}

BEGIN

  DataView := RawDataView;
  DataMgrView := RawDataView;
  Sys.ScreenSaverTime := -240;  { 240 seconds = 4 minutes, <0 = quit }

END.
