(* -------------------------------------------------------------- *)
(*  Transaction entry.                                            *)
(* -------------------------------------------------------------- *)

Procedure FillInDisb;
  { fill in disbursements with those previously used by same name }
  Var
    FillIt : Boolean;
    FillLp : Integer;
    FillDisb, CmpDisb, EdtCmpDisb, MtchDisb : Byte;
    CmpAmt, CalcAmt, RtoAmt1, RtoAmt2 : Real;
    CmpAcct : DisbAcctType;
  Begin
    DisbFilled := False;
    If EdtHdrRec.Name = '' Then Exit;
    FillDisb := 1;  FillIt := not KeyPressed;
    CmpDisb := 0;  CmpAcct := '';
    { more than one disbursements already filled disqualifies }
    While ( FillDisb <= MaxDisb ) and FillIt do
      Begin
        If Amount( EdtDisbArr[ FillDisb ].Amt ) Then
          If ( EdtDisbArr[ FillDisb ].Acct <> '' ) and
             ( CmpDisb = 0 ) Then
            Begin
              { remember index of filled entry }
              CmpAmt := EdtDisbArr[ FillDisb ].Amt;
              CmpAcct := EdtDisbArr[ FillDisb ].Acct;
              CmpDisb := FillDisb;
            End
          Else
            FillIt := FillIt and ( EdtDisbArr[ FillDisb ].Acct = '' );
        Inc( FillDisb );
      End;
    { if valid for filling, search for same name }
    If FillIt and ( CmpDisb > 0 ) Then
      Begin
        { search backwards circularly from current }
        Prompt( 'Searching to fill disbursements . . .' );
        FillLp := TrxNum - 1;
        If FillLp <= 0 Then FillLp := TrxCount;
        While ( not KeyPressed ) and
              ( FillLp <> TrxNum ) and ( ErrMsg = '' ) do
          Begin
            ReadTrx( FillLp );
            If HdrRec.Name = EdtHdrRec.Name Then
              Begin  { name matches }
                { find matching disb of same sign }
                MtchDisb := 0;  FillDisb := 1;
                While ( MtchDisb = 0 ) and ( FillDisb <= MaxDisb ) do
                  If ( DisbArr[ FillDisb ].Acct = CmpAcct ) and
                     ( DisbArr[ FillDisb ].Amt * CmpAmt > 0.0 ) Then
                    MtchDisb := FillDisb
                  Else
                    Inc( FillDisb );
                { if match found, fill in from prior record }
                If MtchDisb > 0 Then
                  Begin
                    RtoAmt2 := Abs( DisbArr[ MtchDisb ].Amt );
                    FillLp := TrxNum + 1;  DisbFilled := True;
                    EdtCmpDisb := CmpDisb;
                    For FillDisb := 1 to MaxDisb do
                      If FillDisb <> MtchDisb Then
                        Begin
                          Inc( EdtCmpDisb );
                          If EdtCmpDisb > MaxDisb Then EdtCmpDisb := 1;
                          { skip over already filled disb }
                          If EdtCmpDisb = CmpDisb Then
                            Begin
                              Inc( EdtCmpDisb );
                              If EdtCmpDisb > MaxDisb Then
                                EdtCmpDisb := 1;
                            End;
                          { calculate ratio to whole amount }
                          CalcAmt := Abs( CmpAmt );
                          RtoAmt1 := Abs( DisbArr[ FillDisb ].Amt );
                          CalcAmt := CalcAmt * ( RtoAmt1 / RtoAmt2 );
                          { swap signs appropriately by field }
                          If DisbArr[ FillDisb ].Amt < 0.0 Then
                            CalcAmt := CalcAmt * -1.0;
                          { round at two decimal places }
                          CalcAmt := Round( CalcAmt * 100.0 ) / 100.0;
                          EdtDisbArr[ EdtCmpDisb ].Amt := CalcAmt;
                          EdtDisbArr[ EdtCmpDisb ].Acct :=
                            DisbArr[ FillDisb ].Acct;
                        End;
                  End;
              End;
            Dec( FillLp );
            If FillLp <= 0 Then FillLp := TrxCount;
          End;
      End;
    If DisbFilled Then
      Begin
        Prompt( 'Disbursements filled . . .' );
        Delay( 300 );  Prompt( '' );
      End;
  End;

{------------------------------------------------------------------}

Procedure CheckAgeLimit( InDate : DateType );
  Begin
    If ( InDate <> '' ) and ( ErrMsg = '' ) and ( MaxAge > 0 ) and
       ( DateInDays( InDate ) > 0 ) and 
       ( DateInDays( DateString ) > 0 ) and
       ( DateInDays( InDate ) < DateInDays( DateString ) - MaxAge ) Then
      ErrMsg := 'Maximum age setting prohibits change';
  End;

{------------------------------------------------------------------}

Procedure EditTrx;
  { edit all fields of a transaction }
  Var
    FldLp, DisbLp, DisbNum : Byte;
    KeyMsg : String40;
    PrevAmt : Real;
    PrevAcct : DisbAcctType;
    CompDate : DateType;
    PrevDisbFilled : Boolean;
  Begin
    PrevDisbFilled := False;
    ReadTrx( TrxNum );  FldLp := 1;  DisbNum := 1;
    PrevAmt := -1.0 * DisbArr[ 1 ].Amt;
    PrevAcct := DisbArr[ 1 ].Acct;
    If HdrRec.PrcDate > HdrRec.Date Then
      ExamineDate( HdrRec.PrcDate )
    Else
      ExamineDate( HdrRec.Date );
    { preserve editing record to avoid overwrite from background tasks }
    EdtHdrRec := HdrRec;  EdtDisbArr := DisbArr;  EdtMemo := MemoRec;
    UnDoHdrRec := HdrRec;  UnDoDisbArr := DisbArr;  UnDoMemo := MemoRec;
    UnDoTrxNum := TrxNum;
    Message( 'Edit transaction.' );
    CompDate := HdrRec.PrcDate;
    If CompDate < HdrRec.Date Then CompDate := HdrRec.Date;
    CheckAgeLimit( CompDate );
    While ( ErrMsg = '' ) and ( not Escape ) and
          ( FldLp <= Length( InFldOrder ) ) do
      Begin
        { enter fields according to field input order specifier string }
        Case InFldOrder[ FldLp ] of
          FlagMk :
            Begin
              Prompt( 'Enter user flag:  ' );
              InpStr( EdtHdrRec.Flag, FlagSize, _String );
              AnyChange := AnyChange or InpChange;
            End;
          DateMk :
            Begin
              Repeat
                Prompt( 'Enter date issued:  ' );
                If PrevDate <> '' Then
                  Begin
                    CrsrXY( 64, 24 );  Write( '<SPACE>=', PrevDate );
                  End;
                CrsrXY( 22, 24 );  SpclInpKeys := ' ';
                InpStr( EdtHdrRec.Date, DateSize, _Date );
                AnyChange := AnyChange or InpChange;
                If InpKey = ' ' Then EdtHdrRec.Date := PrevDate;
              Until InpKey <> ' ';
              ExamineDate( EdtHdrRec.Date );
            End;
          SeqMk :
            Begin
              Prompt( 'Enter sequence identifier:  ' );
              InpStr( EdtHdrRec.Seq, SeqSize, _String );
              AnyChange := AnyChange or InpChange;
            End;
          NameMk :
            Begin
              Prompt( 'Enter name:  ' );
              InpStr( EdtHdrRec.Name, NameSize, _Text );
              AnyChange := AnyChange or InpChange;
              { check for automatic disbursements fill }
              If not PrevDisbFilled Then
                FillInDisb;
              PrevDisbFilled := PrevDisbFilled or DisbFilled;
            End;
          AmtMk :
            Begin
              { various features on up/down/^return keys }
              DisbFilled := False;
              Repeat
                Repeat
                  { get amount }
                  Prompt( 'Enter disbursement ' + I_S( DisbNum ) + ':' );
                  If Amount( PrevAmt ) Then
                    Begin
                      KeyMsg := '<SPACE>=' + R_S( PrevAmt );
                      CrsrXY( 80 - Length( KeyMsg ), 24 );
                      Write( KeyMsg );
                    End;
                  CrsrXY( 24 + 1 + AmtSize + 1, 24 );  Write( AcctMk );
                  Write( Pad( EdtDisbArr[ DisbNum ].Acct, -AcctSize ) );
                  CrsrXY( 24, 24 );  Write( AmtMk );
                  SpclInpKeys := ' ' + UpKy + DnKy + ^E + ^X;
                  InpReal( EdtDisbArr[ DisbNum ].Amt );
                  AnyChange := AnyChange or InpChange;
                  If InpChange Then SetReCalc;
                  If InpKey = ' ' Then
                    Begin
                      ErrMsg := '';
                      EdtDisbArr[ DisbNum ].Amt := PrevAmt;
                      If EdtDisbArr[ DisbNum ].Acct = '' Then
                        EdtDisbArr[ DisbNum ].Acct := PrevAcct;
                    End;
                Until InpKey <> ' ';
                PrevAmt := -1.0 * EdtDisbArr[ DisbNum ].Amt;
                If ( ErrMsg = '' ) and ( not Escape ) and
                   ( InpKey = EntrKy ) Then
                  Begin
                    { get name }
                    HdrRec := EdtHdrRec;  DisbArr := EdtDisbArr;
                    MemoRec := EdtMemo;
                    WriteTrx( TrxNum );  ShowLine( TrxNum );
                    Prompt( 'Enter disbursement ' + I_S( DisbNum ) + ':' );
                    CrsrXY( 24, 24 );  Write( AmtMk );
                    Write( Pad(R_S( EdtDisbArr[ DisbNum ].Amt ),-AmtSize) );
                    CrsrXY( 24 + 1 + AmtSize + 1, 24 );  Write( AcctMk );
                    SpclInpKeys := UpKy + DnKy + ^E + ^X;
                    InpStr( EdtDisbArr[ DisbNum ].Acct, AcctSize, _String );
                    AnyChange := AnyChange or InpChange;
                    If InpChange Then SetReCalc;
                    { check for automatic disbursements fill }
                    If not PrevDisbFilled Then
                      FillInDisb;
                  End;
                PrevAcct := EdtDisbArr[ DisbNum ].Acct;
                HdrRec := EdtHdrRec;  MemoRec := EdtMemo;
                DisbArr := EdtDisbArr;
                WriteTrx( TrxNum );  ShowLine( TrxNum );
                If ( InpKey = DnKy ) or ( InpKey = ^X ) Then
                  Dec( DisbNum )
                Else
                  If ( InpKey = UpKy ) or ( InpKey = ^E ) or
                     ( InpKey = EntrKy ) Then
                    Inc( DisbNum );
                If DisbNum < 1 Then DisbNum := MaxDisb;
                If ( DisbNum > MaxDisb ) and ( InpKey = EntrKy ) Then
                  InpKey := TabKy;
                If DisbNum > MaxDisb Then DisbNum := 1;
              Until Escape or ( ErrMsg <> '' ) or DisbFilled or
                    ( InpKey = TabKy ) or ( InpKey = S_TabKy );
              SortDisbs;  EdtDisbArr := DisbArr;
              WriteTrx( TrxNum );  ShowLine( TrxNum );
              PrevDisbFilled := PrevDisbFilled or DisbFilled;
              For DisbLp := 1 to MaxDisb do
                Begin
                  If EdtDisbArr[ DisbLp ].Acct = '' Then
                    EdtDisbArr[ DisbLp ].Amt := 0.0;
                  If not Amount( EdtDisbArr[ DisbLp ].Amt ) Then
                    EdtDisbArr[ DisbLp ].Acct := '';
                End;
            End;
          PrcDateMk :
            Begin
              Repeat
                Prompt( 'Enter date processed:  ' );
                If PrevDate <> '' Then
                  Begin
                    CrsrXY( 64, 24 );  Write( '<SPACE>=', PrevDate );
                  End;
                CrsrXY( 25, 24 );  SpclInpKeys := ' ';
                InpStr( EdtHdrRec.PrcDate, DateSize, _Date );
                ShowError;
                AnyChange := AnyChange or InpChange;
                If InpKey = ' ' Then EdtHdrRec.PrcDate := PrevDate;
              Until InpKey <> ' ';
              ExamineDate( EdtHdrRec.PrcDate );
            End;
          MemoMk :
            Begin
              Prompt( 'Enter memo:  ' );
              InpStr( EdtMemo, MemoSize, _Memo );
              AnyChange := AnyChange or InpChange;
            End;
        End;
        If InpKey = S_TabKy Then Dec( FldLp ) Else Inc( FldLp );
        If FldLp < 1 Then FldLp := Length( InFldOrder );
        If ( FldLp > Length( InFldOrder ) ) and ( InpKey <> EntrKy ) Then
          FldLp := 1;
        HdrRec := EdtHdrRec;  DisbArr := EdtDisbArr;  MemoRec := EdtMemo;
        WriteTrx( TrxNum );  ShowLine( TrxNum );
      End;
    If ErrMsg = '' Then TranslateLine( FormatLine );
    ShowError;
  End;

{------------------------------------------------------------------}

Procedure InsertTrx;
  { insert a blank transaction in to data }
  Var
    TrxLp : Integer;
  Begin
    If ErrMsg <> '' Then Exit;
    If TrxCount >= TrxMax Then ErrMsg := 'Memory full';
    If TrxNum <= UnDoTrxNum Then UnDoTrxNum := -1;
    Message( 'Inserting . . .' );
    TrxLp := TrxCount;
    While ( ErrMsg = '' ) and ( TrxLp >= TrxNum ) do
      Begin
        ReadTrx( TrxLp );  WriteTrx( TrxLp + 1 );
        Dec( TrxLp );
      End;
    If ErrMsg = '' Then
      Begin
        CreateTrx;  WriteTrx( TrxNum );
      End;
    AnyChange := True;
  End;

{------------------------------------------------------------------}

Procedure DeleteTrx;
  { delete a transaction from data }
  Var
    TrxLp : Integer;
    CompDate : DateType;
  Begin
    If ( ErrMsg <> '' ) or
       ( TrxCount < 1 ) or ( TrxNum > TrxCount ) Then Exit;
    If TrxNum <= UnDoTrxNum Then UnDoTrxNum := -1;
    ReadTrx( TrxNum );
    CompDate := HdrRec.PrcDate;
    If CompDate < HdrRec.Date Then CompDate := HdrRec.Date;
    Message( 'Deleting . . .' );
    CheckAgeLimit( CompDate );
    TrxLp := TrxNum;
    While ( ErrMsg = '' ) and ( TrxLp <= TrxCount ) do
      Begin
        ReadTrx( TrxLp + 1 );  WriteTrx( TrxLp );
        Inc( TrxLp );
      End;
    If ( ErrMsg = '' ) and ( TrxNum < TrxCount ) Then
      Dec( TrxCount );
    AnyChange := True;
  End;

(*----------------------------------------------------------------*)
(*  Disk operations.                                              *)
(* -------------------------------------------------------------- *)

Function ErrorName( Err : Integer ) : String80;
  Var
    ErrNumMsg : String[10];
  Begin
    Str( Err, ErrNumMsg );
    ErrorName := 'Error # ' + ErrNumMsg;
    Case Err of
         0 : ErrorName := 'No error';
         2 : ErrorName := 'File not found';
         3 : ErrorName := 'Path not found';
         4 : ErrorName := 'Too many files open';
         5 : ErrorName := 'Access denied';
         6 : ErrorName := 'Invalid file handle';
        12 : ErrorName := 'Invalid file access code';
        15 : ErrorName := 'Invalid drive number';
        16 : ErrorName := 'Cannot remove current path';
        17 : ErrorName := 'Inter-device rename';
       100 : ErrorName := 'Access past file end';
       101 : ErrorName := 'Disk full';
       102 : ErrorName := 'File not assigned';
       103 : ErrorName := 'File not open';
       104 : ErrorName := 'Not open for input';
       105 : ErrorName := 'Not open for output';
       106 : ErrorName := 'Bad numeric format';
       150 : ErrorName := 'Write protected';
       151 : ErrorName := 'Unknown unit error';
       152 : ErrorName := 'Device not ready';
       154 : ErrorName := 'CRC error in data';
       155 : ErrorName := 'Drive request error';
       156 : ErrorName := 'Seek error';
       157 : ErrorName := 'Unknown disk type';
       158 : ErrorName := 'Sector not found';
       159 : ErrorName := 'Printer out of paper';
       160 : ErrorName := 'Printer off line';
       161 : ErrorName := 'Device read error';
       162 : ErrorName := 'Hardware failure';
       200 : ErrorName := 'Division by zero';
       201 : ErrorName := 'Range check';
       202 : ErrorName := 'Stack overflow';
       203 : ErrorName := 'Heap overflow';
       204 : ErrorName := 'Bad pointer operation';
       205 : ErrorName := 'Numeric overflow';
       206 : ErrorName := 'Numeric underflow';
       207 : ErrorName := 'Bad real number';
       255 : ErrorName := 'User break';
     End;
  End;

{------------------------------------------------------------------}

Procedure ChkErr( OpNm : String40 );
  { fully handle I/O error checking and reporting }
  Var
    Result : Integer;
  Begin
    Result := IOResult;
    If Result <> 0 Then
      If ErrMsg = '' Then
        ErrMsg := ErrorName( Result ) + ' ' + OpNm + ' ' + FileName;
  End;

{------------------------------------------------------------------}

Procedure ReadConf;
  { load configuration file }
  Var
    Cfg : Text;
    JunkResult : Integer;
  Begin
    If ErrMsg <> '' Then Exit;
    FileName := CfgFileName;  Assign( Cfg, FileName );
    Reset( Cfg );  JunkResult := IOResult;
    If JunkResult = 0 Then
      Begin
        ReadLn( Cfg, LdgrFileName );  ReadLn( Cfg, InFldOrder );
        ReadLn( Cfg, OutFldOrder );  ReadLn( Cfg, ShowAcct );
        ReadLn( Cfg, RangeStart );  ReadLn( Cfg, RangeEnd );
        ReadLn( Cfg, ReptType );  ReadLn( Cfg, ReptSort );
        ReadLn( Cfg, ListType );  ReadLn( Cfg, OutSelect );
        ReadLn( Cfg, OutName );  ReadLn( Cfg, FindTarget );
        ReadLn( Cfg, BalCalcMode );  ReadLn( Cfg, SortDir );
        ReadLn( Cfg, DirPath );  ReadLn( Cfg, MaxAge );
        ReadLn( Cfg, ReptAcct );  ReadLn( Cfg, ListShowGroups );
        ReadLn( Cfg, CutFileName );  ReadLn( Cfg, FileMode );
        ReadLn( Cfg, ReptDvsr );  ReadLn( Cfg, ReptAllIncl );
        If Abs( ReptDvsr ) < 0.0001 Then ReptDvsr := 1.0;
        ChkErr( 'reading' );
      End;
    Close( Cfg );  JunkResult := IOResult;
    AdjRange;
  End;

{------------------------------------------------------------------}

Procedure WriteConf;
  { save configuration file }
  Var
    Cfg : Text;
  Begin
    If ErrMsg <> '' Then Exit;
    FileName := CfgFileName;  Assign( Cfg, FileName );
    ReWrite( Cfg );  ChkErr( 'opening' );
    If ErrMsg = '' Then
      Begin
        WriteLn( Cfg, LdgrFileName );  WriteLn( Cfg, InFldOrder );
        WriteLn( Cfg, OutFldOrder );  WriteLn( Cfg, ShowAcct );
        Writeln( Cfg, RangeStart );  Writeln( Cfg, RangeEnd );
        WriteLn( Cfg, ReptType );  WriteLn( Cfg, ReptSort );
        WriteLn( Cfg, ListType );  WriteLn( Cfg, OutSelect );
        Writeln( Cfg, OutName );  Writeln( Cfg, FindTarget );
        WriteLn( Cfg, BalCalcMode );  WriteLn( Cfg, SortDir );
        Writeln( Cfg, DirPath );  Writeln( Cfg, MaxAge );
        Writeln( Cfg, ReptAcct );  Writeln( Cfg, ListShowGroups );
        Writeln( Cfg, CutFileName );  Writeln( Cfg, FileMode );
        Writeln( Cfg, ReptDvsr );  Writeln( Cfg, ReptAllIncl );
        ChkErr( 'writing' );
      End;
    Close( Cfg );  ChkErr( 'closing' );
  End;

{------------------------------------------------------------------}

Procedure ReadFile( Apnd : Boolean );
  { read data file to memory, optionally append }
  Var
    TypeCheck : File of Char;
    LdgrFile_Ascii : Text;
    LdgrFile_Binary : File of TrxRecType;
    TrxRec : TrxRecType;
    EndOfFile : Boolean;
    InLn : String;
    ClrLp : Byte;
    RdCh : Char;
    CurrFileMode : Char;
  Begin
    If ErrMsg <> '' Then Exit;
    If AnyChange and ( not Apnd ) Then
      Begin
        Prompt( 'Changes have been made!  Are you sure:  (N)o  (Y)es' );
        RdCh := 'N';  Select( RdCh, 'NY' );
        If RdCh <> 'Y' Then Exit;
      End;
    AnyChange := Apnd;
    If not Apnd Then UnDoTrxNum := -1;
    For ClrLp := 1 to ScreenLines do
      Begin
        CrsrXY( 1, ClrLp );  ClrEol;
      End;
    Prompt( 'Reading ' + FileName + ' . . .' );
    If not Apnd Then
      Begin  { clear memory }
        TrxCount := 0;  TrxNum := 0;
      End
    Else
      Begin  { append }
        TrxNum := TrxCount;
      End;
    { determine save type from first character in file }
    Assign( TypeCheck, FileName );
    Reset( TypeCheck );  ChkErr( 'opening' );
    If ErrMsg = '' Then
      Begin  Read( TypeCheck, RdCh );  ChkErr( 'reading' );  End;
    Close( TypeCheck );  ChkErr( 'closing' );
    If RdCh = #255 Then CurrFileMode := 'B' Else CurrFileMode := 'A';
    If ErrMsg = '' Then
      Begin
        If CurrFileMode = 'B' Then
          Begin
            Assign( LdgrFile_Binary, FileName );
            Reset( LdgrFile_Binary );
          End
        Else
          Begin
            Assign( LdgrFile_Ascii, FileName );
            Reset( LdgrFile_Ascii );
          End;
        ChkErr( 'opening' );
      End;
    EndOfFile := False;
    While ( ErrMsg = '' ) and ( not EndOfFile ) do
      Begin
        { read text line }
        If CurrFileMode = 'B' Then
          Read( LdgrFile_Binary, TrxRec )
        Else
          ReadLn( LdgrFile_Ascii, InLn );
        ChkErr( 'reading' );
        If ErrMsg = '' Then
          If TrxCount + 1 > TrxMax Then
            ErrMsg := 'Too many lines, truncated'
          Else
            Begin
              { convert line to fields }
              Inc( TrxNum );
              If CurrFileMode = 'B' Then
                Begin
                  HdrRec.Flag := TrxRec.Flag;
                  HdrRec.Date := TrxRec.Date;
                  HdrRec.Seq := TrxRec.Seq;
                  HdrRec.Name := TrxRec.Name;
                  HdrRec.PrcDate := TrxRec.PrcDate;
                  HdrRec.Memo := Nil;
                  MemoRec := TrxRec.Memo;
                  DisbArr := TrxRec.DisbArr;
                End
              Else
                TranslateLine( InLn );
              { show progress }
              ShowStat;
              { show conversion errors if any }
              ShowError;
              { write to memory }
              ExamineDate( HdrRec.PrcDate );
              ExamineDate( HdrRec.Date );
              WriteTrx( TrxNum );
            End;
        If CurrFileMode = 'B' Then
          EndOfFile := EOF( LdgrFile_Binary )
        Else
          EndOfFile := EOF( LdgrFile_Ascii );
      End;
    If CurrFileMode = 'B' Then
      Close( LdgrFile_Binary )
    Else
      Close( LdgrFile_Ascii );
    ChkErr( 'closing' );
    { create a blank if empty }
    If TrxNum = 0 Then
      Begin
        TrxNum := 1;  CreateTrx;  WriteTrx( TrxNum );
      End;
    { position cursor at end }
    If TrxCount > ScreenLines Then
      Begin
        PgStart := TrxCount - ScreenLines + 1;  SetPgEnd;
      End
    Else
      Begin
        PgStart := 1;  SetPgEnd;
      End;
    If ( ErrMsg = '' ) and ( not Apnd ) Then
      Begin
        PrevDate := DateString;
        If FileName <> PrevLdgrFileName Then WriteConf;
        PrevLdgrFileName := FileName;
      End;
    Prompt( '' );  SetReCalc;  UpdateScreen;
  End;

{------------------------------------------------------------------}

Procedure WriteFile( BlkStart, BlkEnd : Integer );
  { write data in memory to disk }
  Var
    LdgrFile_Ascii : Text;
    LdgrFile_Binary : File of TrxRecType;
    TrxRec : TrxRecType;
    InLn : String;
    TrxNumSave : Integer;
  Begin
    If ( ErrMsg <> '' ) or
       ( BlkStart > TrxCount ) or ( BlkEnd < 1 ) Then Exit;
    Prompt( 'Writing ' + FileName + ' . . .' );
    If FileMode = 'B' Then
      Begin
        Assign( LdgrFile_Binary, FileName );
        ReWrite( LdgrFile_Binary );
      End
    Else
      Begin
        Assign( LdgrFile_Ascii, FileName );
        ReWrite( LdgrFile_Ascii );
      End;
    ChkErr( 'opening' );
    TrxNumSave := TrxNum;  TrxNum := BlkStart;
    While ( ErrMsg = '' ) and ( TrxNum <= BlkEnd ) and
          ( TrxNum <= TrxCount ) do
      Begin
        { read record and convert to text line }
        ReadTrx( TrxNum );
        If FileMode = 'B' Then
          Begin
            TrxRec.ModeFlag := #255;
            TrxRec.Flag := HdrRec.Flag;
            TrxRec.Date := HdrRec.Date;
            TrxRec.Seq := HdrRec.Seq;
            TrxRec.Name := HdrRec.Name;
            TrxRec.PrcDate := HdrRec.PrcDate;
            TrxRec.Memo := MemoRec;
            TrxRec.DisbArr := DisbArr;
          End
        Else
          Begin
            InLn := FormatLine;
            { strip trailing only }
            While ( InLn <> '' ) and ( InLn[ Length( InLn ) ] = ' ' ) do
              Delete( InLn, Length( InLn ), 1 );
          End;
        { show progress }
        ShowStat;
        If FileMode = 'B' Then
          Write( LdgrFile_Binary, TrxRec )
        Else
          WriteLn( LdgrFile_Ascii, InLn );
        ChkErr( 'writing' );
        Inc( TrxNum );
      End;
    TrxNum := TrxNumSave;  ShowStat;
    If FileMode = 'B' Then
      Close( LdgrFile_Binary )
    Else
      Close( LdgrFile_Ascii );
    ChkErr( 'closing' );
    If ErrMsg = '' Then
      Begin
        AnyChange := False;
        If FileName <> PrevLdgrFileName Then WriteConf;
        PrevLdgrFileName := FileName;
      End;
    Prompt( '' );
  End;

(* -------------------------------------------------------------- *)
(*  Menus.                                                        *)
(* -------------------------------------------------------------- *)

Procedure EditMenu;
  { select editing operation }
  Var
    PrevNum, JumpLine : Integer;
    EditMenuCh : Char;
    EdtChSet : String10;
    EdtPrompt : String80;
  Begin
    Message( 'Edit Menu.' );
    EdtPrompt := 'Select:  (D)elete  (E)dit  (I)nsert  (J)ump';
    EdtChSet := 'DEIJ';
    If UndoTrxNum = TrxNum Then
      Begin
        EdtPrompt := EdtPrompt + '  (U)ndo';
        EdtChSet := EdtChSet + 'U';
      End;
    Prompt( EdtPrompt );
    EditMenuCh := 'E';  Select( EditMenuCh, EdtChSet );
    Case EditMenuCh of
      'D' : DeleteTrx;
      'E' : EditTrx;
      'I' : InsertTrx;
      'J' :
        Begin
          Message( 'Jump to specified line.' );
          Prompt( 'Enter line number to jump to:  ' );
          JumpLine := TrxCount;  InpInt( JumpLine );
          If not Escape Then
            Begin
              PrevNum := TrxNum;  TrxNum := JumpLine;
              If TrxNum < 1 Then TrxNum := 1;
              If TrxNum > TrxCount Then TrxNum := TrxCount;
              If PrevNum <> TrxNum Then
                Begin
                  SetReCalc;
                  If ( TrxNum > PgEnd ) or ( TrxNum < PgStart ) Then
                    Begin
                      PgStart := TrxNum;  SetPgEnd;
                      UpdateScreen
                    End
                  Else
                    Begin
                      ShowLine( PrevNum );  ShowLine( TrxNum );
                    End;
                End;
            End;
        End;
      'U' :
        Begin
          HdrRec := UnDoHdrRec;  DisbArr := UnDoDisbArr;
          MemoRec := UnDoMemo;
          WriteTrx( TrxNum );  ShowLine( TrxNum );
          AnyChange := True;
        End;
    End;
    SetReCalc;
    If ( EditMenuCh = 'I' ) or ( EditMenuCh = 'D' ) Then UpDateScreen;
  End;

{------------------------------------------------------------------}

Function Str2( I : Integer ) : String10;
  Begin
    I := Abs( I ) mod 100;
    Str2 := Chr( ( I div 10 ) + 48 ) + Chr( ( I mod 10 ) + 48 );
  End;

{------------------------------------------------------------------}

Procedure FSplit( Path : String80;  Var Dir : String80;
                  Var Name : String20;  Var Ext : String20 );
  Var
    Lp : Byte;
  Begin
    Dir := '';  Name := '';  Ext := '';
    If Pos( '\', Path ) > 0 Then
      Begin
        For Lp := Length( Path ) DownTo 1 do
          If Path[ Lp ] = '\' Then
            Dir := Path[ Lp ] + Dir
          Else
            If Dir <> '' Then
              Dir := Path[ Lp ] + Dir;
        Delete( Path, 1, Length( Dir ) );
      End;
    If Pos( '.', Path ) > 0 Then
      Begin
        For Lp := Length( Path ) DownTo 1 do
          If Path[ Lp ] = '.' Then
            Ext := Path[ Lp ] + Ext
          Else
            If Copy( Ext, 1, 1 ) <> '.' Then
              Ext := Path[ Lp ] + Ext;
        Delete( Path, Length( Path ) - Length( Ext ) + 1, Length( Ext ) );
      End;
    Name := Path;
  End;

{------------------------------------------------------------------}

Function DTFormat( InDT : LongInt ) : String20;
  Var
    DT : DateTime;
  Begin
    UnPackTime( InDT, DT );
    DTFormat :=
      Str2(DT.Year) + '/' + Str2(DT.Month) + '/' + Str2(DT.Month) +
      '  ' + Str2(DT.Hour) + ':' + Str2(DT.Min) + ':' + Str2(DT.Sec);
  End;

{------------------------------------------------------------------}

Procedure DoDir;
  Var
    DirRow, ClrDirRow, DirResult : Integer;
    DirCh : Char;
    Dir : SearchRec;
    ShowPath, SplitPath, DirLn : String80;
    SplitName, SplitExt : String20;
  Begin
    SupressScreenUpdate := True;
    Move( Mem[ VAddr : $0000 ], SaveScreen, SizeOf( SaveScreen ) );
    FSplit( DirPath, ShowPath, SplitName, SplitExt );
    FindFirst( DirPath, AnyFile, Dir );
    DirResult := DosError;
    If DirResult <> 0 Then
      ErrMsg := 'No files matching ' + DirPath
    Else
      Begin
        Prompt( 'Press (ESC) to exit directory, ' +
                'or any other key to continue.' );
        TextBackground( Black );  TextColor( LightGreen );
        DirCh := NoKy;  DirRow := 0;
        While ( DirResult = 0 ) and ( not Escape ) do
          Begin
            If DirRow = 0 Then
              Begin
                For ClrDirRow := 1 to ScreenLines do
                  Begin
                    CrsrXY( 1, ClrDirRow );  ClrEol;
                  End;
                CrsrXY( 1, 1 );
              End;
            Inc( DirRow );  CrsrXY( 1, DirRow );
            FSplit( Dir.Name, SplitPath, SplitName, SplitExt );
            If ( SplitExt <> '.' ) and ( SplitExt <> '..' ) and
               ( Copy( SplitExt, 1, 1 ) = '.' ) Then
              Delete( SplitExt, 1, 1 );
            SplitName := Pad( SplitName, -8 ) + ' ' + Pad( SplitExt, -3 );
            If Dir.Attr and Directory <> 0 Then
              DirLn := ShowPath + SplitName + '  <DIR>'
            Else
              DirLn := ShowPath + SplitName + Pad( I_S( Dir.Size ), 9 ) +
                       '  ' + DTFormat( Dir.Time );
            If Length( DirLn ) > 79 Then
              DirLn := Copy( DirLn, 1, 78 ) + #175;
            Writeln( DirLn );
            FindNext( Dir );  DirResult := DosError;
            If DirResult <> 0 Then
              Prompt( 'End of directory.  Press any key.' );
            If ( DirResult <> 0 ) or ( DirRow = ScreenLines ) Then
              Begin
                DirRow := 0;
                CrsrOn;
                While not KeyPressed do;
                While KeyPressed do DirCh := ReadKey;
                Escape := DirCh = #27;
                CrsrOff;
              End;
          End;
      End;
    Move( SaveScreen, Mem[ VAddr : $0000 ], SizeOf( SaveScreen ) );
    SupressScreenUpdate := False;
  End;

{------------------------------------------------------------------}

Procedure DiskMenu;
  { select disk operation }
  Var
    TstFile : File;
    TstResult, JunkResult : Integer;
    DskCh, ConfCh : Char;
    EnterPath : String[50];
    UseName : String[50];
  Begin
    Message( 'Disk menu.' );
    Prompt( 'Select:  (A)ppend  (D)irectory  ' +
            '(L)oad  (M)ode  (S)ave  (W)rite block' );
    DskCh := 'S';  Select( DskCh, 'ADLMSW' );
    Case DskCh of
      'M' :
        Begin
          Message( 'Set file mode.' );
          Prompt( 'Read and write as:  (A)scii  (B)inary' );
          ConfCh := FileMode;  Select( ConfCh, 'AB' );
          If ( ConfCh = 'A' ) or ( ConfCh = 'B' ) Then
            Begin
              If FileMode <> ConfCh Then WriteConf;
              FileMode := ConfCh;
            End;
        End;
      'D' :
        Begin
          Message( 'Directory.' );
          Prompt( 'Enter path:  ' );
          EnterPath := DirPath;  InpStr( EnterPath, 50, _Text );
          If not Escape Then
            Begin
              If EnterPath <> DirPath Then
                Begin
                  DirPath := EnterPath;
                  WriteConf;
                End;
              DoDir;
            End;
        End;
      'A', 'L' :
        Begin
          If DskCh = 'L' Then
            Message( 'Load file.' )
          Else
            Message( 'Append file to current data.' );
          Prompt( 'Enter file name:  ' );
          If DskCh = 'A' Then
            UseName := CutFileName
          Else
            UseName := LdgrFileName;
          InpStr( UseName, 50, _Text );
          If DskCh = 'A' Then
            CutFileName := UseName
          Else
            LdgrFileName := UseName;
          If not Escape Then
            Begin
              If InpChange Then WriteConf;
              FileName := UseName;  ReadFile( DskCh = 'A' );
            End;
         End;
      'S', 'W' :
        Begin
          If DskCh = 'S' Then
            Message( 'Save file.' )
          Else
            Message( 'Write range to file.' );
          Prompt( 'Enter file name:  ' );
          If DskCh = 'W' Then
            UseName := CutFileName
          Else
            UseName := LdgrFileName;
          InpStr( UseName, 50, _Text );
          If DskCh = 'W' Then
            CutFileName := UseName
          Else
            LdgrFileName := UseName;
          If not Escape Then
            Begin
              If InpChange Then WriteConf;
              Assign( TstFile, UseName );
              Reset( TstFile );  TstResult := IOResult;
              Close( TstFile );  JunkResult := IOResult;
              If TstResult = 0 Then
                Begin
                  Prompt( 'Replace file:  (N)o  (Y)es' );
                  ConfCh := 'N';  Select( ConfCh, 'NY' );
                End
              Else
                ConfCh := 'Y';
              If ConfCh = 'Y' Then
                Begin
                  FileName := UseName;
                  If DskCh = 'W' Then
                    WriteFile( RangeStart, RangeEnd )
                  Else
                    WriteFile( 1, TrxCount );
                End;
            End;
        End;
    End;
  End;

{------------------------------------------------------------------}

Procedure QuitMenu;
  { termination confirmation }
  Var
    QuitCh : Char;
  Begin
    Message( 'Quit.' );
    If AnyChange Then
      Begin
        Prompt( 'Changes have been made!  ' );  QuitCh := 'N';
      End
    Else
      Begin
        Prompt( '' );  QuitCh := 'Y';
      End;
    Write( 'Are you sure:  (N)o  (Y)es' );  Select( QuitCh, 'NY' );
    If QuitCh <> 'Y' Then Exit;
    TextColor( White );  TextBackground( Black );
    CrsrXY( 1, 25 );  Writeln;  ClrEol;  CrsrOn;
    Halt;
  End;

{------------------------------------------------------------------}

Procedure FindAcct( AcctIn : DisbAcctType;  ChkGroup : Boolean );
  Var
    TrxLp, DisbLp : Integer;
    AcctFound : Boolean;
  Begin
    { verify valid account }
    Message( 'Checking . . .' );
    TrxLp := 1;  AcctFound := False;
    While ( ErrMsg = '' ) and
          ( TrxLp <= TrxCount ) and ( not AcctFound ) do
      Begin
        ReadTrx( TrxLp );  DisbLp := 1;
        While ( ErrMsg = '' ) and
              ( DisbLp < MaxDisb ) and ( not AcctFound ) do
          Begin
            AcctFound :=
              ( ChkGroup and
                ( Copy( HdrRec.Name, 1, 1 ) = GroupMark ) and
                ( Copy( HdrRec.Name, 2, 255 ) = AcctIn ) ) or
              ( AcctIn = DisbArr[ DisbLp ].Acct );
            Inc( DisbLp );
          End;
        Inc( TrxLp );
      End;
    If not AcctFound Then
      ErrMsg := 'Selected account not found';
  End;

{------------------------------------------------------------------}

Procedure ConfMenu;
  Var
    CfgCh, FldCh, InCalcMode : Char;
    AcctSave : DisbAcctType;
  Begin
    { configuration menu }
    Message( 'Configuration menu.' );
    Prompt( 'Select:  ' +
            '(A)ge limit  (B)alance account  (F)ield order  (I)nfo' );
    CfgCh := 'B';  Select( CfgCh, 'ABFI' );
    Case CfgCh of
      'A' : Begin
              Message( 'Configure transaction edit protection.' );
              Prompt( 'Enter maximum age in days:  ' );
              InpInt( MaxAge );
              If ( MaxAge < 0 ) or ( MaxAge > 3650 ) Then MaxAge := 45;
            End;
      'B' : Begin
              Message( 'Configure balance display.' );
              Prompt( 'Enter name of balance account:  ' );
              AcctSave := ShowAcct;  InpStr( ShowAcct, AcctSize, _String );
              If not Escape Then
                Begin
                  FindAcct( ShowAcct, False );
                  If ErrMsg <> '' Then
                    ShowAcct := AcctSave
                  Else
                    Begin
                      Message( 'Calculation mode.' );
                      Prompt( 'Balance:  (I)ssued  (P)rocessed' );
                      InCalcMode := BalCalcMode;
                      Select( InCalcMode, 'IP' );
                      If not Escape Then
                        Begin
                          SetReCalc;
                          If ShowAcct <> AcctSave Then
                            SortAllDisbs;
                          BalCalcMode := InCalcMode;
                        End;
                    End;
                End;
            End;
      'F' : Begin
              Message( 'Change field order.' );
              Prompt( 'Select:  (D)isplay  (I)nput' );
              FldCh := 'D';  Select( FldCh, 'DI' );
              Case FldCh of
                'I' : EnterFldOrder( InFldOrder );
                'D' : EnterFldOrder( OutFldOrder );
               End;
            End;
      'I' : Begin
              Prompt( 'System information:  Press any key' );
              Message( 'Memory limit ' + I_S( TrxMax ) + ' lines' +
                       ' of ' + I_S( MaxTrx65K ) + '.' );
              PressAnyKey;
            End;
     End;
    If InpChange and ( not Escape ) Then
      Begin
        WriteConf;
        If ( CfgCh = 'B' ) or ( CfgCh = 'F' ) Then
          Begin
            SetReCalc;  UpdateScreen;
          End;
      End;
  End;

{------------------------------------------------------------------}

Procedure FindTrx;
  { find any string in whole ledger }
  Var
    FindLp : Integer;
    FindLn : String;
    Found : Boolean;
  Begin
    Message( 'Find.' );
    Prompt( 'Enter search target:  ' );  InpStr( FindTarget, 50, _Text );
    If Escape Then Exit;
    Message( 'Searching for ' + FindTarget + ' . . .' );
    FindLp := TrxNum + 1;
    If FindLp > TrxCount Then FindLp := 1;
    Found := False;
    While ( ErrMsg = '' ) and
          ( FindLp <> TrxNum ) and ( TrxNum > 0 ) and ( not Found ) do
      Begin
        If KeyPressed Then
          ErrMsg := 'Search interrupted'
        Else
          Begin
            ReadTrx( FindLp );  FindLn := FormatLine;
            Upper( FindLn );
            If Pos( FindTarget, FindLn ) > 0 Then
              Begin
                TrxNum := FindLp;  Found := True;
              End
            Else
              Begin
                Inc( FindLp );  If FindLp > TrxCount Then FindLp := 1;
              End;
          End;
      End;
    If ErrMsg = '' Then
      Begin
        If not Found Then
          Begin
            Prompt( ^G + 'NOT FOUND:  Press any key.' );  PressAnyKey;
          End
        Else
          Begin
            If ( TrxNum > PgEnd ) or ( TrxNum < PgStart ) Then
              Begin
                PgStart := TrxNum;  SetPgEnd;
              End;
            SetReCalc;  UpdateScreen;
          End;
      End;
  End;

{------------------------------------------------------------------}

Procedure SetRange;
  { set range start and end }
  Var
    PrevStart, PrevEnd : Integer;
    RngCh : Char;
  Begin
    Message( 'Mark a range.   (Now: ' + I_S( RangeStart ) +
             '..' + I_S( RangeEnd ) + ')' );
    Prompt( 'Block:  (C)lear  (E)nd  (S)tart' );
    PrevStart := RangeStart;  PrevEnd := RangeEnd;
    RngCh := RangeCh;  Select( RngCh, 'CES' );
    Case RngCh of
      'C' : Begin
              RangeStart := 1;
              RangeEnd := MaxInt;
            End;
      'S' : RangeStart := TrxNum;
      'E' : If TrxNum >= TrxCount Then
              RangeEnd := MaxInt
            Else
              RangeEnd := TrxNum;
    End;
    If RngCh = 'S' Then RangeCh := 'E';
    If RngCh = 'E' Then RangeCh := 'S';
    AdjRange;
    If ( RangeStart <> PrevStart ) or ( RangeEnd <> PrevEnd ) Then
      WriteConf;
  End;

(* -------------------------------------------------------------- *)
(*  Movement routines.                                            *)
(* -------------------------------------------------------------- *)

Procedure LineMove( Dis : Integer );
  Var
    PrevStart, PrevNum : Integer;
  Begin
    SetPgEnd;  PrevStart := PgStart;
    PrevNum := TrxNum;  TrxNum := TrxNum + Dis;
    If TrxNum < 1 Then TrxNum := 1;
    If TrxNum > TrxCount Then
      If Dis = 1 Then
        Begin
          If TrxNum > TrxMax Then
            Begin
              ErrMsg := 'Memory full';
              ShowError;
            End
          Else
            Begin
              TrxNum := TrxCount + 1;
              CreateTrx;  WriteTrx( TrxNum );
            End;
        End
      Else
        TrxNum := TrxCount;
    If TrxNum < PgStart Then PgStart := TrxNum;
    If TrxNum > PgEnd Then PgStart := TrxNum - ScreenLines + 1;
    SetPgEnd;
    If PrevNum <> TrxNum Then SetReCalc;
    If PgStart <> PrevStart Then
      UpdateScreen
    Else
      If PrevNum <> TrxNum Then
        Begin
          ShowLine( PrevNum );  ShowLine( TrxNum );
        End;
  End;

{------------------------------------------------------------------}

Procedure ProcessMove( MoveCh : Char );
  Var
    PrevLeftMgn : Byte;
  Begin
    PrevLeftMgn := LeftMgn;
    Case MoveCh of
      UpKy, ^E : LineMove( -1 );
      DnKy, ^X : LineMove( 1 );
    PgUpKy, ^R : If TrxNum > PgStart Then
                   LineMove( PgStart - TrxNum )
                 Else
                   LineMove( -ScreenLines );
    PgDnKy, ^C : If TrxNum < PgEnd Then
                   LineMove( PgEnd - TrxNum )
                 Else
                   LineMove( ScreenLines );
    C_HmKy, ^W : LineMove( -TrxCount );
   C_EndKy, ^Z : LineMove( TrxCount );
      HmKy, ^A : LeftMgn := 1;
     EndKy, ^F : LeftMgn := MaxLen - 70;
     LftKy, ^S : If LeftMgn > 1 Then
                   If LeftMgn > 8 Then
                     LeftMgn := LeftMgn - 8
                   Else
                     LeftMgn := 1;
     RgtKy, ^D : If LeftMgn < MaxLen - 70 Then
                   If LeftMgn < MaxLen - 70 - 8 Then
                     LeftMgn := LeftMgn + 8
                   Else
                     LeftMgn := MaxLen - 70;
    End;
    If LeftMgn <> PrevLeftMgn Then UpdateScreen;
  End;

{------------------------------------------------------------------}

