{$D-,L-,B-,F-,I-,N-,R-,S-,V-}
{$M 6000,0,655360}

Program Ldgr;

(*
   Revision history:
     91/08/08 Added groups feature for analysis report.
     91/11/12 Interrupting sort give not found message.
     91/11/12 Append overwrites entries at current line.
     91/11/12 Group definitions are adding to balance.
     91/11/12 Output file not closed if report interrupted.
     91/11/12 Cursor lingers on last line after redraw.
     91/11/12 Status line doesn't update for some cursor moves.
     91/11/12 Always show groups first regardless of report type.
     91/11/12 Need easy way of entering reversing disbursement.
     91/11/12 Use run path to find help file.
     91/11/12 Don't sort disbs if balance account doesn't change.
     91/11/12 Improve sorting disbs message, add line status.
     92/01/09 Limitation increased to 2340 from 1040 transactions.
     92/02/01 Binary file type added for faster data load.
     92/03/23 Amount divisor added for analysis reports.
     92/09/29 Minus sign, then space loops on amount entries.
     93/09/12 Added include all option to reports.
     93/09/12 Fixed : character not allowed in file names.
     93/09/12 Fixed cursor position bug in crsron/off.
*)

(*
This program is a modified general ledger.
It provides fields for issued date, sequence identifier,
name, processed date, user flag, memo, and disbursements.
To aid readability, rather than having a large number of columns,
each disbursement is composed of a dollar amount followed by
an account name.  See LDGR.HLP for more information.
The program includes search, sort, and block features, configurable
input and output field orders, total and subtotal display
for any disbursement account, analysis and listing reports.
All data is stored in memory for maximum speed.
Data space is allocated at startup to maximum available.
All access to data space is through read and write procedures.
Transaction records are passed through three primary variables:
  HdrRec { header }, DisbArr { disbursements }, and MemoRec { memo }.
Data file is text for easy external processing.
All configureable parameters are saved and recalled.
User interface is through simple menus.
*)

Uses Dos, Crt;

Const
  ProgName = 'LDGR';
  Ver = '3.6';
  Rev = '93/09/12';
  { encoded copyright }
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
    'fwvs)1sz~';
  CpyRtLp : Byte = 1;
  { maximum transaction and memo capacity }
  MaxTrx65K : Word = 0;
  MaxMemRecs = 2340;
  MaxMemo = 150;
  { extended key handling }
  NoKy = #255;       EscKy = #27;      EntrKy = #13;
  F1Ky = #187;       BkSpcKy = #8;     TabKy = #9;
  S_TabKy = #143;    HmKy = #199;      EndKy = #207;
  PgUpKy = #201;     PgDnKy = #209;    LftKy = #203;
  RgtKy = #205;      UpKy = #200;      DnKy = #208;
  DelKy = #211;      InsKy = #210;     C_HmKy = #247;
  C_EndKy = #245;    C_LftKy = #243;   C_RgtKy = #244;
  { data entry type }
  _Text = 0;
  _String = 1;
  _Real = 2;
  _Date = 3;
  _Memo = 4;
  { other data entry parameters }
  AllowMove : Boolean = True;
  AnyChange : Boolean = False;
  InpChange : Boolean = False;
  NumericSet : Set of Char = [ '0'..'9', '-', '+', '.' ];
  AlphaSet : Set of Char = [ 'A'..'Z', '0'..'9', '-', '+', '.', '/' ];
  DateSet : Set of Char = [ '0'..'9', '/', '-' ];
  MoveSet : String[ 20 ] = '';
  AnyKey : Char = NoKy;
  InpKey : Char = NoKy;
  SpclInpKeys : String[ 10 ] = '';
  MaxLen : Byte = 80;
  { field identifier marks in data file }
  GroupMark = '*';
  FlagMk = '!';
  PrcDateMk = '<';
  SeqMk = '#';
  DateMk = '>';
  NameMk = '"';
  AmtMk = '$';
  AcctMk = '@';
  MemoMk = '''';
  MaxFld = 7;
  MkSet : Set of Char =
    [ FlagMk,DateMk,SeqMk,NameMk,AmtMk,AcctMk,PrcDateMk,MemoMk ];
  { screen size and other aspects }
  ScreenLines = 23;
  PgStart : Integer = 1;
  PgEnd : Integer = ScreenLines;
  LeftMgn : Byte = 1;
  VAddr : Word = $B800;
  MonoMode : Boolean = False;
  Mono : Byte = $07;
  CO80 : Byte = $03;
  FrClr : Byte = Black;
  BkClr : Byte = LightGreen;
  RangeStart : Integer = 0;  { stored }
  RangeEnd : Integer = 32767; { stored }
  RangeCh : Char = 'S';
  { transaction data storage size parameters }
  MaxDisb = 7;
  DateSize = 8;
  FlagSize = 2;
  SeqSize = 6;
  NameSize = 20;
  AmtSize = 9;
  PctSize = 6;
  AcctSize = 8;
  MemoSize = 30;
  { transaction edit and display }
  DfltFldSet : String[ 10 ] = '';
  DfltFldOrder : String[ 10 ] = '';
  InFldOrder : String[ 10 ] = '';  { stored }
  OutFldOrder : String[ 10 ] = '';  { stored }
  LdgrFileName : String[ 50 ] = 'LDGR.LGR';  { stored }
  PrevLdgrFileName : String[ 50 ] = '';
  CutFileName : String[ 50 ] = 'LDGR.CUT';  { stored }
  FileMode : Char = 'A';  { stored }
  ShowAcct : String[ 20 ] = 'CH1';  { stored }
  ReptAcct : String[ 20 ] = 'NET';  { stored }
  ReptDvsr : Real = 1.0;  { stored }
  LineBal : Real = 0.0;
  TotalBal : Real = 0.0;
  TrxMax : LongInt = 0;
  TrxCount : Integer = 0;
  TrxNum : Integer = 1;
  { general }
  FileName : String[ 80 ] = '';
  FindTarget : String[ 50 ] = '';  { stored }
  DirPath : String[ 50 ] = '.\*.*';  { stored }
  BalCalcMode : Char = 'I';  { stored }
  DoReCalc : Boolean = True;
  DoUpdateScreen : Boolean = False;
  SupressScreenUpdate : Boolean = True;
  ErrMsg : String[ 80 ] = '';
  HelpFileName = 'LDGR.HLP';
  UnDoTrxNum : Integer = -1;
  CfgFileName = 'LDGR.CFG';
  PrevDate : String[ 8 ] = '';
  MaxDate : String[ 8 ] = '';
  MaxAge : Integer = 45;  { stored }
  { used by report }
  ReptMaxAcct = 120;
  OutName : String[ 80 ] = 'LDGR.PRN';  { stored }
  _Screen = 'CON';
  _Printer = 'PRN';
  _File = 'LDGR.PRN';
  OutSelect : String[ 10 ] = _Screen;  { stored }
  ReptSort : Char = '6';  { stored }
  ReptAllIncl : Char = 'Y';  { stored }
  ReptType : Char = 'A';  { stored }
  ListType : Char = 'A';  { stored }
  ListShowGroups : Char = 'Y';  { stored }
  SortDir : Char = 'A';  { stored }

Type
  { fundamental types }
  String80 = String[ 80 ];
  String50 = String[ 50 ];
  String40 = String[ 40 ];
  String20 = String[ 20 ];
  String10 = String[ 10 ];
  { transaction storage types }
  FlagType = String[ FlagSize ];
  DateType = String[ DateSize ];
  SeqType = String[ SeqSize ];
  NameType = String[ NameSize ];
  MemoType = String[ MemoSize ];
  DisbAcctType = String[ AcctSize ];
  DisbAmtType = Real;
  ReptDisbAcctType = String[ AcctSize + 1 ];
  DisbRecType =
    Record
      Acct : DisbAcctType;
      Amt : DisbAmtType;
    End;
  DisbArrType = Array[ 1..MaxDisb ] of DisbRecType;  { 91 bytes }
  HdrRecType =
    Record
      Flag : FlagType;
      Date : DateType;
      Seq : SeqType;
      Name : NameType;
      PrcDate : DateType;
      Memo : ^MemoType;
    End;
  TrxRecType =
    Record
      ModeFlag : Char;
      Flag : FlagType;
      Date : DateType;
      Seq : SeqType;
      Name : NameType;
      PrcDate : DateType;
      Memo : MemoType;
      DisbArr : DisbArrType;
    End;
  HdrRecType_1 =
    Record
      Flag : FlagType;
      Date : DateType;
      PrcDate : DateType;
      Memo : ^MemoType;
    End;
  HdrRecType_2 =
    Record
      Seq : SeqType;
      Name : NameType;
    End;
  { primary storage }
  HdrRecMemType_1 = Array [ 1..MaxMemRecs ] of HdrRecType_1;
  HdrRecMemType_2 = Array [ 1..MaxMemRecs ] of HdrRecType_2;
  DisbAcctArrayType = Array [ 1..MaxDisb ] of DisbAcctType;
  DisbAcctRecMemType = Array [ 1..MaxMemRecs ] of DisbAcctType;
  DisbAmtArrayType = Array [ 1..MaxDisb ] of Real;
  DisbAmtRecMemType = Array [ 1..MaxMemRecs ] of DisbAmtType;
  { used bye reports }
  ReptRecType =
    Record
      Acct : ReptDisbAcctType;
      Debit, Credit : Real;
      IsGroup, InGroup : Boolean;
    End;
  ReptArrayType = Array [ 1..ReptMaxAcct ] of ReptRecType;

Var
  { primary data storage areas }
  HdrMem_1 : ^HdrRecMemType_1;
  HdrMem_2 : ^HdrRecMemType_2;
  DisbAcctMem_1 : ^DisbAcctRecMemType;
  DisbAcctMem_2 : ^DisbAcctRecMemType;
  DisbAcctMem_3 : ^DisbAcctRecMemType;
  DisbAcctMem_4 : ^DisbAcctRecMemType;
  DisbAcctMem_5 : ^DisbAcctRecMemType;
  DisbAcctMem_6 : ^DisbAcctRecMemType;
  DisbAcctMem_7 : ^DisbAcctRecMemType;
  DisbAmtMem_1 : ^DisbAmtRecMemType;
  DisbAmtMem_2 : ^DisbAmtRecMemType;
  DisbAmtMem_3 : ^DisbAmtRecMemType;
  DisbAmtMem_4 : ^DisbAmtRecMemType;
  DisbAmtMem_5 : ^DisbAmtRecMemType;
  DisbAmtMem_6 : ^DisbAmtRecMemType;
  DisbAmtMem_7 : ^DisbAmtRecMemType;
  HdrMemSpace_1 : Pointer;
  HdrMemSpace_2 : Pointer;
  DisbAcctMemSpace_1 : Pointer;
  DisbAcctMemSpace_2 : Pointer;
  DisbAcctMemSpace_3 : Pointer;
  DisbAcctMemSpace_4 : Pointer;
  DisbAcctMemSpace_5 : Pointer;
  DisbAcctMemSpace_6 : Pointer;
  DisbAcctMemSpace_7 : Pointer;
  DisbAmtMemSpace_1 : Pointer;
  DisbAmtMemSpace_2 : Pointer;
  DisbAmtMemSpace_3 : Pointer;
  DisbAmtMemSpace_4 : Pointer;
  DisbAmtMemSpace_5 : Pointer;
  DisbAmtMemSpace_6 : Pointer;
  DisbAmtMemSpace_7 : Pointer;
  { primary data passing working storage }
  HdrRec : HdrRecType;
  DisbArr : DisbArrType;
  MemoRec : MemoType;
  DisbAcctArray : DisbAcctArrayType;
  DisbAmtArray : DisbAmtArrayType;
  DisbRec : DisbRecType;
  { used by line translate }
  PrsWd : String80;
  FldType : String40;
  { used by help }
  SaveScreen : Array[ 1..4096 ] of Byte;
  { used by data entry }
  Escape : Boolean;
  { main menu selection }
  MenuCh : Char;
  { used by reports }
  RptArr : ReptArrayType;
  RatioDebit : Real;
  RatioCredit : Real;
  RptAccts : Integer;
  { used by editing }
  DisbFilled : Boolean;
  EdtHdrRec : HdrRecType;
  EdtDisbArr : DisbArrType;
  EdtMemo : MemoType;
  UnDoHdrRec : HdrRecType;
  UnDoDisbArr : DisbArrType;
  UnDoMemo : MemoType;

(* -------------------------------------------------------------- *)
(*  High level procedures called from lower levels.               *)
(* -------------------------------------------------------------- *)

Procedure ProcessMove( MoveCh : Char );  Forward;

Procedure UpdateScreen;  Forward;

Procedure OnLineHelp;  Forward;

Procedure ReCalc;  Forward;

(* -------------------------------------------------------------- *)
(*  String, numeric and date utilities.                           *)
(* -------------------------------------------------------------- *)

Function I_S( InNum : LongInt ) : String40;
  { integer to string quick convert }
  Var
    OutLn : String40;
  Begin
    Str( InNum, OutLn );  I_S := OutLn;
  End;

{------------------------------------------------------------------}

Function R_S( InNum : Real ) : String20;
  { real to string quick convert }
  Var
    OutLn : String20;
  Begin
    Str( InNum:1:2, OutLn );  R_S := OutLn;
  End;

{------------------------------------------------------------------}

Function Amount( InR : Real ) : Boolean;
  { zero comparison ignoring insignificant digits }
  Begin
    Amount := ( InR <= -0.01 ) or ( InR >= 0.01 );
  End;

{------------------------------------------------------------------}

Function PadStr( InLn : String80; Sz : Integer; Ch : Char ) : String80;
  { Pad a string in a field of any character. }
  { size >0 left justifies, <0 right justifies, }
  { size >length of input string truncates }
  Var
    OutLn : String80;
    Ps, AbsSz : Byte;
  Begin
    AbsSz := Abs( Sz );
    If Length( InLn ) < AbsSz Then
      Begin
        FillChar( OutLn[ 1 ], AbsSz, Ch );
        If Sz < 0 Then
          Ps := 1
        Else
          Ps := AbsSz - Length( InLn ) + 1;
        Move( InLn[ 1 ], OutLn[ Ps ], Length( InLn ) );
        OutLn[ 0 ] := Chr( AbsSz );
        PadStr := OutLn;
      End
    Else
      PadStr := Copy( InLn, 1, AbsSz );
  End;

{------------------------------------------------------------------}

Function Pad( InLn : String80; Sz : Integer ) : String80;
  { pad with spaces }
  Begin
    Pad := PadStr( InLn, Sz, ' ' );
  End;

{------------------------------------------------------------------}

Procedure Strip( Var InLn : String );
  Begin
    While ( InLn <> '' ) and ( InLn[ 1 ] = ' ' ) do
      Delete( InLn, 1, 1 );
    While ( InLn <> '' ) and ( InLn[ Length( InLn ) ] = ' ' ) do
      Delete( InLn, Length( InLn ), 1 );
  End;

{------------------------------------------------------------------}

Function Space( Len : Integer ) : String80;
  Var
    OutLn : String80;
  Begin
    OutLn[ 0 ] := Chr( Len );
    FillChar( OutLn[ 1 ], Len, ' ' );
    Space := OutLn;
  End;

{------------------------------------------------------------------}

Procedure Upper( Var InLn : String );
  Var
    UpLp : Byte;
  Begin
    For UpLp := 1 to Length( InLn ) do
      InLn[ UpLp ] := UpCase( InLn[ UpLp ] );
  End;

{------------------------------------------------------------------}

Procedure ExamineDate( InDate : DateType );
  { set PrevDate and MaxDate data entry support variables }
  Begin
    If ( ErrMsg <> '' ) or ( InDate = '' ) Then Exit;
    If InDate > MaxDate Then MaxDate := InDate;
    PrevDate := InDate;
  End;

{------------------------------------------------------------------}

Function DateInDays( DateStr : DateType ) : LongInt;
  { return number of days since 80-01-01 }
  Var
    Yr, Mo, Da, DaTotal : LongInt;
    DaLp, DaResult : Integer;
  Const
    DaysPerMonth : array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    { initialize function to zero, flags conversion error }
    DateInDays := -1;
    { remove invalid characters }
    For DaLp := Length( DateStr ) DownTo 1 do
      If not ( DateStr[ DaLp ] in [ '0'..'9' ] ) Then
        Delete( DateStr, DaLp, 1 );
    { exit if remaining length improper }
    If Length( DateStr ) <> 6 Then Exit;
    { convert digit pairs to numbers }
    Val( Copy( DateStr, 1, 2 ), Yr, DaResult );
    If DaResult <> 0 Then Exit;
    Val( Copy( DateStr, 3, 2 ), Mo, DaResult );
    If DaResult <> 0 Then Exit;
    Val( Copy( DateStr, 5, 2 ), Da, DaResult );
    If DaResult <> 0 Then Exit;
    { range check all values }
    If ( Yr < 0 ) or ( Yr > 99 ) or
       ( Mo < 1 ) or ( Mo > 12 ) or
       ( Da < 1 ) or ( Da > 31 ) or
       ( ( not ( Da in [ 1 .. DaysPerMonth[ Mo ] ] ) ) and
         ( not ( ( Da = 29 ) and ( Mo = 2 ) and ( Yr mod 4 = 0 ) ) ) ) Then
      Exit;
    { convert date to days since 01-01-80 }
    { assume years before 80 are in 21st century }
    If Yr < 80 Then Yr := Yr + 100;
    Yr := Yr - 80;
    { start with leap year days }
    DaTotal := Yr div 4;
    { compensate for fact that the base year, 1980, is a leap year }
    Inc( DaTotal );
    { compensate if current year is leap year, but leap day is not past }
    If ( Yr mod 4 = 0 ) and
       ( ( Mo < 2 ) or ( ( Mo = 2 ) and ( Da < 29 ) ) ) Then
      Dec( DaTotal );
    { add days so far this year }
    For DaLp := 1 to Mo - 1 do
      DaTotal := DaTotal + DaysPerMonth[ DaLp ];
    DaTotal := DaTotal + Da;
    { add days for past years }
    DaTotal := DaTotal + ( Yr * 365 );
    DateInDays := DaTotal;
  End;

(* -------------------------------------------------------------- *)
(*  Screen utilities.                                             *)
(* -------------------------------------------------------------- *)

Procedure CrsrOff;
  { cursor off }
  Var
    Regs : Registers;
  Begin
    Regs.CX := $0F0F;  Regs.Ax := $0100;  Intr( $10, Regs );
  End;

{------------------------------------------------------------------}

Procedure CrsrOn;
  Var
    Regs : Registers;
  { cursor on and set style }
  Begin
    Regs.CX := $0607;  Regs.Ax := $0100;  Intr( $10, Regs );
  End;

{------------------------------------------------------------------}

Procedure CrsrXY( X, Y : Word );
  Var
    Regs : Registers;
  Begin
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $02;  Regs.DH := Y - 1;  Regs.DL := X - 1;
    Intr( $10, Regs );
  End;

{------------------------------------------------------------------}

Function GetVMode : Byte;
  Var
    Regs : Registers;
  Begin
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := 15;
    Intr( $10, Regs );
    GetVMode := Regs.AL;
  End;

{------------------------------------------------------------------}

Procedure SetVMode( M : Byte );
  Var
    Regs : Registers;
  Begin
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $00;  Regs.AL := M;
    Intr( $10, Regs );
  End;

{------------------------------------------------------------------}

Procedure ShowStat;
  Var
    Stat : String40;
  Begin
    { display line number, total lines, line balance, total balance }
    { right side of line 25 }
    TextColor( Black );  TextBackGround( LightGray );
    Stat := I_S( TrxNum ) + '/' + I_S( TrxCount ) + ' ';
    Stat := Stat + ShowAcct + ' $' + R_S( LineBal );
    { display wiggle symbol until recalc is done }
    If DoReCalc Then
      Stat := Stat + #247
    Else
      Stat := Stat + '/';
    CrsrXY( 40, 25 );
    Write( Pad( Stat + R_S( TotalBal ), 39 ) );  ClrEol;
  End;

{------------------------------------------------------------------}

Procedure SetReCalc;
  Begin
    DoReCalc := True;
    ShowStat;
  End;

{------------------------------------------------------------------}

Procedure Prompt( Msg : String80 );
  { display user input prompt }
  Begin
    { all of line 24, leave cursor at end of prompt }
    TextColor( Black );  TextBackGround( LightGray );
    CrsrXY( 1, 24 );  ClrEol;  Write( ' ' + Msg );
  End;

{------------------------------------------------------------------}

Procedure Message( Msg : String80 );
  { display operation name or instructional message }
  Var
    Stat : String40;
  Begin
    { left side of line 25 }
    TextColor( Black );  TextBackGround( LightGray );
    If Length( ' ' + Msg ) > 40 Then
      Msg := Copy( Msg, 1, 36 ) + '...';
    CrsrXY( 1, 25 );  Write( Pad( ' ' + Msg, -40 ) );
  End;

(* -------------------------------------------------------------- *)
(*  Keyboard utilities.                                           *)
(* -------------------------------------------------------------- *)

Function GetKey : Char;
  { get a keypress, but also process recalc and help }
  Var
    Ch : Char;
    XSv, YSv, AttrSv : Byte;
  Begin
    { supress cursor when not on input field }
    XSv := WhereX;  YSv := WhereY;  AttrSv := TextAttr;
    Repeat
      { complete recalc if pending }
      CrsrOn;
      If DoUpdateScreen and ( not KeyPressed ) Then
        Begin
          CrsrOff;  UpdateScreen;
          CrsrXY( XSv, YSv );  CrsrOn;  TextAttr := AttrSv;
        End;
      If DoReCalc and ( not KeyPressed ) Then
        Begin
          ReCalc;
          If not DoReCalc Then
            Begin
              CrsrOff;  ShowStat;
              CrsrXY( XSv, YSv );  CrsrOn;  TextAttr := AttrSv;
            End;
        End;
      Ch := ReadKey;  CrsrOff;
      { get extended keys }
      If Ch = #0 Then
        Begin
          Ch := ReadKey;
          If Ord( Ch ) < 128 Then
            Ch := Chr( Ord( Ch ) + $80 )
          Else
            Ch := NoKy;
        End
      Else
        If Ord( Ch ) > 127 Then Ch := NoKy;
      { remain in loop if help }
      If Ch = F1Ky Then
        Begin
          OnLineHelp;
          CrsrXY( XSv, YSv );  TextAttr := AttrSv;
        End;
    Until Ch <> F1Ky;
    GetKey := Ch;
  End;

{------------------------------------------------------------------}

Procedure Select( Var Ch : Char;  OKChars : String20 );
  { basic menu type selection support }
  Var
    DfltCh : Char;
  Begin
    DfltCh := Ch;  Escape := False;
    OKChars := OKChars + EscKy + EntrKy;
    CrsrXY( 73, 24 );  Write( #174, '   ', #175 );  CrsrXY( 75, 24 );
    If Ch <> NoKy Then
      Begin
        Write( Ch );  CrsrXY( 75, 24 );
      End;
    Repeat
      Repeat
        CrsrXY( 75, 24 );  Ch := UpCase( GetKey );
        If ( Pos( Ch, OKChars ) = 0 ) and
           ( ( not AllowMove ) or ( Pos( Ch, MoveSet ) = 0 ) ) Then
          Write( ^G );
        If ( Ch = #13 ) and ( DfltCh <> NoKy ) Then Ch := DfltCh;
      Until ( Pos( Ch, OKChars ) > 0 ) or
            ( AllowMove and ( Pos( Ch, MoveSet ) > 0 ) );
      If Ch in [ ' '..'~' ] Then
        Begin
          Write( Ch );  CrsrXY( 75, 24 );
        End;
      { allow screen movement while waiting for valid selection }
      If AllowMove and ( Pos( Ch, MoveSet ) > 0 ) Then
        ProcessMove( Ch );
    Until ( not AllowMove ) or ( Pos( Ch, MoveSet ) = 0 );
    Escape := Ch = EscKy;
    { clear line after selection }
    Prompt( '' );
  End;

{------------------------------------------------------------------}

Procedure PressAnyKey;
  Begin
    CrsrOn;
    While KeyPressed do AnyKey := ReadKey;
    While not KeyPressed do;
    CrsrOff;
    While KeyPressed do AnyKey := ReadKey;
    Prompt( '' );
  End;

{------------------------------------------------------------------}

Procedure OnLineHelp;
  Var
    HlpRow, ClrHlpRow : Byte;
    HlpCh : Char;
    HlpLn : String80;
    HlpFile : Text;
    HlpResult : Integer;
    RunPath, SplitPath : String80;
    SplitName, SplitExt : String20;
  Begin
    SupressScreenUpdate := True;
    Move( Mem[ VAddr : $0000 ], SaveScreen, SizeOf( SaveScreen ) );
    Message( 'Help.' );
    Prompt( 'Press (ESC) to exit help, or any other key to continue.' );
    TextBackground( Black );  TextColor( LightGreen );
    RunPath := ParamStr( 0 );
    FSplit( RunPath, SplitPath, SplitName, SplitExt );
    Upper( SplitPath );
    Assign( HlpFile, SplitPath + HelpFileName );
    Reset( HlpFile );  HlpResult := IOResult;
    HlpCh := NoKy;  HlpRow := 0;
    While ( not EOF( HlpFile ) ) and ( HlpResult = 0 ) and
          ( HlpCh <> EscKy ) do
      Begin
        If HlpRow = 0 Then
          Begin
            For ClrHlpRow := 1 to ScreenLines do
              Begin
                CrsrXY( 1, ClrHlpRow );  ClrEol;
              End;
            CrsrXY( 1, 1 );
          End;
        ReadLn( HlpFile, HlpLn );  HlpResult := IOResult;
        If HlpResult = 0 Then
          Begin
            If Copy( HlpLn, 1, 1 ) = ^L Then
              HlpRow := ScreenLines
            Else
              Begin
                Inc( HlpRow );  CrsrXY( 1, HlpRow );
                Write( HlpLn );
              End;
            If EOF( HlpFile ) or ( HlpRow = ScreenLines ) Then
              Begin
                HlpRow := 0;
                While not KeyPressed do;
                While KeyPressed do HlpCh := ReadKey;
              End;
          End;
      End;
    Close( HlpFile );  HlpResult := HlpResult + IOResult;
    If HlpResult <> 0 Then
      Begin
        Message( ^G + 'Trouble accessing help file ' + 
                      SplitPath + HelpFileName );
        Prompt( 'ERROR:  Press any key.' );  PressAnyKey;
      End;
    Move( SaveScreen, Mem[ VAddr : $0000 ], SizeOf( SaveScreen ) );
    SupressScreenUpdate := False;
  End;

{------------------------------------------------------------------}

Procedure ShowError;
  Begin
    If ErrMsg <> '' Then
      Begin
        Message( ^G + ErrMsg + '.' );  ErrMsg := '';
        Prompt( 'ERROR:  Press any key.' );
        PressAnyKey;  Message( '' );
      End;
  End;

(* -------------------------------------------------------------- *)
(*  Data record handling.                                         *)
(* -------------------------------------------------------------- *)

Procedure CreateTrx;
  { make a blank record }
  Begin
    FillChar( HdrRec, SizeOf( HdrRec ), 0 );  MemoRec := '';
    FillChar( DisbArr, SizeOf( DisbArr ), 0 );
  End;

{------------------------------------------------------------------}

Procedure ReadTrx( TrxIdx : Integer );
  { get a record from data storage arrays }
  Begin
    If ( TrxIdx >= 1 ) and ( TrxIdx <= TrxCount ) Then
      Begin
        HdrRec.Flag := HdrMem_1^[ TrxIdx ].Flag;
        HdrRec.Date := HdrMem_1^[ TrxIdx ].Date;
        HdrRec.PrcDate := HdrMem_1^[ TrxIdx ].PrcDate;
        Move( HdrMem_1^[ TrxIdx ].Memo, HdrRec.Memo,
              SizeOf( HdrRec.Memo ) );
        HdrRec.Seq := HdrMem_2^[ TrxIdx ].Seq;
        HdrRec.Name := HdrMem_2^[ TrxIdx ].Name;
        If HdrRec.Memo <> Nil Then
          MemoRec := HdrRec.Memo^
        Else
          MemoRec := '';
        DisbArr[ 1 ].Acct := DisbAcctMem_1^[ TrxIdx ];
        DisbArr[ 1 ].Amt := DisbAmtMem_1^[ TrxIdx ];
        DisbArr[ 2 ].Acct := DisbAcctMem_2^[ TrxIdx ];
        DisbArr[ 2 ].Amt := DisbAmtMem_2^[ TrxIdx ];
        DisbArr[ 3 ].Acct := DisbAcctMem_3^[ TrxIdx ];
        DisbArr[ 3 ].Amt := DisbAmtMem_3^[ TrxIdx ];
        DisbArr[ 4 ].Acct := DisbAcctMem_4^[ TrxIdx ];
        DisbArr[ 4 ].Amt := DisbAmtMem_4^[ TrxIdx ];
        DisbArr[ 5 ].Acct := DisbAcctMem_5^[ TrxIdx ];
        DisbArr[ 5 ].Amt := DisbAmtMem_5^[ TrxIdx ];
        DisbArr[ 6 ].Acct := DisbAcctMem_6^[ TrxIdx ];
        DisbArr[ 6 ].Amt := DisbAmtMem_6^[ TrxIdx ];
        DisbArr[ 7 ].Acct := DisbAcctMem_7^[ TrxIdx ];
        DisbArr[ 7 ].Amt := DisbAmtMem_7^[ TrxIdx ];
      End
    Else
      CreateTrx;
  End;

{------------------------------------------------------------------}

Procedure WriteTrx( TrxIdx : Integer );
  { write a record from data storage arrays }
  Begin
    If ( TrxIdx >= 1 ) and ( TrxIdx <= TrxCount + 1 ) Then
      Begin
        { add a record past end if necessary }
        If TrxIdx > TrxCount Then Inc( TrxCount );
        { memos must be dynamically allocated or released }
        If HdrRec.Memo <> Nil Then
          Begin
            HdrRec.Memo^ := MemoRec;
            If MemoRec = '' Then
              Begin
                Dispose( HdrRec.Memo );  HdrRec.Memo := Nil;
              End
            Else
              HdrRec.Memo^ := MemoRec;
          End
        Else
          If MemoRec <> '' Then
            Begin
              If MemAvail > SizeOf( MemoType ) Then
                Begin
                  New( HdrRec.Memo );  HdrRec.Memo^ := MemoRec;
                End
              Else
                ErrMsg := 'Insufficient memory for memo';
            End;
        HdrMem_1^[ TrxIdx ].Flag := HdrRec.Flag;
        HdrMem_1^[ TrxIdx ].Date := HdrRec.Date;
        HdrMem_1^[ TrxIdx ].PrcDate := HdrRec.PrcDate;
        Move( HdrRec.Memo, HdrMem_1^[ TrxIdx ].Memo,
              SizeOf( HdrMem_1^[ TrxIdx ].Memo ) );
        HdrMem_2^[ TrxIdx ].Seq := HdrRec.Seq;
        HdrMem_2^[ TrxIdx ].Name := HdrRec.Name;
        DisbAcctMem_1^[ TrxIdx ] := DisbArr[ 1 ].Acct;
        DisbAmtMem_1^[ TrxIdx ] := DisbArr[ 1 ].Amt;
        DisbAcctMem_2^[ TrxIdx ] := DisbArr[ 2 ].Acct;
        DisbAmtMem_2^[ TrxIdx ] := DisbArr[ 2 ].Amt;
        DisbAcctMem_3^[ TrxIdx ] := DisbArr[ 3 ].Acct;
        DisbAmtMem_3^[ TrxIdx ] := DisbArr[ 3 ].Amt;
        DisbAcctMem_4^[ TrxIdx ] := DisbArr[ 4 ].Acct;
        DisbAmtMem_4^[ TrxIdx ] := DisbArr[ 4 ].Amt;
        DisbAcctMem_5^[ TrxIdx ] := DisbArr[ 5 ].Acct;
        DisbAmtMem_5^[ TrxIdx ] := DisbArr[ 5 ].Amt;
        DisbAcctMem_6^[ TrxIdx ] := DisbArr[ 6 ].Acct;
        DisbAmtMem_6^[ TrxIdx ] := DisbArr[ 6 ].Amt;
        DisbAcctMem_7^[ TrxIdx ] := DisbArr[ 7 ].Acct;
        DisbAmtMem_7^[ TrxIdx ] := DisbArr[ 7 ].Amt;
      End;
  End;

{------------------------------------------------------------------}

Function FormatLine : String;
  { create a display string out of the current transaction record }
  Var
    OutLn : String;
    FldLp, DisbLp : Byte;
  Begin
    OutLn := '';
    { output fields according to field display order specifier string }
    For FldLp := 1 to Length( OutFldOrder ) do
      Begin
        { pad blank fields except memo }
        OutLn := OutLn + ' ';
        Case OutFldOrder[ FldLp ] of
          FlagMk :
            If HdrRec.Flag = '' Then
              OutLn := OutLn + Space( FlagSize + 1 )
            Else
              OutLn := OutLn +
                Pad( FlagMk + HdrRec.Flag, FlagSize + 1 );
          DateMk :
            If HdrRec.Date = '' Then
              OutLn := OutLn + Space( DateSize + 1 )
            Else
              OutLn := OutLn +
                Pad( DateMk + HdrRec.Date, -( DateSize + 1 ) );
          SeqMk :
            If HdrRec.Seq = '' Then
              OutLn := OutLn + Space( SeqSize + 1 )
            Else
              OutLn := OutLn +
                Pad( SeqMk + HdrRec.Seq, -( SeqSize + 1 ) );
          NameMk :
            If HdrRec.Name = '' Then
              OutLn := OutLn + Space( NameSize + 1 )
            Else
              OutLn := OutLn +
                Pad( NameMk + HdrRec.Name + NameMk, -( NameSize + 1 ) );
          AmtMk :
            Begin
              DisbLp := 2;
              For DisbLp := 1 to MaxDisb do
                Begin
                  DisbRec := DisbArr[ DisbLp ];
                  If Amount( DisbRec.Amt ) Then
                    Begin
                      OutLn := OutLn +
                        Pad( AmtMk + R_S( DisbRec.Amt ), AmtSize + 1 );
                      OutLn := OutLn + ' ' +
                        Pad( AcctMk + DisbRec.Acct, -( AcctSize + 1 ) );
                    End;
                End;
            End;
          PrcDateMk :
            If HdrRec.PrcDate = '' Then
              OutLn := OutLn + Space( DateSize + 1 )
            Else
              OutLn := OutLn +
                Pad( PrcDateMk + HdrRec.PrcDate, -( DateSize + 1 ) );
          MemoMk :
            If HdrRec.Memo <> Nil Then
              OutLn := OutLn + MemoMk + HdrRec.Memo^ + MemoMk;
        End;
      End;
    FormatLine := OutLn;
    If Length( OutLn ) > MaxLen Then MaxLen := Length( OutLn );
  End;

{------------------------------------------------------------------}

Procedure SortDisbs;
  { sort disbursements of current transaction }
  {  put balance account entries first, high to low amounts }
  {  sort remainder alphabetically then high to low amounts }
  Var
    DisbLp, DisbCnt : Integer;
    SortAcct1, SortAcct2 : DisbAcctType;
    SortAmt1, SortAmt2 : DisbAmtType;
    DisbArr2 : DisbArrType;
    Swap : DisbRecType;
    Done : Boolean;
  Begin
    { remove blanks }
    DisbLp := 1;  DisbCnt := 0;
    FillChar( DisbArr2, SizeOf( DisbArr2 ), 0 );
    For DisbLp := 1 to MaxDisb do
      If Amount( DisbArr[ DisbLp ].Amt ) and
         ( DisbArr[ DisbLp ].Acct <> '' ) Then
        Begin
          Inc( DisbCnt );
          DisbArr2[ DisbCnt ] := DisbArr[ DisbLp ];
        End;
    { sort }
    Done := False;
    While not Done do
      Begin
        Done := True;
        For DisbLp := 1 to DisbCnt - 1 do
          Begin
            SortAcct1 := DisbArr2[ DisbLp ].Acct;
            { force balance account to top of sort }
            If SortAcct1 = ShowAcct Then SortAcct1 := '';
            SortAmt1 := DisbArr2[ DisbLp ].Amt;
            SortAcct2 := DisbArr2[ DisbLp + 1 ].Acct;
            If SortAcct2 = ShowAcct Then SortAcct2 := '';
            SortAmt2 := DisbArr2[ DisbLp + 1 ].Amt;
            If ( SortAcct1 > SortAcct2 ) or
               ( ( SortAcct1 = SortAcct2 ) and
                 ( SortAmt1 < SortAmt2 ) ) Then
              Begin
                Done := False;
                Swap := DisbArr2[ DisbLp ];
                DisbArr2[ DisbLp ] := DisbArr2[ DisbLp + 1 ];
                DisbArr2[ DisbLp + 1 ] := Swap;
              End;
          End;  { for/while }
      End;
    DisbArr := DisbArr2;
  End;

{------------------------------------------------------------------}

Procedure SortAllDisbs;
  { sort the disbursements of all transactions }
  Var
    PrevTrxNum, DisbLp : Integer;
    GrpNm : DisbAcctType;
  Begin
    If ErrMsg <> '' Then Exit;
    Message( 'Sorting disbursements . . .' );
    PrevTrxNum := TrxNum;  TrxNum := 1;
    While ( ErrMsg = '' ) and ( TrxNum <= TrxCount ) do
      Begin
        ReadTrx( TrxNum );  ShowStat;
        If ErrMsg = '' Then
          Begin  { read ok }
            SortDisbs;
            WriteTrx( TrxNum );
          End;  { read ok }
        Inc( TrxNum );
      End;
    TrxNum := PrevTrxNum;  ShowStat;
  End;

{------------------------------------------------------------------}

Procedure ParseQuote( Var InLn : String;  Delim : Char );
  { used by line translate }
  Begin
    If ErrMsg <> '' Then Exit;
    PrsWd := '';
    While ( InLn <> '' ) and ( InLn[ 1 ] <> DeLim ) do
      Begin
        PrsWd := PrsWd + InLn[ 1 ];
        Delete( InLn, 1, 1 );
      End;
    If ( InLn <> '' ) and ( InLn[ 1 ] <> Delim ) Then
      ErrMsg := 'Missing closing delimiter'
    Else
      Delete( InLn, 1, 1 );
    If PrsWd = '' Then ErrMsg := 'Quote missing';
  End;

{------------------------------------------------------------------}

Procedure ParseWord( Var InLn : String );
  { used by line translate }
  Begin
    If ErrMsg <> '' Then Exit;
    PrsWd := '';
    While ( InLn <> '' ) and ( InLn[ 1 ] <> ' ' ) and
          ( not ( InLn[ 1 ] in MkSet ) ) do
      Begin
        PrsWd := PrsWd + InLn[ 1 ];
        Delete( InLn, 1, 1 );
      End;
    If PrsWd = '' Then ErrMsg := 'Word missing';
  End;

{------------------------------------------------------------------}

Function DateString : DateType;
  { return system date as YY/MM/DD string }
  Var
    Regs : Registers;
  Begin
    Regs.AH := $2A;  Intr( $21, Regs );
    DateString :=
      Chr( Regs.CX mod 100 div 10 + 48 ) + { Yh }
      Chr( Regs.CX mod 10 + 48 ) +         { Yl }
      Chr( 47 ) +                          { slash }
      Chr( Regs.DH div 10 + 48 ) +         { Mh }
      Chr( Regs.DH mod 10 + 48 ) +         { Ml }
      Chr( 47 ) +                          { slash }
      Chr( Regs.DL div 10 + 48 ) +         { Dh }
      Chr( Regs.DL mod 10 + 48 );          { Dl }
  End;

{------------------------------------------------------------------}

Procedure CheckDate( DateStr : String20 );
  { check valididity of formatted date }
  Var
    Yr, Mo, Da, DTResult : Integer;
  Const
    DaysPerMonth : array [ 1..12 ] of Integer =
      ( 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 );
  Begin
    If ErrMsg <> '' Then Exit;
    If Length( DateStr ) <> 8 Then
      ErrMsg := 'Bad length'
    Else
      Begin
        Val( Copy( DateStr, 1, 2 ), Yr, DTResult );
        Val( Copy( DateStr, 4, 2 ), Mo, DTResult );
        Val( Copy( DateStr, 7, 2 ), Da, DTResult );
        If ( Yr < 0 ) or ( Yr > 99 ) or
           ( Mo < 1 ) or ( Mo > 12 ) or
           ( Da < 1 ) or ( Da > 31 ) or
           ( ( not ( Da in [ 1 .. DaysPerMonth[ Mo ] ] ) ) and
             ( not ( ( Da = 29 ) and ( Mo = 2 ) and
             ( Yr mod 4 = 0 ) ) ) ) Then
          ErrMsg := 'Invalid date';
      End;
  End;

{------------------------------------------------------------------}

Procedure FixDate( Var InWd : String20 );
  { take a date in any format and output YY-MM-DD }
  Var
    ChkLp, SlashCnt : Byte;
    PrevSlash : Boolean;
  Begin
    If InWd = '' Then Exit;
    { check for basic packed format }
    For ChkLp := 1 to Length( InWd ) do
      If InWd[ ChkLp ] = '-' Then
        InWd[ ChkLp ] := '/';
    If ( Pos( '/', InWd ) = 0 ) and ( Length( InWd ) = 6 ) Then
      Begin
        Insert( '/', InWd, 5 );  Insert( '/', InWd, 3 );
      End;
    { prepare to check number of slashes }
    ChkLp := 1;  SlashCnt := 0;  PrevSlash := True;
    While ( ErrMsg = '' ) and ( ChkLp <= Length( InWd ) ) do
      Begin
        If not ( InWd[ ChkLp ] in DateSet ) Then
          ErrMsg := 'Invalid character'
        Else
          If InWd[ ChkLp ] = '/' Then
            Begin
              If PrevSlash Then
                ErrMsg := 'Bad format'
              Else
                Begin
                  Inc( SlashCnt );  PrevSlash := True;
                End;
              InWd[ ChkLp ] := '/';
            End
          Else
            PrevSlash := False;
        Inc( ChkLp );
      End;
    If ( ErrMsg = '' ) and ( SlashCnt <> 2 ) Then
      ErrMsg := 'Missing "/"';
    { pad one digit fields }
    If ErrMsg = '' Then
      If Length( InWd ) < 4 Then
        ErrMsg := 'Date too short'
      Else
        Begin
          If Copy( InWd, 5, 1 ) = '/' Then Insert( '0', InWd, 4 );
          If Length( InWd ) < 8 Then Insert( '0', InWd, 7 );
          { done? now check for validity of value }
          CheckDate( InWd );
        End;
  End;

{------------------------------------------------------------------}

Function ParseDate( Var InLn : String ) : DateType;
  { used by line translate }
  Begin
    ParseDate := '';  ParseWord( InLn );
    If ErrMsg = '' Then FixDate( PrsWd );
    If ErrMsg = '' Then ParseDate := PrsWd;
  End;

{------------------------------------------------------------------}

Function ParseReal( Var InLn : String ) : Real;
  { used by line translate }
  Var
    ChkLp : Byte;
    Result : Integer;
    OutReal : Real;
  Begin
    { initialize }
    ParseReal := 0.0;  ParseWord( InLn );  ChkLp := 1;
    { validity check }
    If ErrMsg = '' Then
      Begin
        While ( ErrMsg = '' ) and ( ChkLp <= Length( PrsWd ) ) do
          Begin
            If not ( PrsWd[ ChkLp ] in NumericSet ) Then
              ErrMsg := 'Invalid character';
            Inc( ChkLp );
          End;
        If ErrMsg = '' Then
          Begin
            Val( PrsWd, OutReal, Result );
            If Result <> 0 Then ErrMsg := 'Bad format';
          End;
      End;
    If ErrMsg = '' Then ParseReal := OutReal;
  End;

{------------------------------------------------------------------}

Procedure TranslateLine( InLn : String );
  { translate a string into field values, used by data file read }
  Var
    FldCh, PrevFldCh : Char;
    ChkLp, DisbLp, PrevDisbIdx : Byte;
    AmtTemp : Real;
  Begin
    If ErrMsg <> '' Then Exit;
    { initialize }
    CreateTrx;  Strip( InLn );
    ErrMsg := '';  FldCh := NoKy;
    If InLn = '' Then Exit;
    While ( ErrMsg = '' ) and ( InLn <> '' ) do
      Begin
        { initialize }
        PrsWd := '';  FldType := '';  Strip( InLn );
        { decode first character as field identifier }
        PrevFldCh := FldCh;  FldCh := InLn[ 1 ];
        { parse appropriate type }
        Case FldCh of
          FlagMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Flag';
              ParseWord( InLn );  ChkLp := 1;
              While ( ErrMsg = '' ) and
                    ( ChkLp <= Length( PrsWd ) ) do
                Begin
                  If not ( PrsWd[ ChkLp ] in AlphaSet ) Then
                    ErrMsg := 'Invalid character';
                  Inc( ChkLp );
                End;
              If ErrMsg = '' Then
                If Length( PrsWd ) > FlagSize Then
                  ErrMsg := 'Item too long';
              If ErrMsg = '' Then HdrRec.Flag := PrsWd;
            End;
          DateMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Date issued';
              HdrRec.Date := ParseDate( InLn );
            End;
          SeqMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Sequence';
              ParseWord( InLn );  ChkLp := 1;
              While ( ErrMsg = '' ) and
                    ( ChkLp <= Length( PrsWd ) ) do
                Begin
                  If not ( PrsWd[ ChkLp ] in AlphaSet ) Then
                    ErrMsg := 'Invalid character';
                  Inc( ChkLp );
                End;
              If ErrMsg = '' Then
                If Length( PrsWd ) > SeqSize Then
                  ErrMsg := 'Item too long';
              If ErrMsg = '' Then HdrRec.Seq := PrsWd;
            End;
          NameMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Name';
              ParseQuote( InLn, NameMk );  ChkLp := 1;
              If ErrMsg = '' Then
                If Length( PrsWd ) > NameSize Then
                  ErrMsg := 'Item too long';
              If ErrMsg = '' Then HdrRec.Name := PrsWd;
            End;
          AmtMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Amount';
              AmtTemp := ParseReal( InLn );  DisbLp := 1;
              While ( DisbLp <= MaxDisb ) and
                    Amount( DisbArr[ DisbLp ].Amt ) do
                Inc( DisbLp );
              If ( DisbLp <= MaxDisb ) Then
                DisbArr[ DisbLp ].Amt := AmtTemp
              Else
                ErrMsg := 'Too many amounts';
            End;
          AcctMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Account';
              ParseWord( InLn );  ChkLp := 1;
              While ( ErrMsg = '' ) and
                    ( ChkLp <= Length( PrsWd ) ) do
                Begin
                  If not ( PrsWd[ ChkLp ] in AlphaSet ) Then
                    ErrMsg := 'Invalid character';
                  Inc( ChkLp );
                End;
              If ErrMsg = '' Then
                If Length( PrsWd ) > AcctSize Then
                  ErrMsg := 'Item too long';
              If ErrMsg = '' Then
                If PrevFldCh <> AmtMk Then
                  ErrMsg := 'No amount preceding'
                Else
                  Begin
                    DisbLp := 1;
                    While ( DisbLp <= MaxDisb ) and
                          ( DisbArr[ DisbLp ].Acct <> '' ) do
                      Inc( DisbLp );
                    If ( DisbLp <= MaxDisb ) Then
                      DisbArr[ DisbLp ].Acct := PrsWd
                    Else
                      ErrMsg := 'Too many accounts';
                  End;
            End;
          PrcDateMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Date processed';
              HdrRec.PrcDate := ParseDate( InLn );
            End;
          MemoMk :
            Begin
              Delete( InLn, 1, 1 );  FldType := 'Memo';
              ParseQuote( InLn, MemoMk );  ChkLp := 1;
              If ErrMsg = '' Then
                If Length( PrsWd ) > MemoSize Then
                  ErrMsg := 'Memo too long';
              If ErrMsg = '' Then MemoRec := PrsWd;
            End;
         End;
        { format detailed error message }
        If ( ErrMsg = '' ) and ( FldCh <> AcctMk ) Then
          If PrevFldCh = AmtMk Then
            Begin
              ErrMsg := 'No account following';
              FldType := 'Amount';
            End;
        If ErrMsg = '' Then
          Begin
            If FldType = '' Then
              ErrMsg := 'Unknown field type';
          End
        Else
          ErrMsg := ErrMsg + ' in ' + FldType + ' field';
      End;
  End;

(* -------------------------------------------------------------- *)
(*  High level support routines.                                  *)
(* -------------------------------------------------------------- *)

Procedure ReCalc;
  { recalculate line and total balances }
  Var
    ReCalcLp : Integer;
    DisbLp : Byte;
    ReCalcLineBal : Real;
    ReCalcTotalBal : Real;
  Begin
    { initialize }
    ReCalcLp := 1;  ReCalcLineBal := 0.0;  ReCalcTotalBal := 0.0;
    While ( not KeyPressed ) and ( ReCalcLp <= TrxCount ) do
      Begin
        ReadTrx( ReCalcLp );
        If HdrRec.Name[ 1 ] <> GroupMark Then
          Begin
            For DisbLp := 1 to MaxDisb do
              If ( DisbArr[ DisbLp ].Acct = ShowAcct ) and
                 ( ( BalCalcMode = 'I' ) or
                   ( HdrRec.PrcDate <> '' ) ) Then
                Begin
                  ReCalcTotalBal :=
                    ReCalcTotalBal + DisbArr[ DisbLp ].Amt;
                  If ReCalcLp <= TrxNum Then
                    ReCalcLineBal :=
                      ReCalcLineBal + DisbArr[ DisbLp ].Amt;
                End;
          End;
        Inc( ReCalcLp );
      End;
    { if finished then clear flag }
    If not KeyPressed Then
      Begin
        LineBal := ReCalcLineBal;  TotalBal := ReCalcTotalBal;
        DoReCalc := False;
      End;
    ReadTrx( TrxNum );
  End;

{------------------------------------------------------------------}

Procedure GetMemory;
  { primary memory size determination and allocation }
  Var
    TrxRecSize, TrxLimit, BiggestRecSize : Word;
    TrxMem : LongInt;
  Begin
    { calculate greatest possible total number of records }
    TrxRecSize := SizeOf( HdrRec ) + SizeOf( DisbArr );
    TrxMem := MemAvail - ( MaxMemo * SizeOf( MemoRec ) );
    TrxMax := TrxMem div TrxRecSize;
    { calculate maximum possible array of largest record type }
    BiggestRecSize := SizeOf( HdrRecType_1 );
    If BiggestRecSize < SizeOf( HdrRecType_2 ) Then
       BiggestRecSize := SizeOf( HdrRecType_2 );
    If BiggestRecSize < SizeOf( DisbAcctType ) Then
       BiggestRecSize := SizeOf( DisbAcctType );
    If BiggestRecSize < SizeOf( DisbAmtType ) Then
       BiggestRecSize := SizeOf( DisbAmtType );
    TrxLimit := 65535 div BiggestRecSize;
    MaxTrx65K := TrxLimit;
    If TrxMax > TrxLimit Then TrxMax := TrxLimit;
    GetMem( HdrMemSpace_1, TrxMax * SizeOf( HdrRecType_1 ) );
    HdrMem_1 :=
      Ptr( Seg( HdrMemSpace_1^ ), Ofs( HdrMemSpace_1^ ) );
    GetMem( HdrMemSpace_2, TrxMax * SizeOf( HdrRecType_2 ) );
    HdrMem_2 :=
      Ptr( Seg( HdrMemSpace_2^ ), Ofs( HdrMemSpace_2^ ) );
    GetMem( DisbAcctMemSpace_1, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_1 :=
      Ptr( Seg( DisbAcctMemSpace_1^ ), Ofs( DisbAcctMemSpace_1^ ) );
    GetMem( DisbAmtMemSpace_1, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_1 :=
      Ptr( Seg( DisbAmtMemSpace_1^ ), Ofs( DisbAmtMemSpace_1^ ) );
    GetMem( DisbAcctMemSpace_2, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_2 :=
      Ptr( Seg( DisbAcctMemSpace_2^ ), Ofs( DisbAcctMemSpace_2^ ) );
    GetMem( DisbAmtMemSpace_2, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_2 :=
      Ptr( Seg( DisbAmtMemSpace_2^ ), Ofs( DisbAmtMemSpace_2^ ) );
    GetMem( DisbAcctMemSpace_3, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_3 :=
      Ptr( Seg( DisbAcctMemSpace_3^ ), Ofs( DisbAcctMemSpace_3^ ) );
    GetMem( DisbAmtMemSpace_3, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_3 :=
      Ptr( Seg( DisbAmtMemSpace_3^ ), Ofs( DisbAmtMemSpace_3^ ) );
    GetMem( DisbAcctMemSpace_4, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_4 :=
      Ptr( Seg( DisbAcctMemSpace_4^ ), Ofs( DisbAcctMemSpace_4^ ) );
    GetMem( DisbAmtMemSpace_4, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_4 :=
      Ptr( Seg( DisbAmtMemSpace_4^ ), Ofs( DisbAmtMemSpace_4^ ) );
    GetMem( DisbAcctMemSpace_5, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_5 :=
      Ptr( Seg( DisbAcctMemSpace_5^ ), Ofs( DisbAcctMemSpace_5^ ) );
    GetMem( DisbAmtMemSpace_5, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_5 :=
      Ptr( Seg( DisbAmtMemSpace_5^ ), Ofs( DisbAmtMemSpace_5^ ) );
    GetMem( DisbAcctMemSpace_6, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_6 :=
      Ptr( Seg( DisbAcctMemSpace_6^ ), Ofs( DisbAcctMemSpace_6^ ) );
    GetMem( DisbAmtMemSpace_6, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_6 :=
      Ptr( Seg( DisbAmtMemSpace_6^ ), Ofs( DisbAmtMemSpace_6^ ) );
    GetMem( DisbAcctMemSpace_7, TrxMax * SizeOf( DisbAcctType ) );
    DisbAcctMem_7 :=
      Ptr( Seg( DisbAcctMemSpace_7^ ), Ofs( DisbAcctMemSpace_7^ ) );
    GetMem( DisbAmtMemSpace_7, TrxMax * SizeOf( DisbAmtType ) );
    DisbAmtMem_7 :=
      Ptr( Seg( DisbAmtMemSpace_7^ ), Ofs( DisbAmtMemSpace_7^ ) );
  End;

{------------------------------------------------------------------}

Procedure Help;
  { command line help }
  Var
    Con : Text;
  Begin
    For CpyRtLp := 1 to 26 do
      If CpyRtLp = 1 Then
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
      Else
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                 Ord( CpyRt[ CpyRtLp - 1 ] ) );
    Assign( Con, '' );  ReWrite( Con );
    Writeln( Con, ProgName + ' ver ' + Ver + ' rev ' + Rev +
                  ' - ' + CpyRt + '.' );
    Writeln( Con, 'Generalized financial ledger and checkbook.' );
    Writeln( Con, 'Menu driven, configurable, with editing,' );
    Writeln( Con, 'reporting, and on line help features.' );
    Writeln( Con, 'Can be run in less than 256K of RAM.' );
    Close( Con );  Halt( 1 );
  End;

(* -------------------------------------------------------------- *)
(*  Data entry routines.                                          *)
(* -------------------------------------------------------------- *)

Procedure InpStr( Var Value : String80;  Len : Integer;  Typ : Byte );
  { full feature data entry routine }
  Var
    ReadStr, PrevValue : String80;
    RdCh, InsCh : Char;
    InpLp, X, StartX, MaxX, XPos : Byte;
    Done : Boolean;
  Begin
    Done := False;
    SpclInpKeys := SpclInpKeys + EntrKy + TabKy + S_TabKy;
    If ( Typ = _Text ) or ( Typ = _Memo ) Then
      InsCh := ' '
    Else
      InsCh := #250;
    { initialize }
    Escape := False;  InpKey := NoKy;
    X := WhereX;  StartX := X;
    { pass in default }
    PrevValue := Value;
    If Typ = _Date Then
      Begin  FixDate( Value );  ErrMsg := '';  End;
    ReadStr := Value;
    { get and process keys }
    While ( not Done ) and ( not Escape ) do
      Begin
        CrsrXY( StartX, WhereY );
        Write( PadStr( ReadStr, -Len, #250 ) );
        CrsrXY( X, WhereY );  XPos := X - StartX + 1;
        MaxX := StartX - 1 + Length( ReadStr );
        RdCh := GetKey;
        If Pos( RdCh, SpclInpKeys ) > 0 Then
          Done := True
        Else
          Case RdCh of
          EscKy : Escape := True;
           HmKy, ^A : X := StartX;
          EndKy, ^F : X := MaxX + 1;
          RgtKy, ^D : If X < MaxX + 1 Then Inc( X );
          LftKy, ^S : If X > StartX Then Dec( X );
          InsKy : If Length( ReadStr ) < Len Then
                    ReadStr := Copy( ReadStr, 1, XPos - 1 ) + InsCh +
                               Copy( ReadStr, XPos, 255 );
          DelKy : If ( Length( ReadStr ) > 0 ) and
                     ( XPos <= Length( ReadStr ) ) Then
                    ReadStr := Copy( ReadStr, 1, XPos - 1 ) +
                               Copy( ReadStr, XPos + 1, 255 );
        BkSpcKy : If Length( ReadStr ) > 0 Then
                    Begin
                      If ( XPos > 1 ) and
                         ( XPos <= Length( ReadStr ) + 1 ) Then
                        Begin
                          ReadStr := Copy( ReadStr, 1, XPos - 2 ) +
                                     Copy( ReadStr, XPos, 255 );
                          Dec( X );
                        End;
                    End;
       ' '..'~' : If XPos <= Len Then
                    Begin
                      If InpKey = NoKy Then ReadStr := '';
                      If Typ <> _Memo Then RdCh := UpCase( RdCh );
                      If ( Typ = _Text ) or ( Typ = _Memo ) or
                         ( ( Typ = _String ) and ( RdCh in AlphaSet ) ) or
                         ( ( Typ = _Real ) and ( RdCh in NumericSet ) ) or
                         ( ( Typ = _Date ) and ( RdCh in DateSet ) ) Then
                        Begin
                          If XPos > Length( ReadStr ) Then
                            ReadStr := ReadStr + RdCh
                          Else
                            ReadStr[ XPos ] := RdCh;
                          Inc( X );
                        End;
                    End;
           End;
        InpKey := RdCh;
      End;
    { remove insert characters }
    If not Escape Then
      Begin
        For InpLp := Length( ReadStr ) DownTo 1 do
          If ReadStr[ InpLp ] = #250 Then
            Delete( ReadStr, InpLp, 1 );
        Strip( ReadStr );  Value := ReadStr;
      End;
    CrsrXY( StartX, WhereY );
    If Typ = _Real Then
      Write( Pad( Value, Len ) )
    Else
      Write( Pad( Value, -Len ) );
    If Typ = _Date Then
      Begin
        FixDate( Value );
        If ErrMsg <> '' Then Value := PrevValue;
      End;
    InpChange := Value <> PrevValue;
    SpclInpKeys := '';
    Prompt( '' );
  End;

{------------------------------------------------------------------}

Procedure InpReal( Var Value : Real );
  { accept a numeric string then validity check for real number }
  Var
    Work : String40;
    Temp : Real;
    Code, Len : Integer;
  Begin
    If ErrMsg <> '' Then Exit;
    If Amount( Frac( Value ) ) Then
      Str( Value:1:2, Work )
    Else
      Str( Value:1:0, Work );
    Len := AmtSize;  InpStr( Work, Len, _Real );
    If not Escape Then
      Begin
        If Work = '' Then Work := '0.0';
        Val( Work, Temp, Code );
        If Code <> 0 Then
          ErrMsg := 'Bad amount input';
        If ErrMsg = '' Then
          If ( Temp < -99999.99 ) or ( Temp > 999999.99 ) Then
            ErrMsg := 'Amount out of range';
        If ErrMsg = '' Then Value := Temp;
      End;
  End;

{------------------------------------------------------------------}

Procedure InpInt( Var I : Integer );
  { accept a numeric string then validity check for integer }
  Var
    R : Real;
  Begin
    R := I + 0.0;  InpReal( R );
    If ErrMsg = '' Then
      Begin
        If R < -MaxInt + 0.0 Then R := -MaxInt + 0.0;
        If R > MaxInt + 0.0 Then R := MaxInt + 0.0;
        I := Trunc( R );
      End;
  End;

{------------------------------------------------------------------}

Procedure EnterFldOrder( Var IOrder : String10 );
  { accept a special field order specification string }
  Var
    FldSet, FldOrderTemp : String20;
    FldLp, FldPos, ShowLp : Byte;
    FldInCh : Char;
  Begin
    FldSet := IOrder;
    If Length( FldSet ) <> Length( DfltFldOrder ) Then
      FldSet := DfltFldOrder
    Else
      For FldLp := 1 to Length( DfltFldOrder ) do
        If Pos( DfltFldOrder[ FldLp ], FldSet ) = 0 Then
          FldSet := DfltFldOrder;
    FldLp := 1;  FldOrderTemp := '';
    While ( FldLp <= Length( FldSet ) - 1 ) and ( not Escape ) do
      Begin
        Message( 'Set field order, select field ' + I_S( FldLp ) + ':' );
        Prompt( '' );
        For ShowLp := 1 to Length( FldSet ) do
          Begin
            Write( '  ' + FldSet[ ShowLp ] );
            Case FldSet[ ShowLp ] of
              FlagMk    : Write( 'flag' );
              DateMk    : Write( 'date' );
              SeqMk     : Write( 'sequence' );
              NameMk    : Write( 'name' );
              AmtMk     : Write( 'amount' );
              PrcDateMk : Write( 'processed' );
              MemoMk    : Write( 'memo' );
            End;
          End;
        FldInCh := FldSet[ 1 ];  Select( FldInCh, FldSet );
        If not Escape Then
          Begin
            FldPos := Pos( FldInCh, FldSet );
            Delete( FldSet, FldPos, 1 );
            FldOrderTemp := FldOrderTemp + FldInCh;
          End;
      End;
    If not Escape Then
      Begin
        FldOrderTemp := FldOrderTemp + FldSet;
        InpChange := IOrder <> FldOrderTemp;
        IOrder := FldOrderTemp;
      End;
  End;

(* -------------------------------------------------------------- *)
(*  Screen painting.                                              *)
(* -------------------------------------------------------------- *)

Procedure ShowLine( TrxIdx : Integer );
  { display a record on appropriate screen line }
  Var
    X : Integer;
    OutLn : String;
  Begin
    X := TrxIdx - PgStart + 1;
    { check if off screen }
    If ( X >= 1 ) and ( X <= ScreenLines ) Then
      Begin
        CrsrXY( 1, X );  ReadTrx( TrxIdx );
        { crop }
        OutLn := Copy( FormatLine, LeftMgn, 79 );
        { highlight current line }
        If TrxIdx = TrxNum Then
          Begin
            TextColor( FrClr );  TextBackGround( BkClr );
          End
        Else
          Begin
            TextColor( White );  TextBackGround( Black );
          End;
        Write( OutLn );  ClrEol;
      End;
  End;

{------------------------------------------------------------------}

Procedure UpdateScreen;
  { display all lines on current page }
  Var
    TrxLp : Integer;
  Begin
    DoUpdateScreen := True;
    If KeyPressed or SupressScreenUpdate Then Exit;
    TrxLp := PgStart;
    ShowLine( TrxNum );
    While ( not KeyPressed ) and
          ( TrxLp <= PgEnd ) do
      Begin
        If TrxLp <> TrxNum Then ShowLine( TrxLp );
        Inc( TrxLp );
      End;
    DoUpdateScreen := KeyPressed;
    ReadTrx( TrxNum );
  End;

{------------------------------------------------------------------}

Procedure SetPgEnd;
  { recalculate page end given start }
  Begin
    PgEnd := PgStart + ScreenLines - 1;
  End;

{------------------------------------------------------------------}

Procedure AdjRange;
  { check and correct for valid range start and end }
  Var
    RngSwap : Integer;
  Begin
    If RangeStart <= RangeEnd Then Exit;
    RngSwap := RangeStart;
    RangeStart := RangeEnd;
    RangeEnd := RngSwap;
  End;

(* -------------------------------------------------------------- *)
(*  Included edit, disk, menus, movement and report procedures.   *)
(* -------------------------------------------------------------- *)

{$I LDGR.INC }
{$I LDGRREPT.INC }

(* -------------------------------------------------------------- *)
(*  MAIN LOGIC                                                    *)
(* -------------------------------------------------------------- *)

Begin

  If ( ParamCount > 0 ) and ( ParamStr( 1 ) = '?' ) Then Help;
  { set different colors and screen address if monochrome }
  MonoMode := GetVMode = $07;
  If MonoMode Then
    Begin
      VAddr := $B000;  FrClr := LightBlue;  BkClr := Black;
    End;
  TextColor( FrClr );  TextBackGround( BkClr );
  { If MonoMode Then SetVMode( $07 ) Else SetVMode( $03 ); }
  GetMemory;  CheckSnow := False;  CheckBreak := False;  CrsrOff;
  TextColor( Black );  TextBackGround( LightGray );
  CrsrXY( 1, 24 );  ClrEol;  CrsrXY( 1, 25 );  ClrEol;
  Message( 'Initializing . . .' );
  DfltFldSet :=
    FlagMk + DateMk + SeqMk + NameMk + AmtMk + PrcDateMk + MemoMk;
  DfltFldOrder := DfltFldSet;
  MoveSet :=
    UpKy + DnKy + LftKy + RgtKy + PgUpKy + PgDnKy +
    C_HmKy + C_EndKy + HmKy + EndKy +
    ^S + ^D + ^A + ^F + ^E + ^X + ^R + ^C + ^W + ^Z;
  InFldOrder := DfltFldSet;  OutFldOrder := DfltFldSet;
  { load configuration }
  ReadConf;  ShowError;
  { check for command line file name }
  If ParamCount > 0 Then LdgrFileName := ParamStr( 1 );
  Upper( LdgrFileName );
  If LdgrFileName <> '' Then
    Begin
      FileName := LdgrFileName;  ReadFile( False );  ShowError;
    End
  Else
    Begin
      CreateTrx;  WriteTrx( TrxNum );
      SetReCalc;  UpdateScreen;
    End;
  Repeat
    SupressScreenUpdate := False;
    Message( 'Main Menu.    (F1)=help  (ESC)=quit' );
    Prompt( 'Select:  ' +
            '(B)lock  (C)onfig  (D)isk  (E)dit  (F)ind  (R)eports' );
    MenuCh := 'E';  Select( MenuCh, 'BCDEFR' );
    Case MenuCh of
      'B' : SetRange;
      'C' : ConfMenu;
      'D' : DiskMenu;
      'E' : EditMenu;
      'F' : FindTrx;
      'R' : ReptMenu;
    EscKy : QuitMenu;
    End;
    ShowError;
  Until False;
End.

