(* -------------------------------------------------------------- *)
(*  Report features.                                              *)
(* -------------------------------------------------------------- *)

Procedure SortLdgr;
  { sort current ledger (Schell-Metzner) }
  Var
    Jump, Gap, Index, Passes, PassCount, Temp : Integer;
    First, Next : String;
    HdrRec1, HdrRec2 : HdrRecType;
    DisbArr1, DisbArr2 : DisbArrType;
    MemoRec1, MemoRec2 : MemoType;
    SeqStr1, SeqStr2 : SeqType;
    SortCh : Char;
  Function CompTrx : Char;
    Var  { true if trx1 >= trx2 }
      FldLp, DisbLp : Integer;
      DisbRec1, DisbRec2 : DisbRecType;
      CompTemp : CHar;
    Begin  { efficiently compare two trx records }
      CompTrx := '=';  CompTemp := '=';  FldLp := 1;
      While ( CompTemp = '=' ) and ( FldLp <= Length( OutFldOrder ) ) do
        Begin
          Case OutFldOrder[ FldLp ] of
            FlagMk :
              If HdrRec1.Flag > HdrRec2.Flag Then
                CompTemp := '>'
              Else
                If HdrRec1.Flag < HdrRec2.Flag Then
                  CompTemp := '<';
            DateMk :
              If HdrRec1.Date > HdrRec2.Date Then
                CompTemp := '>'
              Else
                If HdrRec1.Date < HdrRec2.Date Then
                  CompTemp := '<';
            SeqMk :
              Begin
                SeqStr1 := PadStr( HdrRec1.Seq, SeqSize, '0' );
                SeqStr2 := PadStr( HdrRec2.Seq, SeqSize, '0' );
                If SeqStr1 > SeqStr2 Then
                  CompTemp := '>'
                Else
                  If SeqStr1 < SeqStr2 Then
                    CompTemp := '<';
              End;
            NameMk :
              If HdrRec1.Name > HdrRec2.Name Then
                CompTemp := '>'
              Else
                If HdrRec1.Name < HdrRec2.Name Then
                  CompTemp := '<';
            AmtMk :
              Begin
                DisbLp := 1;
                While ( CompTemp = '=' ) and ( DisbLp <= MaxDisb ) do
                  Begin
                    DisbRec1 := DisbArr1[ DisbLp ];
                    DisbRec2 := DisbArr2[ DisbLp ];
                    If DisbRec1.Acct > DisbRec2.Acct Then
                      CompTemp := '>'
                    Else
                      If DisbRec1.Acct < DisbRec2.Acct Then
                        CompTemp := '<'
                      Else  { names are equal, check amounts }
                        If DisbRec1.Amt > DisbRec2.Amt Then 
                          CompTemp := '>'
                        Else
                          If DisbRec1.Amt < DisbRec2.Amt Then
                            CompTemp := '<';
                    Inc( DisbLp );
                  End;
              End;
            PrcDateMk :
              If HdrRec1.PrcDate > HdrRec2.PrcDate Then
                CompTemp := '>'
              Else
                If HdrRec1.PrcDate < HdrRec2.PrcDate Then
                  CompTemp := '<';
            MemoMk :
              If MemoRec1 > MemoRec2 Then
                CompTemp := '>'
              Else
                If MemoRec1 < MemoRec2 Then
                  CompTemp := '<';
          End;  { case }
          Inc( FldLp );
        End;  { field loop }
      CompTrx := CompTemp;
    End;  { CompTrx function }
  Begin
    Prompt( 'Sort order:  (A)scending  (D)escending' );
    SortCh := SortDir;  Select( SortCh, 'AD' );
    If ( ( SortCh = 'A' ) or ( SortCh = 'D' ) ) and
       ( TrxCount > 1 ) Then
      Begin
        UnDoTrxNum := -1;
        If SortCh <> SortDir Then
          Begin
            SortDir := SortCh;  WriteConf;
          End;
        Message( 'Sorting ledger . . .' );
        TrxNum := 1;  ShowStat;
        Gap := TrxCount div 2;
        Temp := Gap;  Passes := 0;  PassCount := 0;
        While Temp > 0 do  { figure required iterations }
          Begin
            Inc( Passes );  Temp := Temp div 2;
          End;
        While ( Gap > 0 ) and ( ErrMsg = '' ) do
          Begin
            Index := Gap + 1;
            While ( Index <= TrxCount ) and ( ErrMsg = '' ) do
              Begin
                Jump := Index - Gap;
                While ( Jump > 0 ) and ( ErrMsg = '' ) do
                  Begin
                    If KeyPressed Then ErrMsg := 'Sort interrupted';
                    ReadTrx( Jump );
                    HdrRec1 := HdrRec;  DisbArr1 := DisbArr;
                    MemoRec1 := MemoRec;
                    ReadTrx( Jump + Gap );
                    HdrRec2 := HdrRec;  DisbArr2 := DisbArr;
                    MemoRec2 := MemoRec;
                    If ( ( SortDir = 'A' ) and ( CompTrx = '>' ) ) or
                       ( ( SortDir <> 'A' ) and ( CompTrx = '<' ) ) Then
                      Begin
                        HdrRec := HdrRec1;  DisbArr := DisbArr1;
                        MemoRec := MemoRec1;
                        WriteTrx( Jump + Gap );
                        HdrRec := HdrRec2;  DisbArr := DisbArr2;
                        MemoRec := MemoRec2;
                        WriteTrx( Jump );
                        Jump := Jump - Gap;
                      End
                    Else
                      Jump := 0;
                  End;
                Inc( Index );
              End;
            Gap := Gap div 2;
            Inc( PassCount );
            TrxNum := Trunc( ( TrxCount + 0.0 ) * ( PassCount / Passes ) );
            ShowStat;
          End;
        TrxNum := 1;  PgStart := 1;  SetPgEnd;
        SetReCalc;  UpdateScreen;  AnyChange := True;
      End;
  End;

{------------------------------------------------------------------}

Procedure ReptAnalysisAcctPrep;
  { Optionally initialize account array to all accounts. }
  Var
    RptLp, DisbLp, SrchLp : Integer;
    RptAcct : DisbAcctType;
    RptAmt : Real;
  Begin
    If ReptAllIncl <> 'Y' Then Exit;
    If ErrMsg <> '' Then Exit;
    RptLp := 1;
    Message( 'Initializing accounts . . .' );
    While ( ErrMsg = '' ) and ( RptLp <= TrxCount ) do
      Begin
        ReadTrx( RptLp );  DisbLp := 1;
        { loop through disburements in a transaction }
        While ( ErrMsg = '' ) and
              ( DisbLp <= MaxDisb ) do
          Begin
            RptAcct := DisbArr[ DisbLp ].Acct;
            RptAmt := DisbArr[ DisbLp ].Amt;
            { check for valid data }
            If ( RptAcct <> '' ) and
               Amount( RptAmt ) Then
              Begin
                { find matching or first free account slot }
                SrchLp := 1;
                While ( SrchLp <= RptAccts ) and
                      ( RptAcct <> RptArr[ SrchLp ].Acct ) do
                  Inc( SrchLp );
                If SrchLp > ReptMaxAcct Then
                  ErrMsg := 'Too many accounts'
                Else
                  { create new slot }
                  If SrchLp > RptAccts Then
                    Begin
                      RptAccts := SrchLp;
                      RptArr[ SrchLp ].Acct := RptAcct;
                    End;
              End;
            Inc( DisbLp );
          End;
        Inc( RptLp );
      End;
  End;

{------------------------------------------------------------------}

Procedure ReptAnalysisLoad;
  { process totals by disburesment account report }
  { sum amounts into array by account name }
  Var
    RptLp, DisbLp, SrchLp : Integer;
    RptAcct : DisbAcctType;
    RptAmt : Real;
    SomeData : Boolean;
  Begin
    If ErrMsg <> '' Then Exit;
    SomeData := False;  RptLp := RangeStart;
    Message( 'Collecting amounts . . .' );
    While ( ErrMsg = '' ) and
          ( RptLp <= RangeEnd ) and ( RptLp <= TrxCount ) do
      Begin
        ReadTrx( RptLp );  DisbLp := 1;
        { loop through disburements in a transaction }
        While ( ErrMsg = '' ) and
              ( Copy( HdrRec.Name, 1, 1 ) <> GroupMark ) and
              ( DisbLp <= MaxDisb ) do
          Begin
            RptAcct := DisbArr[ DisbLp ].Acct;
            RptAmt := DisbArr[ DisbLp ].Amt;
            { check for valid data }
            If ( RptAcct <> '' ) and
               Amount( RptAmt ) Then
              Begin
                { find matching or first free account slot }
                SrchLp := 1;
                While ( SrchLp <= RptAccts ) and
                      ( RptAcct <> RptArr[ SrchLp ].Acct ) do
                  Inc( SrchLp );
                If SrchLp > ReptMaxAcct Then
                  ErrMsg := 'Too many accounts'
                Else
                  Begin
                    SomeData := True;
                    { create new slot }
                    If SrchLp > RptAccts Then
                      Begin
                        RptAccts := SrchLp;
                        RptArr[ SrchLp ].Acct := RptAcct;
                      End;
                    { distribute appropriately to debits or credits }
                    If RptAmt > 0.0 Then
                      RptArr[ SrchLp ].Credit :=
                        RptArr[ SrchLp ].Credit + RptAmt
                    Else
                      RptArr[ SrchLp ].Debit :=
                        RptArr[ SrchLp ].Debit + RptAmt;
                  End;
              End;
            Inc( DisbLp );
          End;
        Inc( RptLp );
      End;
    If ( ErrMsg = '' ) and ( not SomeData ) Then
      ErrMsg := 'No data in range';
  End;

{------------------------------------------------------------------}

Procedure ReptAnalysisSumGroups;
  Var
    TrxLp, DisbLp, SrchLp, GroupIdx : Integer;
  Begin
    Message( 'Summing groups . . .' );
    RatioDebit := 0.0;  RatioCredit := 0.0;
    TrxLp := 1;
    While ( ErrMsg = '' ) and ( TrxLp <= TrxCount ) do
      Begin  { search for groups }
        ReadTrx( TrxLp );
        If ( ErrMsg = '' ) and
           ( Copy( HdrRec.Name, 1, 1 ) = GroupMark ) Then
          Begin  { group has been found }
            { find group or create }
            GroupIdx := 0;
            For SrchLp := 1 to RptAccts do
              If RptArr[ SrchLp ].IsGroup and
                 ( RptArr[ SrchLp ].Acct =
                   Copy( HdrRec.Name, 2, 255 ) ) Then
                GroupIdx := SrchLp;
            If GroupIdx = 0 Then
              If RptAccts < ReptMaxAcct Then
                Begin
                  Inc( RptAccts );
                  RptArr[ RptAccts ].Acct :=
                    Copy( HdrRec.Name, 2, 255 );
                  RptArr[ RptAccts ].IsGroup := True;
                  GroupIdx := RptAccts;
                End
              Else
                ErrMsg := 'Too many groups';
            If ErrMsg = '' Then
              For DisbLp := 1 to MaxDisb do
                Begin  { disb loop for group }
                  For SrchLp := 1 to RptAccts do
                    If SrchLp <> GroupIdx Then
                      If DisbArr[ DisbLp ].Acct =
                         RptArr[ SrchLp ].Acct Then
                        Begin  { account in group }
                          RptArr[ GroupIdx ].Credit :=
                            RptArr[ GroupIdx ].Credit +
                            ( RptArr[ SrchLp ].Credit *
                              DisbArr[ DisbLp ].Amt );
                          RptArr[ GroupIdx ].Debit :=
                            RptArr[ GroupIdx ].Debit +
                            ( RptArr[ SrchLp ].Debit *
                              DisbArr[ DisbLp ].Amt );
                          RptArr[ SrchLp ].InGroup := True;
                        End;  { account in group }
                End;  { disb loop }
          End;  { is a group }
        Inc( TrxLp );
      End;  { trx loop }
    SrchLp := 1;
    While ( ErrMsg = '' ) and ( SrchLp <= RptAccts ) do
      Begin
        If RptArr[ SrchLp ].Acct = ReptAcct Then
          Begin
            RatioCredit := RptArr[ SrchLp ].Credit;
            RatioDebit := RptArr[ SrchLp ].Debit;
            SrchLp := RptAccts + 1;
          End;
        Inc( SrchLp );
      End;
  End;

{------------------------------------------------------------------}

Procedure ReptAnalysisPrint;
  { print totals by disbursement account report }
  Var
    RptLp, RptRow, RptClr, AcctIdx : Integer;
    RatioBal, Bal, Credit, Debit, PctBal, PctCredit, PctDebit : Real;
    Sort1, Sort2 : Real;
    Acct1, Acct2 : DisbAcctType;
    PrinterName : String80;
    Swap : ReptRecType;
    Done, FirstPass : Boolean;
    OutFile : Text;
    GrpChar : Char;
  Procedure FillReptReals( FillNum : Integer );
    Begin
      PctBal := 0.0;  PctCredit := 0.0;  PctDebit := 0.0;
      Credit := RptArr[ FillNum ].Credit;
      Debit := RptArr[ FillNum ].Debit;
      Bal := RptArr[ FillNum ].Debit + RptArr[ FillNum ].Credit;
      If Amount( RatioBal ) Then
        Begin
          PctBal := Int( Bal * 1000.0 / RatioBal ) / 10.0;
          If PctBal > 999.9 Then PctBal := 999.9;
          If PctBal < -999.9 Then PctBal := -999.9;
        End;
      If Amount( RatioDebit ) Then
        Begin
          PctDebit := Int( Debit * 1000 / RatioDebit ) / 10.0;
          If PctDebit > 999.9 Then PctDebit := 999.9;
          If PctDebit < -999.9 Then PctDebit := -999.9;
        End;
      If Amount( RatioCredit ) Then
        Begin
          PctCredit := Int( Credit * 1000 / RatioCredit ) / 10.0;
          If PctCredit > 999.9 Then PctCredit := 999.9;
          If PctCredit < -999.9 Then PctCredit := -999.9;
        End;
    End;
  Function GetSortReal : Real;
    Begin
      Case ReptSort of
        '1' : GetSortReal := Bal;
        '2' : GetSortReal := PctBal;
        '3' : GetSortReal := Credit;
        '4' : GetSortReal := PctCredit;
        '5' : GetSortReal := Debit;
        '6' : GetSortReal := PctDebit;
        Else GetSortReal := 0.0;
      End;
    End;
  Begin
    If ErrMsg <> '' Then Exit;
    SupressScreenUpdate := True;
    If ( OutSelect = _Screen ) or ( OutSelect = _Printer ) Then
      PrinterName := OutSelect
    Else
      PrinterName := OutName;
    FileName := PrinterName;  Assign( OutFile, PrinterName );
    ReWrite( OutFile );  ChkErr( 'opening' );
    RatioBal := RatioDebit + RatioCredit;
    Message( 'Sorting accounts . . .' );
    Repeat
      { bubble sort report data }
      Done := True;
      For RptLp := 1 to RptAccts - 1 do
        Begin
          FillReptReals( RptLp );  Sort1 := GetSortReal;
          FillReptReals( RptLp + 1 );  Sort2 := GetSortReal;
          Acct1 := RptArr[ RptLp ].Acct;
          Acct2 := RptArr[ RptLp + 1 ].Acct;
          If ( ( ReptSort = 'N' ) and ( Acct1 > Acct2 ) ) or
             ( ( ReptSort <> 'N' ) and
               ( ( Sort1 < Sort2 ) or
                 ( ( Sort1 = Sort2 ) and
                   ( Acct1 > Acct2 ) ) ) ) Then
            Begin
              Swap := RptArr[ RptLp ];
              RptArr[ RptLp ] := RptArr[ RptLp + 1 ];
              RptArr[ RptLp + 1 ] := Swap;
              Done := False;
            End;
        End;
    Until Done;
    Message( 'Printing report . . .' );
    RptLp := 1;  RptRow := 0;
    AcctIdx := 0;  FirstPass := True;
    { print report array }
    While ( ErrMsg = '' ) and ( RptLp <= RptAccts ) do
      Begin
        { this routine causes groups to be shown first }
        Repeat
          Inc( AcctIdx );
          If AcctIdx > RptAccts Then
            Begin
              FirstPass := not FirstPass;
              AcctIdx := 1;
            End;
        Until 
          ( FirstPass and RptArr[ AcctIdx ].IsGroup ) or
          ( ( not FirstPass ) and ( not RptArr[ AcctIdx ].IsGroup ) );
        { get percentages }
        FillReptReals( AcctIdx );
        { handle screen output }
        If ( OutSelect = _Screen )  and ( RptRow >= ScreenLines ) Then
          Begin
            Prompt( 'SCREEN FULL:  Press any key.' );
            PressAnyKey;  RptRow := 0;
            If AnyKey = EscKy Then
              Begin
                Prompt( ^G + 'Report interrupted.' );  Delay( 500 );
                Close( OutFile );  ChkErr( 'closing' );
                Exit;
              End;
          End;
        If RptRow = 0 Then
          Begin
            RptRow := 1;
            If OutSelect = _Screen Then
              Begin
                TextColor( White );  TextBackground( Black );
                For RptClr := ScreenLines DownTo 1 do
                  Begin
                    CrsrXY( 1, RptClr );  ClrEol;
                  End;
              End;
            If ( RangeStart > 1 ) or ( RangeEnd < TrxCount ) Then
              Write( OutFile, '(BLOCK) ' );
            Write( OutFile, 'Disbursements report sorted by ' );
            Case ReptSort of
              '1' : Write( OutFile, 'balance' );
              '2' : Write( OutFile, '% balance' );
              '3' : Write( OutFile, 'credit' );
              '4' : Write( OutFile, '% credit' );
              '5' : Write( OutFile, 'debit' );
              '6' : Write( OutFile, '% debit' );
              Else Write( OutFile, 'name' );
            End;
            Writeln( OutFile, ', % compares to ' + ReptAcct );
            If ( ReptDvsr > 1.0001 ) or ( ReptDvsr < 0.9999 ) Then
              Begin
                Writeln( OutFile, 'Amounts shown are divided by ',
                                  ReptDvsr:1:2 );
                Inc( RptRow );
              End;
            Writeln( OutFile, ' ', Pad( 'ACCT', -AcctSize ) + ' ' +
              Pad( 'BAL', AmtSize ) + ' ' + Pad( '%BAL', PctSize ) + ' ' +
              Pad( '+', AmtSize ) + ' ' + Pad( '%+', PctSize ) + ' ' +
              Pad( '-', AmtSize ) + ' ' + Pad( '%-', PctSize ) );
            WriteLn( OutFile,
                     PadStr( '', 1 + AcctSize + 6 +
                                 3 * AmtSize + 3 * PctSize, '-' ) );
            ChkErr( 'writing' );
            RptRow := RptRow + 3;
          End;
        If RptArr[ AcctIdx ].IsGroup Then
          GrpChar := GroupMark
        Else
          If RptArr[ AcctIdx ].InGroup Then
            GrpChar := ' '
          Else
            GrpChar := 'x';
        If ( ReptDvsr > 1.0001 ) or ( ReptDvsr < 0.9999 ) Then
          Begin
            Bal := Bal / ReptDvsr;
            Credit := Credit / ReptDvsr;
            Debit := Debit / ReptDvsr;
          End;
        WriteLn( OutFile, GrpChar,
                          Pad( RptArr[ AcctIdx ].Acct, -AcctSize ), ' ',
                          Bal:AmtSize:2, ' ', PctBal:PctSize:1, ' ',
                          Credit:AmtSize:2, ' ',
                          PctCredit:PctSize:1, ' ',
                          Debit:AmtSize:2, ' ',
                          PctDebit:PctSize:1 );
        ChkErr( 'writing' );  Inc( RptRow );  Inc( RptLp );
      End;
    Close( OutFile );  ChkErr( 'closing' );
    If ErrMsg = '' Then
      Begin
        Prompt( 'DONE:  Press any key.' );  PressAnyKey;
      End;
    SupressScreenUpdate := False;
  End;

{------------------------------------------------------------------}

Procedure ReptAnalysis;
  { handle all steps in disbursement analysis report }
  Begin
    FillChar( RptArr, SizeOf( RptArr ), 0 );  RptAccts := 0;
    ReptAnalysisAcctPrep;
    ReptAnalysisLoad;
    ReptAnalysisSumGroups;
    ReptAnalysisPrint;
  End;

{------------------------------------------------------------------}

Procedure ReptListing;
  { generate listing of transaction with a running line balance }
  Var
    TempDisbArr : DisbArrType;
    LstRow, RptClr, TrxNumSave, LstLp, DisbLp : Integer;
    PrinterName : String80;
    LstCh : Char;
    Printable : Boolean;
    LstBal : Real;
    OutFile : Text;
    OutLn : String;
  Begin
    If ErrMsg <> '' Then Exit;
    SupressScreenUpdate := True;
    If ( OutSelect = _Screen ) or ( OutSelect = _Printer ) Then
      PrinterName := OutSelect
    Else
      PrinterName := OutName;
    FileName := PrinterName;  Assign( OutFile, PrinterName );
    ReWrite( OutFile );  ChkErr( 'opening' );
    If TrxNum > 1 Then
      Begin
        TrxNumSave := TrxNum;  TrxNum := RangeStart - 1;  SetRecalc;
        Repeat
          If KeyPressed Then LstCh := ReadKey;  ReCalc;
        Until not DoReCalc;
        TrxNum := TrxNumSave;  SetReCalc;  LstBal := LineBal;
      End
    Else
      LstBal := 0.0;
    LstLp := RangeStart;  LstRow := 0;
    While ( ErrMsg = '' ) and
          ( LstLp <= RangeEnd ) and ( LstLp <= TrxCount ) do
      Begin
        ReadTrx( LstLp );
        If Copy( HdrRec.Name, 1, 1 ) = GroupMark Then
          Printable := ListShowGroups = 'Y'
        ELse
          Begin  { check balance effecting }
            Printable := False;
            For DisbLp := 1 to MaxDisb do
              If Amount( DisbArr[ DisbLp ].Amt ) Then
                Begin
                  If ListType = 'A' Then Printable := True;
                  If DisbArr[ DisbLp ].Acct = ShowAcct Then
                    Begin
                      LstBal := LstBal + DisbArr[ DisbLp ].Amt;
                      Printable := True;
                    End;
                End;
          End;  { check balance effecting }
        If Printable Then
          Begin
            If ( OutSelect = _Screen ) and ( LstRow >= ScreenLines ) Then
              Begin
                Prompt( 'SCREEN FULL:  Press any key.' );
                PressAnyKey;  LstRow := 0;
                If AnyKey = EscKy Then
                  Begin
                    Prompt( ^G + 'Report interrupted.' );  Delay( 500 );
                    Close( OutFile );  ChkErr( 'closing' );
                    Exit;
                  End;
              End;
            If LstRow = 0 Then
              Begin
                LstRow := 1;
                If OutSelect = _Screen Then
                  Begin
                    TextColor( White );  TextBackground( Black );
                    For RptClr := ScreenLines DownTo 1 do
                      Begin
                        CrsrXY( 1, RptClr );  ClrEol;
                      End;
                  End;
                If ( RangeStart > 1 ) or ( RangeEnd < TrxCount ) Then
                  Write( OutFile, '(BLOCK) ' );
                If ListType = 'A' Then
                  Writeln( OutFile, 'Full transaction listing.  Balance ',
                                    'of ', ShowAcct, ' account shown.' )
                Else
                  Writeln( OutFile, 'Listing of transactions effecting ',
                                    ShowAcct, ' account.' );
                WriteLn( OutFile );  ChkErr( 'writing' );
                LstRow := LstRow + 2;
              End;
            OutLn := Pad( R_S( LstBal ), AmtSize ) + ' ' + FormatLine;
            If ( OutSelect = _Screen ) and ( Length( OutLn ) > 79 ) Then
              OutLn := Copy( OutLn, 1, 78 ) + #175;
            Writeln( OutFile, OutLn );
            ChkErr( 'writing' );
            Inc( LstRow );
          End;
        Inc( LstLp );
      End;
    Close( OutFile );  ChkErr( 'closing' );
    If ErrMsg = '' Then
      Begin
        Prompt( 'DONE:  Press any key.' );  PressAnyKey;
      End;
    SupressScreenUpdate := False;
  End;

{------------------------------------------------------------------}

Procedure SelectOutput;
  { select report destination }
  Var
    OutCh : Char;
  Begin
    Message( 'Specify ouput device for report.' );
    Prompt( 'Select:  (F)ile  (P)rinter  (S)creen' );
    If OutSelect = _File Then OutCh := 'F';
    If OutSelect = _Printer Then OutCh := 'P';
    If OutSelect = _Screen Then OutCh := 'S';
    Select( OutCh, 'FPS' );
    Case OutCh of
      'F' : Begin
              OutSelect := _File;
              Prompt( 'Enter file name:  ' );
              InpStr( OutName, 50, _String );
            End;
      'P' : OutSelect := _Printer;
      'S' : OutSelect := _Screen;
     End;
    If OutCh <> EscKy Then WriteConf;
  End;

{------------------------------------------------------------------}

Procedure SetFormat;
  { select type of report }
  Var
    InReptType, InReptSort, InpReptAllIncl, InListType : Char;
    AcctSave : DisbAcctType;
  Begin
    Message( 'Report format.' );
    Prompt( 'Select:  (A)nalysis  (L)isting' );
    InReptType := ReptType;  Select( InReptType, 'AL' );
    If not Escape Then
      Begin
        ReptType := InReptType;
        Case ReptType of
          'A' : Begin
                  Prompt( 'Sort by:  (N)ame (1)Bal (2)%Bal ' +
                          '(3)Crdt (4)%Crdt (5)Dbt (6)%Dbt' );
                  InReptSort := ReptSort;
                  Select( InReptSort, 'N123456' );
                  If not Escape Then
                    Begin
                      ReptSort := InReptSort;
                      Prompt( 'Enter comparison account or group:  ' );
                      AcctSave := ReptAcct;
                      InpStr( ReptAcct, AcctSize, _String );
                      If not Escape Then
                        Begin
                          FindAcct( ReptAcct, True );
                          Message( 'Report format.' );
                          If ErrMsg <> '' Then
                            ReptAcct := AcctSave
                          Else
                            Begin
                              Prompt( 'Enter factor by which ' +
                                      'to divide amounts:  ' );
                              InpReal( ReptDvsr );
                              If not Escape Then
                                Begin
                                  InpReptAllIncl := ReptAllIncl;
                                  Prompt( 'Include all accounts?  ' );
                                  Select( InpReptAllIncl, 'YN' );
                                  If not Escape Then
                                    ReptAllIncl := InpReptAllIncl;
                                End;
                            End;
                        End;
                    End;
                End;
          'L' : Begin
                  Prompt( 'List:  (A)ll  (B)alance effecting' );
                  InListType := ListType;  Select( InListType, 'AB' );
                  If not Escape Then
                    Begin
                      ListType := InListType;
                      Prompt( 'Include group definitions in list:  ' +
                              '(N)o  (Y)es' );
                      Select( ListShowGroups, 'NY' );
                    End;
                End;
         End;
        WriteConf;
      End;
  End;

{------------------------------------------------------------------}

Procedure ReptMenu;
  { report menu }
  Var
    RptCh : Char;
  Begin
    Message( 'Report menu.' );
    Prompt( 'Select:  (F)ormat  (O)utput  (P)rint  (S)ort' );
    RptCh := 'P';  Select( RptCh, 'FOPS' );
    Case RptCh of
      'F' : SetFormat;
      'O' : SelectOutput;
      'P' : If ReptType = 'A' Then
              ReptAnalysis
            Else
              ReptListing;
      'S' : SortLdgr;
     End;
    If RptCh <> EscKy Then
      Begin
        SetReCalc;  UpdateScreen;
      End;
  End;

{------------------------------------------------------------------}

