{$B-,D-,F-,I-,L-,N-,R-,S-,V-}
{$M 8000,0,32000}

Program Letters;

Uses
  Crt, Dos, Graph, Drivers, Fonts;

Var
  DevMaxX, DevMaxY, MaxColor : Word;     { maximum screen resolution }
  QuitLn : String[ 10 ];
  Ch : Char;
  Con : Text;
  DelayDur : Integer;
  ExitSave : Pointer;

Const
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      'fwvs)1sz~';
  CpyRtLp : Byte = 1;

{--------------------------------------------------------------------}
{ Cleanup at termination.                                            }
{--------------------------------------------------------------------}

Procedure TidyExit;
  Begin
    ExitProc := ExitSave;
    RestoreCRTMode;
    Halt;
  End;

{--------------------------------------------------------------------}
{ Initialize graphics, halt if failed.                               }
{--------------------------------------------------------------------}

Procedure InitGraphics;
  Var
    PathToDriver : String;  { Stores the DOS path to *.BGI & *.CHR }
    Result, GraphMode, GraphDriver : Integer;  { graphics device driver }
  Begin
    DirectVideo := False;
    GraphDriver := CGA;
    GraphMode := CGAC0;
    InitGraph( GraphDriver, GraphMode, 'C:\TP\GRAPH' );
    Result := GraphResult;
    If Result <> grOK Then
      Begin
        Writeln( 'Graphics initialization error: ',
                 GraphErrorMsg( Result ) );
        Halt( 1 );
      End;
    MaxColor := GetMaxColor;  { Get the maximum allowable drawing color }
    DevMaxX := GetMaxX;          { Get screen resolution values }
    DevMaxY := GetMaxY;
    ClearDevice;      {clearscreen}
  End; { Initialize }

{--------------------------------------------------------------------}
{ Consume keyboard buffer.                                           }
{--------------------------------------------------------------------}

Procedure EatKeys;
  Var
    EKCh : Char;
  Begin
    While KeyPressed do EKCh := ReadKey;
  End;

{--------------------------------------------------------------------}
{ Do stuff with a character.                                         }
{--------------------------------------------------------------------}

Procedure ShowChar;
  Var
    ChSize : Integer;
    ShowCh : Char;
  Begin
    ShowCh := Ch;
    OutTextXY( 1, 1, '' );
    SetColor( 1 );
    For ChSize := 1 to 8 do
      Begin
        Sound( 100 * ( ChSize * 2 ) );
        SetUserCharSize( ChSize, 1, ChSize, 1 );
        SetTextStyle( TriplexFont, HorizDir, UserCharSize );
        OutText( ShowCh );
        Delay( DelayDur );
        NoSound;
        Delay( DelayDur );
      End;
    SetColor( 2 );
    Ch := #0;
    While not ( Ch in [ 'A'..'Z', '0'..'9' ] ) do
      Begin
        ChSize := 1 + Random( 2 );
        SetUserCharSize( ChSize, 1, ChSize, 1 );
        SetTextStyle( TriplexFont, HorizDir, UserCharSize );
        OutTextXY( 1 + Random( DevMaxX ), 1 + Random( DevMaxY ), ShowCh );
        Delay( DelayDur * 2 );
        If KeyPressed Then Ch := UpCase( ReadKey );
        EatKeys;
      End;
  End;

{--------------------------------------------------------------------}

Procedure CheckQuit( Ch : Char );
  Begin
    If not ( Ch in [ 'A'..'Z' ] ) Then Exit;
    If Length( QuitLn ) = 4 Then
      QuitLn := Copy( QuitLn, 2, 3 );
    QuitLn := QuitLn + Ch;
    If QuitLn = 'QUIT' Then TidyExit;
  End;

{--------------------------------------------------------------------}

Procedure SetDelay;
  Var
    Regs : Registers;
    Sec10, EndSec10, Sec10Cnt : Byte;
    SpeedCnt : LongInt;
  Begin
    { perform a timed loop }
    Sec10Cnt := 0;  SpeedCnt := 0;
    Regs.AH := $2C;  MSDos( Regs );  { get time }
    Sec10 := Regs.DL div 10;
    { wait for next 1/10 second }
    EndSec10 := Sec10;
    Repeat
      Regs.AH := $2C;  MSDos( Regs );
      Sec10 := Regs.DL div 10;
    Until Sec10 <> EndSec10;
    EndSec10 := Sec10;
    { ooop for 4/10 seconds }
    Repeat
      Inc( SpeedCnt );
      Regs.AH := $2C;  MSDos( Regs );
      Sec10 := Regs.DL div 10;
      If Sec10 <> EndSec10 Then
        Begin
          EndSec10 := Sec10;
          Inc( Sec10Cnt );
        End;
    Until Sec10Cnt = 4;
    If SpeedCnt <= 300 Then
      DelayDur := 0   { probable 8086/8088 }
    Else
      If SpeedCnt < 480 Then
        DelayDur := 15  { fast 8088 or slow 286 }
      Else
        DelayDur := 30;  { other }
  End;

{--------------------------------------------------------------------}

Procedure CheckCGA;
  Var
    Regs : Registers;
  Begin
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := 15;
    Intr( $10, Regs );
    If Regs.AL and $0F = $07 Then
      Begin
        Writeln( 'CGA graphics capability required.' );
        Halt( 1 );
      End;
  End;

{--------------------------------------------------------------------}

Begin

  If ParamCount > 0 Then
    Begin
      For CpyRtLp := 1 to 26 do
        If CpyRtLp = 1 Then
          CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
        Else
          CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                   Ord( CpyRt[ CpyRtLp - 1 ] ) );
      Assign( Con, '' );  ReWrite( Con );
      Writeln( Con, 'LETTERS ver 3.1 rev 91/06/29 - ', CpyRt, '.' );
      Writeln( Con, 'Colorfully display the alphabet for toddlers.' );
      Writeln( Con, '  Draw a letter or number pressed, sweeping' );
      Writeln( Con, '    from small to large with accompanying sound.' );
      Writeln( Con, 'CGA compatible graphics is required.' );
      Close( Con );
      Halt( 1 );
    End;
  CheckCGA;
  Randomize;
  QuitLn := '';
  If RegisterBGIfont(@TriplexFontProc) < 0 Then
    Begin
      Writeln( 'Graphics initialization error.' );
      Halt( 1 );
    End;
  If RegisterBGIdriver( @CGADriverProc ) < 0 Then
    Begin
      Writeln( 'Graphics initialization error.' );
      Halt( 1 );
    End;
  InitGraphics;
  ExitSave := ExitProc;
  {$F+}
  ExitProc := @TidyExit;
  {$F-}
  SetBkColor( Blue );
  SetUserCharSize( 1, 1, 1, 1 );
  SetTextStyle( TriplexFont, HorizDir, UserCharSize );
  OutTextXY( 1, 1, 'Press a letter  -' );
  OutTextXY( 1, 80, 'Type  QUIT  to quit' );
  SetDelay;
  Repeat
    Ch := UpCase( ReadKey );
    CheckQuit( Ch );
  Until Ch in [ 'A'..'Z', '0'..'9' ];
  EatKeys;
  ClearDevice;
  Repeat
    ShowChar;
    ClearDevice;
    CheckQuit( Ch );
  Until True = False;

End.



