(*********************************************************************
**********************************************************************
****                                                              ****
****                       MENUMAN.PAS                            ****
****                                                              ****
****                      Menu system.                            ****
****                                                              ****
**********************************************************************
**********************************************************************)

UNIT MenuMan;


{$I LIBRARY.CFG }

INTERFACE


Uses Dos, Crt,
     Global, Sounds, Math, Strings, Video, Time, Key, Disk, Init, InputMan;


CONST
{ use when assembling special lines }
  _BarLeft = #2;
  _BarRight = #3;
  _SkipLine = #4;
{ set these to preferences }
  TurboSearchMsg = 'TURBO-Search Active';
  TurboSearchTimeout = 3000;
  MenuBorderStyle = 1;  { outside frame type }
  MenuLoColor = Yellow;  { unhilited line }
  MenuHiColor = White;  { hilited character }
  MenuReverseColor = LightGray * $10;  { hilited line }
  MenuBorderColor = Yellow;  { outside frame }
  MenuStatusColor = White;  { title, turbo search messages, status arrows }
{ menu selection modes }
  _NoSearch = 0;
  _TurboSearch = 10;
  _LetterSearch = 20;
  _LetterExecute = 30;
  _DiskMenu = 100;
{ block marking }
  MenuBlock : Integer = 0;
  MenuAllMarked : Boolean = False;
{ QuickMenu support }
  QuickMenuUsed : Boolean = True;
  QuickMenuSetup : Boolean = False;
  MaxQuickMenuLines = 25;
{ MenuKey returns ASCII of last key pressed in menu system. }
  MenuKey : Char = NoKey;
{ MenuPrevKey returns the prior MenuKey. }
  MenuPrevKey : Char = NoKey;
{ MenuKeys returns one or more unknown keys (i.e. F2, ^K^S) }
{ Menu will make an error tone if a non-null MenuKeys is passed in, }
{ allowing detection of keys used neither by the menu, nor main loop. }
  MenuKeys : String2 = '';
{ MenuWrap allows movement off the end of a meno to wrap back to the top. }
  MenuWrap : Boolean = True;
{ TurboSearchPending allows a turbo search to be interrupted by a keypress }
{ then continued after keypress is handled. }
  TurboSearchPending : Boolean = False;


TYPE
  MenuWhichLetterArrayType = Array [ 0..65534 ] of Byte;
  MenuMarksArrayType = Array [ 0..65534 ] of Char;
  MenuOrderArrayType = Array [ 0..32766 ] of Word;


VAR
  MenuForceRefresh : Boolean;
  MenuSelectMoved : Boolean;  { true for one cycle after select is moved }
  MenuSelect : LongInt;
  MenuLowItem, MenuHiItem : Integer;
  MenuPrevLowItem, MenuPrevSelect : Integer;
  MenuData : ^Byte;   { pointer to data array }
  MenuWX1, MenuWY1, MenuWidth, MenuHeight, MenuTopMargin : Integer;
  MenuDrawFrame : Boolean;
  MenuTitle : String30;
  MenuMin, MenuMax, MenuSpacing : Integer;
  MenuShowMin, MenuShowMax : Integer;
{ Activate the menu "tick marks" feature by first using MenuSetup,
    then point the MenuMarks pointer at an array of character
    by simply using MenuWhichLetter := @your_tickmark_array;
    Disable this feature with MenuMarks := NIL; }
  MenuMarks : ^MenuMarksArrayType;
{ Activate the menu "which letter" feature by first using MenuSetup,
    then point the MenuWhichLetter pointer at an array of pick letter indices
    by simply using MenuWhichLetterPointer := @your_index_array;
    Disable this feature with MenuWhichLetter := NIL;
    This array specifies for each menu line, which letter is highlighted,
    and will select that line, in letter search and execute modes. }
  MenuWhichLetter : ^MenuWhichLetterArrayType;
{ Activate the sorted menu feature by first using MenuSetup,
    then point the MenuWOrder pointer at an array of words
    by simply using MenuOrder := @your_order_array;
    Disable this feature with MenuOrder := NIL;
    This order specifies a translation between the lines on a menu,
    and where each line of data is actually located, where
    menu line n uses data located at [ MenuOrder[n] ] }
  MenuOrder : ^MenuOrderArrayType;
  MenuFirstCol : Byte;  { left margin depending on marking enabled }
{ Set MenuViewOnly after menu setup to display a menu without selector. }
  MenuViewOnly : Boolean;
{ Selection mode of menu, see menu selection mode constants above. }
  MenuMode : Integer;
{ one item cache for disk menus }
  MenuPrevSeekElement : LongInt;
  MenuPrevSeekRecord : GenlString;
{ These are set by the MenuRecAdj procedure to calculate displayed field. }
  MenuRecOfs, MenuRecPad : Integer;
{ MenuFileName is the name of the file to be used in place of a data array. }
  MenuFileName : GenlString;
{ MenuFile is the file to be used in place of a data array. }
  MenuFile : File;
{ MenuResult returns I/O results when a disk menu is used. }
  MenuResult : Integer;
{ This is where the Turbo Search string is assembled }
  TurboSearchStr : SmallString;
{ SkipUpward sets which direction to go when skipping blank lines. }
  SkipUpward : Boolean;
{ QuickMenuLn is a string array which can be used to quickly build small }
{ menus.  It is not used by the menu system, it is only for the programmer. }
  QuickMenuLn : Array [ 1..MaxQuickMenuLines ] of String50;
  QuickMenuLines : Word;  { tracks number of quick menu lines }
{ QuickMenuLtr is an array of bytes which can be used as a MenuWhichLetter }
{ index array to quickly build small menus.  Like QuickMenuLn above. }
  QuickMenuLtr : Array [ 1..MaxQuickMenuLines ] of Byte;
{ EditMenuType stores type and range configuration for EditMenu. }
  EditMenuType : Array [ 1..MaxQuickMenuLines ] of
    Record  FieldType:Integer;  Min, Max:Real;  FieldData:Pointer;  End;
  EditMenuPos : Byte;  { start of data portion of an EditMenu line }
{ Like QuickMenuLn, Choice and PrevChoice are included for programmer use. }
  PrevChoice, Choice : Integer;

{ Note that many numeric expressions may seem redundant or wasteful,
  such as "MenuWY1 + MenuTopMargin + MenuHeight - 1 + 1" where "- 1 + 1"
  is obviously nonsensical.  This format is used so that the expression
  follows the logical method for determining the desired screen coordinates
  rather than being an optimized form.  The Turbo Pascal compiler will
  optimize constant expressions like this one anyway.
  In the above case, the "- 1" changes MenuHeight from a height into an
  offset so that the last line can be calculated, then "+ 1" gives the
  coordinate of the bottom of the menu frame, which is outside of the
  menu window by one row/column on all side. }

{ setup a menu in detail }
PROCEDURE MenuSetup( Title : GenlString;  Var Data;
                     Width, Max, Mode, X, Y, Height, Spacing : Integer );

{ display a field within a record }
PROCEDURE MenuRecAdj( RecSize, RecOfs : Integer );

{ return a menu selection }
FUNCTION  Menu : Integer;

{ show line sensitive bottom line help }
PROCEDURE ShowLineMsg( WhichLine : Integer;  Msg : String40 );

{ skip to any line in menu }
PROCEDURE MenuMove( Distance : Integer );

{ return video coordinates of start of current menu line }
FUNCTION  MenuAtY : Integer;
FUNCTION  MenuAtX : Integer;

{ add a line to the quick menu }
PROCEDURE QuickMenuLine( MenuLn : String30 );

{ return a selection from a quick menu }
FUNCTION QuickMenu( MenuName : SmallString ) : Integer;


PROCEDURE EditMenuLine( MenuLn : String30;
                        InType : Integer;  Var InVar );

FUNCTION EditMenu( MenuName : SmallString ) : Integer;


IMPLEMENTATION

{-------------------------------------------------------------------------}

(* MENUSETUP configures the next menu to be presented:
    TITLE is displayed at the top of the menu.
    DATA is the array of strings to be shown.
    WIDTH is the byte size of the array element string type.
    MAX is the number of selections available. ( Data[ 1..Max ] )
      Minimum displayed choice is set by passing Data as Data[ min ].
    MODE sets selection mode, use with mode name constants:
      _NoSearch = no highlights or searches, cursor keys only, enter selects.
      _TurboSearch = no hilites, turbo searches, enter selects.
      _LetterSearch = letter hilites, letter searches, enter selects.
      _LetterExecute = same as above, but valid search or enter selects.
      In letter modes, first letter is highlighted if MenuWhichLetter = NIL,
      otherwise MenuWhichLetter used as array of highlight indices.
    X, Y is the top left corner of the menu window.
      Use zero for X or Y for automatic centering.
    HEIGHT is the height of the menu window in rows.
      Use zero for height to automatically use all available space.
    SPACING sets spacing of menu lines, 1 is single, 2 is double.
      Use zero to automatically set double or single given available space.

   In addition to this basic setup, several other options can be enabled
   by setting values in certain global variables after using MenuSetup:
     MenuMarks, which by providing a character array,
       will allow the setting of 'tick' marks next to menu selections,
     MenuWhichLetter, which by providing a byte array of indices,
       will highlight any desired character of each menu line.
     MenuViewOnly, inhibits the display of a selector bar.
     MenuForceRefresh, which causes a forced refresh of current menu window.
     MenuWrap, set false to prevent movement off top or bottom of menu.
   See the variables section for details.
   Also, several internal variables can be accessed for special needs:
     MenuSelect, returns the current line of the menu, whether or not
       a selection has been made.
     MenuWX1, MenuWX2, MenuWidth, MenuHeight, return the window settings.
     MenuSelectMoved, tells whether a movement has been made in the menu. *)

(* Also see MenuRecAdj which allows the display of a single field
   from within a record.  *)


PROCEDURE MenuSetup( Title : GenlString;  Var Data;
                     Width, Max, Mode, X, Y, Height, Spacing : Integer );
  VAR  FileNameVar : String absolute Data;
  BEGIN
    { prepare pointer to data, and flag whether from disk or memory }
    MenuResult := OK;
    If TestDigit( Mode, _DiskMenu ) = _DiskMenu
      THEN BEGIN
      { initialize disk menu data pointer }
      { In case previous menu was from disk, close it }
        {$I-} Close( MenuFile ); {$I+} JunkResult := IOResult;  { ignored }
        If not FileExists( MenuFileName )
          THEN MenuResult := _FileNotFound;
        MenuPrevSeekElement := 0;
      { flag data array as nil }
        MenuData := NIL;  MenuFileName := FileNameVar;
      END
      ELSE BEGIN
      { initialize memory array menu data pointer }
        MenuData := Ptr( Seg( Data) , Ofs( Data) );
      { flag menu file name as null }
        MenuFileName := '';
      END;
  { set basic menu configuration globals to setup arguments }
    MenuWidth := Width;
    MenuMax := Max;
    MenuMode := Mode mod 100;
    MenuWX1 := X;
    MenuWY1 := Y;
    MenuTopMargin := TopMargin;
    MenuHeight := Height;
    MenuSpacing := Spacing;
    MenuTitle := Title;
    Strip( MenuTitle );
    If MenuTitle <> ''  { force at least one space on each side of title }
      THEN MenuTitle := ' ' + Abrv( MenuTitle, MenuWidth - 3 ) + ' ';
  { flag setup confirmed, ready to draw }
    MenuDrawFrame := True;
  { initialize operational global variables }
    MenuBlock := 0;
    MenuMin := 1;  MenuShowMin := MenuMin;
    MenuSelectMoved := True;
    MenuAllMarked := False;
    MenuLowItem := MenuMin;  MenuPrevLowItem := MenuLowItem;
    MenuSelect := MenuMin;  MenuPrevSelect := MenuSelect;
    MenuKey := NoKey;  MenuPrevKey := NoKey;
    MenuKeys := '';
    TurboSearchStr := '';
    TurboSearchPending := False;
  { initialize extended features to defaults }
    MenuForceRefresh := True;
    MenuWhichLetter := NIL;
    MenuMarks := NIL;
    MenuOrder := NIL;
    MenuRecOfs := 0;  MenuRecPad := 0;
    MenuViewOnly := False;
    MenuWrap := True;
  END;

{-------------------------------------------------------------------------}

(* MENURECADJ allows the menu system to use an array of records.
    It specifies where in the record to find the displayable string.
    Set RecOfs to the number of bytes in the record before the string,
    and set RecSize to the starting byte # of the displayable string.
    Use MenuRecAdj after using MenuSetup  *)

PROCEDURE MenuRecAdj( RecSize, RecOfs : Integer );
  BEGIN
    MenuRecOfs := RecOfs - 1;
    MenuRecPad := RecSize - RecOfs - MenuWidth + 1;
  END;

{-------------------------------------------------------------------------}

{ Used internally to range adjust the menu selector bar. }

PROCEDURE CheckMenuSelect;
  BEGIN
    If MenuSelect > MenuMax
      THEN If MenuWrap
             THEN MenuSelect := MenuMin ELSE MenuSelect := MenuMax;
    If MenuSelect < MenuMin
      THEN If MenuWrap
             THEN MenuSelect := MenuMax ELSE MenuSelect := MenuMin;
  END;

{-------------------------------------------------------------------------}

{ Used internally to return the string for any requested menu line. }

FUNCTION MenuItem( Element : LongInt ) : GenlString;
  VAR  RecAdj, Point, PointSeg, PointOfs : LongInt;  S : GenlString;
       ItemResult : Integer;
  BEGIN
    CheckStack;
    If MenuOrder <> NIL Then Element := MenuOrder^[ Element ];
    RecAdj := Element * MenuRecOfs + ( Element - 1 ) * MenuRecPad;
    Point := ( Element - 1 ) * MenuWidth + RecAdj;
    If MenuData <> NIL
      THEN BEGIN  { memory }
        Point := Ofs( MenuData^ ) + Point;
      { Point is converted to a segment and offset to avoid any risk }
      { of the point value exceding the word range allowed as an offset }
      { when accessing the Mem array. }
        PointSeg := ( Point and $FFFFFFF0 ) shr 4;
        PointOfs := Point and $0000000F;
        ChkPtr( S, SizeOf( S ), 'menu item' );
        Move( Mem[ Seg( MenuData^ ) + PointSeg : PointOfs ], S, MenuWidth );
      END  { memory }
      ELSE BEGIN
        { check if same element as last request }
        If Element = MenuPrevSeekElement
          THEN S := MenuPrevSeekRecord
          ELSE BEGIN
            { otherwise get from disk }
            {$I-} Seek( MenuFile, Point ); {$I+} ItemResult := IOResult;
            If ItemResult = 0
              THEN BEGIN
                {$I-} BlockRead( MenuFile, S, MenuWidth ); {$I+}
                ItemResult := IOResult;
               END;
            If ItemResult <> 0
              THEN BEGIN  S := '';  Element := 0;  END;
            MenuPrevSeekElement := Element;
            MenuPrevSeekRecord := S;
            If MenuResult = 0 THEN MenuResult := ItemResult;
          END;
      END;
    MenuItem := S;
  END;

{-------------------------------------------------------------------------}

{ Through this procedure, a calling program can place the menu selector }
{ bar at any desired distance from its current position. }
{ For example:  MenuWrap := False;  MenuMove( -MaxInt );  forces home. }
{   MenuMove( 1 ) is one line down, MenuMove( -1 ) is one line up. }

PROCEDURE MenuMove( Distance : Integer );
  VAR  Done, IsBlank : Boolean;  PrevSelect : Integer;
       ItemStr : GenlString;
  BEGIN
  { This section skips over blank lines }
    PrevSelect := MenuSelect;
    MenuSelect := MenuSelect + Distance;
    CheckMenuSelect;
    Repeat
      ItemStr := MenuItem( MenuSelect );
      IsBlank := ( ItemStr <> '' ) and ( ItemStr[ 1 ] = _SkipLine );
      If IsBlank
        THEN If SkipUpward
               THEN Dec( MenuSelect ) ELSE Inc( MenuSelect );
      CheckMenuSelect;
      Done := ( not IsBlank ) or  { skipping finished }
              ( MenuSelect = PrevSelect ) or  { or all blank }
              ( ( not MenuWrap ) and  { or limit hit }
                ( ( SkipUpward and ( MenuSelect = MenuMin ) ) or
                  ( ( not SkipUpward ) and ( MenuSelect = MenuMax ) )
                ) );
      If IsBlank and ( not BlankLine( ItemStr ) )
        THEN RemoteLn( Copy( ItemStr, 2, Length( ItemStr ) - 1 ) );
    Until Done;
  END;

{-------------------------------------------------------------------------}

{ return video coordinates of start of current menu line }

FUNCTION  MenuAtY : Integer;
  BEGIN
    MenuAtY := ( MenuSelect - MenuLowItem ) * MenuSpacing
               + MenuWY1 + MenuTopMargin;
  END;

{-------------------------------------------------------------------------}

{ return video coordinates of start of current menu line }

FUNCTION  MenuAtX : Integer;
  BEGIN
    MenuAtX := MenuWX1 + MenuFirstCol - 1;
  END;

{-------------------------------------------------------------------------}

{ This is part of the QuickMenu support.  If this is the first usage since }
{ a quick menu was displayed, the quick system is reset.  The number of }
{ quick menu lines is incremented, and the passed line is placed at the }
{ end of the quick menu display string array being built. }

PROCEDURE QuickMenuLine( MenuLn : String30 );
  BEGIN
    If QuickMenuUsed
      THEN BEGIN
        QuickMenuLines := 0;
        QuickMenuUsed := False;  QuickMenuSetup := False;
      END;
    If QuickMenuLines >= MaxQuickMenuLines
      THEN QuickMenuLines := MaxQuickMenuLines - 1;
    Inc( QuickMenuLines );
    QuickMenuLn[ QuickMenuLines ] := MenuLn;
  END;

{-------------------------------------------------------------------------}

{ This procedure, along with the EditMenu procedure, using the }
{ QuickMenu system resources, allows the rapid construction of a record }
{ editor.  Simply put _SkipLine at the beginning of non-editable lines. }
{ The Line passed is a field name which will be displayed.  Do not }
{ include a colon.  Second argument is the fields variable type }
{ described using the input manager's type identifying system. }
{ The third argument is the field itself, which must be global enough }
{ that its address and value will not change for the duration of Editing. }
{ If the allowed range of a numeric field is non standard, precede the }
{ MenuLine procedure with a Range statement (see InputMan). }
{ This feature uses one variable, EditMenuType. }

PROCEDURE EditMenuLine( MenuLn : String30;
                        InType : Integer;  Var InVar );
  BEGIN
    QuickMenuLine( MenuLn );
    EditMenuType[ QuickMenuLines ].Min := RangeMin;
    EditMenuType[ QuickMenuLines ].Max := RangeMax;
    RangeMin := 0.0;  RangeMax := 0.0;
    EditMenuType[ QuickMenuLines ].FieldType := InType;
    EditMenuType[ QuickMenuLines ].FieldData :=
      Ptr( Seg( InVar ), Ofs( InVar ) );
  END;

{-------------------------------------------------------------------------}

{ The quick menu mode defaults to _LetterExecute. }
{ Modifications to the quick menu setup defaults should be made in the }
{ application program's menu selection loop, just after the first call of }
{ QuickMenu.  On the first iteration, quick menu does not return a choice, }
{ it uses this first pass to do the default setup.  Therefore, an }
{ applications program may change the default by including a block such as: }
{   "If QuickMenuSetup and ( not QuickMenuUsed ) Then" (change the defaults) }
{ immediately after the call, such as "Choice := QuickMenu( '' );" }

FUNCTION QuickMenu( MenuName : SmallString ) : Integer;
  VAR  LenLoop, QuickWid : Byte;
  BEGIN
    If not QuickMenuSetup
      THEN BEGIN
        QuickWid := 0;
        For LenLoop := 1 to QuickMenuLines do
          If PrintLen( QuickMenuLn[ LenLoop ] ) > QuickWid
            THEN QuickWid := PrintLen( QuickMenuLn[ LenLoop ] );
        MenuSetup( MenuName { title },
                   QuickMenuLn { data },
                   QuickWid + 1 { width },
                   QuickMenuLines { max },
                   _LetterExecute { mode },
                   0 { x }, 0 { y },
                   0 { height }, 0 { spacing } );
        MenuRecAdj( SizeOf( QuickMenuLn[ 1 ] ), 1 );
        QuickMenu := 0;  MenuKey := NoKey;
        QuickMenuSetup := True;
        Exit;
      END;
    QuickMenuUsed := True;
    QuickMenu := Menu;
  END;

{-------------------------------------------------------------------------}

PROCEDURE RefreshEditMenu;
  VAR  ShowLoop : Byte;
  BEGIN
  { assemble field name and contents }
    For ShowLoop := 1 to QuickMenuLines do
      If Copy( QuickMenuLn[ ShowLoop ], 1, 1 ) = _SkipLine
        THEN QuickMenuLn[ ShowLoop ] :=  { read only }
               Pad( QuickMenuLn[ ShowLoop ], -1 - EditMenuPos ) + ' | ' +
               PadShow( EditMenuType[ ShowLoop ].FieldData^,
                        EditMenuType[ ShowLoop ].FieldType )
        ELSE QuickMenuLn[ ShowLoop ] :=  { editable }
               Pad( QuickMenuLn[ ShowLoop ], 0 - EditMenuPos ) + ' : ' +
               PadShow( EditMenuType[ ShowLoop ].FieldData^,
                        EditMenuType[ ShowLoop ].FieldType );
    MenuForceRefresh := True;
  END;

{-------------------------------------------------------------------------}

PROCEDURE PutMenuName;
  VAR CJ : Byte;
  BEGIN
    If MenuTitle = '' THEN Exit;
    CJ := ( ( MenuWidth - 1 ) - Length( MenuTitle ) ) div 2 + 1;
    WindowTitle( MenuWX1 + CJ + MenuFirstCol - 2,
                 MenuWY1 + MenuTopMargin - 1,
                 MenuStatusColor, MenuTitle );
  END;

{-------------------------------------------------------------------------}

FUNCTION EditMenu( MenuName : SmallString ) : Integer;
  VAR  LenLoop : Byte;
  BEGIN
    If not QuickMenuSetup
      THEN BEGIN  { first call }
        EditMenuPos := 0;
        For LenLoop := 1 to QuickMenuLines do
          If PrintLen( QuickMenuLn[ LenLoop ] ) > EditMenuPos
            THEN EditMenuPos := PrintLen( QuickMenuLn[ LenLoop ] );
        RefreshEditMenu;
        EditMenu := QuickMenu( MenuName );
        MenuMode := _NoSearch;
        BottomLn( Btm_EditMenu );
      END  { first call }
      ELSE BEGIN  { subsequent calls }
        EditMenu := QuickMenu( MenuName );
        If ( MenuKey in PrintableCharSet ) or ( MenuKey = EnterKey ) and
           ( MenuSelect >= 1 ) and ( MenuSelect <= QuickMenuLines )
          THEN BEGIN  { accept data }
            If MenuKey = EnterKey
              THEN PreKey := NoKey ELSE PreKey := MenuKey;
            Window( 1, 1, 80, 25 );
            Range( EditMenuType[ MenuSelect ].Min,
                   EditMenuType[ MenuSelect ].Max );
            GotoXY( MenuAtX + EditMenuPos, MenuAtY );
            Enter( EditMenuType[ MenuSelect ].FieldData^,
                   EditMenuType[ MenuSelect ].FieldType );
            RefreshEditMenu;
          END;  { accept data }
      END;  { subsequent calls }
  END;

{-------------------------------------------------------------------------}

(*  MENU returns the positive subscript starting at one
      of the element selected from a menu,
      or returns zero if an invalid key is pressed.
      MenuKey always contains the last key pressed.  *)

FUNCTION Menu : Integer;

  VAR
    ValidKey : Boolean;
    StartingSelect, CJ : Integer;
    XSave, YSave, WX1Save, WY1Save, WX2Save, WY2Save : Word;
    AttrSave, CursorSave, BlockLoop, RecSz : Word;
    FileSz : LongInt;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  PROCEDURE MenuBottomMsg;
  { This displays various bottom of menu messages for Turbo search. }
    VAR  TCh : Char;  CJ : Integer;  Pending : String6;
    BEGIN
      { replace any obliterated bottom border }
      TCh := BorderChars[ MenuBorderStyle ].B;
      WindowTitle( MenuWX1, MenuWY1 + MenuTopMargin + MenuHeight - 1 + 1,
                   MenuBorderColor,
                   NonTextChar +
                   Rep( ' ', Length( TurboSearchStr ) ) +
                   Rep( TCh, MenuWidth - 1 - Length( TurboSearchStr ) ) );
      If MenuMode <> _TurboSearch THEN Exit;
      { if no search, place the "Turbo-Search active" message }
      CJ := ( ( MenuWidth - 1 ) - Length( TurboSearchMsg ) ) div 2 + 1;
      If TurboSearchStr = ''
        THEN WindowTitle( MenuWX1 + CJ + MenuFirstCol - 2,
                          MenuWY1 + MenuTopMargin + MenuHeight - 1 + 1,
                          MenuStatusColor,
                          NonTextChar + TurboSearchMsg )
        ELSE BEGIN
          { otherwise, display the current search string }
          TCh := TurboSearchStr[ Length( TurboSearchStr ) ];
          { and ellipsis dots if the search is still in progress }
          If TurboSearchPending
            THEN Pending := '...' ELSE Pending := '';
          WindowTitle( MenuWX1 + 1,
                       MenuWY1 + MenuTopMargin + MenuHeight - 1 + 1,
                       MenuStatusColor,
                       NonTextChar + TurboSearchStr + Pending );
        END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  PROCEDURE DoTurboSearch( SearchChar : Char );
    VAR  FindLoop, EndSearch : Integer;
         Found : Boolean;  ItemStr, SearchStr : GenlString;
         Word1, Word2, Word3, Str2 : SmallString;
    BEGIN
      CheckStack;
    { This allows a turbo search to be interrupted by a keypress }
    { Check if a pending search has been terminated }
      If ( SearchChar = TabKey ) and ( TurboSearchStr = '' )
        THEN BEGIN  TurboSearchPending := False;  Exit;  END;
    { double space, or first character space are not allowed }
      If ( ( ( TurboSearchStr <> '' ) and
             ( TurboSearchStr[ Length( TurboSearchStr ) ] = ' ' ) ) or
           ( TurboSearchStr = '' ) ) and
         ( SearchChar = ' ' ) THEN Exit;
      TurboSearchPending := True;
      Found := False;
      If SearchChar <> TabKey
        THEN BEGIN  { normally, start at current line }
          TurboSearchStr := TurboSearchStr + SearchChar;
          Remote( SearchChar );  RemoteForceNL := True;
          MenuBottomMsg;  FindLoop := MenuSelect - 1;
        END
        ELSE FindLoop := MenuSelect;  { if tab, start one line advanced }
      If not MenuWrap
        THEN EndSearch := MenuMax
        ELSE If FindLoop >= MenuMin        { search ends }
               THEN EndSearch := FindLoop  { one line before where started }
               ELSE EndSearch := MenuMax;  { or last line if started at top }
      { parse search string in advance for speed }
      SearchStr := TurboSearchStr;
      Word1 := ParseStr( SearchStr );
      Word2 := ParseStr( SearchStr );
      Word3 := ParseStr( SearchStr );
      REPEAT   { look for matching item }
        Inc( FindLoop );
        If FindLoop > MenuMax THEN FindLoop := MenuMin;
        ItemStr := Upper( MenuItem( FindLoop ) );  Strip( ItemStr );
        Found := ( ItemStr <> '' ) and ( ItemStr[ 1 ] <> _SkipLine ) and
                 ( Word1 = Copy( ParseStr(ItemStr), 1, Length(Word1) ) ) and
                 ( Word2 = Copy( ParseStr(ItemStr), 1, Length(Word2) ) ) and
                 ( Word3 = Copy( ParseStr(ItemStr), 1, Length(Word3) ) );
      UNTIL Found or ( FindLoop = EndSearch ) or KeyPressed;
      { check if search was finished }
      TurboSearchPending :=
         ( not Found ) and ( FindLoop <> EndSearch ) and KeyPressed;
      If Found THEN MenuSelect := FindLoop;
      If not TurboSearchPending
        THEN BEGIN
        { If a turbo search is unsuccesful, remove the last character }
        { typed, from the turbo search string }
          If Found
            THEN DitSound
            ELSE BEGIN
              ValidKey := False;
              Delete( TurboSearchStr, Length( TurboSearchStr ), 1 );
            END;
          MenuBottomMsg;
        END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  FUNCTION DoLetterSearch( SearchChar : Char ) : Boolean;
    VAR  Index, FindLoop, EndSearch : Integer;
         Found : Boolean;  S : GenlString;
    BEGIN
      Found := False;  SearchChar := SearchChar;
      If MenuMode = _LetterExecute
        THEN FindLoop := MenuSelect - 1  { start at current line }
        ELSE FindLoop := MenuSelect;     { start at next line }
      If MenuWrap
        THEN EndSearch := FindLoop
        ELSE EndSearch := MenuMax;
      If EndSearch < MenuMin Then EndSearch := MenuMax;
      REPEAT   { look for matching item }
        Inc( FindLoop );
        If FindLoop > MenuMax THEN FindLoop := MenuMin;
        S := MenuItem( FindLoop );
        If ( S <> '' ) and ( S[ 1 ] <> _SkipLine )
          THEN BEGIN
            { choose which letter to compare given mode }
            If MenuWhichLetter = NIL
              THEN Index := 1
              ELSE BEGIN
                ChkPtr( MenuWhichLetter^[ FindLoop ], 1,
                        'menu search index' );
                If ( MenuWhichLetter^[ FindLoop ] < 1 ) or
                   ( MenuWhichLetter^[ FindLoop ] > Length( S ) )
                  THEN Index := 1
                  ELSE Index := MenuWhichLetter^[ FindLoop ];
              END;
            Found := UpCase( S[ Index ] ) = SearchChar;
          END;
      UNTIL Found or ( FindLoop = EndSearch );
      If Found THEN MenuSelect := FindLoop;
      DoLetterSearch := Found;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

{ This procedure, used by MenuRefresh, puts a menu line on the screen. }
{ Handling all display features: skipped, bar, ticks, highlighted, ... }

  PROCEDURE ShowItem( Item : Integer );
    VAR  L : Integer;  S : GenlString;  RemoteTick : String2;
         ItemY, BarStart, BarEnd, MenuAttr, MenuLtrIdx : Byte;
         SkipThisLine : Boolean;
    BEGIN
      If ( Item < MenuLowItem ) or ( Item > MenuHiItem ) then Exit;
      S := MenuItem( Item );
      If S <> ''
        THEN BEGIN
          { check if only part of line is highlighted }
          BarStart := Pos( _BarRight, S );
          If BarStart = 0
            THEN BarStart := 1
            ELSE Delete( S, BarStart, 1 );
          BarEnd := Pos( _BarLeft, S );
          If BarEnd = 0
            THEN BarEnd := MenuWidth - 1
            ELSE Delete( S, BarEnd, 1 );
        END
        ELSE BEGIN
          BarStart := 1;  BarEnd := MenuWidth - 1;
        END;
      { tailor the data for line width }
      SkipThisLine := Copy( S, 1, 1 ) = _SkipLine;
      If SkipThisLine THEN Delete( S, 1, 1 );
      S := Pad( S, 1 - MenuWidth );
      { determine screen position }
      ItemY := ( Item - MenuLowItem ) * MenuSpacing
               + MenuWY1 + MenuTopMargin;
      { check for tick marks }
      If MenuMarks = NIL
        THEN RemoteTick := ''
        ELSE BEGIN
          ChkPtr( MenuMarks^[ Item ], 1, 'menu show mark' );
          RemoteTick := MenuMarks^[ Item ] + ' ';
          If not Sys.RemoteOnly
            THEN VWrite( MenuWX1 + 1, ItemY, MenuLoColor, MenuMarks^[Item] );
        END;
      { find highlighted letter }
      If ( MenuMode <> _LetterSearch ) and ( MenuMode <> _LetterExecute ) or
         SkipThisLine
        THEN MenuLtrIdx := 0  { no hilites }
        ELSE If MenuWhichLetter = NIL
               THEN MenuLtrIdx := 1  { hilite first }
               ELSE BEGIN
                 CHkPtr( MenuWhichLetter^[ Item ], 1, 'menu letter index' );
                 MenuLtrIdx := MenuWhichLetter^[ Item ];  { hilite indexed }
               END;
      { output to remote if selected line }
      If Item = MenuSelect
        THEN RemoteLn( RemoteTick +
                       Copy( S, BarStart, BarEnd - BarStart + 1 ) );
      If Sys.RemoteOnly THEN Exit;
      { output line, one character at a time }
      For L := 1 to Length( S ) do
        BEGIN
          If L = MenuLtrIdx
            THEN MenuAttr := MenuHiColor  { hilited letter }
            ELSE If ( Item = MenuSelect ) and
                    ( L >= BarStart ) and ( L <= BarEnd )
                   THEN MenuAttr := MenuReverseColor  { selected line }
                   ELSE MenuAttr := MenuLoColor;  { basic line }
          VWrite( MenuWX1 + MenuFirstCol - 1 + L - 1,
                  ItemY, MenuAttr, NonTextChar + S[ L ] )
        END;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  PROCEDURE ShowStatus( PrevKey : Char );
  { This procedure does the arrow or character in the top right corner }
    VAR  Arrow : Char;  Color : Byte;
    BEGIN
      If ( MenuSelect > MenuShowMin ) and
         ( MenuSelect < MenuShowMax )
        THEN Arrow := UpDnArChar
        ELSE If MenuSelect = MenuShowMin
               THEN Arrow := DnArChar ELSE Arrow := UpArChar;
      If MenuHeight = 1
        THEN BEGIN
          Arrow := BorderChars[ MenuBorderStyle ].T;
          Color := MenuBorderColor;
        END ELSE Color := MenuStatusColor;
      If PrevKey = ^Q THEN Arrow := 'Q';
      If PrevKey = ^K THEN Arrow := 'K';
      WindowTitle( MenuWX1 + MenuFirstCol - 1 + MenuWidth - 1,
                   MenuWY1 + MenuTopMargin - 1, Color,
                   NonTextChar + Arrow );
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  PROCEDURE ScrollAdjust;
    { This makes sure that no matter where MenuSelect is placed, }
    { a valid page and line will be displayed. }
    BEGIN
      MenuMove( 0 );  { skip over blank ranges }
      If MenuMax <= MenuHeight
        THEN BEGIN  { all on screen }
          MenuLowItem := 1;
          MenuHiItem := MenuMax;
          Exit;
        END;  { all on screen }
      If MenuSelect < MenuLowItem
        THEN BEGIN  { select over top }
          MenuLowItem := MenuSelect;
          MenuHiItem := MenuLowItem + ( MenuHeight - 1 ) div MenuSpacing;
        END;  { select over top }
     If MenuSelect > MenuHiItem
       THEN BEGIN  { select below bottom }
         MenuHiItem := MenuSelect;
         MenuLowItem := MenuHiItem - ( MenuHeight - 1 ) div MenuSpacing;
       END;  { select below bottom }
     MenuHiItem := MenuLowItem + MenuHeight - 1;
     If MenuHIItem > MenuMax THEN MenuHiItem := MenuMax;
  END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  PROCEDURE MenuRefresh;
    { redraws selector bar, or entire contents of menu window }
    VAR  RefLoop :Integer;
    BEGIN
      ScrollAdjust;  ShowStatus( NoKey );
      RefLoop := MenuLowItem;
      If ( MenuPrevLowItem = MenuLowItem ) and ( not MenuForceRefresh )
        THEN BEGIN  { no scrolling }
          ShowItem( MenuPrevSelect );
          ShowItem( MenuSelect );
        END  { no scrolling }
        ELSE BEGIN  { refresh all }
          While ( RefLoop <= MenuMax ) and ( RefLoop <= MenuHiItem ) Do
            BEGIN
              ShowItem( RefLoop );
              Inc( RefLoop );
            END;  { refresh all }
        END;
      MenuPrevLowItem := MenuLowItem;
      MenuPrevSelect := MenuSelect;
      MenuForceRefresh := False;
    END;

{- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -}

  PROCEDURE FigureFrame;
    { determines size of menu frame }
    VAR  MaxHeight : Byte;
    BEGIN
      If MenuWX1 = 0
        THEN MenuWX1 := ( 80 div 2 ) - ( ( MenuWidth - 1 + 4 ) div 2 ) + 1;
      MaxHeight := 25 - MenuTopMargin - 2;
      If MenuHeight = 0
        THEN If MenuMax < MaxHeight - 1
               THEN MenuHeight := MenuMax
               ELSE MenuHeight := MaxHeight - 1;
      If MenuSpacing = 0
        THEN If MenuMax * 2 > MaxHeight
               THEN MenuSpacing := 1
               ELSE BEGIN
                 MenuSpacing := 2;  MenuHeight := MenuHeight * 2 - 1;
               END;
      If MenuWY1 = 0
        THEN MenuWY1 := MaxHeight div 2 - MenuHeight div 2 + 2;
    END;

{-------------------------------------------------------------------------}

PROCEDURE FindShowLimits;
  { Determines upper and lower selectable limits. }
  { This compensates for top or bottom skipped lines. }
  VAR  SelectSave : Integer;  WrapSave, SkipSave : Boolean;
  BEGIN
    { save current setup }
    SelectSave := MenuSelect;  WrapSave := MenuWrap;
    SkipSave := SkipUpward;  MenuWrap := True;
    { find bottom limit by wrapping over top }
    MenuSelect := MenuMin;  SkipUpward := True;  MenuMove( -1 );
    MenuShowMax := MenuSelect;
    { find top limit by wrapping off bottom }
    MenuSelect := MenuMax;  SkipUpward := False;  MenuMove( 1 );
    MenuShowMin := MenuSelect;
    { restore prior setup }
    MenuSelect := SelectSave;  MenuWrap := WrapSave;
    { range check current position }
    If MenuSelect < MenuShowMin THEN MenuSelect := MenuShowMin;
    If MenuSelect > MenuShowMax THEN MenuSelect := MenuShowMax;
  END;

{-------------------------------------------------------------------------}

{ This is the heart of the menu system, the main block. }
{ Except for a few refresh support lines, and the frame draw section }
{ at the top, this long procedure is entirely a keypress handler. }

  Label MenuExit;

  BEGIN
    GetVAttr( XSave, YSave, WX1Save, WY1Save, WX2Save, WY2Save,
              AttrSave, CursorSave );
    Cursor( Off );
    StartingSelect := MenuSelect;
    MenuPrevKey := MenuKey;
    SkipUpward := False;
    CheckStack;
    ValidKey := True;
    If MenuDrawFrame
      THEN BEGIN  { create menu frame }
        MenuResult := 0;
        If MenuMarks = NIL
          THEN MenuFirstCol := 2 else MenuFirstCol := 4;
      { open file if disk menu }
        If MenuData = NIL
          THEN BEGIN  { open }
            Assign( MenuFile, MenuFileName );
            {$I-} Reset( MenuFile, 1 ); {$I+} MenuResult := IOResult;
            If MenuResult = 0
              THEN BEGIN  { check record size }
                RecSz := MenuRecOfs + MenuRecPad + MenuWidth;
                FileSz := FileSize( MenuFile );
                If ( FileSz = 0 ) or ( FileSz mod RecSz <> 0 )
                  THEN MenuResult := -99
                  ELSE BEGIN  { check file size }
                    If ( MenuResult = 0 ) and
                       ( ( MenuMax = 0 ) or ( MenuMax > FileSz div RecSz ) )
                      THEN MenuMax := FileSz div RecSz;
                  END;  { check file size }
              END;  { check record size }
          END;  { open }
        If MenuResult <> 0 THEN MenuMax := 0;
        If MenuMax < 1 THEN Exit;
        FigureFrame;
        BorderColor := MenuBorderColor;
        BorderStyle := MenuBorderStyle;
        FrameType := Frame_FullWindow;
        Frame( MenuWX1, MenuTopMargin + MenuWY1,
               MenuWX1 + MenuWidth - 1 + MenuFirstCol - 1,
               MenuTopMargin + MenuWY1 + MenuHeight - 1 );
        KeyLogComment( 'menu' + MenuTitle );
        PutMenuName;
        MenuHiItem := MenuLowItem + MenuHeight;
        MenuDrawFrame := False;
        TurboSearchStr := '';  MenuBottomMsg;
      END;
    If MenuForceRefresh
      THEN BEGIN
        FindShowLimits;  MenuRefresh;
        StartingSelect := MenuSelect;
      END;
    If MenuKeys <> ''  { un-handled extended keys }
      THEN BEGIN
        OopSound;  MenuKeys := '';  ShowStatus( NoKey );
        MenuPrevKey := NoKey;
      END;
    REPEAT
    { If pending, service turbo search while no key is pressed. }
      If TurboSearchPending
        THEN DoTurboSearch( TabKey );
      SetScreenSaver;
      SetTimer( 2, TurboSearchTimeout );
      REPEAT CheckScreenSaver;
      UNTIL KeyPressed or
            ( TimerElapsed( 2 ) and
              ( ( MenuPrevKey <> NoKey ) or ( TurboSearchStr <> '' ) ) );
    { This clears turbo search after several seconds idle. }
      If TimerElapsed( 2 ) and
         ( ( MenuPrevKey <> NoKey ) or ( TurboSearchStr <> '' ) )
        THEN BEGIN
          If TurboSearchStr <> ''
            THEN BEGIN  TurboSearchStr := '';  MenuBottomMsg;  END;
          If MenuPrevKey <> NoKey
            THEN BEGIN  MenuPrevKey := NoKey;  ShowStatus( NoKey );  END;
        END;
    UNTIL KeyPressed;
    MenuKey := UpCase( GetKey );
  { Note that WordStart keys are supported in parallel to cursor keys. }
    If MenuPrevKey in [ ^Q, ^K, ^O ]
      THEN MenuKey := ControlKey( MenuKey );
    If ( ( MenuPrevKey = ^Q ) and
         ( not ( MenuKey in [ ^R, ^C ] ) ) ) or
       ( ( MenuPrevKey = ^K ) and
         ( ( MenuMarks = NIL ) or
           ( not ( MenuKey in [ ^B, ^K, ^H, ^Q ] ) ) ) )
       or ( MenuPrevKey = ^O )
       THEN BEGIN
         MenuKeys := MenuPrevKey + MenuKey;
         ValidKey := False;  Goto MenuExit;
       END;
    If ( MenuPrevKey = ^Q ) and ( MenuKey = ^R )
      THEN MenuKey := HomeKey;  { ^QR = top }
    If ( MenuPrevKey = ^Q ) and ( MenuKey = ^C )
      THEN MenuKey := EndKey;  { ^QC = bottom }
    If ( MenuPrevKey = ^K ) and ( MenuKey = ^B )  { ^KB = block start }
      THEN BEGIN  MenuBlock := 0;  MenuKey := F8Key;  END;
    If ( MenuPrevKey = ^K ) and ( MenuKey = ^K )  { ^KK = block end }
      THEN If MenuBlock = 0
             THEN BEGIN  OopSound;  MenuKey := NoKey;  END
             ELSE MenuKey := F8Key;
    If ( MenuPrevKey = ^K ) and ( MenuKey = ^H )
      THEN BEGIN  { ^KH = block hide/reveal }
        If MenuAllMarked
          THEN MenuKey := F9Key ELSE MenuKey := F10Key;
        MenuAllMarked := not MenuAllMarked;
      END;
    If ( MenuPrevKey = ^K ) and ( MenuKey = ^Q )
      THEN MenuKey := ESCKey;  { ^KQ = quit }
    If MenuKey = ^L THEN MenuKey := TabKey;
  { Keypress interpretation begins here. }
    If ( MenuMode = _NoSearch ) and
       ( ( MenuKey in PrintableCharSet ) or
         ( MenuKey = EnterKey ) )
      THEN BEGIN   { no search, exit on keypress }
        If MenuKey = EnterKey
          THEN DeeDeeSound ELSE DitSound;
        Goto MenuExit;
      END;  { no search, exit on keypress }
    If MenuKey = ESCKey
      THEN BEGIN  { ESC is always quit }
        If MenuMarks = NIL THEN ZoopSound ELSE EnterSound;
        Goto MenuExit;
      END;  { ESC is always quit }
    If MenuKey = EnterKey
      THEN BEGIN  { enter key - exit with selection}
        EnterSound;
        Goto MenuExit;
      END;  { enter key }
    If (MenuKey in PrintableCharSet) and
       (MenuKey <> '+') and (MenuKey <> '-')
      THEN BEGIN  { alpha numeric key }
        If MenuMode = _TurboSearch
          THEN DoTurboSearch( MenuKey )
          ELSE BEGIN  { letter search }
            If not DoLetterSearch( MenuKey )
              THEN BEGIN
                ValidKey := False;  Goto MenuExit;
              END
              ELSE BEGIN  { letter found }
                If MenuMode = _LetterExecute
                  THEN BEGIN  { accept }
                    BleepSound;
                    Goto MenuExit;
                  END;  { accept }
                If MenuMode = _LetterSearch
                  THEN BEGIN  { letter search }
                    DeeDeeSound;  Goto MenuExit;
                  END;  { letter search }
              END;  { letter found }
          END;  { letter search }
        Goto MenuExit;  { turbo search }
      END;  { alpha numeric key }
    Case MenuKey of    { up/down menu movement }
    ^K, ^Q, ^O : BEGIN  DitSound;  ShowStatus( MenuKey );  END;
  ^E, UpArKey : BEGIN
               SkipUpward := True;
               If MenuMax < 2
                 THEN OopSound
                 ELSE BEGIN
                   If MenuSelect - 1 >= MenuMin
                     THEN BEGIN  DiSound;  Dec( MenuSelect );  END  { up }
                     ELSE If MenuWrap
                            THEN BEGIN
                              DiSound;  MenuSelect := MenuMax;
                            END ELSE OopSound;
                 END;
             END;
  ^X, DnArKey : BEGIN
               If MenuMax < 2
                 THEN OopSound
                 ELSE BEGIN
                   If MenuSelect + 1 <= MenuMax
                     THEN BEGIN  DiSound;  Inc( MenuSelect );  END  { down }
                     ELSE If MenuWrap
                            THEN BEGIN
                              DiSound;  MenuSelect := MenuMin;
                            END ELSE OopSound;
                 END;
             END;
  ^R, PgUpKey : BEGIN
               SkipUpward := True;
               DiSound;
               If MenuSelect - MenuHeight + 1 < MenuMin
                 THEN MenuSelect := MenuMin
                 ELSE MenuSelect := MenuSelect - MenuHeight + 1;  { pg up }
             END;
  ^C, PgDnKey : BEGIN
               DiSound;
               If MenuSelect + MenuHeight - 1 > MenuMax
                 THEN MenuSelect := MenuMax
                 ELSE MenuSelect := MenuSelect + MenuHeight - 1;  { pg down }
             END;
   HomeKey : BEGIN
               DiSound;
               MenuSelect := MenuMin;  { top of array }
             END;
    EndKey : BEGIN
                DiSound;
                MenuSelect := MenuMax;  { end of array }
             END;
 BackSpaceKey : If MenuMode <> _TurboSearch
               THEN OopSound
               ELSE BEGIN
                 DooSound;  TurboSearchStr := '';  MenuBottomMsg;
               END;
    TabKey : If ( MenuMode <> _TurboSearch ) or ( TurboSearchStr = '' )
               THEN OopSound
               ELSE DoTurboSearch( MenuKey );
     F8Key : If MenuMarks = NIL
               THEN OopSound
               ELSE BEGIN  { block mark }
                 ChkPtr( MenuMarks^, 1, 'menu mark block' );
                 BleepSound;
                 If MenuBlock = 0
                   THEN MenuBlock := MenuSelect
                   ELSE BEGIN
                     If MenuBlock > MenuSelect
                          THEN For BlockLoop := MenuSelect to MenuBlock do
                                 MenuMarks^[ BlockLoop ] := TickChar
                          ELSE For BlockLoop := MenuBlock to MenuSelect do
                                 MenuMarks^[ BlockLoop ] := TickChar;
                     MenuBlock := 0;
                   END;
                 MenuForceRefresh := True;
                 Goto MenuExit;
               END;  { block mark }
     '+'   : If MenuMarks = NIL then
               OopSound
             ELSE
               BEGIN  { mark selection }
                 ChkPtr( MenuMarks^[ MenuSelect ], 1, 'menu mark' );
                 MenuMarks^[ MenuSelect ] := TickChar;
                 DitSound;  MenuMove( 1 );  Goto MenuExit;
               END;  { mark selection }
     '-'   : If MenuMarks = NIL then
               OopSound
             ELSE
               BEGIN  { unmark selection }
                 ChkPtr( MenuMarks^[ MenuSelect ], 1, 'menu mark' );
                 MenuMarks^[ MenuSelect ] := UnTickChar;
                 DitSound;  MenuMove( 1 );  Goto MenuExit;
               END;  { unmark selection }
     F9Key : If MenuMarks = NIL
               THEN OopSound
               ELSE BEGIN  { unmark all }
                 ChkPtr( MenuMarks^[ MenuMin ], MenuMax, 'menu unmark all' );
                 BleepSound;
                 For BlockLoop := MenuMin to MenuMax do
                   MenuMarks^[ BlockLoop ] := UnTickChar;
                 MenuForceRefresh := True;
                 Goto MenuExit;
               END;  { unmark all }
    F10Key : If MenuMarks = NIL
               THEN OopSound
               ELSE BEGIN  { mark all }
                 ChkPtr( MenuMarks^[ MenuMin ], MenuMax, 'menu mark all' );
                 BleepSound;
                 For BlockLoop := MenuMin to MenuMax do
                   MenuMarks^[ BlockLoop ] := TickChar;
                 MenuForceRefresh := True;
                 Goto MenuExit;
               END;  { mark all }
         ELSE ValidKey := False; { invalid extended key }
     End;  {Case}
    MenuExit:   { menu system recycle }
    SetVAttr( XSave, YSave, WX1Save, WY1Save, WX2Save, WY2Save,
              AttrSave, CursorSave );
    MenuSelectMoved := MenuSelect <> StartingSelect;
    If MenuSelectMoved or MenuForceRefresh
      THEN MenuRefresh;
    If ValidKey
      THEN BEGIN  Menu := MenuSelect;  MenuKeys := '';  END
      ELSE BEGIN
        Menu := 0;
        If MenuKeys = '' THEN MenuKeys := MenuKey;
      END;
    Escape := MenuKey = EscKey;
  END;   { menu procedure }

{-------------------------------------------------------------------------}

{ Use in loop with Menu to show special bottom prompt }
{ when the selector bar is placed of specifc menu lines. }
{ A message for line 0 will always be put.  Use this as the first }
{ of a series of ShowLineMsg statements to set the default. }
{ To have the default clear any other messages, the default must be }
{ a field of spaces as long as the longest non-default message. }

PROCEDURE ShowLineMsg( WhichLine : Integer;  Msg : String40 );
  BEGIN
    If MenuViewOnly or ( not MenuSelectMoved ) or
       ( ( WhichLine <> MenuSelect ) and ( WhichLine <> 0 ) )
      THEN Exit;
    VWrite( 71 - Length( Msg ), 25, Reverse, Msg );
  END;

{-------------------------------------------------------------------------}

BEGIN

END.


