{$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 16384,0,655360}

Program Repl;

USES Dos;

TYPE
  String80 = String[80];
  BigBuffer = Array [ 1..64000 ] of Char;

CONST
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;
  ProgramName = 'REPL';
  Version = '3.2';
  Revision = '91/06/29';
  MaxFiles = 100;
  RgtMgn = 80;
  GoodChars = [ 'a'..'z', 'A'..'Z', '0'..'9', '_' ];
  TempFileSpec = 'REPL.$$$';
  BackFileSpec = 'REPL.BAK';
  NL = #13#10;
  Tab = '  ';
  Spaces = '                                        ';  { 40 }
          { 1234567890123456789012345678901234567890 }

VAR
  JoinBegin, NewProcedure : Boolean;
  InBuf, OutBuf : ^BigBuffer;
  InPoint, InSize, OutSize, OutCnt, FindPoint : LongInt;
  FileName : Array [ 1..MaxFiles ] of String80;
  ComLine, ComFiles, Path, TempStr, Target, Replace, SaveLn : String;
  Big, Terse, WholeWord, IgnoreCase,
  Confirm, Confirmed, Quote : Boolean;
  FileCount, Files, Loop, Result, EOLPoint : LongInt;
  Line, Matches, Changes : LongInt;
  TotalMatches, TotalChanges, TotalChangedFiles : LongInt;
  LineStart : LongInt;
  Ch : Char;
  BigInFile, BigOutFile, TempFile : Text;
  InFile, OutFile : File;
  SubTab : String[10];


{-------------------------------------------------------------------------}

FUNCTION GetKey:Char;
  VAR  Regs:Registers;
  BEGIN
    FillChar( Regs, SizeOf( Regs ), 0 );
    Regs.AH := $08;  { ConInputNoEcho }
    MsDos( Regs );
    GetKey := UpCase( Chr( Regs.AL ) );
  END;

{-------------------------------------------------------------------------}

PROCEDURE Stop( Msg:String );
  BEGIN
    Writeln( Msg, '  Program terminated.' );
    Halt( 2 );
  END;

{-------------------------------------------------------------------------}

FUNCTION ShowFile:String80;
  BEGIN  ShowFile := 'FILE: ' + FileName[ FileCount ];  END;

{-------------------------------------------------------------------------}

PROCEDURE UpLine( VAR InStr:String );
  VAR L:Byte;
  BEGIN  For L := 1 to Length(InStr) do InStr[L] := UpCase(InStr[L]);  END;

{-------------------------------------------------------------------------}

FUNCTION IntString( InNum:LongInt ):String80;
  VAR  OutStr:String80;
  BEGIN  Str( InNum, OutStr );  IntString := OutStr;  END;

{-------------------------------------------------------------------------}

PROCEDURE DeLead( VAR Str:String );
  BEGIN
    While ( Str <> '' ) and
          ( Str[1] in [ ' ', #10, #13, ^I ] ) do Delete( Str, 1, 1 );
  END;

{-------------------------------------------------------------------------}

PROCEDURE DeTrail( VAR Str:String );
  BEGIN
    While ( Str <> '' ) and
          ( Str[Length(Str)] in [ ' ', #10, #13, ^I ] )
          do Delete( Str, Length(Str), 1 );
  END;

{-------------------------------------------------------------------------}

PROCEDURE Strip( VAR Str:String );
  BEGIN  DeLead( Str );  DeTrail( Str );  END;

{-------------------------------------------------------------------------}

PROCEDURE ParsePath( FullName:String80; VAR Path, Name:String80 );
  BEGIN
    Path := '';  Name := '';
    While ( FullName <> '' ) and ( FullName[ Length(FullName) ] <> '\' ) do
      BEGIN
        Name := FullName[ Length(FullName) ] + Name;
        Delete( FullName, Length(FullName), 1 );
      END;
    Path := FullName;
  END;

{-------------------------------------------------------------------------}

PROCEDURE GetFiles( FullName:String80 );
  VAR  Path, Name:String80;  Dir:SearchRec;
  BEGIN
    ParsePath( FullName, Path, Name );
    FindFirst( FullName, AnyFile, Dir );
    If DosError <> 0 THEN Stop( 'No files matching "' + FullName + '".' );
    While DosError = 0 do
      BEGIN
        If ( Dir.Attr = $00 ) or ( Dir.Attr = $20 )  { normal or archived }
          THEN BEGIN
            If Files = MaxFiles THEN
              Stop( ShowFile + NL + '  Too many files specified.' );
            If Files = 1 Then
              Writeln( 'Scanning directory . . .' );
            Inc( Files );
            FileName[ Files ] := Path + Dir.Name;
            UpLine( FileName[ Files ] );
          END;
        FindNext( Dir );
      END;
  END;

{-------------------------------------------------------------------------}

PROCEDURE SortFileNames;
  VAR  Swap:String80;  Done:Boolean;  DelLoop, Base:Word;
  BEGIN
    Base := 1;
    If Files > 1 Then
      Writeln( 'Sorting directory . . .' );
    Repeat   { Remove duplicates. }
      Done := True;  Loop := Base + 1;
      While ( Loop < Files ) and Done do
        BEGIN
          Done := FileName[ Base ] <> FileName[ Loop ];
          Inc( Loop );
        END;
      If Base + 1 >= Files
        THEN Done := True
        ELSE If Done
               THEN BEGIN  Inc( Base );  Done := Base >= Files;  END
               ELSE BEGIN
                 Files := Files - 1;
                 For DelLoop := Loop to Files do
                   FileName[ DelLoop ] := FileName[ DelLoop + 1 ];
               END;
    Until Done;
    If Files > 1 then
      Repeat
        Done := True;
        For Loop := 1 to Files - 1 do
          If FileName[ Loop ] > FileName[ Loop + 1 ]
            THEN BEGIN
              Swap := FileName[ Loop ];
              FileName[ Loop ] := FileName[ Loop + 1 ];
              FileName[ Loop + 1 ] := Swap;
              Done := False;
            END;
      Until Done;
  END;

{-------------------------------------------------------------------------}

PROCEDURE CheckHelp;
  BEGIN
    If ( ParamCount = 0 ) or
       ( ( ParamCount = 1 ) and ( ParamStr(1) = '?' ) )
      THEN BEGIN
        For CpyRtLp := 1 to 26 do
          If CpyRtLp = 1 Then
            CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
          Else
            CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                     Ord( CpyRt[ CpyRtLp - 1 ] ) );
        Writeln( ProgramName, ' ver ', Version,
                 ' rev ', Revision, ' - ', CpyRt, '.' );
        Writeln( 'Run an automatic search and replace through',
                 ' specified files.' );
        Writeln( '  Can also convert files between CR and CRLF formats.' );
        Writeln( ProgramName, ' [options] ["]target["] ' + 
                 '["]replacement["] file(s)' );
        Writeln( '  File(s) can be one or a series of names or',
                 ' wildcard specifiers.' );
        Writeln( 'Options:' );
        Writeln( '  -B  Force on disk processing.' );
        Writeln( '  -G  Globally replace without confirmation.' );
        Writeln( '  -I  Ignore case.' );
        Writeln( '  -T  Output terse changes readout.' );
        Writeln( '  -W  Match whole words only,',
                 ' separated by spaces or symbols.' );
        Writeln( 'Target/replacement:' );
        Writeln( '  Enclose the text in quotes to include',
                 ' imbedded spaces.' );
        Writeln( '  Program does not support wildcard match or ',
                 'replacement text.' );
        Writeln( ProgramName, ' +CR {or} -CR filename [output_filename]' );
        Writeln( '  Strip or add carriage returns after line feeds.' );
        Writeln( '  Omit output file name for in place conversion.' );
        Writeln( 'This program is intended for use only on ASCII files.' );
        Writeln( 'Files under 64K are processed in memory, larger files',
                 ' use disk.' );
        Writeln( 'Program returns errorlevel if no changes made.' );
        Halt( 1 );
      END
      ELSE If ( ParamCount < 1 ) then Stop( 'Insufficient parameters.' );
  END;

{-------------------------------------------------------------------------}

FUNCTION WholePos( Target, Source : String ):Byte;
  VAR  Point:Byte;
  BEGIN
    Point := Pos( Target, Source );
    If ( Point > 0 ) and
       ( ( Point - 1 + Length(Target) >= Length(Source) ) or
         ( Source[ Point + Length(Target) ] in [ #13, #10, ' ' ] ) ) and
       ( ( Point = 1 ) or ( Source[ Point - 1 ] in [ #13, #10, ' ' ] ) )
       THEN WholePos := Point ELSE WholePos := 0;
  END;

{-------------------------------------------------------------------------}

PROCEDURE ProcessLine( VAR VarLn:String );
  VAR  Show, Point:LongInt;  UpLn, InLn, OutLn : String;
       Matched:Boolean;  KeyStr:String80;
  BEGIN
    InLn := VarLn;  UpLn := InLn;  OutLn := '';
    If IgnoreCase THEN UpLine( UpLn );
    While InLn <> '' do
      BEGIN  { Consume line. }
        Point := Pos( Target, UpLn );
        If Point = 0
          THEN BEGIN  OutLn := OutLn + InLn;  InLn := '';  END
          ELSE BEGIN  { Target found. }
            Matched := Point > 0;
            If Matched and WholeWord THEN
              If ( ( Point > 1 ) and
                   ( UpLn[ Point-1 ] in GoodChars ) ) or
                 ( ( Point + Length(Target) - 1 < Length(UpLn) ) and
                   ( UpLn[ Point+Length(Target) ] in GoodChars ) )
                THEN Matched := False;
            If Matched
              THEN BEGIN  { Matched. }
                Inc( Matches );
                Confirmed := not Confirm;
                If not Confirmed
                  THEN BEGIN  { Confirmed. }
                    Show := Length( VarLn ) - Length( InLn ) + Point;
                    Write( Line:5, '> ' + Copy( VarLn, 1, Show - 1 ) );
                    Write( #16 + Copy( VarLn, Show, Length(Target) ) + #17 );
                    Write( Copy( VarLn, Show + Length(Target),
                    EOLPoint - Show - Length(Target) + 1 ) );
                    Write( ' ', #175 );
                    Repeat
                      Ch := GetKey;
                      If ( not ( Ch in [ 'Y', 'N' ] ) ) THEN Write( ^G );
                    Until Ch in [ 'Y', 'N' ];
                    Writeln( UpCase( Ch ) );
                    Confirmed := Ch = 'Y';
                  END;  { Confirm. }
              END;  { Matched. }
            If Matched and Confirmed
              THEN BEGIN
                OutLn := OutLn + Copy( InLn, 1, Point - 1 ) + Replace;
                Inc( Changes );
              END
              ELSE OutLn := OutLn +
                            Copy( InLn, 1, Point + Length(Target) - 1 );
          END;  { Target found. }
        If InLn <> ''
          THEN BEGIN
            InLn := Copy( InLn, Point + Length(Target), 255 );
            UpLn := Copy( UpLn, Point + Length(Target), 255 );
          END;
      END;  { Consume line. }
    VarLn := OutLn;
  END;

{*************************************************************************}

Procedure CheckCRLFProcess;
  Type
    String50 = String[50];
  Const
    InPlace : Boolean = False;
    Cnt : LongInt = 0;
    AddNL : Boolean = False;
    PriorCR : Boolean = False;
    MaxNames = 200;
    NameCount : LongInt = 0;
    InBufSize = 512;
    OutBufSize = InBufSize * 2;
  Var
    FileLoop : LongInt;
    Names : Array [ 1..MaxNames ] of String[12];
    Mask : String[79];
    Path : string[67];
    Name : string[8];
    Ext : string[4];
    Dir : SearchRec;
    InFile, OutFile : File;
    InBuf : Array[ 1..InBufSize ] of Char;
    InSize : LongInt;
    OutBuf : Array[ 1..OutBufSize ] of Char;
    OutSize : LongInt;
    ProcLp : LongInt;
    Pct, PrevPct : Real;
    DotPos : LongInt;
    InName, OutName : String50;
    Ch : Char;
    Lp : Byte;
    UCParam : String[20];
    AddCR : Boolean;
  Procedure ChkErr( Msg : String50 );
    Begin
      If IOResult = 0 Then Exit;
      If AddNL Then Writeln;
      Writeln( Msg );  Halt( 1 );
    End;
  Procedure GetDirArr( FileName : String50 );
    Begin
      FSplit( FileName, Path, Name, Ext );
      Mask := Name + Ext;
      FindFirst( Path + Mask, Archive, Dir );
      While DosError = 0 do
        Begin
          If NameCount >= MaxNames Then
            Begin
              Writeln( 'Too many files matching "' + FileName + '"' );
              Halt( 1 );
            End;
          Inc( NameCount );
          Names[ NameCount ] := Dir.Name;
          FindNext( Dir );
        End;
      If NameCount < 1 Then
        Begin
          Writeln( 'No files matching "' + FileName + '"' );
          Halt( 1 );
        End;
    End;
  Begin
    If ParamCount < 2 Then Exit;
    UCParam := ParamStr( 1 );
    For Lp := 1 to Length( UCParam ) do
      UCParam[ Lp ] := UpCase( UCParam[ Lp ] );
    If ( UCParam <> '+CR' ) and ( UCParam <> '-CR' ) Then Exit;
    AddCR := UCParam = '+CR';
    InName := ParamStr( 2 );
    If ( ParamCount > 2 ) and
       ( ( Pos( '?', InName ) > 0 ) or ( Pos( '*', InName ) > 0 ) ) Then
      Begin
        Writeln( 'Only in place conversions allowed with wildcards.' );
        Halt( 1 );
      End;
    GetDirArr( InName );
    For FileLoop := 1 to NameCount do
      Begin
        InName := Path + Names[ FileLoop ];
        If ParamCount > 2 Then
          OutName := ParamStr( 3 )
        Else
          Begin
            InPlace := True;
            OutName := TempFileSpec;
          End;
        Assign( InFile, InName );
        Reset( InFile, 1 );  ChkErr( InName + ' not found.' );
        PrevPct := 0;
        Assign( OutFile, OutName );
        ReWrite( OutFile, 1 );  ChkErr( 'Cannot create ' + OutName );
        OutSize := 0;
        While not EOF( InFile ) do
          Begin
            BlockRead( InFile, InBuf, InBufSize, Word(InSize) );
            ChkErr( 'Trouble reading ' + InName );
            For ProcLp := 1 to InSize do
              If AddCR Then
                Begin  { add }
                  Ch := InBuf[ ProcLp ]; 
                  If ( Ch = #10 ) and ( not PriorCR ) Then
                    Begin
                      Inc( Cnt );
                      Inc( OutSize );
                      OutBuf[ OutSize ] := #13;
                    End;
                  Inc( OutSize );
                  OutBuf[ OutSize ] := Ch;
                  PriorCR := Ch = #13;
                End  { add }
              Else
                Begin  { remove }
                  Ch := InBuf[ ProcLp ]; 
                  If Ch <> #13 Then
                    Begin
                      Inc( OutSize );
                      OutBuf[ OutSize ] := Ch;
                    End
                  Else
                    Inc( Cnt );
                End;  { remove }
            If OutSize > 0 Then
              Begin
                BlockWrite( OutFile, OutBuf, Word(OutSize) );
                ChkErr( 'Trouble writing ' + OutName );
                OutSize := 0;
              End;
            Pct := ( FilePos( InFile ) / FileSize( InFile ) ) * 100;
            If ( PrevPct = 0 ) or ( Pct - PrevPct >= 1.0 ) Then
              Begin  { show progress }
                AddNL := True;
                Write( #13, Pct:1:0,' % done.' );
                PrevPct := Pct;
              End;
          End;
        If OutSize > 0 Then
          Begin
            BlockWrite( OutFile, OutBuf, Word(OutSize) );
            ChkErr( 'Trouble writing ' + OutName );
          End;
        Writeln;  AddNL := False;
        Close( InFile );  ChkErr( 'Trouble closing ' + InName );
        Close( OutFile );  ChkErr( 'Trouble closing ' + OutName );
        If InPlace Then
          Begin
            Erase( InFile );  ChkErr( 'Trouble erasing ' + InName );
            Rename( OutFile, InName );
            ChkErr( 'Trouble renaming ' + OutName );
          End;
        Write( 'Done.  File: ', InName, '  ',
               Cnt, ' carriage returns ' );
        If AddCR Then
          Writeln( 'added.' )
        Else
          Writeln( 'removed.' );
      End;
    If Cnt = 0 Then
      Halt
    Else
      Halt( 1 );
  End;


{*************************************************************************}

BEGIN

  CheckHelp;
  Loop := 2;
  Move( Mem[ PrefixSeg : $80 ], ComLine, SizeOf( ComLine ) );
  Delete( ComLine, 1, 1 );
{ Parse out the options. }
  IgnoreCase := False;  Confirm := True;  WholeWord := False;
  Terse := False;  Big := False;
  CheckCRLFProcess;
  If ComLine[ 1 ] = '-'
    THEN BEGIN  { Options. }
      Delete( ComLine, 1, 1 );
      While ( ComLine[ 1 ] <> ' ' ) and ( ComLine <> '' ) do
        BEGIN  { Option loop. }
          Case UpCase( ComLine[ 1 ] ) of
          'B': BEGIN
                 Big := True;
                 Delete( ComLine, 1, 1 );
                END;
          'G': BEGIN
                 Confirm := False;
                 Delete( ComLine, 1, 1 );
                END;
          'I': BEGIN
                 IgnoreCase := True;
                 Delete( ComLine, 1, 1 );
                END;
          'T': BEGIN
                 Terse := True;
                 Delete( ComLine, 1, 1 );
                END;
          'W': BEGIN
                 WholeWord := True;
                 Delete( ComLine, 1, 1 );
                END;
          ELSE Stop( 'Unknown option  ''-' + ComLine[ 1 ] + '''' );
         End;  { Option case. }
        END;  { Option loop. }
    END;  { Options. }

{ Parse out the target string. }
  Target := '';  Quote := False;
  If ComLine[ 1 ] = ' ' THEN Delete( ComLine, 1, 1 );
  If ComLine = '' THEN Stop( 'No target specified.' );
  If ComLine[ 1 ] = '"'
    THEN BEGIN  Quote := True;  Delete( ComLine, 1, 1 );  END;
  While ( ComLine <> '' ) and
        ( ( Quote and ( ComLine[ 1 ] <> '"' ) ) or
          ( ( not Quote ) and ( ComLine[ 1 ] <> ' ' ) ) ) do
        BEGIN
          Target := Target + ComLine[ 1 ];
          Delete( ComLine, 1, 1 );
        END;
  If Quote and ( ( ComLine = '' ) or ( ComLine[ 1 ] <> '"' ) )
    THEN Stop( 'Missing close quote in target.' );
  If Quote THEN Delete( ComLine, 1, 1 );

{ Parse out the replace string. }
  Replace := '';  Quote := False;
  If ComLine[ 1 ] = ' ' THEN Delete( ComLine, 1, 1 );
  If ComLine = '' THEN Stop( 'No replacement specified.' );
  If ComLine[ 1 ] = '"'
    THEN BEGIN  Quote := True;  Delete( ComLine, 1, 1 );  END;
  While ( ComLine <> '' ) and
        ( ( Quote and ( ComLine[ 1 ] <> '"' ) ) or
          ( ( not Quote ) and ( ComLine[ 1 ] <> ' ' ) ) ) do
        BEGIN
          Replace := Replace + ComLine[ 1 ];
          Delete( ComLine, 1, 1 );
        END;
  If Quote and ( ( ComLine = '' ) or ( ComLine[ 1 ] <> '"' ) )
    THEN Stop( 'Missing close quote in replacement.' );
  If Quote THEN Delete( ComLine, 1, 1 );

{ Parse out the file names. }
  Files := 0;
  If ComLine[ 1 ] = ' ' THEN Delete( ComLine, 1, 1 );
  If ComLine = '' THEN Stop( 'No files specified.' );
  ComFiles := ComLine;  UpLine( ComFiles );
  REPEAT
    TempStr := '';
    While ( ComLine <> '' ) and ( ComLine[1] <> ' ' ) do
      BEGIN
        TempStr := TempStr + ComLine[ 1 ];
        Delete( ComLine, 1, 1 );
      END;
    GetFiles( TempStr );
    If ComLine[ 1 ] = ' ' THEN Delete( ComLine, 1, 1 );
  UNTIL ComLine = '';
  SortFileNames;

{ Output program startup message. }
  Write( 'REPL ' );
  If IgnoreCase or ( not Confirm ) or WholeWord or
     Terse or Big THEN Write( ' -' );
  If Big THEN Write( 'b' );
  If not Confirm THEN Write( 'g' );
  If IgnoreCase THEN Write( 'i' );
  If Terse THEN Write( 't' );
  If WholeWord THEN Write( 'w' );
  Write( ' "' + Target + '" "' + Replace + '"' );
  Writeln( ' ' + ComFiles );

{ Run the replacement. }
  JoinBegin := False;  NewProcedure := False;  SubTab := '';
  If not Big THEN BEGIN  New( InBuf );  New( OutBuf );  END;
  TotalMatches := 0;  TotalChanges := 0;  TotalChangedFiles := 0;
  If IgnoreCase THEN UpLine(Target);
  For FileCount := 1 to Files do
    BEGIN  { File loop. }
      Line := 0;  Matches := 0;  Changes := 0;
      InSize := 0;  InPoint := 1;  OutSize := 1;
      If not Big
        THEN BEGIN  { Determine if file is big. }
          Assign( InFile, FileName[ FileCount ] );
          Reset( InFile, 1 );  Result := IOResult;
          If Result <> 0 THEN
            Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                  ', trouble opening for size assesment.' );
          Big := FileSize( InFile ) > SizeOf( BigBuffer );
          Close( InFile );  Result := IOResult;
          If Result <> 0 THEN
            Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                  ', trouble closing after size assesment.' );
        END;  { Determine if file is big. }
      If Big
        THEN BEGIN
          Assign( BigInFile, FileName[ FileCount ] );
          Reset( BigInFile );
          Result := IOResult;
          If Result <> 0 THEN
            Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                  ', trouble opening for line input.' );
        END
        ELSE BEGIN
          Assign( InFile, FileName[ FileCount ] );
          Reset( InFile, 1 );
          Result := IOResult;
          If Result <> 0 THEN
            Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                  ', trouble opening for block input.' );
        END;
      If not Big
        THEN BEGIN
          BlockRead( InFile, InBuf^, SizeOf( BigBuffer ), Word(InSize) );
          Result := IOResult;
          If ( Result = 0 ) and ( InSize <= 1 ) THEN Result := 1000;
          If ( Result = 0 ) and ( InSize >= SizeOf( BigBuffer ) )
            THEN Result := 1001;
          If Result <> 0 THEN
            Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                  ', trouble reading input block.' );
        END;
      If Big
        THEN BEGIN
          Assign( BigOutFile, TempFileSpec );
          ReWrite( BigOutFile );  Result := IOResult;
          If Result <> 0 THEN
            Stop( 'FILE: ' + TempFileSpec + NL +
                  '  Error #' + IntString(Result) +
                  ', trouble creating temporary file.' );
        END;
      Writeln( ShowFile );
      While ( ( not Big ) and ( InPoint <= InSize ) ) or
            ( Big and ( not EOF( BigInFile ) ) ) do
        BEGIN  { Convert file. }
          If Big
            THEN BEGIN
              ReadLn( BigInFile, TempStr );  Result := IOResult;
              If Result <> 0 THEN
                Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                      ', trouble reading input line.' );
              TempStr := TempStr + NL;
              EOLPoint := Length( TempStr ) - 2;
            END
            ELSE BEGIN  { Non-big read. }
              FindPoint := InPoint;
              EOLPoint := FindPoint;
              While ( FindPoint <= InSize ) and
                    ( not ( InBuf^[ FindPoint ] in [ #13, #10 ] ) ) do
                    Inc( FindPoint );
              EOLPoint := FindPoint - EOLPoint;
              If ( FindPoint <= InSize ) and
                 ( InBuf^[ FindPoint ] in [ #13, #10 ] )
                 THEN BEGIN
                   TempStr := TempStr + InBuf^[ FindPoint ];
                   Inc( FindPoint );
                 END;
              If ( FindPoint <= InSize ) and ( FindPoint > 1 ) and
                 ( InBuf^[ FindPoint - 1 ] = #13 ) and
                 ( InBuf^[ FindPoint ] = #10 )
                 THEN BEGIN
                   TempStr := TempStr + InBuf^[ FindPoint ];
                   Inc( FindPoint );
                 END;
              Move( InBuf^[ InPoint ], TempStr[1], EOLPoint );
              TempStr[0] := Chr( EOLPoint );
              TempStr := TempStr + NL;
              InPoint := FindPoint;
              EOLPoint := Length( TempStr ) - 2;
            END;  { Non-big read. }
          Inc( Line );
          SaveLn := TempStr;
          Write( '' );
          ProcessLine( TempStr );
{          If not Terse then Write( TempStr ); }
          If Big
            THEN BEGIN
              Write( BigOutFile, TempStr );  Result := IOResult;
              If Result <> 0 THEN
                Stop( 'FILE: ' + TempFileSpec + NL +
                      '  Error #' + IntString(Result) +
                      ', trouble writing output line.' );
            END
            ELSE BEGIN  { Non-big write. }
              If OutSize + Length( TempStr ) + 0.0 > SizeOf( BigBuffer )
                THEN Stop( ShowFile + NL + '  Output exceeds capacity.' +
                           '  No changes saved.' );
              Move( TempStr[1], OutBuf^[ OutSize ], Length( TempStr ) );
              OutSize := OutSize + Length( TempStr );
            END;
        END;  { Convert file. }
      If Terse and ( Matches = 0 ) and ( Changes = 0 )
        THEN Writeln;
      If Confirm and ( Files > 1 ) and
         ( ( not Terse ) or ( Terse and ( Matches > 0 ) ) )
        THEN Write( '  MATCHES: ', Matches );
      If ( Files > 1 ) and
         ( ( not Terse ) or ( Terse and ( Changes > 0 ) ) )
         THEN Write( '  CHANGES: ', Changes );
      If ( Files > 1 ) and
         ( ( not Terse ) or
           ( Terse and ( ( Changes > 0 ) or ( Matches > 0 ) ) ) )
         THEN Writeln;
      If ( ( Terse and ( Files = 1 ) ) ) and
         ( ( Changes > 0 ) or ( Matches > 0 ) )
        THEN Write( '  ' );
      TotalMatches := TotalMatches + Matches;
      TotalChanges := TotalChanges + Changes;
      If Big THEN Close( BigInFile ) ELSE Close( InFile );
      Result := IOResult;
      If Result <> 0 THEN
        Stop( ShowFIle + NL + '  Error #' + IntString(Result) +
              ', trouble closing source file.'  );
      If Big
        THEN BEGIN
          Close( BigOutFile );  Result := IOResult;
          If Result <> 0 THEN
            Stop( 'FILE: ' + TempFileSpec + NL +
                  '  Error #' + IntString(Result) +
                  ', trouble closing temporary file.' );
        END;
      If Changes > 0
        THEN BEGIN  { File processed, some changes made. }
          Assign( TempFile, BackFileSpec );
          Erase( TempFile );  Result := IOResult;
          If Big
            THEN Rename( BigInFile, BackFileSpec )
            ELSE Rename( InFile, BackFileSpec );
          Result := IOResult;
          If Result <> 0 THEN
            Stop( 'FILE: ' + BackFileSpec + NL +
                  '  Error #' + IntString(Result) +
                  ', trouble renaming source file.' );
          If Big
            THEN BEGIN
              Rename( BigOutFile, FileName[ FileCount ] );
              Result := IOResult;
              If Result <> 0 THEN
                Stop( 'FILE: ' + TempFileSpec + NL +
                      '  Error #' + IntString(Result) +
                      ', trouble renaming temporary file to ''' +
                      FileName[ FileCount ] + '''' );
            END
            ELSE BEGIN  { Non-big flush. }
              Assign( OutFile, FileName[ FileCount ] );
              ReWrite( OutFile, 1 );  Result := IOResult;
              If Result <> 0 THEN
                Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                      ', trouble opening block output.' );
              BlockWrite( OutFile, OutBuf^, Word(OutSize - 1), Word(OutCnt) );
              Result := IOResult;
              If ( Result = 0 ) and
                 ( OutSize - 1 <> OutCnt ) THEN Result := 1000;
              If Result <> 0 THEN
                Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                      ', trouble writing block.' );
              Close( OutFile );  Result := IOResult;
              If Result <> 0 THEN
                Stop( ShowFile + NL + '  Error #' + IntString(Result) +
                      ', trouble closing block write.' );
            END;  { Non-big flush. }
          If Big THEN Erase( BigInFile ) ELSE Erase( InFile );
          Result := IOResult;
          Inc( TotalChangedFiles );
        END  { File processed, some changes made. }
        ELSE BEGIN
          If Big
            THEN BEGIN
              Assign( TempFile, '\REPLACE.TMP' );
              Erase( TempFile );  Result := IOResult;
            END;
        END;
    END;  { File loop. }
  If TotalChanges > 0 Then
    Begin
      Write( 'TOTAL  --->  FILES: ', TotalChangedFiles, ' of ', Files );
      If Confirm THEN Write( '  MATCHES: ', TotalMatches );
      Writeln( '  CHANGES: ', TotalChanges );
      Halt;
    End
  Else
    Begin 
      Writeln( 'No changes made.  ', Files, ' files.' );
      Halt( 1 );
    End;

END.






