(***********************************************************************
************************************************************************
****                                                                ****
****                           SOUNDS                               ****
****                                                                ****
****           Various keypress and event sound effects.            ****
****                                                                ****
************************************************************************
***********************************************************************)


Unit Sounds;

{$I LIBRARY.CFG }

INTERFACE


Uses Dos, Crt, Global;

{ Sys.SoundOn globally enables the extended sounds, except the error tones }
{ OopSound, ClaxonSound, and SirenSound }

Procedure Tone( Freq,                   { frequency, in hertz }
                Dur : Word              { duration, in milliseconds }   );
Procedure BleepSound;
Procedure SpecialSound;
Procedure GrikSound;
Procedure ClaxonSound;
Procedure SirenSound;
Procedure ContinueSound;
Procedure DeeDeeSound;
Procedure DiSound;
Procedure DitSound;
Procedure DooSound;
Procedure EnterSound;
Procedure AttentionSound;
Procedure NoKaySound;
Procedure OKaySound;
Procedure OopSound;
Procedure ZoopSound;



IMPLEMENTATION


{----------------------------------------------------------------------}
{ Procedure Tone                                                       }
{                                                                      }
{ used by the other sound routines to produce sound                    }
{----------------------------------------------------------------------}

Procedure Tone( Freq,                   { frequency, in hertz }
                Dur : Word              { duration, in milliseconds }   );

Begin
  If Sys.SoundOn and ( not Sys.RemoteOnly ) Then
    Begin
      Sound( Freq );
      Delay( Dur );
      NoSound
    End
End;



{----------------------------------------------------------------------}
{ Procedure BleepSound                                                 }
{                                                                      }
{ produce a sound to indicate pressing a command key                   }
{----------------------------------------------------------------------}

Procedure BleepSound;

Begin
  Tone( 2000, 100 );
  Tone( 4000, 100 )
End;


{----------------------------------------------------------------------}
{ Procedure SpecialSound                                               }
{                                                                      }
{ play a chromatic scale to echo special operations                    }
{----------------------------------------------------------------------}

Procedure SpecialSound;

Begin
  Tone( 784, 40 );
  Tone( 831, 40 );
  Tone( 880, 40 );
  Tone( 932, 40 );
  Tone( 986, 40 )
End;


{----------------------------------------------------------------------}
{ Procedure GrikSound                                                  }
{                                                                      }
{ a special chirp/croak used for macro key operations                  }
{----------------------------------------------------------------------}

Procedure GrikSound;

var
  GrikLoop : Word;

Begin
  For GrikLoop := 1 To 3 Do
    Begin
      Tone( 7000, 8 );
      Tone( 1900, 8 );
    End;
End;


{----------------------------------------------------------------------}
{ Procedure ClaxonSound                                                }
{                                                                      }
{ produce a sound to indicate trouble                                  }
{ (e.g. program error, i/o device error, alien invasion, etc.)         }
{----------------------------------------------------------------------}

Procedure ClaxonSound;

var
  ClaxLoop : Word;

Begin
  If Sys.Remote Then Remote( ^G );
  If Sys.RemoteOnly Then Exit;
  For ClaxLoop := 1 To 70 Do
    Begin
      If Odd( ClaxLoop ) Then
          Sound( 1000 )
      Else
          Sound( 250 );
      Delay( 10 )
    End;
  NoSound
End;


{----------------------------------------------------------------------}
{ Procedure SirenSound                                                 }
{                                                                      }
{ produce a sound to indicate trouble                                  }
{ (e.g. program error, i/o device error, alien invasion, etc.)         }
{----------------------------------------------------------------------}

Procedure SirenSound;

var  Loop : Integer;  B : Boolean;

Begin
  If Sys.Remote Then Remote( ^G );
  If Sys.RemoteOnly Then Exit;
  Loop := 150;
  While Loop <= 1800 do
    BEGIN  Sound( Loop );  Delay( 1 );  Loop := Loop + 2;  END;
  NoSound;  Delay( 500 );
End;



{----------------------------------------------------------------------}
{ Procedure ContinueSound;                                             }
{                                                                      }
{ produce a sound to indicate mode dismissal                           }
{----------------------------------------------------------------------}

Procedure ContinueSound;

Begin
  Tone( 1700, 50 );
  Tone( 1200, 50 )
End;



{----------------------------------------------------------------------}
{ Procedure DeeDeeSound                                                }
{                                                                      }
{ produce a sound to indicate menu selection made                      }
{----------------------------------------------------------------------}

Procedure DeeDeeSound;

Begin
  Tone( 1500, 30 );
  Delay( 30 );
  Tone( 1500, 30 )
End;



{----------------------------------------------------------------------}
{ Procedure DiSound                                                    }
{                                                                      }
{ produce a sound to indicate movement to new menu item                }
{----------------------------------------------------------------------}

Procedure DiSound;

Begin
  Tone( 2000, 4 )
End;



{----------------------------------------------------------------------}
{ Procedure DitSound;                                                  }
{                                                                      }
{ produce a sound to indicate a keypress                               }
{----------------------------------------------------------------------}

Procedure DitSound;

Begin
  Tone( 2000, 40 )
End;



{----------------------------------------------------------------------}
{ Procedure DooSound                                                   }
{                                                                      }
{ produce a sound to echo use of the backspace key                     }
{----------------------------------------------------------------------}

Procedure DooSound;

Begin
  Tone( 880, 50 );
  Tone( 831, 50 );
  Tone( 784, 50 )
End;



{----------------------------------------------------------------------}
{ Procedure EnterSound                                                 }
{                                                                      }
{ produce a sound to echo use of the enter key                         }
{----------------------------------------------------------------------}

Procedure EnterSound;

Begin
  Tone( 422, 50 );
  Tone( 502, 50 );
  Tone( 670, 50 )
End;



{----------------------------------------------------------------------}
{ Procedure AttentionSound;                                            }
{                                                                      }
{ produce a sound indicating the system needs attention                }
{----------------------------------------------------------------------}

Procedure AttentionSound;

Begin
  Tone( 5000, 60 )
End;



{----------------------------------------------------------------------}
{ Procedure NoKaySound;                                                }
{                                                                      }
{ produce a sound to indicate a 'no' user response to a Y/N prompt     }
{----------------------------------------------------------------------}

Procedure NoKaySound;

var
  Loop : Integer;

Begin
  For Loop := 1 to 6 do
      If Odd( Loop ) Then
          Tone( 350, 25 )
      Else
          Tone( 300, 25 )
End;



{----------------------------------------------------------------------}
{ Procedure OKaySound                                                  }
{                                                                      }
{ produce a sound to indicate a 'yes' user response to a Y/N prompt    }
{----------------------------------------------------------------------}

Procedure OKaySound;

Begin
  Tone( 100, 25 );
  Tone( 200, 25 );
  Tone( 400, 25 );
  Tone( 800, 20 );
  Tone( 1600, 20 );
  Tone( 3200, 15 );
  Tone( 6400, 15 )
End;



{----------------------------------------------------------------------}
{ Procedure OopSound                                                   }
{                                                                      }
{ produce a sound to indicate a bad key press                          }
{----------------------------------------------------------------------}

Procedure OopSound;

Begin
  If Sys.Remote Then Remote( ^G );
  If Sys.RemoteOnly Then Exit;
  Sound( 400 );
  Delay( 250 );
  NoSound
End;



{----------------------------------------------------------------------}
{ Procedure ZoopSound                                                  }
{                                                                      }
{ produce a sound to echo use of the escape key                        }
{----------------------------------------------------------------------}

Procedure ZoopSound;

var
  Loop : Integer;

Begin
  If Sys.SoundOn and ( not Sys.RemoteOnly ) Then
    Begin
      For Loop := 12 DownTo 6 Do
        Begin
          Sound( Loop * 100 );
          Delay( 20 )
        End;
      NoSound
    End
End;
{----------------------------------------------------------------------}

Begin
  Sys.SoundOn := True;
End.


