{$A+,B-,D-,E-,F-,I-,L-,N-,O-,R-,S-,V-}
{$M 16384,0,0}

{ SPLICE ver 3.1 rev 91/06/29 }
{ Added blocking ver 3.2 rev 96/05/31 }


Program Splice;

CONST
  CpyRt : String[ 26 ] =   { encoded '(C) Andrew Bedno - Chicago' }
      #168#155#102#119#161#173#118#142#115#146#41#162#163#127#138#129 +
      #49#141#115#163#165#129#122#126#134#136;
  CpyRtLp : Byte = 1;
{ MaxBuf = 128;  1996.05.31 }
  MaxBuf = 1024;
  DoSplit : Boolean = False;
  DoMerge : Boolean = False;

TYPE
  String80  = String[80];

VAR
  Result    : Integer;
  Source    : File;
  Target    : File;
  ByteFile  : File of Byte;
  Block     : Array[1..MaxBuf * 11] of Byte;
  EndBlock  : LongInt;
  Size      : LongInt;
  BlockSize : LongInt;
  PieceName : String[60];
  FileName  : String[60];
  BlocksPerSplit : LongInt;
  Pieces    : LongInt;
  Blocks    : LongInt;
  Remainder : LongInt;
  Loop, I, X : Integer;
  R         : Real;
  V         : LongInt;
  St        : String[20];
  Ch        : Char;
  ReadSize : Word;

Procedure Help;
  Begin
    For CpyRtLp := 1 to 26 do
      If CpyRtLp = 1 Then
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 )
      Else
        CpyRt[ CpyRtLp ] := Chr( Ord( CpyRt[ CpyRtLp ] ) - 128 +
                                 Ord( CpyRt[ CpyRtLp - 1 ] ) );
    Writeln( 'SPLICE ver 3.2 rev 96/05/31 - ', CpyRt, '.' );
    Writeln( 'Split, merge, or excerpt from large files.' );
    Writeln( '  Split a large file into several smaller files' );
    Writeln( '    or merge those parts back into an original.' );
    Writeln( '  Also excerpt any section from a file.' );
    Writeln( '    or generate an arbitrary size test file.' );
    Writeln( '  Useful for telecommunicating or examining large files.' );
    Writeln( 'SPLICE /S filename pieces' );
    Writeln( '  Split file into pieces smaller files.' );
    Writeln( 'SPLICE /M filename pieces' );
    Writeln( '  Merge pieces smaller files into larger file.' );
    Writeln( '  Filename refers to the whole.' );
    Writeln( '  Pieces use name with ext changed to .P00, .P01, ...' );
    Writeln( '  From 2 to 99 pieces can be created.' );
    Writeln( '  Size of all but last piece mod 128 = 0.' );
    Writeln( 'SPLICE /C filename offset size' );
    Writeln( '  Cut size bytes from filename starting at offset byte.' );
    Writeln( '  Output appended to SPLICE.$$$ file.' );
    Writeln( 'SPLICE /F size' );
    Writeln( '  Create SPLICE.$$$ size byte fake text file.' );
    Halt( 1 );
  End;


Procedure Fatal( St : String80 );
  Begin
    Writeln( St );
    Halt(1);
  End;

Procedure Trouble( St : String80 );
  Begin
    Result := IOResult;
    If Result > 0 then
      Begin
        Writeln;
        Fatal('Trouble ' + St + ' file.');
      End;
  End;

Procedure GetParams;
  Begin
    If ParamCount < 3 then Help;
    DoSplit := ( ParamStr( 1 ) = '/S' ) or ( ParamStr( 1 ) = '/s' );
    DoMerge := ( ParamStr( 1 ) = '/M' ) or ( ParamStr( 1 ) = '/m' );
    If ( not DoSplit ) and ( not DoMerge ) Then
      Fatal( 'Invalid operation.' );
    St := ParamStr(3);
    Val( St, V, Result );
    If Result > 0 then Fatal('Invalid number of pieces.');
    Pieces := V;
    If ( Pieces < 2 ) or ( Pieces > 99 ) then
      Fatal( 'Number of pieces out of range 2 <= pieces <= 99');
    FileName := ParamStr(2);
  End;


Procedure MakePieceName;
  Var
    PosDot : Integer;
    PNm : String[10];
  Begin
    PosDot := Pos( '.', FileName );
    If PosDot = 0 Then
      PieceName := FileName
    Else
      PieceName := Copy( FileName, 1, PosDot - 1 );
    Str( Loop - 1, PNm );
    If Length( PNm ) = 1 Then PNm := '0' + PNm;
    PieceName := PieceName + '.P' + PNm;
  End;


Procedure OpenPiece;
  Begin
    MakePieceName;
    If DoSplit Then
      Write( #13, 'Writing piece ', Loop, ' of ', Pieces )
    Else
      Write( #13, 'Reading piece ', Loop, ' of ', Pieces );
    Assign( Target, PieceName );
    If DoSplit Then
      ReWrite( Target, 1 )
    Else
      Reset( Target, 1 );
    Trouble('opening piece');
  End;


Procedure Split;
  Begin
    Assign( Source, FileName );
    Reset( Source, 1 );
    Result := IOResult;
    If Result > 0 then Fatal('Source file not found.');
    Size := FileSize( Source );
    Blocks := Size div MaxBuf div Pieces;
    If Blocks < 1 then Fatal('File too small.');
    Remainder := Size - ( Blocks * Pieces * MaxBuf );
    Result := IOResult;
    If Result > 0 then Fatal('Source file not found.');
    Loop := 1;
    While ( Result = 0 ) and ( Loop <= Pieces ) do
    Begin  { read whole blocks }
      EndBlock := Blocks;
      If ( Loop = Pieces ) and ( Remainder > 0 ) Then Inc( EndBlock );
      OpenPiece;
      For V := 1 to EndBlock do
      Begin  { piece write loop }
        If V > Blocks Then  { round up to next 128 }
          ReadSize := Remainder
        Else
          ReadSize := MaxBuf;
{       FillChar( Block, SizeOf( Block ), 0 );  1996.05.31 }
        BlockRead( Source, Block, ReadSize, I );
        Trouble('reading source');
        If ( V <= Blocks ) and ( ReadSize <> I ) Then
          Fatal( 'Size error reading piece file.' );
        BlockWrite( Target, Block, ReadSize );
        Trouble('writing piece');
      End;  { piece write loop }
      Close( Target );
      Trouble('closing piece');
      Loop := Loop + 1;
    End;  { reading whole file }
    Close( Source );
    Trouble('closing source');
    Writeln;
  End;


Procedure Merge;
  Begin
    Assign( Source, FileName );
    Erase( Source );  Result := IOResult;
    ReWrite( Source, 1 );  Result := IOResult;
    If Result > 0 then Fatal('Source file not found.');
    Loop := 1;
    While ( Result = 0 ) and ( Loop <= Pieces ) do
    Begin  { read whole blocks }
      OpenPiece;
      While ( not EOF( Target ) ) do
      Begin  { piece write loop }
        BlockSize := FileSize( Target ) - FilePos( Target );
        If BlockSize > 128 Then BlockSize := 128;
        BlockRead( Target, Block, BlockSize, I );
        Trouble('reading piece');
        If BlockSize <> I then Fatal('Trouble block reading piece file.');
        BlockWrite( Source, Block, I, X );
        Trouble('writing whole');
        If X <> I then Fatal('Trouble block writing whole file.');
      End;  { piece write loop }
      Close( Target );
      Trouble('closing piece');
      Loop := Loop + 1;
    End;  { reading whole file }
    Close( Source );
    Trouble('closing source');
    Writeln;
  End;


Procedure MakeFake;
  Type
    String40 = String[40];
  Var
    NumStr    : String[20];
    OutStr    : String[80];
    OutFile   : Text;
    LineCnt   : LongInt;
    ByteCount : LongInt;
  Const
    TargetSize : LongInt = 0;
  Procedure ErrTst( St : String40 );
    Var
      Result : Integer;
    Begin
      If IOResult = 0 Then Exit;
      Writeln( 'Trouble ' + St + ' test file.' );
      Close( OutFile );  Result := IOResult;
      Halt( 1 );
    End;
  Begin
    Val( ParamStr( 2 ), TargetSize, Result );
    If ( Result <> 0 ) or ( TargetSize < 1 ) Then
      Begin
        Writeln( 'Invalid file size specified.' );
        Halt( 1 );
      End;
    Writeln( 'Creating ', TargetSize, ' byte file . . .' );
    ByteCount := 0;  LineCnt := 0;
    Assign( OutFile, 'SPLICE.$$$' );
    ReWrite( OutFile );  ErrTst( 'creating' );
    While ByteCount < TargetSize do
      Begin
        Inc( LineCnt );
        Str( LineCnt, NumStr );
        OutStr := 'This is line number ' + NumStr;
        Str( ByteCount, NumStr );
        OutStr := OutStr + ', starting byte number ' + NumStr + '.';
        ByteCount := ByteCount + Length( OutStr ) + 2;
        Writeln( OutFile, OutStr );  ErrTst( 'writing' );
        If LineCnt mod 100 = 0 Then
          Write( #13, LineCnt, ' lines, ', ByteCount, ' bytes.' );
      End;
    Str( LineCnt:5, NumStr );
    OutStr := 'DONE: ' + NumStr + ' lines, ';
    ByteCount := ByteCount + 34;
    Str( ByteCount:6, NumStr );
    OutStr := OutStr + NumStr + ' bytes.';
    Writeln( OutFile, OutStr );
    ErrTst( 'writing' );
    Close( OutFile );  ErrTst( 'closing' );
    Writeln( #13, 'DONE:  ', LineCnt, ' lines, ', ByteCount, ' bytes.' );
  End;


Procedure DoCut;
  Type
    String40 = String[40];
  Const
    BytOfs : LongInt = 0;
    BytSize : LongInt = 0;
    BufSz = 512;
    PrevPct : Byte = 0;
  Var
    Pct : Byte;
    OutFile, InFile : File;
    InSz, SzLp : LongInt;
    Buf : Array[ 1..BufSz ] of Byte;
  Procedure ErrTst( St : String40 );
    Var
      Result : Integer;
    Begin
      If IOResult = 0 Then Exit;
      Writeln( 'Trouble ' + St + ' file.' );
      Close( OutFile );  Result := IOResult;
      Halt( 1 );
    End;
  Begin
    If ParamCount < 4 Then
      Begin
        Writeln( 'Insufficient parameters.' );
        Halt( 1 );
      End;
    Assign( InFile, ParamStr( 2 ) );
    Reset( InFile, 1 );  ErrTst( 'opening input' );
    Assign( OutFile, 'SPLICE.$$$' );
    Reset( OutFile, 1 );  Result := IOResult;
    If Result <> 0 Then
      Begin
        ReWrite( OutFile, 1 );  ErrTst( 'creating output' );
      End
    Else
      Begin
        Seek( OutFile, FileSize( OutFile ) );
        ErrTst( 'appending output' );
      End;
    Val( ParamStr( 3 ), BytOfs, Result );
    If ( Result <> 0 ) or ( BytOfs < 0 ) Then
      Begin
        Writeln( 'Invalid offset value.' );
        Halt( 1 );
      End;
    If BytOfs >= FileSize( InFile ) Then
      Begin
        Writeln( 'Offset is beyond end of file.' );
        Halt( 1 );
      End;
    Val( ParamStr( 4 ), BytSize, Result );
    If ( Result <> 0 ) or ( BytSize < 1 ) Then
      Begin
        Writeln( 'Invalid size value.' );
        Halt( 1 );
      End;
    If BytSize + BytOfs - 1 >= FileSize( InFile ) Then
      Begin
        Writeln( 'Size and offset go beyond end of file.' );
        BytSize := FileSize( InFile ) - BytOfs;
        Writeln( 'Size reduced to ', BytSize );
      End;
    Seek( InFile, BytOfs );  ErrTst( 'positioning input' );
    SzLp := 1;
    While SzLp <= BytSize do
      Begin
        If BytSize - SzLp + 1 >= BufSz Then
          InSz := BufSz
        Else
          InSz := BytSize - SzLp + 1;
        SzLp := SzLp + InSz;
        BlockRead( InFile, Buf, InSz );
        ErrTst( 'reading input' );
        BlockWrite( OutFile, Buf, InSz );
        ErrTst( 'writing output' );
        Pct := Round( SzLp * 100 / BytSize );
        If Pct <> PrevPct Then
          Begin
            PrevPct := Pct;
            Write( #13, Pct:3,  '% done.' );
          End;
      End;
    Writeln;
  End;


Begin

  If ( ParamCount > 0 ) and
     ( ( ParamStr( 1 ) = '/C' ) or ( ParamStr( 1 ) = '/c' ) ) Then
    DoCut
  Else
    If ( ParamCount > 0 ) and
       ( ( ParamStr( 1 ) = '/F' ) or ( ParamStr( 1 ) = '/f' ) ) Then
      MakeFake
    Else
      Begin
        GetParams;
        If DoSplit Then
          Split
        Else
          Merge;
      End;

End.


