(*********************************************************************
**********************************************************************
****                                                              ****
****                      STRINGS.PAS                             ****
****                                                              ****
****        String utilities.  Formatting, conversions, etc.      ****
****                                                              ****
**********************************************************************
**********************************************************************)


Unit Strings;


{$I LIBRARY.CFG }


INTERFACE


Uses Dos, Crt, Global, Sounds, Math;

type
  String200 = String[200];
  String175 = String[175];
  String150 = String[150];
  String125 = String[125];
  String100 = String[100];
  String80 = String[80];
  String53 = String[53];
  String50 = String[50];
  String40 = String[40];
  String30 = String[30];
  String25 = String[25];
  String20 = String[20];
  String15 = String[15];
  String10 = String[10];
  String8 = String[8];
  String6 = String[6];
  String5 = String[5];
  String4 = String[4];
  String2 = String[2];
  String1 = String[1];
  CharSet = Set of Char;

const
  DefaultParseCharSet : Set of Char = [ ' ', ',', ';' ];
  CommaStr : String1 = ',';
  DecimalStr : String1 = '.';


var
  ParseCharSet : Set of Char;
  TempSt : BigString;



Function PadString
        (
                S : BigString;             { string to be padded }
                Ch : Char;              { character to pad with }
                Width : Integer         { width of field }
        ) : BigString;                     { result }

Function Pad
        (
                S : BigString;             { string to be padded }
                Width : Integer         { width of field }
        ) : BigString;                     { result }

Function Abrv
        (
                S : BigString;             { string to be abbreaviated }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

Function Rep
        (
                Ch : Char;              { character to be repeated }
                Count : Integer         { length of resultant string }
        ) : BigString;                     { result }

Function Space
        (
                Count : Integer           { length of resultant string }
        ) : BigString;                     { result }

Procedure StripString
        (
            var  S : BigString;             { string to be stripped }
                Ch : Char               { character to be removed }
        );

Procedure DeTrailString
        (
            var  S : BigString;            { string to be stripped }
            Ch : Char
        );

Procedure DeTrail
        (
            var  S : BigString             { string to be stripped }
        );

Procedure Strip
        (
            var  S : BigString              { string to be stripped }
        );

Function Upper
        (
                S : BigString
        ) : BigString;

Function LowCase
        (         Ch : Char
        ) : Char;

Function ControlKey
        (         Ch : Char
        ) : Char;

Function Capital
        (         S : BigString
        ) : BigString;

Function CenterString
        (
                S : BigString;             { string to be centered }
                Ch : Char;              { character to pad with }
                Width : Integer          { width of field }
        ) : BigString;                     { result }

Function Center
        (
                S : BigString;             { string to be centered }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

Function CJ
        (
                S : BigString;             { string to be centered }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

Function Commas
        (
                S : String40
        ) : String40;

Function I_S
        (
                InNum : LongInt
        ) : String10;

Function R_S
        (
                InNum : Real;
                Decimal : Integer
        ) : String10;

Function B_S
        (
                InB : Boolean
        ) : String5;

Function S_I
        (
                InStr : SmallString
        ) : LongInt;

Function S_R
        (
                InStr : SmallString
        ) : Real;

Function OKMsg
        (
                Result : Integer
        ) : SmallString;

Function IsAre
        (
                InNum : LongInt
        ) : String5;

Function Plural
        (
                InNum : LongInt
        ) : String2;

Function ParseStr
       (
                  var InStr : BigString  { source string }
       ) : BigString;   { returned substring }

Function ParseInt
       (
                  var InStr : BigString  { source string }
       ) : LongInt;   { returned value }

Function ParseReal
       (
                  var InStr : BigString  { source string }
       ) : Real;   { returned value }

Function ParamScan( Target : SmallString ) : Integer;

Function FullName : String40;

Function BlankLine( InStr : String150 ) : Boolean;

Function PrintLen( InStr : String150 ) : Integer;


IMPLEMENTATION

{----------------------------------------------------------------------}
{ Function PadString                                                   }
{                                                                      }
{ pad string S to length Abs(Width)                                    }
{ if S is shorter than Abs(Width), truncate it to that length          }
{ otherwise, pad it with Ch (on the left if >0, else on the right)     }
{----------------------------------------------------------------------}

Function PadString
        (
                S : BigString;             { string to be padded }
                Ch : Char;              { character to pad with }
                Width : Integer         { width of field }
        ) : BigString;                     { result }

Var AbsWidth : Integer;

Begin
  If Width = 0 Then S := ''
    Else
      Begin
        ChkPtr( S, Abs( Width ), 'string pad' );
        If Length( S ) < Abs(Width) Then
          Begin
            If Width < 0 Then
                FillChar( S[ Length(S) + 1 ], Abs(Width) - Length(S), Ch )
              Else
                Begin
                  Move( S[ 1 ], S[ Width - Length(S) + 1 ], Length(S) );
                  FillChar( S[ 1 ], Width - Length(S), Ch )
                End
          End;
      End;
  S[ 0 ] := Chr( Abs( Width ) );
  PadString := S
End;



{----------------------------------------------------------------------}
{ Function Pad                                                         }
{                                                                      }
{ pad string S to length Abs(Width)                                    }
{ if S is shorter than Abs(Width), truncate it to that length          }
{ otherwise, pad it with spaces (on the left if >0, else on the right) }
{----------------------------------------------------------------------}

Function Pad
        (
                S : BigString;             { string to be padded }
                Width : Integer         { width of field }
        ) : BigString;                     { result }

Begin
  Pad := PadString( S, ' ', Width );
End;



{----------------------------------------------------------------------}
{ Function Abrv                                                        }
{                                                                      }
{ abbreviate string to Width characters                                }
{ if string is shorter than Width, return it unchanged, otherwise,     }
{ truncate it to Width - 1 characters, and append a period             }
{----------------------------------------------------------------------}

Function Abrv
        (
                S : BigString;             { string to be abbreaviated }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

Begin
  If Width < 0 Then Exit;
  If Length( S ) > Width Then
    Begin
      S[ 0 ] := Chr( Width );
      S[ Width ] := '.'
    End;
  Abrv := S
End;



{----------------------------------------------------------------------}
{ Function Rep                                                         }
{                                                                      }
{ make a string of the character Ch repeated Count times               }
{----------------------------------------------------------------------}

Function Rep
        (
                Ch : Char;              { character to be repeated }
                Count : Integer            { length of resultant string }
        ) : BigString;                     { result }

var
  S : BigString;

Begin
  If Count < 0 Then Exit;
  S[ 0 ] := Chr( Count );
  FillChar( S[ 1 ], Count, Ch );
  Rep := S;
End;



{----------------------------------------------------------------------}
{ Function Space                                                       }
{                                                                      }
{ make a string of spaces, of length Count                             }
{----------------------------------------------------------------------}

Function Space
        (
                Count : Integer            { length of resultant string }
        ) : BigString;                     { result }

Begin
  Space := Rep( ' ', Count );
End;



{----------------------------------------------------------------------}
{ Procedure DeTrailString                                              }
{                                                                      }
{ remove any selected trailing characters from a string                }
{----------------------------------------------------------------------}

Procedure DeTrailString
        (
            var  S : BigString;            { string to be stripped }
            Ch : Char
        );

var
  Index : Integer;

Begin
  Index := Length( S );  If Index = 0 Then Exit;
  While ( S[ Index ] = Ch ) and ( Index > 0 ) Do
      Index := Index - 1;
  S[ 0 ] := Chr( Index );
End;



{----------------------------------------------------------------------}
{ Procedure DeTrail                                                    }
{                                                                      }
{ remove trailing spaces from a string                                 }
{----------------------------------------------------------------------}

Procedure DeTrail
        (
            var  S : BigString             { string to be stripped }
        );

Begin
  DeTrailString( S, ' ' );
End;



{----------------------------------------------------------------------}
{ Procedure StripString                                                }
{                                                                      }
{ remove any leading or trailing occurences of the character Ch, from  }
{ the string S                                                         }
{----------------------------------------------------------------------}

Procedure StripString
        (
            var  S : BigString;             { string to be stripped }
                Ch : Char               { character to be removed }
        );

var
  Index : Integer;

Begin
  Index := Length( S );  If Index = 0 Then Exit;
  While ( S[ Index ] = Ch ) and ( Index > 0 ) Do
      Index := Index - 1;
  If Index > 0 Then
    Begin
      S[ 0 ] := Chr( Index );
      Index := 1;
      While ( S[ Index ] = Ch ) and ( Index < Length( S ) ) Do
          Index := Index + 1;
      If Index > 1 Then
          Delete( S, 1, Index - 1 );
    End;
End;



{----------------------------------------------------------------------}
{ Procedure Strip                                                      }
{                                                                      }
{ remove any leading or trailing spaces from the string S              }
{----------------------------------------------------------------------}

Procedure Strip
        (
            var  S : BigString              { string to be stripped }
        );

var
  Index : Integer;

Begin
  StripString( S, ' ' );
End;



{----------------------------------------------------------------------}
{ Function Upper                                                       }
{                                                                      }
{ convert string to upper case                                         }
{----------------------------------------------------------------------}

Function Upper
        (
                S : BigString
        ) : BigString;

var
  Index : Integer;

Begin
  For Index := 1 To Length( S ) Do
      S[ Index ] := UpCase( S[ Index ] );
  Upper := S
End;



{----------------------------------------------------------------------}
{ Function LowCase                                                     }
{                                                                      }
{ convert character to lower case, opposite of built in UpCase         }
{----------------------------------------------------------------------}

Function LowCase
        (         Ch : Char
        ) : Char;

Begin
  If Ch in [ 'A'..'Z' ] Then
    LowCase := Chr( Ord( Ch ) + ( Ord( 'a' ) - Ord( 'A' ) ) )
  Else
    LowCase := Ch;
End;



{----------------------------------------------------------------------}
{ Function ControlKey                                                  }
{                                                                      }
{ convert a character into a control character                         }
{----------------------------------------------------------------------}

Function ControlKey
        (         Ch : Char
        ) : Char;

Begin
  If Ch in [ 'A'..'Z', 'a'..'z' ] Then
    ControlKey := Chr( Ord( UpCase( Ch ) ) - Ord( 'A' ) + 1 )
  Else ControlKey := Ch;
End;



{----------------------------------------------------------------------}
{ Function Capital                                                     }
{                                                                      }
{ return string with first letter capitalized                          }
{----------------------------------------------------------------------}

Function Capital
        (         S : BigString
        ) : BigString;

Begin
  If S <> '' Then
      S[ 1 ] := UpCase( S[ 1 ] );
  Capital := S;
End;



{----------------------------------------------------------------------}
{ Function CenterString                                                }
{                                                                      }
{ center string S in a field of Width characters                       }
{ if S is longer than Width, truncate it, otherwise, pad it with Ch    }
{----------------------------------------------------------------------}

Function CenterString
        (
                S : BigString;             { string to be centered }
                Ch : Char;              { character to pad with }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

var
  Result : BigString;  CenterPos : Integer;

Begin
  If Width < 0 Then Exit;
  If Length( S ) = Width Then
      CenterString := S
  Else
  If Length( S ) > Width Then
    Begin
      S[ 0 ] := Chr( Width );
      CenterString := S
    End
  Else
  If Width - Length( S ) = 1 Then
      CenterString := S + Ch
  Else
    Begin
      Result[ 0 ] := Chr( Width );
      CenterPos := ( Width - Length(S) ) div 2;
      ChkPtr( Result[ CenterPos ], Width, 'string center' );
      FillChar( Result[ 1 ], Width, Ch );
      Move( S[ 1 ], Result[ CenterPos ], Length(S) );
      CenterString := Result
    End
End;



{----------------------------------------------------------------------}
{ Function Center                                                      }
{                                                                      }
{ center string S in a field of Width characters                       }
{ if S is longer than Width, truncate it, otherwise, pad it with spaces}
{----------------------------------------------------------------------}

Function Center
        (
                S : BigString;             { string to be centered }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

Begin
  Center := CenterString( S, ' ', Width );
End;



{----------------------------------------------------------------------}
{ Function CJ                                                          }
{                                                                      }
{ center string in Width characters, by prepending spaces, same        }
{ as the function Center, with trailing spaces removed from result     }
{----------------------------------------------------------------------}

Function CJ
        (
                S : BigString;             { string to be centered }
                Width : Integer            { width of field }
        ) : BigString;                     { result }

var
  Index : Integer;

Begin
  If Width < 0 Then Exit;
  If Length( S ) > Width Then
      S[ 0 ] := Chr( Width )
  Else
    Begin
      Index := Width + LongIntMath - Length( S );
      If Index > 1 Then
        Begin
          Index := Index div 2;
          ChkPtr( S, Length( S ) + Index, 'string CJ' );
          Move( S[ 1 ], S[ Index + 1 ], Length( S ) );
          FillChar( S[ 1 ], Index, ' ' );
          S[ 0 ] := Chr( Length( S ) + Index );
        End;
    End;
  CJ := S
End;



{----------------------------------------------------------------------}
{ Function Commas                                                      }
{                                                                      }
{ take a string, assumed to be a number, and add commas before the     }
{ decimal point                                                        }
{----------------------------------------------------------------------}

Function Commas
        (
                S : String40
        ) : String40;

var
  OutSt : String40;
  Point : Integer;
  Loop, Len : Integer;

Begin
  Strip( S );
  If Length( S ) < 4 Then
      OutSt := S
  Else
    Begin  { overall length > 3 }
      OutSt := '';
      Point := Pos( '.', S );
      If Point = 0 Then
          Len := Length( S )
      Else
        Begin  { copy decimal part }
          Len := Point - 1;
          OutSt := Copy( S, Point, Length( S ) - Len )
        End;  { copy decimal part }
      Loop := Len;
      Point := 0;
      While Loop > 0 do
        Begin   { process mantissa }
          If ( Point >= 3 ) and ( S[ Loop ] in [ '1'..'9' ] ) Then
            Begin
              OutSt := CommaStr + OutSt;
              Point := 0
            End;
          Inc( Point );
          OutSt := S[ Loop ] + OutSt;
          Dec( Loop );
        End    { process mantissa }
    End;   { overall length > 3 }
  Commas := OutSt
End;



{----------------------------------------------------------------------}
{ Function I_S                                                         }
{                                                                      }
{ convert an integer into a string                                     }
{----------------------------------------------------------------------}

Function I_S
        (
                InNum : LongInt
        ) : String10;

var
  S : String40;

Begin
  Str( InNum, S );
  Strip( S );
  I_S := S
End;



{----------------------------------------------------------------------}
{ Function R_S                                                         }
{                                                                      }
{ convert a real into a string                                         }
{----------------------------------------------------------------------}

Function R_S
        (
                InNum:Real;
                Decimal:Integer
        ) : String10;

var
  S : String40;

Begin
  Str( InNum : 1 : Decimal, S );
  Strip( S );
  R_S := S
End;



{----------------------------------------------------------------------}
{ Function B_S                                                         }
{                                                                      }
{ convert a Boolean into a string                                      }
{----------------------------------------------------------------------}

Function B_S
        (
                InB : Boolean
        ) : String5;

Begin
  If InB Then
      B_S := 'TRUE'
  Else
      B_S := 'FALSE'
End;



{----------------------------------------------------------------------}
{ Function S_I                                                         }
{                                                                      }
{ convert a string into an integer, bad conversion returns zero        }
{ and sets global ConversionError boolean                              }
{----------------------------------------------------------------------}

Function S_I
        (
                InStr : SmallString
        ) : LongInt;

var
  OutNum : LongInt;  NumLoop, Code : Integer;

Begin
  NumLoop := Length( InStr );
  While NumLoop > 0 do
    BEGIN
      If not ( InStr[ NumLoop ] in [ '0'..'9', '-' ] ) then
        Delete( InStr, NumLoop, 1 );
      Dec( NumLoop );
    END;
  Val( InStr, OutNum, Code );
  ConversionError := Code <> 0;
  If ConversionError then OutNum := 0;
  S_I := OutNum
End;



{----------------------------------------------------------------------}
{ Function S_R                                                         }
{                                                                      }
{ convert a string into a real, bad conversion returns zero            }
{ and sets global ConversionError boolean                              }
{----------------------------------------------------------------------}

Function S_R
        (
                InStr : SmallString
        ) : Real;

var
  OutNum : Real;  NumLoop, Code : Integer;

Begin
  NumLoop := Length( InStr );
  While NumLoop > 0 do
    BEGIN
      If not ( InStr[ NumLoop ] in [ '0'..'9', '.', 'E', 'e', '-' ] ) then
        Delete( InStr, NumLoop, 1 );
      Dec( NumLoop );
    END;
  Val( InStr, OutNum, Code );
  ConversionError := Code <> 0;
  If ConversionError then OutNum := 0.0;
  S_R := OutNum
End;



{----------------------------------------------------------------------}
{ Function OKMsg                                                       }
{                                                                      }
{ return the string 'OK.', or the string for an error message          }
{----------------------------------------------------------------------}

Function OKMsg
        (
                Result : Integer
        ) : SmallString;

Begin
  If Result = 0 Then
    OKMsg := ' OK.'
  Else
    OKMsg := ', error # ' + I_S( Result ) + '.'
End;



{----------------------------------------------------------------------}
{ Function IsAre                                                       }
{                                                                      }
{ return the string 'is' for a number of 1, return 'are' otherwise     }
{----------------------------------------------------------------------}

Function IsAre
        (
                InNum : LongInt
        ) : String5;

Begin
  If InNum = 1 Then
    IsAre := 'is'
  Else
    IsAre := 'are'
End;



{----------------------------------------------------------------------}
{ Function Plural                                                      }
{                                                                      }
{ return the string '' for a number of 1, return 's' otherwise         }
{----------------------------------------------------------------------}

Function Plural
        (
                InNum : LongInt
        ) : String2;

Begin
  If InNum = 1 Then
    Plural := ''
  Else
    Plural := 's'
End;


{----------------------------------------------------------------------}
{ Function ParseStr                                                    }
{                                                                      }
{ Given set of valid separator characters,                             }
{ returns substring parsed from main string upto first separator.      }
{ Removes substring from source string, separator is discarded.        }
{ Global ParseCharSet is used for separators, a calling procedure can  }
{ change this set, it is reset to DefaultParseCharSet after every use. }
{----------------------------------------------------------------------}

Function ParseStr
        (
                var InStr : BigString  { source string }
        ) : BigString;   { returned substring }

var
  ParseLoop : Integer;

Begin
  While ( InStr <> '' ) and  { strip leading parse characters }
        ( InStr[ 1 ] in ParseCharSet ) do
    Delete( InStr, 1, 1 );
  ParseLoop := 1;  { init char counter }
  While ( not ( InStr[ ParseLoop ] in ParseCharSet ) ) and
        ( ParseLoop <= Length( InStr ) ) do
    Inc( ParseLoop );  { loop along string until separator }
  ParseStr := Copy( InStr, 1, ParseLoop - 1 );
  InStr := Copy( InStr, ParseLoop + 1, 255 );
  ParseCharSet := DefaultParseCharSet;
End;



{----------------------------------------------------------------------}
{ Function ParseInt                                                    }
{                                                                      }
{ Given set of valid separator characters,                             }
{ returns integer parsed from main string upto first separator.        }
{ Removes substring from source string, separator is discarded.        }
{ Bad conversion returns zero and sets global ConversionError boolean  }
{----------------------------------------------------------------------}

Function ParseInt
       (
                  var InStr : BigString  { source string }
       ) : LongInt;   { returned value }

var
  TempStr : String30;

Begin
  TempStr := ParseStr( InStr );
  ParseInt := S_I( TempStr );
End;



{----------------------------------------------------------------------}
{ Function ParseReal                                                   }
{                                                                      }
{ Same as ParseInt (above) but returns real                            }
{----------------------------------------------------------------------}

Function ParseReal
       (
                  var InStr : BigString  { source string }
       ) : Real;   { returned value }

var
  TempStr : String30;

Begin
  TempStr := ParseStr( InStr );
  ParseReal := S_R( TempStr );
End;



{----------------------------------------------------------------------}
{ Function ParamScan                                                   }
{                                                                      }
{ Returns the number of a command line argument matching the target.   }
{ Returns zero if no match.  This procedure is case insensitive.       }
{----------------------------------------------------------------------}

Function ParamScan( Target : SmallString ) : Integer;

var  ParamLoop : Integer;  Param : SmallString;

Begin
  ParamScan := 0;
  If ParamCount < 1 Then Exit;
  Target := Upper( Target );
  For ParamLoop := 1 to ParamCount do
    Begin
      Param := Upper( ParamStr( ParamLoop ) );
      If Param = Target THEN ParamScan := ParamLoop;
    End;
End;



{----------------------------------------------------------------------}
{ Function FullName                                                    }
{                                                                      }
{ Returns a full program name                                          }
{ assembled from the name, version and revision.                       }
{----------------------------------------------------------------------}

Function FullName:String40;

Begin
  FullName := Sys.ProgramName + ', ' + Sys.Version + ', ' + Sys.Revision;
End;


{----------------------------------------------------------------------}
{ Function BlankLine                                                   }
{                                                                      }
{ returns true if line is empty or only spaces                         }
{----------------------------------------------------------------------}

Function BlankLine( InStr : String150 ) : Boolean;

Begin
  Strip( InStr );  BlankLine := InStr = '';
End;


{----------------------------------------------------------------------}
{ Function PrintLen                                                    }
{                                                                      }
{ Returns number of displayable/printable characters in string.        }
{----------------------------------------------------------------------}

Function PrintLen( InStr : String150 ) : Integer;

Var  PLoop, PCnt : Integer;

Begin
  PCnt := 0;
  For PLoop := 1 to Length( InStr ) do
    If InStr[ PLoop ] in PrintableCharSet Then Inc( PCnt );
  PrintLen := PCnt;
End;


{--------------------------------------------------------------------------}

Begin
  Sys.DebugEnabled := ( ParamScan( '/D' ) > 0 ) and ( not Sys.Demo );
  ParseCharSet := DefaultParseCharSet;
End.


