(*********************************************************************
**********************************************************************
****                                                              ****
****                        VIDEO.PAS                             ****
****                                                              ****
****   Basic to mid-range screen access and redirection support.  ****
****                                                              ****
**********************************************************************
**********************************************************************)

Unit Video;

{$I LIBRARY.CFG }

INTERFACE

Uses
  Dos, Crt, Global, Sounds, Math, Strings;

TYPE
  BorderCharRecordType =
    Record  TL, TR, BL, BR, T, B, L, R : Char;  End;

CONST
  BorderChars : Array [ 0..2 ] of BorderCharRecordType =
    ( ( { thin line }
        TL: #218;  TR: #191;  BL: #192;  BR: #217;
        T: #196;  B: #196;  L: #179;  R: #179 ),
      ( { double line }
        TL: #201;  TR: #187;  BL: #200;  BR: #188;
        T: #205;  B: #205;  L: #186;  R: #186 ),
      ( { solid blocks }
        TL: #219;  TR: #219;  BL: #219;  BR: #219;
        T: #219;  B: #219;  L: #219;  R: #219 ) );
  Brdr_Thin = 0;
  Brdr_Dbl = 1;
  Brdr_Solid = 2;
  DefaultBorderStyle = 0;
  DefaultBorderColor = Yellow;
  MessageBorderStyle = 2;
  MessageBorderColor = LightRed;
  MessageTextColor   = White;
  Reverse        = LightGray * $10;   { white (bright) background }
  DimReverse     = DarkGray * $10;    { gray (dim) background }
  Bright         = White;             { white (bright) text }
  Normal         = Yellow;            { yellow (normal) text }
  Dim            = DarkGray;          { dim (gray) text }
  Underline      = LightBlue;         { normal (light blue) underline }
  DimUnderline   = Blue;              { dim (blue) underline }
  RtArChar       = #26;
  LfArChar       = #27;
  LfRtArChar     = #29;
  UpArChar       = #24;
  DnArChar       = #25;
  UpDnArChar     = #18;
  TickChar       = #251;
  UnTickChar     = #32;
  VBar           = #179;
  DblHBar        = #205;
{ Use NonTextChar as first character in a VWrite string to supress }
{ output to STDOUT. }
  NonTextChar    = #255;
  ScreenWidth    = 80;
  ScreenHeight   = 25;
  EnterSign      = #17#196#217;
  On             = True;  { use with Cursor() }
  Off            = False;
  BigTitle       = True;  { use with ClearScreen() }
  SmallTitle     = False;
  ScreenCleared : Boolean = False;  { true if screen cleared at least once }
  Frame_NoWindow = 0;     { use with Frame() }
  Frame_FullWindow = 1;
  Frame_InsetWindow = 2;
  FrameType : Byte = Frame_InsetWindow;
  BorderColor : Byte = DefaultBorderColor;
  BorderStyle : Byte = DefaultBorderStyle;
  _Mono = $07;
  TopMargin : Byte = 6;  { number of header lines }
{ encoded 'RSP Electronics, Inc.' }
{ encoded 'RSP Electronics, Inc.' }
  ScreenDeveloper : String30 =
      #209#209#205#156#192#230#222#219#235#232#228#226#220#213#228#156 +
      #143#183#219#207#153;
  MaxScreenStack     = 4;  { maximum number of pushed screens }
  ScreenStackPointer : Byte = 0;  { push screen stack pointer }
  Btm_LetterMenu =
    ' '#24#25' Move Through Choices   <RET> Select   <ESC> Exit';
  Btm_QuickMenu =
    ' Select Using Highlighted Letter   <ESC> Exit';
  Btm_PickMenu =
    ' '#24#25' Move through list   <RET> Select   <ESC> Exit';
  Btm_EditMenu =
    ' '#24#25' Move through list   Type in data   <ESC> Exit';
  Btm_MarkMenu =
    ' '#24#25' Move   <+>/<-> Mark/unmark   <RET> Finish   <ESC> Exit';
  Btm_DataEntry =
    ' '#27#26' Move   Type in data   <RET> Finish   <ESC> Exit';
  FrameX1 : Word = 0;  FrameX2 : Word = 0;  { most recent frame }
  FrameY1 : Word = 0;  FrameY2 : Word = 0;

TYPE
  ScreenType = Array [ 1..ScreenHeight, 1..ScreenWidth ] of
    Record  Ch : Char;  Attr : Byte;  End;
  SaveStackRecord =
    Record
      SaveScreen : ^ScreenType;
      AtX, AtY, WX1, WY1, WX2, WY2, Attr, CursorSave,
      FX1, FX2, FY1, FY2, TopMgn : Word;
    End;

VAR
  VMode : Byte;
  Screen : ^ScreenType;
  ScreenStack : Array [ 1..MaxScreenStack ] of SaveStackRecord;
  VSeg, VOfs  : Word;
  PrevPercent : Real;
  ConLnBuffer : Array [ 1 .. 255 ] of Char; {buffer for CON device driver}
  BottomLnSave : GenlString;
  VideoExitSave : Pointer;
  Con : Text;  { predefined and opened console output device }

Procedure SetCursor( CursorSet : Word );
Procedure Cursor( COn : Boolean );
Procedure ReverseVideo;
Procedure NormVideo;
Procedure LowVideo;
Procedure HighVideo;
Procedure ShowTime;
Procedure BottomLn( BLn : GenlString );
Procedure SetVideo;
Procedure ScreenChkPtr( Var InPtr;  PtrSize : LongInt;  Msg : GenlString );
Procedure VWrite( X, Y, Attr : Integer;  S : GenlString );
Procedure Process( Msg1, Msg2 : String40 );
Procedure ClearLine( Y : Integer );
Procedure ClearScreen( FullSize : Boolean );
Procedure GetVAttr( VAR X, Y, WX1, WY1, WX2, WY2, Attr, CursorIn : Word );
Procedure SetVAttr( X, Y, WX1, WY1, WX2, WY2, Attr, CursorIn : Word );
Procedure SaveBottomLn;
Procedure PushScreen;
Procedure PopScreen;
Procedure PutScreen( ScreenNum : Byte );
Procedure Frame( X1, Y1, X2, Y2 : Integer );
Procedure WindowTitle( X, VY:Word;  Attr:Word;  Title:GenlString );
Procedure FullWindow;
Procedure MeasureMessage( Msg : String;
                          var Width, Height, LastX : Integer );
Procedure StatusWindow;
Procedure ShowPercent( Percent : Real );
Function  GetCursor : Word;
Procedure AssignCon(Var ConFile : Text);
Function  ConOpen(Var ConFile : TextRec) : Integer;
Function  ConWrite(Var ConFile : TextRec) : Integer;
Function  ConVoid(Var ConFile : TextRec) : Integer;


IMPLEMENTATION

{---------------------------------------------------------------------------}

Function GetCursor : Word;
  BEGIN
    Regs.AX := $0300;
    Regs.BX := $0000;
    Intr( $10, Regs );
    GetCursor := Regs.CX;
  END;

{---------------------------------------------------------------------------}

Procedure SetCursor( CursorSet : Word );
  BEGIN
    If Sys.RemoteOnly THEN Exit;
    Regs.AX := $0100;
    Regs.CX := CursorSet;
    Intr( $10, Regs );
  END;

{---------------------------------------------------------------------------}

Procedure Cursor( COn : Boolean );
  VAR  CursorVal : Word;
  Begin
    CursorVal := GetCursor;
    If COn
      THEN CursorVal := CursorVal and $DFFF { on, reset bit 5 }
      ELSE CursorVal := CursorVal or $2000; { off, set bit 5 }
    SetCursor( CursorVal );
  End;

{---------------------------------------------------------------------------}

Procedure ReverseVideo;

BEGIN
  TextColor( Black );
  TextBackground( White );
END;

{---------------------------------------------------------------------------}

Procedure NormVideo;

BEGIN
  TextColor( Yellow );
  TextBackground( Black );
END;

{---------------------------------------------------------------------------}

Procedure LowVideo;
  BEGIN  TextColor( LightGray );  TextBackground( Black );  END;

{---------------------------------------------------------------------------}

Procedure HighVideo;
  BEGIN  TextColor( Yellow ); TextBackground( Black );  END;

{---------------------------------------------------------------------------}
{ Procedure ShowTime                                                        }
{                                                                           }
{ Displays the current system time in the bottom left-hand corner of the    }
{ screen if the Sys.ShowTimeEnabled flag is true.                           }
{---------------------------------------------------------------------------}

Procedure ShowTime;
  Var Hours, Minutes, Seconds : Word;
  Begin
    If Sys.RemoteOnly or ( not Sys.ShowTimeEnabled ) THEN Exit;
    ScreenChkPtr( Screen^[ ScreenHeight, 72 ], 16, 'screen time' );
    Hours := Mem[ $40 : $6E ];   { hours }
    Minutes := MemW[ $40 : $6C ] div 1092;  { minutes }
    Seconds := MemW[ $40 : $6C ] mod 1092 div 18;  { sec }
    Screen^[ ScreenHeight, 72 ].Ch :=
      Chr( Hours div 10 + Ord( '0' ) );
    Screen^[ ScreenHeight, 73 ].Ch :=
      Chr( Hours mod 10 + Ord( '0' ) );
    Screen^[ ScreenHeight, 74 ].Ch := ':';
    Screen^[ ScreenHeight, 75 ].Ch :=
      Chr( Minutes div 10 + Ord( '0' ) );
    Screen^[ ScreenHeight, 76 ].Ch :=
      Chr( Minutes mod 10 + Ord( '0' ) );
    Screen^[ ScreenHeight, 77 ].Ch := ':';
    Screen^[ ScreenHeight, 78 ].Ch :=
      Chr( Seconds div 10 + Ord( '0' ) );
    Screen^[ ScreenHeight, 79 ].Ch :=
      Chr( Seconds mod 10 + Ord( '0' ) );
  End;

{---------------------------------------------------------------------------}
{ Procedure BottomLn                                                        }
{                                                                           }
{ Writes a message such as command instructions to the bottom line of the   }
{ screen.                                                                   }
{---------------------------------------------------------------------------}

Procedure BottomLn( BLn : GenlString );
  BEGIN
    If Sys.RemoteOnly THEN Exit;
    VWrite( 1, ScreenHeight, Reverse,
            NonTextChar + ' ' + Pad( BLn, 1 - ScreenWidth ) );
    ShowTime;
  END;

{-----------------------------------------------------------------------}
{ Procedure SetVideo                                                    }
{                                                                       }
{ Detects monochrome or color adaptor, sets mode and screen address.    }
{-----------------------------------------------------------------------}

Procedure SetVideo;
  BEGIN
    VOfs := 0;  Regs.AX := $0F00;  Intr($10, Regs );  { get CRT mode }
    VMode := Regs.AX and $00FF;
    If VMode = _Mono
      THEN VSeg := $B000 ELSE VSeg := $B800;
    Screen := Ptr( VSeg, VOfs );
  END;

{------------------------------------------------------------------------}
{ Procedure ScreenChkPtr                                                 }
{                                                                        }
{ Checks the pointer to see if it is in the range of the video memory.   }
{ If the pointer is outside video memory the program halts.              }
{------------------------------------------------------------------------}

Procedure ScreenChkPtr( Var InPtr;  PtrSize : LongInt;  Msg : GenlString );

  BEGIN
    Msg := ' ' + Msg + ' in ' + FullOperation;
    If PtrSize < 1 Then
      ChkErr( -110, 'x', Msg, '' );
    If PtrSize > 64000 Then
      ChkErr( -111, 'x', Msg, '' );
    If PtrInt( Ptr( Seg( InPtr ), Ofs( InPtr ) ) ) < $B0000 Then
      ChkErr( -112, 'x', Msg, '' );  { start of mono }
    If PtrInt( Ptr( Seg( InPtr ), Ofs( InPtr ) ) ) + PtrSize > $B8FA0 Then
      ChkErr( -113, 'x', Msg, '' );  { end of color }
  END;

{------------------------------------------------------------------------}
{ Procedure VWrite                                                       }
{                                                                        }
{ Writes a string directly to the video memory, and so to the screen     }
{ starting at the position (X,Y) with the screen attributes passed as a  }
{ parameter.                                                             }
{------------------------------------------------------------------------}

Procedure VWrite( X, Y, Attr : Integer;  S : GenlString );
  VAR  XLoop : Integer;  NonTextPos : Byte;
  BEGIN
    If ( X < 1 ) or ( X > ScreenWidth ) or
       ( Y < 1 ) or ( Y > ScreenHeight ) or
       ( S = '' ) THEN Exit;
    NonTextPos := Pos( NonTextChar, S );
    If NonTextPos = 0
      THEN RemoteLn( S )
      ELSE Delete( S, NonTextPos, 1 );
    If Sys.RemoteOnly THEN Exit;
    If X + Length( S ) - 1 > ScreenWidth
      THEN S[ 0 ] := Chr( ScreenWidth - X + 1 );
    ScreenChkPtr( Screen^[ Y, X ], Length( S ) * 2, 'screen write' );
    For XLoop := X to X + Length( S ) - 1 do
      BEGIN
        Screen^[ Y, XLoop ].Ch := S[ XLoop - X + 1 ];
        Screen^[ Y, XLoop ].Attr := Attr;
      END;
  END;

{---------------------------------------------------------------------------}
{ Procedure Process                                                         }
{                                                                           }
{ Pops up a window on the screen with a message of maximum 40 character.    }
{---------------------------------------------------------------------------}

Procedure Process( Msg1, Msg2 : String40 );
  VAR  Wid, Height : Integer;  FX1, FX2, FY1, FY2 : Word;
  BEGIN
    If ( Msg1 = '' ) and ( Msg2 = '' ) THEN Exit;
    If ( Msg1 = '' ) or ( Msg2 = '' )
      THEN Height := 0 ELSE Height := 2;
    Wid := Length( Msg1 );
    If Wid < Length( Msg2 ) THEN Wid := Length( Msg2 );
    FX1 := FrameX1;  FX2 := FrameX2;  FY1 := FrameY1;  FY2 := FrameY2;
    FrameType := Frame_NoWindow;
    Frame( 3, TopMargin + 3, 5 + Wid, TopMargin + 3 + Height );
    FrameX1 := FX1;  FrameX2 := FX2;  FrameY1 := FY1;  FrameY2 := FY2;
    RemoteLn( Msg1 + ' - ' + Msg2 );
    If Sys.RemoteOnly THEN Exit;
    If Msg1 <> ''
      THEN VWrite( 3, 3 + TopMargin, White,
                   NonTextChar + ' ' + Pad( Msg1, 0 - ( Wid + 2 ) ) );
    If Msg2 <> ''
      THEN BEGIN
        Msg2 := CenterString( ' ' + Msg2 + ' ', #196, Wid + 3 );
        Msg2 := #195 + Msg2 + #180;
        VWrite( 2, 3 + Height + TopMargin, BorderColor,
                NonTextChar +  Msg2 );
      END;
  END;


{---------------------------------------------------------------------------}
{ Procedure ClearScreen                                                     }
{                                                                           }
{ Clears the screen and puts up the standard background screen with site    }
{ name, Process name, etc.                                                  }
{---------------------------------------------------------------------------}

Procedure ClearLine( Y : Integer );
  Var
    L : Byte;
    ScreenLn : Array[ 0..79 ] of Integer;
  Begin
    ScreenChkPtr( Screen^, SizeOf( Screen ), 'screen clear' );
    For L := 0 to 79 do ScreenLn[ L ] := $07B0;  { light grey grid }
    Move( ScreenLn, Screen^[ Y, 1 ], 160 );
  End;


{---------------------------------------------------------------------------}
{ Procedure ClearScreen                                                     }
{                                                                           }
{ Clears the screen and puts up the standard background screen with site    }
{ name, Process name, etc.                                                  }
{---------------------------------------------------------------------------}

Procedure ClearScreen( FullSize : Boolean );
  VAR  L : Byte;  S : GenlString;
       ScreenLn : Array[ 0..79 ] of Integer;
       PL : Byte;
  BEGIN LowVideo;
    ScreenCleared := True;
    Sys.ShowTimeEnabled := True;
    RemoteLn( Sys.ProcessName );
    If Sys.RemoteOnly THEN Exit;
    IF not FullSize
      THEN BEGIN
        TopMargin := 1;
        VWrite( 1, 1, Reverse,
                NonTextChar + Center( Sys.ProcessName, ScreenWidth ) );
      END
      ELSE BEGIN
        TopMargin := 6;
        VWrite( 1, 1, Reverse,
                NonTextChar + #201 + Rep( #205, 78 ) + #187 );   { top }
        S := ScreenDeveloper;
        For L := 1 to Length( S ) do  { decode developer name }
          S[ L ] := Chr( Ord( S[ L ] ) + L - 128 );
        S := #186 + ' ' + Pad( S, -28 ) +            { 30 }
             Pad( Sys.SystemType, -48 ) + ' ' + #186;  { 50 }
        VWrite( 1, 2, Reverse, NonTextChar + S );
        S := #186 + ' ' + Pad( Sys.ComputerName, -12 ) +    { 14 }
             Center( Sys.ProcessName, 58 ) +            { 58 }
             Pad( Sys.Version, 6 ) + ' ' + #186;     { 8 }
        VWrite( 1, 3, Reverse, NonTextChar + S );
        VWrite( 1, 4, Reverse,
                NonTextChar + #204 + Rep( #205, 78 ) + #185 );    { middle }
        VWrite( 1, 5, Reverse,
                NonTextChar + #186 + Center( Sys.PlantName, 78 ) + #186 );
        VWrite( 1, 6, Reverse,
                NonTextChar + #200 + Rep( #205, 78 ) + #188 );   { bottom }
      END;
    For L := 0 to 79 do ScreenLn[ L ] := $07B0;  { light grey grid }
    ScreenChkPtr( Screen^, SizeOf( Screen ), 'screen clear' );
    For L := TopMargin + 1 to 24 do
      Move( ScreenLn, Screen^[ L, 1 ], 160 );
    NormVideo;  BottomLn( ' ' );
    Window( 1, 1, ScreenWidth, ScreenHeight );  GotoXY( 1, 1 );
  END;

{---------------------------------------------------------------------------}
{ Procedure GetVAttr                                                        }
{                                                                           }
{ Collects and returns all the current video attributes.                    }
{---------------------------------------------------------------------------}

Procedure GetVAttr( VAR X, Y, WX1, WY1, WX2, WY2, Attr, CursorIn : Word );
  BEGIN
    X := WhereX;  Y := WhereY;
    WX1 := Lo( WindMin ) + 1;  WX2 := Lo( WindMax ) + 1;
    WY1 := Hi( WindMin ) + 1;  WY2 := Hi( WindMax ) + 1;
    Attr := TextAttr;  CursorIn := GetCursor;
  END;

{---------------------------------------------------------------------------}
{ Procedure SetVAttr                                                        }
{                                                                           }
{ Sets all video attributes.                                                }
{---------------------------------------------------------------------------}

Procedure SetVAttr( X, Y, WX1, WY1, WX2, WY2, Attr, CursorIn : Word );
  BEGIN
    If Sys.RemoteOnly THEN Exit;
    Window( WX1, WY1, WX2, WY2 );
    TextAttr := Attr;  SetCursor( CursorIn );
    GotoXY( X, Y );
  END;

{---------------------------------------------------------------------------}
{ Procedure SaveBottomLn                                                    }
{                                                                           }
{ Saves the current bottom line message.                                    }
{---------------------------------------------------------------------------}

Procedure SaveBottomLn;
  Var
    BtmLoop, Wid : Byte;
  Begin
    If Sys.RemoteOnly Then Exit;
    If Sys.ShowTimeEnabled Then
      Wid := ScreenWidth - 10
    Else
      Wid := ScreenWidth;
    For BtmLoop := 1 to Wid do
      BottomLnSave[ BtmLoop ] :=
        Screen^[ ScreenHeight ][ BtmLoop ].Ch;
    BottomLnSave[ 0 ] := Char( Wid );
  End;


{---------------------------------------------------------------------------}
{ Procedure PushScreen                                                      }
{                                                                           }
{ Saves an exact copy of the current screen on a stack.                     }
{---------------------------------------------------------------------------}

Procedure PushScreen;
  BEGIN
    If Sys.RemoteOnly or
       ( MaxAvail < SizeOf( Screen^ ) ) or
       ( ScreenStackPointer >= MaxScreenStack ) THEN Exit;
    Inc( ScreenStackPointer );
    With ScreenStack[ ScreenStackPointer ] do
      BEGIN
        New( SaveScreen );
        ChkPtr( SaveScreen^, SizeOf( SaveScreen^ ), 'screen push' );
        ScreenChkPtr( Screen^, SizeOf( Screen^ ), 'screen push' );
        SaveScreen^ := Screen^;
        GetVAttr( AtX, AtY, WX1, WY1, WX2, WY2, Attr, CursorSave );
        FX1 := FrameX1;  FX2 := FrameX2;  FY1 := FrameY1;  FY2 := FrameY2;
        TopMgn := TopMargin;
      END;
  END;

{---------------------------------------------------------------------------}
{ Procedure PutScreen                                                       }
{                                                                           }
{ Displays a previously saved screen from anywhere in the screen stack.     }
{---------------------------------------------------------------------------}

Procedure PutScreen( ScreenNum : Byte );
  BEGIN
    If ( ScreenNum < 1 ) or ( ScreenNum > MaxScreenStack ) THEN Exit;
    With ScreenStack[ ScreenNum ] do
      BEGIN
        If Sys.RemoteOnly or ( SaveScreen = NIL ) THEN Exit;
        ChkPtr( SaveScreen^, SizeOf( SaveScreen^ ), 'screen put' );
        ScreenChkPtr( Screen^, SizeOf( Screen^ ), 'screen put' );
        Screen^ := SaveScreen^;
        SetVAttr( AtX, AtY, WX1, WY1, WX2, WY2, Attr, CursorSave );
        FrameX1 := FX1;  FrameX2 := FX2;  FrameY1 := FY1;  FrameY2 := FY2;
        TopMargin := TopMgn;
      END;
  END;

{---------------------------------------------------------------------------}
{ Procedure PopScreen                                                       }
{                                                                           }
{ Replaces the current screen with the screen most recently saved using     }
{ PushScreen.                                                               }
{---------------------------------------------------------------------------}

Procedure PopScreen;
  VAR  XLoop : Byte;
  BEGIN
    If ( ScreenStackPointer <= 0 ) THEN Exit;
    With ScreenStack[ ScreenStackPointer ] do
      BEGIN
        If SaveScreen = NIL THEN Exit;
        PutScreen( ScreenStackPointer );
        ShowTime;
        Dispose( SaveScreen );  SaveScreen := NIL;
      END;
    Dec( ScreenStackPointer );
  END;

{---------------------------------------------------------------------------}
{ Procedure Frame                                                           }
{                                                                           }
{ Draws a frame around the specified region and declares a text window.     }
{---------------------------------------------------------------------------}

Procedure Frame( X1, Y1, X2, Y2 : Integer );

  VAR  FL, MaxHeight : Word;
  BEGIN
    If ( BorderStyle < 1 ) or ( BorderStyle > 2 ) THEN BorderStyle := 0;
    With BorderChars[ BorderStyle ] do
      BEGIN
      { force coordinates into range }
        If X1 < 2 THEN X1 := 2;
        If X2 > ScreenWidth - 1 THEN X2 := ScreenWidth - 1;
        If Y1 < 2 THEN Y1 := 2;
        If Y2 > ScreenHeight - 1 THEN Y2 := ScreenHeight - 1;
        If ( X2 - X1 < 0 ) or ( Y2 - Y1 < 0 ) THEN Exit;
        If FrameType <> Frame_NoWindow
          THEN BEGIN
            Window( X1, Y1, X2, Y2 );  NormVideo;  ClrScr;
            If FrameType = Frame_InsetWindow
              THEN Window( X1 + 1, Y1 + 1, X2 - 1, Y2 - 1 );
            GotoXY( 1, 1 );
          END;
        FrameX1 := X1 - 1;  FrameX2 := X2 + 1;
        FrameY1 := Y1 - 1;  FrameY2 := Y2 + 1;
        VWrite( X1 - 1, Y1 - 1, BorderColor, NonTextChar + TL );
        VWrite( X2 + 1, Y1 - 1, BorderColor, NonTextChar + TR );
        VWrite( X1 - 1, Y2 + 1, BorderColor, NonTextChar + BL );
        VWrite( X2 + 1, Y2 + 1, BorderColor, NonTextChar + BR );
        { draw top and bottom }
        VWrite( X1, Y1 - 1, BorderColor,
                NonTextChar + Rep( T, X2 - X1 + 1 ) );
        VWrite( X1, Y2 + 1, BorderColor,
                NonTextChar + Rep( B, X2 - X1 + 1 ) );
        For FL := Y1 to Y2 do
          BEGIN  { draw left and right }
            VWrite( X1 - 1, FL, BorderColor, NonTextChar + L );
            VWrite( X2 + 1, FL, BorderColor, NonTextChar + R );
          END;
        BorderColor := DefaultBorderColor;
        BorderStyle := DefaultBorderStyle;
        FrameType := Frame_InsetWindow;
      END;
  END;

{---------------------------------------------------------------------------}
{ Procedure WindowTitle                                                     }
{                                                                           }
{ Write a message on the top or bottom of a frame.                          }
{ X is video coord., or use zero for automatic centering in current frame.  }
{ Y is video coord., use FrameY1/FrameY2 for top/bottom of current frame.   }
{---------------------------------------------------------------------------}

Procedure WindowTitle( X, VY:Word;  Attr:Word;  Title:GenlString );
  VAR  VX : Word;
  BEGIN
    If ( Title = '' ) or
       ( FrameX1 = 0 ) or ( FrameX2 = 0 ) or
       ( FrameY1 = 0 ) or ( FrameY2 = 0 ) THEN Exit;
    If X > 0
      THEN VX := X
      ELSE BEGIN
        If FrameX2 - FrameX1 + 1 < Length( Title )
          THEN Title[ 0 ] := Chr( FrameX2 - FrameX1 + 1 );
        VX := FrameX1 +
              ( FrameX2 - FrameX1 + 1 - Length( Title ) ) div 2;
      END;
    VWrite( VX, VY, Attr, Title )
  END;


{---------------------------------------------------------------------------}
{ Procedure FullWindow                                                      }
{                                                                           }
{ Make full screen available for display.                                   }
{---------------------------------------------------------------------------}

Procedure FullWindow;
  BEGIN
    Sys.ShowTimeEnabled := False;
    TextColor( LightGray );  TextBackground( Black );
    Window( 1, 1, ScreenWidth, ScreenHeight );  ClrScr;
  END;

{---------------------------------------------------------------------------}
{ Procedure MeasureMessage                                                  }
{                                                                           }
{ Measures the screen space which will be required by a message.  Width and }
{ Height are the longest line, and number of lines, and LastX is the length }
{ of the last line.                                                         }
{---------------------------------------------------------------------------}

Procedure MeasureMessage( Msg : String;
                          var Width, Height, LastX : Integer );
  var  MsgLoop, WidthCount : Word;
  Begin
    Width := 0;  Height := 0;  LastX := 0;
    WidthCount := 0;  MsgLoop := 1;
    While MsgLoop <= Length( Msg ) do
      BEGIN    { loop through string }
        If Msg[ MsgLoop ] in PrintableCharSet
          THEN BEGIN
            Inc( WidthCount );
            LastX := WidthCount;
          END;
        If ( Msg[ MsgLoop ] = #13 )  and
           ( Msg[ MsgLoop + 1 ] = #10 )  and
           ( MsgLoop + 1 <= Length( Msg ) )  then
           BEGIN    { CR/LF found }
             Inc( Height );
             If WidthCount > Width then Width := WidthCount;
             WidthCount := 0;
           END;     { CR/LF }
        Inc( MsgLoop );
      END;     { loop through string }
    If WidthCount > Width then Width := WidthCount;
  END;

{---------------------------------------------------------------------------}
{ Procedure StatusWindow                                                    }
{                                                                           }
{ Draws a general purpose box in the middle of the screen.                  }
{---------------------------------------------------------------------------}

Procedure StatusWindow;
  Begin
    Frame( 24, 9, 58, 14 );
  End;

{----------------------------------------------------------------------}
{ Procedure ShowPercent                                                }
{                                                                      }
{ This procedure displays a completion percentage for any operation.   }
{ Initialize it by using ShowPercent( 0.0 ), the feed it percentages   }
{ in a processing loop, ie. ShowPercent( FilePos(x) / FileSize(x) )    }
{ ShowPercent is window independent, and updates only every 10 %       }
{----------------------------------------------------------------------}

Procedure ShowPercent( Percent : Real );
  BEGIN
    Percent := Percent * 100;
    If Percent = 0.0
      THEN BEGIN
        Disp( '    0 % done. ' );
        PrevPercent := 0.0;
      END
      ELSE If ( ( Percent >= 100.0 ) and ( PrevPercent < 100.0 ) ) or
              ( Percent - PrevPercent >= 10.0 )
             THEN BEGIN
               Disp( ^H^H^H^H^H {#} +
                      ^H^H^H { % } + ^H^H^H^H {done} + ^H^H {. } +
                      Pad( R_S( Percent, 0 ), 5 ) + ' % done. ' );
               PrevPercent := Percent;
             END;
  END;

{---------------------------------------------------------------------------}
{ Procedure CheckRemote                                                     }
{                                                                           }
{ This procedure checks if the program's output is being redirected         }
{ by first clearing all of a check character from the screen, then doing    }
{ a console write of this character through MSDOS, then checking the screen }
{ to see if the character appeared.  The check character, arbitrarily       }
{ chosen, is ASCII 16, a right pointing triangle.                           }
{ Sys.Remote flag is set depending on this redirection check.               }
{---------------------------------------------------------------------------}

Procedure CheckRemote;
  VAR  XLoop, YLoop : Word;  RemoteResult : Integer;
  BEGIN
    If Sys.ForceRemote THEN Exit;
    ScreenChkPtr( Screen^, SizeOf( Screen^ ), 'screen remote' );
    For XLoop := 1 to ScreenWidth do  { clear check character }
      For YLoop := 1 to ScreenHeight do
        If Screen^[ YLoop, XLoop ].Ch = #16
          THEN Screen^[ YLoop, XLoop ].Ch := #32;
    RemoteLn( #16 );  { print check character to StdOut }
    Sys.Remote := True;
    For XLoop := 1 to ScreenWidth do  { scan for check character }
      For YLoop := 1 to ScreenHeight do
        If Screen^[ YLoop, XLoop ].Ch = #16
          THEN Sys.Remote := False;
    Sys.RemoteOnly := Sys.Remote and Sys.RemoteOnly;
  END;

{---------------------------------------------------------------------------}
{ Function ConVoid                                                          }
{                                                                           }
{ Dummy Function.                                                           }
{---------------------------------------------------------------------------}

Function ConVoid( Var ConFile : TextRec ) : Integer;

Begin
  ConVoid := 0;
End;

{---------------------------------------------------------------------------}
{ Function ConWrite                                                         }
{                                                                           }
{ This gets called whenever there is a call to writeln(con,..) or           }
{ write(con,..) and handles all the output to this predefined screen device.}
{---------------------------------------------------------------------------}

Function ConWrite( Var ConFile : TextRec ) : Integer;

  Var
    ConStr, ConStr1 : BigString;

  Begin
    ConWrite := 0;
    With ConFile do
      Begin
        If  BufPos = 0  then
          Begin
            BufPos := 0;  BufEnd := 0;  Exit;
          End;
        ChkPtr( ConStr1[ 1 ], BufPos - 1, 'con buffer write' );
        Move( BufPtr^, ConStr1[ 1 ], BufPos - 1 );
        ConStr1[ 0 ] := Chr( BufPos - 1 );
        BufPos := 0;  BufEnd := 0;
        { break into shorter lines and display }
        Repeat
          ParseCharSet := [LF];
          ConStr := ParseStr( ConStr1 );
          If Copy( ConStr, Length( ConStr ), 1 ) = CR then
            Delete( ConStr, Length( ConStr ), 1 );
          DispLn( ConStr );
        Until ConStr1 = '';
      End;
  End;

{---------------------------------------------------------------------------}
{ Function ConOpen                                                          }
{                                                                           }
{ Whenever there is a call to open the predefined con device, this function }
{ is executed.  It sets all the fields of the con filehandle record.        }
{---------------------------------------------------------------------------}

Function ConOpen( Var ConFile : TextRec ) : integer;

  Begin
    With ConFile do
      Begin
        If Mode = fmInput then
          Begin
            InOutFunc := @ConVoid;
            FlushFunc := @ConVoid;
          End
        else
          Begin
            Mode := fmOutput;
            InOutFunc := @ConWrite;
            FlushFunc := @ConWrite;
          End;
        CloseFunc := @ConVoid;
      End;
    ConOpen := 0;
  End;


{---------------------------------------------------------------------------}
{ Procedure AssignCon                                                       }
{                                                                           }
{ Initializes the predefined con device.                                    }
{---------------------------------------------------------------------------}

Procedure AssignCon (var ConFile : Text);

  Begin
    With TextRec(ConFile) do
      Begin
        Handle    := $FFFF;
        Mode      := fmClosed;
        BufSize   := SizeOf( ConLnBuffer );
        BufPtr    := @ConLnBuffer;
        OpenFunc  := @ConOpen;
        Name[ 0 ] := #0;
      End;
  End;

{---------------------------------------------------------------------------}
{ Procedure CloseCon                                                        }
{                                                                           }
{ Close the predefined con device.                                          }
{---------------------------------------------------------------------------}

Procedure CloseCon;
  Var
    Junk : Integer;
  Begin
    ExitProc := VideoExitSave;
    {$I-} Close( Con ); {$I+} Junk := IOResult;
  End;


{-------------------------------------------------------------------}

BEGIN

  SetVideo;
  ChkPtr( ScreenStack, SizeOf( ScreenStack ), 'screen stack' );
  FillChar( ScreenStack, SizeOf( ScreenStack ), 0 );
  CheckRemote;
  VideoExitSave := ExitProc;
  ExitProc := @CloseCon;
  AssignCon(con);  {set up and open CON device driver}
  Rewrite(con);

END.




